#!/usr/local/apps/python/2.7/bin/python

#!/Library/Frameworks/Python.framework/Versions/Current/bin/python

# Generation of dependencies for Fortran source code with modules
# Author: Tomas Wilhelmsson ECMWF

import argparse
import re
import os
import sys

# Configureable constants
fortran_extensions = set(['.f','.f90','.f95','.f03','.f08','.for','.ftn','.fpp'])
module_extensions = set(['.mod'])
includes_to_ignore = set(['mpif.h','omp_lib.h'])
modules_to_ignore = set(['omp_lib','ieee_arithmetic','ieee_exceptions',
                         'iso_c_binding','iso_fortran_env','iso_features'])

# Search patterns
cpp_include_pattern = re.compile('^ *#include +"([\w/\.]+)"')
fortran_include_pattern = re.compile('^ *include +"([\w/\.]+)"')
fortran_use_pattern = re.compile('^ *use +(\w+)',re.IGNORECASE)
fortran_module_pattern = re.compile('^ *module +(\w+)',re.IGNORECASE)

# Functions
def expand_incs(incs, incincs):
    """ Recursively expand incs from incincs """ 
    all = set([])
    for inc in incs:
        if incincs.has_key(inc):
            all |= expand_incs(incincs[inc], incincs)
    return all | incs

# From Python Cookbook, 2nd ed, section 2.18 
def search_file(filename, search_path, pathsep=os.pathsep):
    """ Given a search path, find file with requested name """
    for path in search_path.split(pathsep):
        candidate = os.path.join(path, filename)
        if os.path.isfile(candidate):
            return candidate
    return None

def add_dirinc(src, inc, incs, incincs, path):
    if inc not in includes_to_ignore:
        dirinc = search_file(inc, path)
        if dirinc:
            incs.setdefault(src, set()).add(dirinc)
            incincs[dirinc] = True
        else:
            sys.exit(src + ': include file ' + inc + ' not found')

def main():
    # Parse command line
    parser = argparse.ArgumentParser(description='Generate make dependencies.')
    parser.add_argument('sourcefiles', nargs='*')
    parser.add_argument('-I', default=[''], action='append',
                        help='include directory to search')
    parser.add_argument('-f', default='dependencies.mk',
                        help='output file (default: %(default)s)')

    args = parser.parse_args()
    source_files = set(args.sourcefiles)
    include_path = os.pathsep.join(args.I)
    outfile = args.f

    incs = {}
    uses = {}
    mods = {}
    incincs = {}
    
    # Scan source files
    for file in source_files:
        with open(file) as f:
            for line in f:
                # Find CPP #include statements
                match = cpp_include_pattern.match(line)
                if match:
                    add_dirinc(file, match.group(1), incs, incincs, include_path)

                # Only for Fortran
                if os.path.splitext(file)[-1].lower() in fortran_extensions:
                    # Find Fortran include statements
                    match = fortran_include_pattern.match(line)
                    if match:
                        add_dirinc(file, match.group(1), incs, incincs, include_path)
 
                    # Find use of modules
                    match = fortran_use_pattern.match(line)
                    if match:
                        use = match.group(1).lower()
                        if use not in modules_to_ignore:
                            uses.setdefault(file, set()).add(use)
                                
                    # Find module declarations
                    match = fortran_module_pattern.match(line)
                    if match:
                        mod = match.group(1).lower()            
                        mods[mod] = os.path.splitext(file)[0] + '.o'

    # Scan include files for more include files
    found_more_incs = True
    while found_more_incs:
        found_more_incs = False
        includefiles = incincs.keys()

        for file in includefiles:
            if incincs[file] and file not in includes_to_ignore:
                incincs[file] = set()
                dirfile = search_file(file, include_path)
                if dirfile:
                    with open(dirfile) as f:
                        for line in f:
                            # Find #include statements
                            match = cpp_include_pattern.match(line)
                            if match:
                                inc = match.group(1)
                                dirinc = search_file(inc, include_path)
				if dirinc:
                                    incincs.setdefault(file, set()).add(dirinc)
                                    if not incincs.has_key(dirinc):
                                        incincs[dirinc] = True
                                        found_more_incs = True
                                else:
                                    sys.exit('Include file ' + inc + ' not found')
                else:
                    sys.exit('Include file ' + file + ' not found')
                
    allused = set()
    for use in uses.values():
        allused |= use

    # Look for modules in include path
    for mod in allused:
        if mod not in mods:
            for ext in module_extensions:
                for case in [str.lower, str.upper]:
                    file = case(mod) + ext
                    dirfile = search_file(file, include_path)
                    if dirfile:
                        mods[mod] = dirfile
                        break

    # Now ready to write dependencies
    with open(outfile, 'w') as f:
        f.write('# DO NOT EDIT --- auto-generated file\n\n')
        for file in source_files:
            target = os.path.splitext(file)[0] + '.o'
            line = target + ' : ' + file

            # Add includes
            if incs.has_key(file):
                incs[file] = expand_incs(incs[file], incincs)
                for inc in incs[file]:
                    dirinc = search_file(inc, include_path)
                    if dirinc:
                        line = line + ' ' + dirinc
                    else:
                        sys.exit('Include file ' + inc + ' not found')

            # Add modules
            if uses.has_key(file):
                for use in uses[file]:
                    if mods.has_key(use):
                        if mods[use] != target: 
                            line = line + ' ' + mods[use]
                    else:
                        sys.exit(file +': cannot find module: ' + use)

            f.write(line + '\n')

if __name__ == '__main__':
    main()

# TODO
# Add basic #ifdef parsing
