/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include "FieldL95.h"

#include <cstring>
#include <string>

#include "logger/Logger.h"
#include "oops/base/Geometry.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/abor1_cpp.h"
#include "util/asser1.h"

#include "lorenz95/L95Fortran.h"
#include "lorenz95/GomL95.h"
#include "lorenz95/LocsL95.h"

// -----------------------------------------------------------------------------
namespace lorenz95 {
// -----------------------------------------------------------------------------
FieldL95::FieldL95(const oops::Geometry & resol, const util::DateTime & time)
  : time_(time)
{
  resol_ = resol.config().getInt("resol");
  l95field_create_f90(&flds_, resol_);
}
// -----------------------------------------------------------------------------
FieldL95::FieldL95(const FieldL95 & other)
  : resol_(other.resol_), time_(other.time_)
{
  l95field_create_f90(&flds_, resol_);
  const FieldF90 * oflds = other.flds_;
  l95field_copy_f90(&flds_, &oflds);
}
// -----------------------------------------------------------------------------
FieldL95::~FieldL95() {
  l95field_delete_f90(&flds_);
}
// -----------------------------------------------------------------------------
FieldL95 & FieldL95::operator=(const FieldL95 & rhs) {
  const FieldF90 * rhsflds = rhs.flds_;
  l95field_copy_f90(&flds_, &rhsflds);
  time_ = rhs.time_;
  return *this;
}
// -----------------------------------------------------------------------------
void FieldL95::zero() {
  l95field_zero_f90(&flds_);
}
// -----------------------------------------------------------------------------
void FieldL95::interp(const LocsL95 & locs, GomL95 & gom) const {
  const int no = locs.nobs();
  const double * flocs = locs.toFortran();
  const FieldF90 * flds = flds_;
  l95field_interpol_tl_f90(&flds, no, flocs,
                           gom.nobs(), gom.toFortran(), gom.current());
}
// -----------------------------------------------------------------------------
void FieldL95::interpAD(const LocsL95 & locs, const GomL95 & gom) {
  const int no = locs.nobs();
  const double * flocs = locs.toFortran();
  l95field_interpol_ad_f90(&flds_, no, flocs,
                           gom.nobs(), gom.toFortran(), gom.current());
}
// -----------------------------------------------------------------------------
void FieldL95::changeResolution(const FieldL95 & rhs) {
  const FieldF90 * rhflds = rhs.flds_;
  l95field_resol_f90(&flds_, &rhflds);
  time_ = rhs.time_;
}
// -----------------------------------------------------------------------------
void FieldL95::add(const FieldL95 & rhs) {
  ASSERT(rhs.time_ == time_);
  const FieldF90 * rhflds = rhs.flds_;
  l95field_add_f90(&flds_, &rhflds);
}
// -----------------------------------------------------------------------------
void FieldL95::diff(const FieldL95 & x1, const FieldL95 & x2) {
  ASSERT(x1.time_ == time_);
  ASSERT(x2.time_ == time_);
  const FieldF90 * xx1 = x1.flds_;
  const FieldF90 * xx2 = x2.flds_;
  l95field_diff_f90(&flds_, &xx1, &xx2);
}
// -----------------------------------------------------------------------------
void FieldL95::axpy(const double & zz, const FieldL95 & rhs,
                    const bool checktime) {
  ASSERT(!checktime || rhs.time_ == time_);
  const FieldF90 * rhflds = rhs.flds_;
  l95field_axpy_f90(&flds_, zz, &rhflds);
}
// -----------------------------------------------------------------------------
void FieldL95::read(const util::Config & config) {
  const std::string filename(config.getData("filename"));
  size_t len = filename.size()+1;
  char * cfilename = new char[len];
  strncpy(cfilename, filename.c_str(), len);

  char * ctime = new char[21];

  l95field_read_f90(len, cfilename, &flds_, ctime);

  const util::DateTime tt((std::string)ctime);
  const util::DateTime tc(config.getData("date"));
  if (tc != tt) {
    ABORT("FieldL95:read: date and data file inconsistent.");
  }
  time_ = tt;

  delete [] cfilename;
  delete [] ctime;
}
// -----------------------------------------------------------------------------
void FieldL95::write(const util::Config & config) const {
  std::string dir = config.getData("datadir");
  std::string exp = config.getData("exp");
  std::string type = config.getData("type");
  std::string filename = dir+"/"+exp+"."+type;

  if (type == "ens") {
    std::string memb = config.getData("member");
    filename += "."+memb;
  }

  if (type == "fc" || type == "ens") {
    const util::DateTime antime(config.getData("date"));
    filename += "."+antime.toString();
    const util::Duration step = time_ - antime;
    std::string cppstep = step.toString();
    filename += "."+cppstep;
  }

  std::string cpptime = time_.toString();
  if (type == "an") {
    filename += "."+cpptime;
  }
//  filename += ".l95";

  size_t lenfilename = filename.size()+1;
  char * cfilename = new char[lenfilename];
  strncpy(cfilename, filename.c_str(), lenfilename);

  size_t lencpptime = cpptime.size()+1;
  char * ctime = new char[lencpptime];
  strncpy(ctime, cpptime.c_str(), lencpptime);

  const FieldF90 * flds = flds_;
  l95field_write_f90(lenfilename, cfilename, &flds, ctime);
  delete [] cfilename;
  delete [] ctime;
}
// -----------------------------------------------------------------------------
void FieldL95::prtNorms(const std::string & sgrep) const {
  double zmin, zmax, zavg;
  const FieldF90 * flds = flds_;
  l95field_norm_f90(&flds, zmin, zmax, zavg);
  LOG(Info) << sgrep << " Valid time: " << validTime();
  LOG(Info) << sgrep << " Min=" << zmin << ", Max=" << zmax << ", Average=" << zavg;
}
// -----------------------------------------------------------------------------

}  // namespace lorenz95
