/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include "IncrementL95.h"

#include <string>

#include <boost/shared_ptr.hpp>

#include "logger/Logger.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/asser1.h"

#include "lorenz95/ErrorCovarianceL95.h"
#include "lorenz95/FieldL95.h"
#include "lorenz95/GomL95.h"
#include "lorenz95/L95Fortran.h"
#include "lorenz95/LocsL95.h"
#include "lorenz95/ModelError.h"
#include "lorenz95/ModelL95.h"
#include "lorenz95/StateL95.h"
#include "lorenz95/TrajectoryL95.h"

namespace lorenz95 {

// -----------------------------------------------------------------------------
/// Constructor, destructor
// -----------------------------------------------------------------------------
IncrementL95::IncrementL95(const oops::Geometry & resol, const oops::Variables &,
                           const util::DateTime & vt)
  : FieldL95(resol, vt)
{
  FieldL95::zero();
  LOG(Trace) << "IncrementL95::IncrementL95 created.";
}
// -----------------------------------------------------------------------------
IncrementL95::IncrementL95(const IncrementL95 & dx) : FieldL95(dx) {
  LOG(Trace) << "IncrementL95 copy-created.";
}
// -----------------------------------------------------------------------------
IncrementL95::~IncrementL95() {
  LOG(Trace) << "IncrementL95 destructed";
}
// -----------------------------------------------------------------------------
/// Basic operators
// -----------------------------------------------------------------------------
void IncrementL95::diff(const StateL95 & x1, const StateL95 & x2) {
  ASSERT(this->validTime() == x1.validTime());
  ASSERT(this->validTime() == x2.validTime());
  FieldL95::diff(x1, x2);
}
// -----------------------------------------------------------------------------
void IncrementL95::accumul(const double & zz, const StateL95 & xx) {
  FieldL95::axpy(zz, xx, false);
}
// -----------------------------------------------------------------------------
IncrementL95 & IncrementL95::operator=(const IncrementL95 & rhs) {
  FieldL95::operator=(rhs);
  return *this;
}
// -----------------------------------------------------------------------------
IncrementL95 & IncrementL95::operator+=(const IncrementL95 & rhs) {
  ASSERT(this->validTime() == rhs.validTime());
  FieldF90 * fld = FieldL95::toFortran();
  const FieldF90 * rhsflds = rhs.toFortran();
  l95field_add_f90(&fld, &rhsflds);
  return *this;
}
// -----------------------------------------------------------------------------
IncrementL95 & IncrementL95::operator-=(const IncrementL95 & rhs) {
  ASSERT(this->validTime() == rhs.validTime());
  FieldF90 * fld = FieldL95::toFortran();
  const FieldF90 * rhsflds = rhs.toFortran();
  l95field_sub_f90(&fld, &rhsflds);
  return *this;
}
// -----------------------------------------------------------------------------
IncrementL95 & IncrementL95::operator*=(const double & zz) {
  FieldF90 * fld = FieldL95::toFortran();
  l95field_mul_f90(&fld, zz);
  return *this;
}
// -----------------------------------------------------------------------------
void IncrementL95::zero(const util::Duration & shift) {
  FieldL95::zero();
  FieldL95::timeUpdate(shift);
}
// -----------------------------------------------------------------------------
void IncrementL95::axpy(const double & zz, const IncrementL95 & rhs,
                        const bool check) {
  ASSERT(!check || this->validTime() == rhs.validTime());
  FieldL95::axpy(zz, rhs, check);
}

// -----------------------------------------------------------------------------
void IncrementL95::schur_product_with(const IncrementL95 & dx) {
  FieldF90 * field1 = this->toFortran();
  const FieldF90 * field2 = dx.toFortran();
  l95field_schur_f90(&field1, &field2);
}

// -----------------------------------------------------------------------------
void IncrementL95::random() {
  FieldF90 * fld = FieldL95::toFortran();
  l95field_random_f90(&fld);
}
// -----------------------------------------------------------------------------
double IncrementL95::dot_product_with(const IncrementL95 & other) const {
  const FieldF90 * rh1 = this->toFortran();
  const FieldF90 * rh2 = other.toFortran();
  double zz;
  l95field_prod_f90(&rh1, &rh2, zz);
  return zz;
}
// -----------------------------------------------------------------------------
/// Run forecast TL and AD
// -----------------------------------------------------------------------------
void IncrementL95::initTL(const ModelL95 &, const TrajectoryL95 &) {}
// -----------------------------------------------------------------------------
void IncrementL95::initAD(const ModelL95 &, const TrajectoryL95 &) {}
// -----------------------------------------------------------------------------
void IncrementL95::stepTL(const ModelL95 & model, const TrajectoryL95 & traj,
                          const ModelError &) {
  const ModelConfig * ftnModel = model.toFortran();
  FieldF90 * flds = FieldL95::toFortran();
  const TrajF90 * ftraj = traj.get(validTime());
  l95_propagate_tl_f90(&ftnModel, &flds, &ftraj);
  FieldL95::timeUpdate(model.timestep());
}
// -----------------------------------------------------------------------------
void IncrementL95::stepAD(const ModelL95 & model, const TrajectoryL95 & traj,
                          ModelError &) {
  const ModelConfig * ftnModel = model.toFortran();
  FieldL95::timeUpdate(-model.timestep());
  FieldF90 * flds = FieldL95::toFortran();
  const TrajF90 * ftraj = traj.get(validTime());
  l95_propagate_ad_f90(&ftnModel, &flds, &ftraj);
}
// -----------------------------------------------------------------------------
/// Interpolate to observation location
// -----------------------------------------------------------------------------
void IncrementL95::interpolateTL(const LocsL95 & locs, GomL95 & cols) const {
  FieldL95::interp(locs, cols);
}
// -----------------------------------------------------------------------------
void IncrementL95::interpolateAD(const LocsL95 & locs, const GomL95 & cols) {
  FieldL95::interpAD(locs, cols);
}
// -----------------------------------------------------------------------------

}  // namespace lorenz95
