/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include "ObsTable.h"

#include <algorithm>
#include <fstream>
#include <iostream>
#include <map>
#include <memory>
#include <sstream>
#include <string>
#include <vector>

#include "logger/Logger.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/abor1_cpp.h"
#include "util/asser1.h"

#include "lorenz95/LocsL95.h"
#include "lorenz95/ObsVec1D.h"

using std::string;
using std::endl;
using std::ifstream;
using std::ofstream;
using std::stringstream;

// -----------------------------------------------------------------------------
namespace lorenz95 {
// -----------------------------------------------------------------------------

ObsTable::ObsTable(const util::Config & config) {
  if (config.elementExists("ObsFileIn")) {
    nameIn_  = config.getData("ObsFileIn");
    string filename = nameIn_;
    otOpen(filename);
  } else {
    nameIn_.clear();
  }
  nameOut_ = config.getData("ObsFileOut");
  LOG(Trace) << "ObsTable::ObsTable read observations.";
}

// -----------------------------------------------------------------------------

ObsTable::~ObsTable() {
  LOG(Trace) << "ObsTable::~ObsTable nameOut=" << nameOut_;
  if (nameOut_ != "null") {
    LOG(Trace) << "ObsTable::~ObsTable() saving nameOut = " << nameOut_;
    otWrite(nameOut_);
  }
}

// -----------------------------------------------------------------------------

void ObsTable::createObs(const std::vector<util::DateTime> & times,
                         const std::vector<double> & locs) {
  ASSERT(locs.size() == times.size());
  times_ = times;
  locations_ = locs;
}

// -----------------------------------------------------------------------------

void ObsTable::putdb(const string & col, const ObsVec1D & vec) {
  ASSERT(vec.nobs() == nobs());
  ASSERT(data_.find(col) == data_.end());
  data_.insert(std::pair<string, ObsVec1D>(col, vec));
}

// -----------------------------------------------------------------------------

ObsVec1D * ObsTable::getdb(const string & col) const {
  std::map<string, ObsVec1D>::const_iterator ic = data_.find(col);
  ASSERT(ic != data_.end());
  return new ObsVec1D(ic->second);
}

// -----------------------------------------------------------------------------

LocsL95 * ObsTable::locations(const util::DateTime & t1,
                              const util::DateTime & t2) const {
  std::vector<int> olist = timeSelect(t1, t2);
  const int nobs = olist.size();
  std::vector<double> locs(nobs);
  for (int jobs = 0; jobs < nobs; ++jobs) locs[jobs]=locations_[olist[jobs]];
  return new LocsL95(locs);
}

// -----------------------------------------------------------------------------

std::vector<int> ObsTable::timeSelect(const util::DateTime & t1,
                                      const util::DateTime & t2) const {
  std::vector<int> mask;
  for (int jobs = 0; jobs < nobs(); ++jobs)
    if (times_[jobs] > t1 && times_[jobs] <= t2) mask.push_back(jobs);
  return mask;
}

// -----------------------------------------------------------------------------
//  ObsTable Private Methods
// -----------------------------------------------------------------------------

void ObsTable::otOpen(const string & filename) {
  ifstream fin(filename.c_str());
  if (!fin.is_open()) ABORT("ObsTable::otOpen: Error opening file");

  int ncol, nobs;
  fin >> ncol;
  LOG(Trace) << "GroupData ncol=" << ncol;

  std::vector<string> colnames;
  for (int jc = 0; jc < ncol; ++jc) {
    string col;
    fin >> col;
    colnames.push_back(col);
  }
  LOG(Trace) << "GroupData reading columns:";
  for (int jc = 0; jc < ncol; ++jc)
    LOG(Trace) << colnames[jc];

  fin >> nobs;
  LOG(Trace) << "GroupData nobs=" << nobs;
  locations_.resize(nobs);
  ObsVec1D newcol(nobs);
  for (int jc = 0; jc < ncol; ++jc) {
    ASSERT(data_.find(colnames[jc]) == data_.end());
    data_.insert(std::pair<string, ObsVec1D>(colnames[jc], newcol));
  }

  times_.clear();
  int jjj;
  for (int jobs = 0; jobs < nobs; ++jobs) {
    fin >> jjj;
    ASSERT(jjj == jobs);
    string sss;
    fin >> sss;
    DateTime ttt(sss);
    times_.push_back(ttt);
    fin >> locations_[jobs];
    for (std::map<string, ObsVec1D>::iterator jo = data_.begin();
         jo != data_.end(); ++jo)
      fin >> jo->second(jobs);
  }

  fin.close();
  LOG(Trace) << "ObsTable::ot_read done";
}

// -----------------------------------------------------------------------------

void ObsTable::otWrite(const string & filename) const {
  LOG(Trace) << "ObsTable::otWrite writing " << filename;
  ofstream fout(filename.c_str());
  if (!fout.is_open()) ABORT("ObsTable::otWrite: Error opening file");

  int ncol = data_.size();
  fout << ncol << endl;

  for (std::map<string, ObsVec1D>::const_iterator jo = data_.begin();
       jo != data_.end(); ++jo)
    fout << jo->first << endl;

  int nobs = times_.size();
  fout << nobs << endl;
  for (int jobs = 0; jobs < nobs; ++jobs) {
    fout << jobs;
    fout << "  " << times_[jobs];
    fout << "  " << locations_[jobs];
    for (std::map<string, ObsVec1D>::const_iterator jo = data_.begin();
         jo != data_.end(); ++jo)
      fout << "  " << jo->second(jobs);
    fout << endl;
  }

  fout.close();
  LOG(Trace) << "ObsTable::otWrite done";
}

// -----------------------------------------------------------------------------

}  // namespace lorenz95
