/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include "ObsVec1D.h"

#include <boost/foreach.hpp>

#include "logger/Logger.h"
#include "util/asser1.h"
#include "util/random_vector_f90.h"

namespace lorenz95 {
// -----------------------------------------------------------------------------
ObsVec1D::ObsVec1D(const ObsVec1D & other, const bool copy): data_(other.nobs())
{
  if (copy) data_ = other.data_;
}
// -----------------------------------------------------------------------------
ObsVec1D & ObsVec1D::operator= (const ObsVec1D & rhs) {
  data_ = rhs.data_;
  return *this;
}
// -----------------------------------------------------------------------------
ObsVec1D & ObsVec1D::operator*= (const double & zz) {
  BOOST_FOREACH(double & val, data_) val *= zz;
  return *this;
}
// -----------------------------------------------------------------------------
ObsVec1D & ObsVec1D::operator+= (const ObsVec1D & rhs) {
  ASSERT(data_.size() == rhs.data_.size());
  for (int jj = 0; jj < nobs(); ++jj) data_[jj] += rhs.data_[jj];
  return *this;
}
// -----------------------------------------------------------------------------
ObsVec1D & ObsVec1D::operator-= (const ObsVec1D & rhs) {
  ASSERT(data_.size() == rhs.data_.size());
  for (int jj = 0; jj < nobs(); ++jj) data_[jj] -= rhs.data_[jj];
  return *this;
}
// -----------------------------------------------------------------------------
ObsVec1D & ObsVec1D::operator*= (const ObsVec1D & rhs) {
  ASSERT(data_.size() == rhs.data_.size());
  for (int jj = 0; jj < nobs(); ++jj) data_[jj] *= rhs.data_[jj];
  return *this;
}
// -----------------------------------------------------------------------------
ObsVec1D & ObsVec1D::operator/= (const ObsVec1D & rhs) {
  ASSERT(data_.size() == rhs.data_.size());
  for (int jj = 0; jj < nobs(); ++jj) data_[jj] /= rhs.data_[jj];
  return *this;
}
// -----------------------------------------------------------------------------
void ObsVec1D::zero() {
  BOOST_FOREACH(double & val, data_) val = 0.0;
}
// -----------------------------------------------------------------------------
void ObsVec1D::invert() {
  BOOST_FOREACH(double & val, data_) val = 1.0/val;
}
// -----------------------------------------------------------------------------
void ObsVec1D::axpy(const double & zz, const ObsVec1D & rhs) {
  ASSERT(data_.size() == rhs.data_.size());
  for (int jj = 0; jj < nobs(); ++jj) data_[jj] += zz * rhs.data_[jj];
}
// -----------------------------------------------------------------------------
void ObsVec1D::random() {
  util::random_vector_f90(data_.size(), &data_[0]);
}
// -----------------------------------------------------------------------------
double ObsVec1D::dot_product_with(const ObsVec1D & other) const {
  ASSERT(data_.size() == other.data_.size());
  double zz = 0.0;
  for (int jj = 0; jj < nobs(); ++jj) zz += data_[jj] * other.data_[jj];
  return zz;
}
// -----------------------------------------------------------------------------
void ObsVec1D::minMaxAvg(double & zmin, double & zmax, double & zavg) const {
  ASSERT(nobs()>0);
  zmin = data_[0];
  zmax = data_[0];
  zavg = 0.0;
  BOOST_FOREACH(const double & val, data_) {
    if (val < zmin) zmin = val;
    if (val > zmax) zmax = val;
    zavg += val;
  }
  zavg /= nobs();
}
// -----------------------------------------------------------------------------
}  // namespace lorenz95
