/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include "ObservationL95.h"

#include <string>
#include <vector>

#include "logger/Logger.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/asser1.h"

#include "lorenz95/GomL95.h"
#include "lorenz95/LocsL95.h"
#include "lorenz95/NoObsTraj.h"
#include "lorenz95/ObsBias.h"
#include "lorenz95/ObsTable.h"
#include "lorenz95/ObsVec1D.h"

// -----------------------------------------------------------------------------
namespace lorenz95 {
// -----------------------------------------------------------------------------
std::string ObservationL95::obsname_ = "Lorenz 95";
// -----------------------------------------------------------------------------

ObservationL95::ObservationL95(const util::Config & config) {
  LOG(Trace) << "ObservationL95::ObservationL95 start";

// Get Obs Handler
  const util::Config dataConfig(config, "ObsData");
  obsdb_.reset(new ObsTable(dataConfig));

  LOG(Trace) << "ObservationL95::ObservationL95 done";
}

// -----------------------------------------------------------------------------

ObservationL95::~ObservationL95() {}

// -----------------------------------------------------------------------------

void ObservationL95::obsEquiv(const GomL95 & gom, ObsVec1D & ovec,
                              const ObsBias &) const {
  for (int jj = 0; jj < gom.nobs(); ++jj) {
    const int ii = gom.getobs(jj);
    ovec(ii)=gom(jj);
  }
}

// -----------------------------------------------------------------------------

void ObservationL95::obsEquiv(const GomL95 & gom, ObsVec1D & ovec,
                              const ObsBias &, NoObsTraj &) const {
  for (int jj = 0; jj < gom.nobs(); ++jj) {
    const int ii = gom.getobs(jj);
    ovec(ii)=gom(jj);
  }
}

// -----------------------------------------------------------------------------

void ObservationL95::obsEquivTL(const GomL95 & gom, ObsVec1D & ovec,
                                const ObsBias &, const NoObsTraj &) const {
  for (int jj = 0; jj < gom.nobs(); ++jj) {
    const int ii = gom.getobs(jj);
    ovec(ii) = gom(jj);
  }
}

// -----------------------------------------------------------------------------

void ObservationL95::obsEquivAD(GomL95 & gom, const ObsVec1D & ovec,
                                ObsBias &, const NoObsTraj &) const {
  for (int jj = 0; jj < gom.nobs(); ++jj) {
    const int ii = gom.getobs(jj);
    gom(jj) = ovec(ii);
  }
}

// -----------------------------------------------------------------------------

LocsL95 * ObservationL95::locations(const util::DateTime & t1,
                                    const util::DateTime & t2) const {
  return obsdb_->locations(t1, t2);
}

// -----------------------------------------------------------------------------

GomL95 * ObservationL95::getGOM(const util::DateTime & t1,
                                const util::DateTime & t2) const {
  std::vector<int> mask = obsdb_->timeSelect(t1, t2);
  return new GomL95(mask);
}

// -----------------------------------------------------------------------------

ObsVec1D * ObservationL95::read(const std::string & name) const {
  ObsVec1D * pvec = obsdb_->getdb(name);
  return pvec;
}

// -----------------------------------------------------------------------------

void ObservationL95::save(const std::string & name, const ObsVec1D & ov) const {
  obsdb_->putdb(name, ov);
}

// -----------------------------------------------------------------------------

ObsVec1D * ObservationL95::distribute(const util::Config & config,
          const util::DateTime & bgn, const util::DateTime & end) {
  util::Duration first(config.getData("begin"));
  util::Duration last(end-bgn);
  if (config.elementExists("end")) {
    last = util::Duration(config.getData("end"));
  }
  util::Duration freq(config.getData("obs_frequency"));

  int nobstimes = 0;
  util::Duration step(first);
  while (step <= last) {
    ++nobstimes;
    step += freq;
  }

  int nobs_locations = config.getInt("obs_density");
  int nobs = nobs_locations*nobstimes;
  double dx = 1.0/static_cast<double>(nobs_locations);
  LOG(Trace) << "ObservationL95:distribute nobs=" << nobs;

  std::vector<util::DateTime> times;
  times.reserve(nobs);

  std::vector<double> locs(nobs);
  ObsVec1D obserr(nobs);

  int iobs = 0;
  util::DateTime now(bgn);
  step = first;
  while (step <= last) {
    now = bgn + step;
    LOG(Debug) << "distribute now=" << now;
    for (int jobs = 0; jobs < nobs_locations; ++jobs) {
      double xpos = jobs*dx;
      times.push_back(now);
      locs[iobs] = xpos;
      obserr(iobs) = 0.4;
      ++iobs;
    }
    step += freq;
  }
  ASSERT(iobs == nobs);

  LOG(Trace) << "distribute before obsdb create";
  obsdb_->createObs(times, locs);
  LOG(Trace) << "distribute after obsdb create";
  obsdb_->putdb("ObsErr", obserr);

  return new ObsVec1D(nobs);
}

// -----------------------------------------------------------------------------

}  // namespace lorenz95
