!
! Copyright 2011 ECMWF
! 
! This software was developed at ECMWF for evaluation
! and may be used for academic and research purposes only.
! The software is provided as is without any warranty.
! 
! This software can be used, copied and modified but not
! redistributed or sold. This notice must be reproduced
! on each copy made.
!

!> Fortran module to handle L95 background and model error covariance
module L95BKGERR

use l95fields
use random_vectors_gauss_mod
use iso_c_binding
use c_logger_mod
use kinds

implicit none

!> Background and model error covariance data for the L95 model
type l95_b_config
  real(kind=kind_real) :: sigb
  real(kind=kind_real), pointer :: sqrtb(:)
end type l95_b_config
private

! ------------------------------------------------------------------------------
contains
! ------------------------------------------------------------------------------

subroutine c_l95bkg_setup(c_config,sigb,rlen,c_resol) bind(c,name='l95bkg_setup_f90')

implicit none
type(c_ptr), intent(inout) :: c_config
real(c_double), intent(in) :: sigb
real(c_double), intent(in) :: rlen
integer(c_int), intent(in) :: c_resol

integer :: i,j,nresol
integer :: ifax(13)
character(len=160) :: record
real(kind=kind_real), allocatable :: trigs(:), work(:), struct_fn(:)
real(kind=kind_real) :: z
type(l95_b_config), pointer :: config

nresol = c_resol

allocate(config)

config%sigb = sigb

!--- Calculate spectral weights for correlations:
!--- First we construct the structure function in grid space and FFT it

allocate(struct_fn(nresol+2))
do j=1,nresol
  z = (sqrt(0.5_kind_real)/rlen)*real(min(j-1,1+nresol-j),kind_real)
  struct_fn(j) = exp(-z*z)
enddo

allocate (trigs(3*nresol/2+1))
allocate(work(nresol+2))
call set99 (trigs,ifax,nresol)
call fft991 (struct_fn,work,trigs,ifax,1,nresol+2,nresol,1,-1)
deallocate(work)
deallocate(trigs)

if (mod(nresol,2)/=0) then
  write(record,*) "c_l95bkg_setup: number of gridpoints=",nresol
  call c_logger(logger_error,trim(record))
  call abor1_ftn("c_l95bkg_setup: odd number of grid points")
endif

!--- We are after sqrt(B) or sqrt(Q), so square-root the coefficents.
!--- The factor N ensures we get the struture function back if we apply
!--- B or Q to a unit delta function.

allocate(config%sqrtb(0:nresol/2))

do i=0,nresol/2
  config%sqrtb(i) = sqrt(max(0.0_kind_real, real(nresol,kind_real)*struct_fn(1+2*i)))
enddo

deallocate(struct_fn)

!--- return the config structure to c++

c_config = c_loc(config)

return
end subroutine c_l95bkg_setup

! ------------------------------------------------------------------------------

subroutine c_l95bkg_delete(c_config) bind(c,name='l95bkg_delete_f90')

type(c_ptr), intent(inout) :: c_config

type(l95_b_config), pointer :: config

call c_f_pointer(c_config, config)

deallocate(config%sqrtb)
deallocate(config)

end subroutine c_l95bkg_delete

! ------------------------------------------------------------------------------

subroutine c_l95bkg_mult(c_fldin,c_fldout,c_config) &
         & bind(c,name='l95bkg_mult_f90')

implicit none
type(c_ptr), intent(in)    :: c_fldin
type(c_ptr), intent(inout) :: c_fldout
type(c_ptr), intent(in)    :: c_config

integer :: kk
real(c_double), allocatable :: vect(:)
type(l95field), pointer :: fldin
type(l95field), pointer :: fldout
type(l95_b_config), pointer :: config

call c_f_pointer(c_fldin, fldin)
call c_f_pointer(c_fldout,fldout)
call c_f_pointer(c_config, config)

kk = fldout%resol
allocate(vect(kk))

vect(:) = 0.0_kind_real
call c_l95bkg_sqrt_mult_ad(c_fldin,kk,vect,c_config)
call c_l95bkg_sqrt_mult(c_fldout,kk,vect,c_config)

deallocate(vect)

return
end subroutine c_l95bkg_mult

! ------------------------------------------------------------------------------

subroutine c_l95bkg_inv_mult(c_fldin,c_fldout,c_config) &
         & bind(c,name='l95bkg_inv_mult_f90')

implicit none
type(c_ptr), intent(in)    :: c_fldin
type(c_ptr), intent(inout) :: c_fldout
type(c_ptr), intent(in)    :: c_config

integer :: kk
real(c_double), allocatable :: vect(:)
type(l95field), pointer :: fldin
type(l95field), pointer :: fldout
type(l95_b_config), pointer :: config

call c_f_pointer(c_fldin, fldin)
call c_f_pointer(c_fldout,fldout)
call c_f_pointer(c_config, config)

kk = fldout%resol
allocate(vect(kk))

call c_l95bkg_sqrt_inv_mult(kk,vect,c_fldin,c_config)
fldout%x(:) = 0.0_kind_real
call c_l95bkg_sqrt_inv_mult_ad(kk,vect,c_fldout,c_config)

deallocate(vect)

return
end subroutine c_l95bkg_inv_mult

! ------------------------------------------------------------------------------

subroutine c_l95bkg_randomize(c_config, c_fld) bind(c,name='l95bkg_randomize_f90')

implicit none
type(c_ptr), intent(in)    :: c_config
type(c_ptr), intent(inout) :: c_fld

integer :: kk
real(c_double), allocatable :: vect(:)
type(l95field), pointer :: fld
call c_f_pointer(c_fld,fld)

kk = fld%resol
allocate(vect(kk))
call random_vector_gauss(vect(:))
call c_l95bkg_sqrt_mult(c_fld,kk,vect,c_config)
deallocate(vect)

return
end subroutine c_l95bkg_randomize

! ------------------------------------------------------------------------------

subroutine c_l95bkg_sqrt_mult(c_fld,kk,vect,c_config) &
         & bind(c,name='l95bkg_sqrt_mult_f90')

implicit none
type(c_ptr), intent(inout) :: c_fld
integer(c_int), intent(in) :: kk
real(c_double), intent(in) :: vect(kk)
type(c_ptr), intent(in)    :: c_config

integer :: jj, iri, m
integer :: ifax(13)
real(kind=kind_real), allocatable :: trigs(:), xext(:), work(:)
type(l95field), pointer :: fld
type(l95_b_config), pointer :: config

call c_f_pointer(c_fld,fld)
call c_f_pointer(c_config, config)

!--- multiply by the square-root of the correlation matrix

allocate(trigs(3*kk/2+1))
allocate(xext(kk+2))
allocate(work(kk+2))

call set99 (trigs,ifax,kk)

xext(1:kk) = vect(:)
xext(kk+1:kk+2) = 0.0_kind_real

call fft991 (xext,work,trigs,ifax,1,kk+2,kk,1,-1)

do m=0,kk/2
  do iri=1,2
    xext(2*m+iri) = xext(2*m+iri) * config%sqrtb(m)
  enddo
enddo

call fft991 (xext,work,trigs,ifax,1,kk+2,kk,1,+1)
fld%x(:) = xext(1:kk)

deallocate(work)
deallocate(xext)
deallocate(trigs)

!--- multiply by the standard deviation

do jj=1,fld%resol
  fld%x(jj) = config%sigb * fld%x(jj)
enddo

return
end subroutine c_l95bkg_sqrt_mult

! ------------------------------------------------------------------------------

subroutine c_l95bkg_sqrt_mult_ad(c_fld,kk,vect,c_config) &
         & bind(c,name='l95bkg_sqrt_mult_ad_f90')

implicit none
type(c_ptr), intent(in)       :: c_fld
integer(c_int), intent(in)    :: kk
real(c_double), intent(inout) :: vect(kk)
type(c_ptr), intent(in)    :: c_config

integer :: jj, iri, m
integer :: ifax(13)
real(kind=kind_real), allocatable :: trigs(:), xext(:), work(:)
type(l95field), pointer :: fld
type(l95_b_config), pointer :: config

call c_f_pointer(c_fld,fld)
call c_f_pointer(c_config, config)

!--- multiply by the standard deviation

allocate(xext(kk+2))

do jj=1,fld%resol
  xext(jj) = config%sigb * fld%x(jj)
enddo

!--- multiply by the square-root of the correlation matrix

allocate(trigs(3*kk/2+1))
allocate(work(kk+2))

call set99 (trigs,ifax,kk)

xext(kk+1:kk+2) = 0.0_kind_real

call fft991 (xext,work,trigs,ifax,1,kk+2,kk,1,-1)

do m=0,kk/2
  do iri=1,2
    xext(2*m+iri) = xext(2*m+iri) * config%sqrtb(m)
  enddo
enddo

call fft991 (xext,work,trigs,ifax,1,kk+2,kk,1,+1)
vect(:) = vect(:) + xext(1:kk)

deallocate(work)
deallocate(xext)
deallocate(trigs)

return
end subroutine c_l95bkg_sqrt_mult_ad

! ------------------------------------------------------------------------------

subroutine c_l95bkg_sqrt_inv_mult(kk,vect,c_fld,c_config) &
         & bind(c,name='l95bkg_sqrt_inv_mult_f90')

implicit none
integer(c_int), intent(in)    :: kk
real(c_double), intent(inout) :: vect(kk)
type(c_ptr), intent(in)       :: c_fld
type(c_ptr), intent(in)    :: c_config

integer :: jj, iri, m
integer :: ifax(13)
real(kind=kind_real), allocatable :: trigs(:), xext(:), work(:)
type(l95field), pointer :: fld
type(l95_b_config), pointer :: config

call c_f_pointer(c_fld,fld)
call c_f_pointer(c_config, config)

!--- divide by the standard deviation

do jj=1,fld%resol
  vect(jj) = fld%x(jj) / config%sigb
enddo

!--- multiply by the square-root of the inverse correlation matrix

allocate(trigs(3*kk/2+1))
allocate(xext(kk+2))
allocate(work(kk+2))

call set99 (trigs,ifax,kk)

xext(1:kk) = vect(:)
xext(kk+1:kk+2) = 0.0_kind_real

call fft991 (xext,work,trigs,ifax,1,kk+2,kk,1,-1)

do m=0,kk/2
  do iri=1,2
    if (config%sqrtb(m) == 0.0_kind_real) then
      xext(2*m+iri) = 0.0_kind_real
    else
      xext(2*m+iri) = xext(2*m+iri) / config%sqrtb(m)
    endif
  enddo
enddo

call fft991 (xext,work,trigs,ifax,1,kk+2,kk,1,+1)
vect(:) = xext(1:kk)

deallocate(work)
deallocate(xext)
deallocate(trigs)

return
end subroutine c_l95bkg_sqrt_inv_mult

! ------------------------------------------------------------------------------

subroutine c_l95bkg_sqrt_inv_mult_ad(kk,vect,c_fld,c_config) &
         & bind(c,name='l95bkg_sqrt_inv_mult_ad_f90')

implicit none
integer(c_int), intent(in) :: kk
real(c_double), intent(in) :: vect(kk)
type(c_ptr), intent(inout) :: c_fld
type(c_ptr), intent(in)    :: c_config

integer :: jj, iri, m
integer :: ifax(13)
real(kind=kind_real), allocatable :: trigs(:), xext(:), work(:)
type(l95field), pointer :: fld
type(l95_b_config), pointer :: config

call c_f_pointer(c_fld,fld)
call c_f_pointer(c_config, config)

!--- multiply by the square-root of the inverse correlation matrix

allocate(trigs(3*kk/2+1))
allocate(xext(kk+2))
allocate(work(kk+2))

call set99 (trigs,ifax,kk)

xext(1:kk) = vect(:)
xext(kk+1:kk+2) = 0.0_kind_real

call fft991 (xext,work,trigs,ifax,1,kk+2,kk,1,-1)

do m=0,kk/2
  do iri=1,2
    if (config%sqrtb(m) == 0.0_kind_real) then
      xext(2*m+iri) = 0.0_kind_real
    else
      xext(2*m+iri) = xext(2*m+iri) / config%sqrtb(m)
    endif
  enddo
enddo

call fft991 (xext,work,trigs,ifax,1,kk+2,kk,1,+1)

deallocate(work)
deallocate(trigs)

!--- divide by the standard deviation

do jj=1,fld%resol
  fld%x(jj) = fld%x(jj) + xext(jj) / config%sigb
enddo

deallocate(xext)

return
end subroutine c_l95bkg_sqrt_inv_mult_ad

! ------------------------------------------------------------------------------

end module L95BKGERR
