!
! Copyright 2011 ECMWF
! 
! This software was developed at ECMWF for evaluation
! and may be used for academic and research purposes only.
! The software is provided as is without any warranty.
! 
! This software can be used, copied and modified but not
! redistributed or sold. This notice must be reproduced
! on each copy made.
!

!> Fortran module handling L95 model fields

module l95fields

use iso_c_binding
use string_f_c_mod
use random_vectors_mod
use kinds

implicit none
private
public :: l95field

! ------------------------------------------------------------------------------

!> Fortran derived type to represent a lorenz95::L95Field
type :: l95field
  integer           :: resol ! resolution
  real(kind=kind_real), allocatable :: x(:)  ! 1D field
end type l95field

! ------------------------------------------------------------------------------
contains
! ------------------------------------------------------------------------------

subroutine c_l95_setup(c_self, c_resol) bind(c,name='l95field_create_f90')

implicit none
type(c_ptr), intent(inout) :: c_self
integer(c_int), intent(in) :: c_resol

type(l95field), pointer :: self

allocate(self)
self%resol = c_resol
allocate(self%x(self%resol))
c_self = c_loc(self)

return
end subroutine c_l95_setup

! ------------------------------------------------------------------------------

subroutine self_diff(c_self,c_f1,c_f2) bind(c,name='l95field_diff_f90')
implicit none
type(c_ptr), intent(inout) :: c_self
type(c_ptr), intent(in)    :: c_f1
type(c_ptr), intent(in)    :: c_f2

type(l95field), pointer :: self
type(l95field), pointer :: f1
type(l95field), pointer :: f2
call c_f_pointer(c_self,self)
call c_f_pointer(c_f1  ,f1)
call c_f_pointer(c_f2  ,f2)

if (self%resol/=f1%resol) call abor1_ftn("l95fields:diff: error resol")
self%x(:) = f1%x(:) - f2%x(:)

return
end subroutine self_diff

! ------------------------------------------------------------------------------

subroutine c_l95_delete(c_self) bind(c,name='l95field_delete_f90')

implicit none
type(c_ptr), intent(inout) :: c_self

type(l95field), pointer :: self
call c_f_pointer(c_self,self)
if (allocated(self%x)) deallocate(self%x)
deallocate(self)

return
end subroutine c_l95_delete

! ------------------------------------------------------------------------------

subroutine copy(c_self,c_other) bind(c,name='l95field_copy_f90')
implicit none
type(c_ptr), intent(inout) :: c_self
type(c_ptr), intent(in)    :: c_other

type(l95field), pointer :: self, other

call c_f_pointer(c_other,other)
if (c_associated(c_self)) then
  call c_f_pointer(c_self,self)
  if (self%resol /= other%resol) then
    deallocate(self%x)
    self%resol = other%resol
    allocate(self%x(self%resol))
  endif
else
  allocate(self)
  self%resol = other%resol
  allocate(self%x(self%resol))
endif
self%x(:)=other%x(:)

c_self = c_loc(self)

return
end subroutine copy

! ------------------------------------------------------------------------------

subroutine self_zero(c_self) bind(c,name='l95field_zero_f90')
implicit none
type(c_ptr), intent(inout) :: c_self

type(l95field), pointer :: self
call c_f_pointer(c_self,self)
self%x(:) = 0.0_kind_real

return
end subroutine self_zero

! ------------------------------------------------------------------------------

subroutine self_assign(c_self,c_rhs) bind(c,name='l95field_assign_f90')
implicit none
type(c_ptr), intent(inout) :: c_self
type(c_ptr), intent(in)    :: c_rhs

type(l95field), pointer :: self
type(l95field), pointer :: rhs
call c_f_pointer(c_self,self)
call c_f_pointer(c_rhs ,rhs)

if (self%resol/=rhs%resol) call abor1_ftn("l95fields:assign: error resol")
self%x(:) = rhs%x(:)

return
end subroutine self_assign

! ------------------------------------------------------------------------------

subroutine self_add(c_self,c_rhs) bind(c,name='l95field_add_f90')
implicit none
type(c_ptr), intent(inout) :: c_self
type(c_ptr), intent(in)    :: c_rhs

type(l95field), pointer :: self
type(l95field), pointer :: rhs
call c_f_pointer(c_self,self)
call c_f_pointer(c_rhs ,rhs)

if (self%resol/=rhs%resol) call abor1_ftn("l95fields:add: error resol")
self%x(:) = self%x(:) + rhs%x(:)

return
end subroutine self_add

! ------------------------------------------------------------------------------

subroutine diff_incr(c_lhs,c_x1,c_x2) bind(c,name='l95field_diff_incr_f90')
implicit none
type(c_ptr), intent(inout) :: c_lhs
type(c_ptr), intent(in)    :: c_x1
type(c_ptr), intent(in)    :: c_x2

type(l95field), pointer :: lhs
type(l95field), pointer :: x1
type(l95field), pointer :: x2
call c_f_pointer(c_lhs,lhs)
call c_f_pointer(c_x1 ,x1)
call c_f_pointer(c_x2 ,x2)

if (x1%resol/=x2%resol) call abor1_ftn("l95fields:add: error resol 1")
if (lhs%resol/=x1%resol) call abor1_ftn("l95fields:add: error resol 2")
lhs%x(:) = x1%x(:) - x2%x(:)

return
end subroutine diff_incr

! ------------------------------------------------------------------------------

subroutine self_sub(c_self,c_rhs) bind(c,name='l95field_sub_f90')
implicit none
type(c_ptr), intent(inout) :: c_self
type(c_ptr), intent(in)    :: c_rhs

type(l95field), pointer :: self
type(l95field), pointer :: rhs
call c_f_pointer(c_self,self)
call c_f_pointer(c_rhs ,rhs)

if (self%resol/=rhs%resol) call abor1_ftn("l95fields:sub: error resol")
self%x(:) = self%x(:) - rhs%x(:)

return
end subroutine self_sub

! ------------------------------------------------------------------------------

subroutine self_mul(c_self,zz) bind(c,name='l95field_mul_f90')
implicit none
type(c_ptr), intent(inout) :: c_self
real(c_double), intent(in) :: zz

type(l95field), pointer :: self
call c_f_pointer(c_self,self)

self%x(:) = zz * self%x(:)

return
end subroutine self_mul

! ------------------------------------------------------------------------------

subroutine self_axpy(c_self,zz,c_rhs) bind(c,name='l95field_axpy_f90')
implicit none
type(c_ptr), intent(inout) :: c_self
real(c_double), intent(in) :: zz
type(c_ptr), intent(in)    :: c_rhs

type(l95field), pointer :: self
type(l95field), pointer :: rhs
call c_f_pointer(c_self,self)
call c_f_pointer(c_rhs ,rhs)

if (self%resol/=rhs%resol) call abor1_ftn("l95fields:axpy: error resol")
self%x(:) = self%x(:) + zz * rhs%x(:)

return
end subroutine self_axpy

! ------------------------------------------------------------------------------

subroutine self_schur(c_self, c_rhs) bind(c,name='l95field_schur_f90')
implicit none
type(c_ptr), intent(inout) :: c_self
type(c_ptr), intent(in)    :: c_rhs

type(l95field), pointer :: self
type(l95field), pointer :: rhs
call c_f_pointer(c_self,self)
call c_f_pointer(c_rhs ,rhs)

if (self%resol/=rhs%resol) call abor1_ftn("l95fields:axpy: error resol")
self%x(:) = self%x(:)* rhs%x(:)

return
end subroutine self_schur

! ------------------------------------------------------------------------------

subroutine field_prod(c_fld1,c_fld2,zprod) bind(c,name='l95field_prod_f90')
implicit none
type(c_ptr), intent(in)       :: c_fld1, c_fld2
real(c_double), intent(inout) :: zprod

integer :: jj
type(l95field), pointer :: fld1, fld2
call c_f_pointer(c_fld1,fld1)
call c_f_pointer(c_fld2,fld2)

if (fld1%resol/=fld2%resol) call abor1_ftn("l95fields:dot_prod: error resol")

zprod=0.0_kind_real
do jj=1,fld1%resol
  zprod = zprod + fld1%x(jj) * fld2%x(jj)
enddo

return
end subroutine field_prod

! ------------------------------------------------------------------------------

subroutine field_resol(c_self,c_rhs) bind(c,name='l95field_resol_f90')
implicit none
type(c_ptr), intent(inout) :: c_self
type(c_ptr), intent(in)    :: c_rhs
type(l95field), pointer :: self, rhs
real(kind=kind_real) :: xx,zz
integer :: jj,ii
call c_f_pointer(c_self,self)
call c_f_pointer(c_rhs,rhs)

if (rhs%resol==self%resol) then
  self%x(:)=rhs%x(:)
else
  zz=real(rhs%resol,kind_real)/real(self%resol,kind_real)
  do jj=1,self%resol
    xx=real(jj-1,kind_real)*zz
    ii=int(xx+0.5_kind_real)+1
    self%x(jj)=rhs%x(ii)
  enddo
endif

return
end subroutine field_resol

! ------------------------------------------------------------------------------

subroutine field_read(klen,c_file,c_self,c_time) bind(c,name='l95field_read_f90')
use c_logger_mod, only : c_logger, logger_trace
implicit none
type(c_ptr), intent(inout) :: c_self
integer(c_int), intent(in) :: klen
character(kind=c_char,len=1), intent(in)    :: c_file(klen)
character(kind=c_char,len=1), intent(inout) :: c_time(21)
character(len=klen-1) :: fname
character(len=20) :: ftime
integer :: iin, ires

type(l95field), pointer :: self
call c_f_pointer(c_self,self)
call c_f_string (c_file,fname)

call c_logger(logger_trace,'l95field_read: filename='//fname)

iin=99
open(unit=iin, file=fname, form='formatted')

read(iin,*)ires
if (self%resol/=ires) then
  self%resol=ires
  if (allocated(self%x)) deallocate(self%x)
  allocate(self%x(self%resol))
endif

read(iin,*)ftime
call f_c_string(ftime,c_time)

read(iin,*)self%x(:)

close(iin)

return
end subroutine field_read

! ------------------------------------------------------------------------------

subroutine field_write(klen,c_file,c_self,c_time) bind(c,name='l95field_write_f90')

use c_logger_mod, only : c_logger, logger_info
implicit none
type(c_ptr), intent(in)    :: c_self
integer(c_int), intent(in) :: klen
character(kind=c_char,len=1), intent(in) :: c_file(klen)
character(kind=c_char,len=1), intent(in) :: c_time(21)
character(len=klen-1) :: fname
character(len=20) :: ftime
integer :: iout
integer :: j
character(len=80) :: record

type(l95field), pointer :: self
call c_f_pointer(c_self,self)

call c_f_string (c_file,fname)
call c_f_string (c_time,ftime)
call c_logger(logger_info,'l95field_write: filename='//fname//', time='//ftime)

do j=1,size(self%x)
  write(record,*) self%x(j)
  call c_logger(logger_info,trim(record))
enddo

iout=99
open(unit=iout, file=fname, form='formatted')
write(iout,*)self%resol
write(iout,*)ftime
write(iout,*)self%x(:)
close(iout)

return
end subroutine field_write

! ------------------------------------------------------------------------------

subroutine field_norm(c_self,pmin,pmax,pavg) bind(c,name='l95field_norm_f90')
implicit none
type(c_ptr), intent(in)     :: c_self
real(c_double), intent(inout) :: pmin,pmax,pavg

type(l95field), pointer :: self
call c_f_pointer(c_self,self)

if (.not.allocated(self%x)) call abor1_ftn("l95fields:field_norm: not allocated")
if (self%resol<1) call abor1_ftn("l95fields:field_norm: error resol")
pmin=minval(self%x(:))
pmax=maxval(self%x(:))
pavg=sum   (self%x(:))/real(self%resol,kind_real)

return
end subroutine field_norm

! ------------------------------------------------------------------------------

subroutine field_interpol_tl(c_self, npts, locs, ngom, gom, indx) &
         & bind(c,name='l95field_interpol_tl_f90')
implicit none
type(c_ptr), intent(in)       :: c_self
integer(c_int), intent(in)    :: npts
real(c_double), intent(in)    :: locs(npts)
integer(c_int), intent(in)    :: ngom
real(c_double), intent(inout) :: gom(ngom)
integer(c_int), intent(inout) :: indx

real(kind=kind_real) :: dx
integer :: jj,ii
type(l95field), pointer :: self
call c_f_pointer(c_self,self)

dx=1.0_kind_real/real(self%resol,kind_real)
do jj=1,npts
  if (locs(jj)<0.0_kind_real .or. locs(jj)>1.0_kind_real) then
    call abor1_ftn("l95fields: error location")
  endif
  ii=int(locs(jj)/dx+0.5_kind_real)+1
  if (ii<1 .or. ii>self%resol) call abor1_ftn("l95fields: error location")
  indx=indx+1
  gom(indx)=self%x(ii)
enddo

return
end subroutine field_interpol_tl

! ------------------------------------------------------------------------------

subroutine field_interpol_ad(c_self, npts, locs, ngom, gom, indx) &
         & bind(c,name='l95field_interpol_ad_f90')
implicit none
type(c_ptr), intent(inout)    :: c_self
integer(c_int), intent(in)    :: npts
real(c_double), intent(in)    :: locs(npts)
integer(c_int), intent(in)    :: ngom
real(c_double), intent(inout) :: gom(ngom)
integer(c_int), intent(inout) :: indx

real(kind=kind_real) :: dx
integer :: jj,ii
type(l95field), pointer :: self
call c_f_pointer(c_self,self)

dx=1.0_kind_real/real(self%resol,kind_real)
do jj=npts,1,-1
  ii=int(locs(jj)/dx+0.5_kind_real)+1
  indx=indx+1
  self%x(ii)=self%x(ii)+gom(ngom-indx+1)
enddo

return
end subroutine field_interpol_ad

! ------------------------------------------------------------------------------

subroutine c_l95_random(c_self) bind(c,name='l95field_random_f90')

implicit none
type(c_ptr), intent(inout) :: c_self
type(l95field), pointer :: self
call c_f_pointer(c_self,self)

call random_vector(self%x(:))

return
end subroutine c_l95_random

! ------------------------------------------------------------------------------

end module l95fields
