!
! Copyright 2011 ECMWF
! 
! This software was developed at ECMWF for evaluation
! and may be used for academic and research purposes only.
! The software is provided as is without any warranty.
! 
! This software can be used, copied and modified but not
! redistributed or sold. This notice must be reproduced
! on each copy made.
!

!> Fortran module for the L95 model
module LORENZ95

use iso_c_binding
use c_config_mod
use l95fields
use kinds

implicit none
private

! ------------------------------------------------------------------------------

!> Configuration data for the L95 model
type :: l95_config
  real(kind=kind_real) :: f
  real(kind=kind_real) :: dt       ! non-dimensional timestep
  real(kind=kind_real) :: dt0      ! dimensional timestep (seconds)
  integer :: resol ! resolution
end type l95_config

! ------------------------------------------------------------------------------

!> Trajectory data for the L95 model
type :: l95_trajectory
  real(kind=kind_real), allocatable :: xtraj(:,:)
end type l95_trajectory

! ------------------------------------------------------------------------------
contains
! ------------------------------------------------------------------------------

subroutine c_l95_setup(c_config, c_resol, c_model) bind(c,name='l95_setup_f90')

use duration_mod
use c_logger_mod

implicit none
type(c_ptr), intent(in)    :: c_config
type(c_ptr), intent(in)    :: c_resol
type(c_ptr), intent(inout) :: c_model

type(l95_config), pointer :: config
type(duration) :: dtstep
character(len=20) :: ststep
character(len=160) :: record

allocate(config)

config%f = c_config_get_real(c_config,"f")
config%resol = c_config_get_int(c_resol,"resol")

ststep = c_config_get_data(c_config,len(ststep),"tstep")
write(record,*)'ststep=',ststep
call c_logger (logger_info,trim(record))
call duration_create(trim(ststep), dtstep)
config%dt0 = duration_seconds(dtstep)
config%dt = config%dt0/432000.0_kind_real
write(record,*)'dt0=',config%dt0,' dt=',config%dt
call c_logger (logger_info,trim(record))
call duration_delete(dtstep)

write(record,*)'c_l95_nl_setup: f=',config%f,', dt0=',config%dt0, &
             & ' dt=',config%dt,' resol=',config%resol
call c_logger (logger_info,trim(record))

c_model = c_loc(config)

return
end subroutine c_l95_setup

! ------------------------------------------------------------------------------

subroutine c_l95_delete(c_config) bind(c,name='l95_delete_f90')
use c_logger_mod, only : c_logger, logger_trace

implicit none
type(c_ptr), intent(inout) :: c_config

type(l95_config), pointer :: config

call c_f_pointer(c_config,config)
deallocate(config)
call c_logger(logger_trace,'l95_delete_f90: done')

return
end subroutine c_l95_delete

! ------------------------------------------------------------------------------

subroutine c_l95_propagate(c_config,c_fld) bind(c,name='l95_propagate_f90')

implicit none
type(c_ptr), intent(in)    :: c_config
type(c_ptr), intent(inout) :: c_fld

type(l95_config), pointer :: config
type(l95field), pointer :: fld

call c_f_pointer(c_config,config)
call c_f_pointer(c_fld,fld)
if (fld%resol/=config%resol) then
  call abor1_ftn ('c_l95_propagate: Resolution Error')
endif

call dopropagate(config, fld)

return
end subroutine c_l95_propagate

! ------------------------------------------------------------------------------

subroutine c_l95_create_traj(c_config,c_fld,c_traj) bind(c,name='l95_create_traj_f90')

implicit none
type(c_ptr), intent(in)    :: c_config
type(c_ptr), intent(inout) :: c_fld
type(c_ptr), intent(inout) :: c_traj

type(l95_config), pointer :: config
type(l95field), pointer :: fld
type(l95_trajectory), pointer :: traj

call c_f_pointer(c_config,config)
call c_f_pointer(c_fld,fld)
if (fld%resol/=config%resol) then
  call abor1_ftn ('c_l95_create_traj: Resolution Error')
endif

allocate(traj)
allocate(traj%xtraj(fld%resol,4))

call dopropagate(config, fld, traj)

c_traj = c_loc(traj)

return
end subroutine c_l95_create_traj

! ------------------------------------------------------------------------------

subroutine dopropagate(config,fld,traj)

implicit none

type(l95_config), intent(inout) :: config
type(l95field), intent(inout) :: fld
type(l95_trajectory), optional, intent(inout) :: traj

real(kind=kind_real) :: zt
real(kind=kind_real), allocatable :: xx(:),zz(:),qq(:)
logical :: lltraj

lltraj=present(traj)

allocate(xx(fld%resol),zz(fld%resol),qq(fld%resol))

zz(:)=fld%x(:)
if (lltraj) traj%xtraj(:,1)=zz(:)
call l95rhs(fld%resol,zz,config,qq)
xx(:)=qq(:)

zz(:)=fld%x(:)+0.5_kind_real*qq(:)
if (lltraj) traj%xtraj(:,2)=zz(:)
call l95rhs(fld%resol,zz,config,qq)
xx(:)=xx(:)+2.0_kind_real*qq(:)

zz(:)=fld%x(:)+0.5_kind_real*qq(:)
if (lltraj) traj%xtraj(:,3)=zz(:)
call l95rhs(fld%resol,zz,config,qq)
xx(:)=xx(:)+2.0_kind_real*qq(:)

zz(:)=fld%x(:)+qq(:)
if (lltraj) traj%xtraj(:,4)=zz(:)
call l95rhs(fld%resol,zz,config,qq)
xx(:)=xx(:)+qq(:)

zt=1.0_kind_real/6.0_kind_real
fld%x(:)=fld%x(:)+zt*xx(:)

deallocate(xx,zz,qq)

return
end subroutine dopropagate

! ------------------------------------------------------------------------------

subroutine c_l95_propagate_tl(c_config,c_fld,c_traj) bind(c,name='l95_propagate_tl_f90')

implicit none
type(c_ptr), intent(in)    :: c_config
type(c_ptr), intent(inout) :: c_fld
type(c_ptr), intent(in)    :: c_traj

type(l95field), pointer :: fld
type(l95_config), pointer :: config
type(l95_trajectory), pointer :: traj
real(kind=kind_real), allocatable :: xx(:),zz(:),qq(:)
real(kind=kind_real) :: zt

call c_f_pointer(c_config,config)
call c_f_pointer(c_traj,traj)
call c_f_pointer(c_fld,fld)
if (fld%resol/=config%resol) then
  call abor1_ftn('c_l95_propagate_tl: Resolution Error')
endif
allocate(xx(fld%resol),zz(fld%resol),qq(fld%resol))

zz(:)=fld%x(:)
call l95rhs_tl(fld%resol,zz,config,qq,traj%xtraj(:,1))
xx(:)=qq(:)

zz(:)=fld%x(:)+0.5_kind_real*qq(:)
call l95rhs_tl(fld%resol,zz,config,qq,traj%xtraj(:,2))
xx(:)=xx(:)+2.0_kind_real*qq(:)

zz(:)=fld%x(:)+0.5_kind_real*qq(:)
call l95rhs_tl(fld%resol,zz,config,qq,traj%xtraj(:,3))
xx(:)=xx(:)+2.0_kind_real*qq(:)

zz(:)=fld%x(:)+qq(:)
call l95rhs_tl(fld%resol,zz,config,qq,traj%xtraj(:,4))
xx(:)=xx(:)+qq(:)

zt=1.0_kind_real/6.0_kind_real
fld%x(:)=fld%x(:)+zt*xx(:)

deallocate(xx,zz,qq)
return
end subroutine c_l95_propagate_tl

! ------------------------------------------------------------------------------

subroutine c_l95_propagate_ad(c_config,c_fld,c_traj) bind(c,name='l95_propagate_ad_f90')

implicit none
type(c_ptr), intent(in)    :: c_config
type(c_ptr), intent(inout) :: c_fld
type(c_ptr), intent(in)    :: c_traj

type(l95field), pointer :: fld
type(l95_config), pointer :: config
type(l95_trajectory), pointer :: traj
real(kind=kind_real), allocatable :: xx(:),zz(:),qq(:)
real(kind=kind_real) :: zt

call c_f_pointer(c_config,config)
call c_f_pointer(c_traj,traj)
call c_f_pointer(c_fld,fld)
if (fld%resol/=config%resol) then
  call abor1_ftn('c_l95_propagate_ad: Resolution Error')
endif
allocate(xx(fld%resol),zz(fld%resol),qq(fld%resol))

zt=1.0_kind_real/6.0_kind_real
xx(:)=zt*fld%x(:)

qq(:)=xx(:)
call l95rhs_ad(fld%resol,zz,config,qq,traj%xtraj(:,4))
fld%x(:)=fld%x(:)+zz(:)
qq(:)=zz(:)

qq(:)=qq(:)+2.0_kind_real*xx(:)
call l95rhs_ad(fld%resol,zz,config,qq,traj%xtraj(:,3))
fld%x(:)=fld%x(:)+zz(:)
qq(:)=0.5_kind_real*zz(:)

qq(:)=qq(:)+2.0_kind_real*xx(:)
call l95rhs_ad(fld%resol,zz,config,qq,traj%xtraj(:,2))
fld%x(:)=fld%x(:)+zz(:)
qq(:)=0.5_kind_real*zz(:)

qq(:)=qq(:)+xx(:)
call l95rhs_ad(fld%resol,zz,config,qq,traj%xtraj(:,1))
fld%x(:)=fld%x(:)+zz(:)

deallocate(xx,zz,qq)
return
end subroutine c_l95_propagate_ad

! ------------------------------------------------------------------------------

subroutine l95rhs(kn,x,config,dx)
implicit none
integer, intent(in) :: kn
real(kind=kind_real), intent(in)    :: x(kn)
type(l95_config), intent(in) :: config
real(kind=kind_real), intent(inout) :: dx(kn)
real(kind=kind_real) :: dxdt
integer :: jj,jm2,jm1,jp1

do jj=1,kn
  jm2=jj-2; if (jm2<1) jm2=jm2+kn
  jm1=jj-1; if (jm1<1) jm1=jm1+kn
  jp1=jj+1; if (jp1>kn) jp1=jp1-kn
  dxdt = -x(jm2)*x(jm1) + x(jm1)*x(jp1) - x(jj) + config%f
  dx(jj) = config%dt * dxdt
enddo

return
end subroutine l95rhs

! ------------------------------------------------------------------------------

subroutine l95rhs_tl(kn,x,config,dx,xtra)
implicit none
integer, intent(in) :: kn
real(kind=kind_real), intent(in)    :: x(kn)
type(l95_config), intent(in) :: config
real(kind=kind_real), intent(inout) :: dx(kn)
real(kind=kind_real), intent(in)    :: xtra(kn)
real(kind=kind_real) :: dxdt
integer :: jj,jm2,jm1,jp1

do jj=1,kn
  jm2=jj-2; if (jm2<1) jm2=jm2+kn
  jm1=jj-1; if (jm1<1) jm1=jm1+kn
  jp1=jj+1; if (jp1>kn) jp1=jp1-kn
  dxdt = - x(jm2)*xtra(jm1) - xtra(jm2)*x(jm1) &
       & + x(jm1)*xtra(jp1) + xtra(jm1)*x(jp1) - x(jj)
  dx(jj) = config%dt * dxdt
enddo

return
end subroutine l95rhs_tl

! ------------------------------------------------------------------------------

subroutine l95rhs_ad(kn,x,config,dx,xtra)
implicit none
integer, intent(in) :: kn
real(kind=kind_real), intent(inout) :: x(kn)
type(l95_config), intent(in) :: config
real(kind=kind_real), intent(inout) :: dx(kn)
real(kind=kind_real), intent(in)    :: xtra(kn)
real(kind=kind_real) :: dxdt
integer :: jj,jm2,jm1,jp1

x=0.0_kind_real
do jj=1,kn
  jm2=jj-2; if (jm2<1) jm2=jm2+kn
  jm1=jj-1; if (jm1<1) jm1=jm1+kn
  jp1=jj+1; if (jp1>kn) jp1=jp1-kn
  dxdt = config%dt * dx(jj)
  x(jm2) = x(jm2) - dxdt*xtra(jm1)
  x(jm1) = x(jm1) - dxdt*xtra(jm2)
  x(jm1) = x(jm1) + dxdt*xtra(jp1)
  x(jp1) = x(jp1) + dxdt*xtra(jm1)
  x(jj)  = x(jj) - dxdt
enddo
dx=0.0_kind_real

return
end subroutine l95rhs_ad

! ------------------------------------------------------------------------------

subroutine c_l95_delete_traj(c_traj) bind(c,name='l95_delete_traj_f90')

implicit none
type(c_ptr), intent(inout) :: c_traj
type(l95_trajectory), pointer :: traj

if (c_associated(c_traj)) then
  call c_f_pointer(c_traj,traj)
  if (associated(traj)) then
    if (allocated(traj%xtraj)) deallocate(traj%xtraj)
    deallocate(traj)
  endif
  c_traj = c_null_ptr
endif

return
end subroutine c_l95_delete_traj

! ------------------------------------------------------------------------------

end module LORENZ95
