/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include "model/ObsStreamQG.h"

#include <cmath>

#include "logger/Logger.h"
#include "oops/base/NullObsAux.h"
#include "oops/base/NullObsTraj.h"
#include "model/GomQG.h"
#include "model/LocQG.h"
#include "model/ObsHelpQG.h"
#include "model/ObsTrajQG.h"
#include "model/ObsVecQG.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/asser1.h"

// -----------------------------------------------------------------------------
namespace qg {
// -----------------------------------------------------------------------------

ObsStreamQG::ObsStreamQG(const util::Config & config) : obsname_("Stream") {
  LOG(Trace) << "ObsStreamQG start " << obsname_;

  const util::Config dataConfig(config, "ObsData");
  obsdb_ = ObsHelpQG::getHelper(dataConfig);

  F90help * fhelp = obsdb_->toFortran();
  const util::Config * configc = &config;
  qg_stream_setup_f90(&f90stream_, &configc, &fhelp);
  LOG(Trace) << "ObsStreamQG created " << obsname_;
}

// -----------------------------------------------------------------------------

ObsStreamQG::~ObsStreamQG() {
  qg_stream_delete_f90(&f90stream_);
  ObsHelpQG::clearHelper(obsdb_);
}

// -----------------------------------------------------------------------------

void ObsStreamQG::obsEquiv(const GomQG & gom, ObsVecQG & ovec,
                           const oops::NullObsAux &) const {
  F90ovec * fov = ovec.toFortran();
  const F90goms * fgom = gom.toFortran();
  qg_stream_equiv_f90(&fgom, &fov);
}

// -----------------------------------------------------------------------------

void ObsStreamQG::obsEquiv(const GomQG & gom, ObsVecQG & ovec,
                           const oops::NullObsAux &, ObsTrajQG &) const {
  F90ovec * fov = ovec.toFortran();
  const F90goms * fgom = gom.toFortran();
  qg_stream_equiv_f90(&fgom, &fov);
}

// -----------------------------------------------------------------------------

void ObsStreamQG::obsEquivTL(const GomQG & gom, ObsVecQG & ovec,
                             const oops::NullObsAux &, const ObsTrajQG &) const {
  F90ovec * fov = ovec.toFortran();
  const F90goms * fgom = gom.toFortran();
  qg_stream_equiv_tl_f90(&fgom, &fov);
}

// -----------------------------------------------------------------------------

void ObsStreamQG::obsEquivAD(GomQG & gom, const ObsVecQG & ovec,
                             oops::NullObsAux &, const ObsTrajQG &) const {
  const F90ovec * fov = ovec.toFortran();
  F90goms * fgom = gom.toFortran();
  qg_stream_equiv_ad_f90(&fgom, &fov);
}

// -----------------------------------------------------------------------------

GomQG * ObsStreamQG::getGOM(const util::DateTime & t1,
                            const util::DateTime & t2) const {
  const util::DateTime * p1 = &t1;
  const util::DateTime * p2 = &t2;
  GomQG * pgom = new GomQG();
  F90goms * fgom = pgom->toFortran();
  qg_stream_getgom_f90(&f90stream_, &p1, &p2, &fgom);
  return pgom;
}

// -----------------------------------------------------------------------------

LocQG * ObsStreamQG::locations(const util::DateTime & t1,
                               const util::DateTime & t2) const {
  const util::DateTime * p1 = &t1;
  const util::DateTime * p2 = &t2;
  F90locs * floc;
  qg_stream_locations_f90(&f90stream_, &p1, &p2, &floc);
  LocQG * ploc = new LocQG(floc);
  return ploc;
}

// -----------------------------------------------------------------------------

ObsVecQG * ObsStreamQG::read(const std::string & name) const {
  ObsVecQG * pvec = obsdb_->getdb(obsname_, name);
  return pvec;
}

// -----------------------------------------------------------------------------

void ObsStreamQG::save(const std::string & name, const ObsVecQG & ov) const {
  obsdb_->putdb(obsname_, name, ov);
}

// -----------------------------------------------------------------------------

ObsVecQG * ObsStreamQG::distribute(const util::Config & config,
                                   const util::DateTime & t1, const util::DateTime & t2) {
  const util::Config * configc = &config;
  const util::Duration first(config.getData("begin"));
  const util::DateTime start(t1 + first);
  const util::Duration freq(config.getData("obs_period"));
  int nobstimes = 0;
  util::DateTime now(start);
  while (now <= t2) {
    ++nobstimes;
    now += freq;
  }
  const util::DateTime * bgn = &start;
  const util::Duration * stp = &freq;
  F90ovec * fvec;
  qg_stream_generate_f90(&f90stream_, &configc, &bgn, &stp, nobstimes, &fvec);
  return new ObsVecQG(fvec);
}

// -----------------------------------------------------------------------------
}  // namespace qg
