/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include "model/ObsWSpeedQG.h"

#include <cmath>

#include "logger/Logger.h"
#include "oops/base/NullObsAux.h"
#include "oops/base/NullObsTraj.h"
#include "model/GomQG.h"
#include "model/LocQG.h"
#include "model/ObsHelpQG.h"
#include "model/ObsTrajQG.h"
#include "model/ObsVecQG.h"
#include "model/QgFortran.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/asser1.h"

// -----------------------------------------------------------------------------
namespace qg {
// -----------------------------------------------------------------------------

ObsWSpeedQG::ObsWSpeedQG(const util::Config & config) : obsname_("WSpeed") {
  LOG(Trace) << "ObsWSpeedQG start " << obsname_;

  const util::Config dataConfig(config, "ObsData");
  obsdb_ = ObsHelpQG::getHelper(dataConfig);

  F90help * fhelp = obsdb_->toFortran();
  const util::Config * configc = &config;
  qg_wspeed_setup_f90(&f90wspeed_, &configc, &fhelp);
  LOG(Trace) << "ObsWSpeedQG created " << obsname_;
}

// -----------------------------------------------------------------------------

ObsWSpeedQG::~ObsWSpeedQG() {
  qg_wspeed_delete_f90(&f90wspeed_);
  ObsHelpQG::clearHelper(obsdb_);
}

// -----------------------------------------------------------------------------

ObsTrajQG * ObsWSpeedQG::getTraj() const {
  ObsTrajQG * traj = new ObsTrajQG();
  F90goms * ftra = traj->toFortran();
  qg_wspeed_gettra_f90(&f90wspeed_, &ftra);
  LOG(Trace) << "ObsWSpeedQG ObsTrajQG created";
  return traj;
}

// -----------------------------------------------------------------------------

void ObsWSpeedQG::obsEquiv(const GomQG & gom, ObsVecQG & ovec,
                           const oops::NullObsAux &) const {
  F90ovec * fov = ovec.toFortran();
  const F90goms * fgom = gom.toFortran();
  qg_wspeed_eqv_f90(&fgom, &fov);
}

// -----------------------------------------------------------------------------

void ObsWSpeedQG::obsEquiv(const GomQG & gom, ObsVecQG & ovec,
                           const oops::NullObsAux &, ObsTrajQG & traj) const {
  const F90goms * fgom = gom.toFortran();
  F90ovec * fov  = ovec.toFortran();
  F90goms * ftra = traj.toFortran();
  qg_wspeed_tra_f90(&fgom, &fov, &ftra);
}

// -----------------------------------------------------------------------------

void ObsWSpeedQG::obsEquivTL(const GomQG & gom, ObsVecQG & ovec,
                             const oops::NullObsAux &,
                             const ObsTrajQG & traj) const {
  F90ovec * fov = ovec.toFortran();
  const F90goms * fgom = gom.toFortran();
  const F90goms * ftra = traj.toFortran();
  qg_wspeed_equiv_tl_f90(&fgom, &fov, &ftra);
}

// -----------------------------------------------------------------------------

void ObsWSpeedQG::obsEquivAD(GomQG & gom, const ObsVecQG & ovec,
                             oops::NullObsAux &, const ObsTrajQG & traj) const {
  F90goms * fgom = gom.toFortran();
  const F90ovec * fov  = ovec.toFortran();
  const F90goms * ftra = traj.toFortran();
  qg_wspeed_equiv_ad_f90(&fgom, &fov, &ftra);
}

// -----------------------------------------------------------------------------

GomQG * ObsWSpeedQG::getGOM(const util::DateTime & t1,
                            const util::DateTime & t2) const {
  const util::DateTime * p1 = &t1;
  const util::DateTime * p2 = &t2;
  GomQG * pgom = new GomQG();
  F90goms * fgom = pgom->toFortran();
  qg_wspeed_getgom_f90(&f90wspeed_, &p1, &p2, &fgom);
  return pgom;
}

// -----------------------------------------------------------------------------

LocQG * ObsWSpeedQG::locations(const util::DateTime & t1,
                               const util::DateTime & t2) const {
  const util::DateTime * p1 = &t1;
  const util::DateTime * p2 = &t2;
  F90locs * floc;
  qg_wspeed_locations_f90(&f90wspeed_, &p1, &p2, &floc);
  LocQG * ploc = new LocQG(floc);
  return ploc;
}

// -----------------------------------------------------------------------------

ObsVecQG * ObsWSpeedQG::read(const std::string & name) const {
  ObsVecQG * pvec = obsdb_->getdb(obsname_, name);
  return pvec;
}

// -----------------------------------------------------------------------------

void ObsWSpeedQG::save(const std::string & name, const ObsVecQG & ov) const {
  obsdb_->putdb(obsname_, name, ov);
}

// -----------------------------------------------------------------------------

ObsVecQG * ObsWSpeedQG::distribute(const util::Config & config,
                                 const util::DateTime & t1, const util::DateTime & t2) {
  const util::Config * configc = &config;
  const util::Duration first(config.getData("begin"));
  const util::DateTime start(t1 + first);
  const util::Duration freq(config.getData("obs_period"));
  int nobstimes = 0;
  util::DateTime now(start);
  while (now <= t2) {
    ++nobstimes;
    now += freq;
  }
  const util::DateTime * bgn = &start;
  const util::Duration * stp = &freq;
  F90ovec * fvec;
  qg_wspeed_generate_f90(&f90wspeed_, &configc, &bgn, &stp, nobstimes, &fvec);
  return new ObsVecQG(fvec);
}

// -----------------------------------------------------------------------------
}  // namespace qg
