/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include "model/QgErrorCovariance.h"

#include <cmath>

#include "logger/Logger.h"
#include "model/QgFields.h"
#include "model/QgFortran.h"
#include "model/QgGeometry.h"
#include "model/QgIncrement.h"
#include "model/QgModel.h"
#include "model/QgState.h"
#include "oops/base/Geometry.h"
#include "oops/base/Variables.h"
#include "util/Config.h"

// -----------------------------------------------------------------------------
namespace qg {
// -----------------------------------------------------------------------------

QgErrorCovariance::QgErrorCovariance(const oops::Geometry & resol,
                                     const util::Config & bkgErrCovarConfig) {
  time_ = util::DateTime(bkgErrCovarConfig.getData("date"));
  const util::Config * configc = &bkgErrCovarConfig;
  QgGeometry geom(resol);
  const F90geom * fgeo = geom.toFortran();

  qg_b_setup_f90(&ftnConfig_, &configc, &fgeo);

  util::Config ctlvars(bkgErrCovarConfig, util::ConfigEmpty());
  ctlvars.addElement("/", "variables");
  ctlvars.setElement("variables", "ci");
  ctrlvars_.reset(new oops::Variables(ctlvars));

  LOG(Trace) << "QgErrorCovariance created";
}

// -----------------------------------------------------------------------------

QgErrorCovariance::~QgErrorCovariance() {
  qg_b_delete_f90(&ftnConfig_);
}

// -----------------------------------------------------------------------------

void QgErrorCovariance::linearize(const QgState &, const oops::Geometry & resol,
                                  const util::Config & tlm) {
  model_.reset(new QgModel(resol, tlm));
  geom_.reset(new oops::Geometry(resol));
}

// -----------------------------------------------------------------------------

void QgErrorCovariance::mult(const QgIncrement & dxin,
                             QgIncrement & dxout) const {
  const F90flds * dx1 = dxin.getFields().toFortran();
  F90flds * dx2 = dxout.getFields().toFortran();
  qg_b_mult_f90(&ftnConfig_, &dx1, &dx2);
}

// -----------------------------------------------------------------------------

void QgErrorCovariance::inv_mult(const QgIncrement & dxin,
                                 QgIncrement & dxout) const {
  const F90flds * dx1 = dxin.getFields().toFortran();
  F90flds * dx2 = dxout.getFields().toFortran();
  qg_b_inv_mult_f90(&ftnConfig_, &dx1, &dx2);
}

// -----------------------------------------------------------------------------

void QgErrorCovariance::randomize(QgIncrement & dx) const {
  F90flds * fdx = dx.getFields().toFortran();
  qg_b_randomize_f90(&ftnConfig_, &fdx);
}

// -----------------------------------------------------------------------------

}  // namespace qg
