/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include "model/QgFields.h"

#include <cmath>
#include <map>
#include <string>

#include "model/GomQG.h"
#include "model/LocQG.h"
#include "model/QgFortran.h"
#include "model/QgGeometry.h"
#include "model/VarQG.h"
#include "util/Config.h"
#include "util/DateTime.h"

// -----------------------------------------------------------------------------
namespace qg {
// -----------------------------------------------------------------------------
QgFields::QgFields(const QgGeometry & geom, const VarQG & vars,
                   const util::DateTime & time):
  geom_(new QgGeometry(geom)), vars_(new VarQG(vars)), time_(time)
{
  const F90vars * fvars = vars_->toFortran();
  const F90geom * f90geo = geom_->toFortran();
  qg_field_create_f90(&flds_, &f90geo, &fvars);
}
// -----------------------------------------------------------------------------
QgFields::QgFields(const QgFields & other)
  : geom_(other.geom_), vars_(other.vars_), time_(other.time_)
{
  const F90geom * f90geo = geom_->toFortran();
  const F90vars * fvars = vars_->toFortran();
  qg_field_create_f90(&flds_, &f90geo, &fvars);
  qg_field_copy_f90(&flds_, &other.flds_);
}
// -----------------------------------------------------------------------------
QgFields::~QgFields() {
  qg_field_delete_f90(&flds_);
}
// -----------------------------------------------------------------------------
QgFields & QgFields::operator=(const QgFields & rhs) {
  qg_field_copy_f90(&flds_, &rhs.flds_);
  time_ = rhs.time_;
  return *this;
}
// -----------------------------------------------------------------------------
QgFields & QgFields::operator+=(const QgFields & rhs) {
  qg_field_self_add_f90(&flds_, &rhs.flds_);
  return *this;
}
// -----------------------------------------------------------------------------
QgFields & QgFields::operator-=(const QgFields & rhs) {
  qg_field_self_sub_f90(&flds_, &rhs.flds_);
  return *this;
}
// -----------------------------------------------------------------------------
QgFields & QgFields::operator*=(const double & zz) {
  qg_field_self_mul_f90(&flds_, zz);
  return *this;
}
// -----------------------------------------------------------------------------
void QgFields::zero(const util::Duration & shift) {
  qg_field_zero_f90(&flds_);
  time_ += shift;
}
// -----------------------------------------------------------------------------
void QgFields::axpy(const double & zz, const QgFields & rhs) {
  qg_field_axpy_f90(&flds_, zz, &rhs.flds_);
}
// -----------------------------------------------------------------------------
void QgFields::random() {
  qg_field_random_f90(&flds_);
}
// -----------------------------------------------------------------------------
double QgFields::dot_product_with(const QgFields & fld2) const {
  double zz;
  qg_field_dot_prod_f90(&flds_, &fld2.flds_, zz);
  return zz;
}

// -----------------------------------------------------------------------------
void QgFields::schur_product_with(const QgFields & dx) {
    qg_field_self_schur_f90(&flds_, &dx.flds_);
}
// -----------------------------------------------------------------------------
void QgFields::interpolate(const LocQG & locs, GomQG & gom) const {
  const F90flds * flds = flds_;
  const F90locs * floc = locs.toFortran();
        F90goms * fgom = gom.toFortran();
  qg_field_interp_tl_f90(&flds, &floc, &fgom);
}
// -----------------------------------------------------------------------------
void QgFields::interpolateTL(const LocQG & locs, GomQG & gom) const {
  const F90flds * flds = flds_;
  const F90locs * floc = locs.toFortran();
        F90goms * fgom = gom.toFortran();
  qg_field_interp_tl_f90(&flds, &floc, &fgom);
}
// -----------------------------------------------------------------------------
void QgFields::interpolateAD(const LocQG & locs, const GomQG & gom) {
  const F90locs * floc = locs.toFortran();
  const F90goms * fgom = gom.toFortran();
  qg_field_interp_ad_f90(&flds_, &floc, &fgom);
}
// -----------------------------------------------------------------------------
void QgFields::changeResolution(const QgFields & rhs) {
  qg_field_change_resol_f90(&flds_, &rhs.flds_);
  time_ = rhs.time_;
}
// -----------------------------------------------------------------------------
void QgFields::add(const QgFields & rhs) {
  qg_field_add_incr_f90(&flds_, &rhs.flds_);
}
// -----------------------------------------------------------------------------
void QgFields::diff(const QgFields & x1, const QgFields & x2) {
  qg_field_diff_incr_f90(&flds_, &x1.flds_, &x2.flds_);
}
// -----------------------------------------------------------------------------
void QgFields::read(const util::Config & config) {
  const util::Config * conf = &config;
  util::DateTime * dtp;  // The DateTime is created in the Fortran
  qg_field_read_file_f90(&flds_, &conf, &dtp);
  time_ = *dtp;
  delete dtp;
}
// -----------------------------------------------------------------------------
void QgFields::write(const util::Config & config) const {
  const util::Config * conf = &config;
  const F90flds * flds = flds_;
  const util::DateTime * dtp = new util::DateTime(time_);
  qg_field_write_file_f90(&flds, &conf, &dtp);
  delete dtp;
}
// -----------------------------------------------------------------------------
void QgFields::prtNorms(const std::string & sgrep) const {
  const F90flds * flds = flds_;
  qg_field_gpnorm_f90(&flds, sgrep.size(), sgrep.c_str() );
}
// -----------------------------------------------------------------------------
}  // namespace qg
