/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include "model/QgIncrement.h"

#include <string>

#include "logger/Logger.h"
#include "oops/base/NullModelAux.h"
#include "oops/base/Geometry.h"
#include "oops/base/Variables.h"
#include "model/GomQG.h"
#include "model/LocQG.h"
#include "model/QgErrorCovariance.h"
#include "model/QgFields.h"
#include "model/QgGeometry.h"
#include "model/QgModel.h"
#include "model/QgState.h"
#include "model/QgTrajectory.h"
#include "model/VarQG.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/abor1_cpp.h"
#include "util/asser1.h"

namespace qg {

// -----------------------------------------------------------------------------
/// Constructor, destructor
// -----------------------------------------------------------------------------
QgIncrement::QgIncrement(const oops::Geometry & resol, const oops::Variables & vars,
                         const util::DateTime & vt) {
  QgGeometry geom(resol);
  VarQG variables(vars);
  fields_.reset(new QgFields(geom, variables, vt));
  fields_->zero();
  LOG(Trace) << "QgIncrement constructed.";
}
// -----------------------------------------------------------------------------
QgIncrement::QgIncrement(const QgIncrement & dx)
  : fields_(new QgFields(*dx.fields_))
{
  LOG(Trace) << "QgIncrement copy-created.";
}
// -----------------------------------------------------------------------------
QgIncrement::~QgIncrement() {
  LOG(Trace) << "QgIncrement destructed";
}
// -----------------------------------------------------------------------------
/// Basic operators
// -----------------------------------------------------------------------------
void QgIncrement::diff(const QgState & x1, const QgState & x2) {
  ASSERT(this->validTime() == x1.validTime());
  ASSERT(this->validTime() == x2.validTime());
  x1.diff(x2, *fields_);
}
// -----------------------------------------------------------------------------
QgIncrement & QgIncrement::operator=(const QgIncrement & rhs) {
  *fields_ = *rhs.fields_;
  return *this;
}
// -----------------------------------------------------------------------------
QgIncrement & QgIncrement::operator+=(const QgIncrement & dx) {
  ASSERT(this->validTime() == dx.validTime());
  *fields_ += *dx.fields_;
  return *this;
}
// -----------------------------------------------------------------------------
QgIncrement & QgIncrement::operator-=(const QgIncrement & dx) {
  ASSERT(this->validTime() == dx.validTime());
  *fields_ -= *dx.fields_;
  return *this;
}
// -----------------------------------------------------------------------------
QgIncrement & QgIncrement::operator*=(const double & zz) {
  *fields_ *= zz;
  return *this;
}
// -----------------------------------------------------------------------------
void QgIncrement::zero(const util::Duration & shift) {
  fields_->zero(shift);
}
// -----------------------------------------------------------------------------
void QgIncrement::axpy(const double & zz, const QgIncrement & dx,
                       const bool check) {
  ASSERT(!check || this->validTime() == dx.validTime());
  fields_->axpy(zz, *dx.fields_);
}
// -----------------------------------------------------------------------------
void QgIncrement::accumul(const double & zz, const QgState & xx) {
  xx.xpay(zz, *fields_);
}

// -----------------------------------------------------------------------------
void QgIncrement::schur_product_with(const QgIncrement & dx) {
	fields_->schur_product_with(*dx.fields_);
}

// -----------------------------------------------------------------------------
void QgIncrement::random() {
  fields_->random();
}
// -----------------------------------------------------------------------------
double QgIncrement::dot_product_with(const QgIncrement & other) const {
  return dot_product(*fields_, *other.fields_);
}
// -----------------------------------------------------------------------------
/// Interactions with State
// -----------------------------------------------------------------------------
void QgIncrement::updateFields(QgFields & fld) const {
  fld.add(*fields_);  // Not += because only some fields in increment
}
// -----------------------------------------------------------------------------
/// Run forecast TL and AD
// -----------------------------------------------------------------------------
void QgIncrement::initTL(const QgModel & tlm, const QgTrajectory & traj) {
  tlm.initTL(*fields_, traj);
}
// -----------------------------------------------------------------------------
void QgIncrement::initAD(const QgModel & tlm, const QgTrajectory & traj) {
  tlm.initAD(*fields_, traj);
}
// -----------------------------------------------------------------------------
void QgIncrement::stepTL(const QgModel & tlm, const QgTrajectory & traj,
                         const oops::NullModelAux & merr) {
  tlm.stepTL(*fields_, merr, traj);
  fields_->timeUpdate(tlm.timestep());
}
// -----------------------------------------------------------------------------
void QgIncrement::stepAD(const QgModel & tlm, const QgTrajectory & traj,
                         oops::NullModelAux & merr) {
  fields_->timeUpdate(-tlm.timestep());
  tlm.stepAD(*fields_, merr, traj);
}
// -----------------------------------------------------------------------------
/// Interpolate to observation location
// -----------------------------------------------------------------------------
void QgIncrement::interpolateTL(const LocQG & locs, GomQG & cols) const {
  fields_->interpolateTL(locs, cols);
}
// -----------------------------------------------------------------------------
void QgIncrement::interpolateAD(const LocQG & locs, const GomQG & cols) {
  fields_->interpolateAD(locs, cols);
}
// -----------------------------------------------------------------------------
/// Interpolate full fields
// -----------------------------------------------------------------------------
void QgIncrement::changeResolution(const QgIncrement & rhs) {
  fields_->changeResolution(*rhs.fields_);
}
// -----------------------------------------------------------------------------
/// I/O and diagnostics
// -----------------------------------------------------------------------------
void QgIncrement::read(const util::Config & files) {
  fields_->read(files);
}
// -----------------------------------------------------------------------------
void QgIncrement::write(const util::Config & files) const {
  fields_->write(files);
}
// -----------------------------------------------------------------------------
void QgIncrement::prtNorms(const std::string & cgrep) const {
  LOG(Trace) << cgrep << " Valid time: " << validTime();
  fields_->prtNorms(cgrep);
}
// -----------------------------------------------------------------------------

}  // namespace oops
