/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef QG_MODEL_QGINCREMENT_H_
#define QG_MODEL_QGINCREMENT_H_

#include <string>

#include <boost/scoped_ptr.hpp>
#include <boost/shared_ptr.hpp>

#include "model/QgFields.h"
#include "model/QgModel.h"
#include "oops/base/GeneralizedDeparture.h"
#include "util/dot_product.h"
#include "util/DateTime.h"
#include "util/Duration.h"

namespace util {
  class Config;
  class NullModelAux;
}

namespace oops {
  class Geometry;
  class Variables;
}

namespace qg {
  class GomQG;
  class LocQG;
  class QgErrorCovariance;
  class QgState;
  class QgTrajectory;

/// Increment Class: Difference between two states
/*!
 *  Some fields that are present in a State may not be present in
 *  an Increment. The Increment contains everything that is needed by
 *  the tangent-linear and adjoint models, and has methods to propagate
 *  itself using these models.
 */

// -----------------------------------------------------------------------------

class QgIncrement : public oops::GeneralizedDeparture {
 public:

/// Constructor, destructor
  QgIncrement(const oops::Geometry &, const oops::Variables &, const util::DateTime &);
  QgIncrement(const QgIncrement &);
  virtual ~QgIncrement();

/// Basic operators
  void diff(const QgState &, const QgState &);
  QgIncrement & operator =(const QgIncrement &);
  QgIncrement & operator+=(const QgIncrement &);
  QgIncrement & operator-=(const QgIncrement &);
  QgIncrement & operator*=(const double &);
  void zero(const util::Duration & shift = util::Duration(0) );
  void random();
  void axpy(const double &, const QgIncrement &, const bool check = true);
  double dot_product_with(const QgIncrement &) const;
  void schur_product_with(const QgIncrement &);

/// Interpolate to observation location
  void interpolateTL(const LocQG &, GomQG &) const;
  void interpolateAD(const LocQG &, const GomQG &);

/// Interpolate full fields
  void changeResolution(const QgIncrement &);

/// Update State with Increment
  void updateFields(QgFields &) const;

/// I/O and diagnostics
  void read(const util::Config &);
  void write(const util::Config &) const;
  void prtNorms(const std::string &) const;
  const util::DateTime validTime() const {return fields_->time();}

/// Access to data... Could we do without that?
// Only used inside error covariance matrices
  QgFields & getFields() {return *fields_;}
  const QgFields & getFields() const {return *fields_;}

 protected:
  void initTL(const QgModel &, const QgTrajectory &);
  void initAD(const QgModel &, const QgTrajectory &);
  void stepTL(const QgModel &, const QgTrajectory &, const oops::NullModelAux &);
  void stepAD(const QgModel &, const QgTrajectory &, oops::NullModelAux &);

  void accumul(const double &, const QgState &);

/// Data
 private:
  boost::scoped_ptr<QgFields> fields_;
};

// -----------------------------------------------------------------------------
}  // namespace oops

#endif  // QG_MODEL_QGINCREMENT_H_
