/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include "model/QgModel.h"

#include "logger/Logger.h"
#include "oops/base/Geometry.h"
#include "oops/base/NullModelAux.h"
#include "model/QgFields.h"
#include "model/QgFortran.h"
#include "model/QgGeometry.h"
#include "model/QgTrajectory.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/asser1.h"

using oops::NullModelAux;

namespace qg {
// -----------------------------------------------------------------------------
QgModel::QgModel(const oops::Geometry & resol, const util::Config & model)
  : geom_(resol)
{
  LOG(Trace) << "QgModel::QgModel";
  tstep_ = util::Duration(model.getData("tstep"));
  LOG(Trace) << "QgModel: tstep is " << tstep_;

  QgGeometry geom(geom_);
  const F90geom * f90geo = geom.toFortran();

  const util::Config * configc = &model;
  qg_setup_f90(&configc, &f90geo, &qgConfig_);

  util::Config modelvars(model, util::ConfigEmpty());
  modelvars.addElement("/", "variables");
  modelvars.setElement("variables", "nl");
  vars_.reset(new oops::Variables(modelvars));

  util::Config tlvars(model, util::ConfigEmpty());
  tlvars.addElement("/", "variables");
  tlvars.setElement("./variables", "tl");
  varsTL_.reset(new oops::Variables(tlvars));

  LOG(Trace) << "QgModel created";
}
// -----------------------------------------------------------------------------
QgModel::~QgModel() {
  qg_delete_f90(&qgConfig_);
}
// -----------------------------------------------------------------------------
void QgModel::init(QgFields & xx) const {
  F90flds * state = xx.toFortran();
  qg_prepare_integration_f90(&qgConfig_, &state);
}
// -----------------------------------------------------------------------------
void QgModel::initTL(QgFields & dx, const QgTrajectory &) const {
  F90flds * incr = dx.toFortran();
  qg_prepare_integration_tl_f90(&qgConfig_, &incr);
}
// -----------------------------------------------------------------------------
void QgModel::initAD(QgFields & dx, const QgTrajectory &) const {
  F90flds * incr = dx.toFortran();
  qg_prepare_integration_ad_f90(&qgConfig_, &incr);
}
// -----------------------------------------------------------------------------
void QgModel::step(QgFields & xx, const NullModelAux &) const {
  F90flds * state = xx.toFortran();
  qg_propagate_f90(&qgConfig_, &state);
}
// -----------------------------------------------------------------------------
F90traj * QgModel::stepTraj(QgFields & xx) const {
  F90flds * state = xx.toFortran();
  F90traj * ftraj = 0;
  qg_prop_traj_f90(&qgConfig_, &state, &ftraj);
  ASSERT(ftraj != 0);
  return ftraj;
}
// -----------------------------------------------------------------------------
void QgModel::stepTL(QgFields & dx, const NullModelAux &,
                     const QgTrajectory & traj) const {
  F90flds * incr = dx.toFortran();
  const util::DateTime tt = dx.time();
  const F90traj * ftraj = traj.get(tt);
  qg_propagate_tl_f90(&qgConfig_, &incr, &ftraj);
}
// -----------------------------------------------------------------------------
void QgModel::stepAD(QgFields & dx, NullModelAux &, const QgTrajectory & traj) const {
  F90flds * incr = dx.toFortran();
  const util::DateTime tt = dx.time();
  const F90traj * ftraj = traj.get(tt);
  qg_propagate_ad_f90(&qgConfig_, &incr, &ftraj);
}
// -----------------------------------------------------------------------------
}  // namespace qg
