/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include "model/QgState.h"

#include <string>

#include "logger/Logger.h"
#include "oops/base/NullModelAux.h"
#include "oops/base/Geometry.h"
#include "oops/base/Variables.h"
#include "model/GomQG.h"
#include "model/LocQG.h"
#include "model/QgFields.h"
#include "model/QgIncrement.h"
#include "model/QgModel.h"
#include "model/QgTrajectory.h"
#include "model/VarQG.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/asser1.h"

namespace qg {

// -----------------------------------------------------------------------------
/// Constructor, destructor
// -----------------------------------------------------------------------------
QgState::QgState(const oops::Geometry & resol, const oops::Variables & vars,
                 const util::DateTime & vt)
{
  QgGeometry geom(resol);
  VarQG variables(vars);
  fields_.reset(new QgFields(geom, variables, vt));

  LOG(Trace) << "QgState::QgState created.";
}
// -----------------------------------------------------------------------------
QgState::QgState(const oops::Geometry & resol, const util::Config & file)
{
  QgGeometry geom(resol);
// Should get variables from file. YT
  util::Config modelvars(file, util::ConfigEmpty());
  modelvars.addElement("/", "variables");
  modelvars.setElement("variables", "cv");
  oops::Variables vars(modelvars);
  VarQG variables(vars);
// Should get variables from file. YT
  fields_.reset(new QgFields(geom, variables, util::DateTime()));
  fields_->read(file);

  LOG(Trace) << "QgState::QgState created and read in.";
}
// -----------------------------------------------------------------------------
QgState::QgState(const QgState & xx) : fields_(new QgFields(*xx.fields_)) {
  LOG(Trace) << "QgState::QgState copy-created.";
}
// -----------------------------------------------------------------------------
QgState::~QgState() {
  LOG(Trace) << "QgState::QgState destructed.";
}
// -----------------------------------------------------------------------------
/// Basic operators
// -----------------------------------------------------------------------------
QgState & QgState::operator=(const QgState & rhs) {
  *fields_ = *rhs.fields_;
  return *this;
}
// -----------------------------------------------------------------------------
/// Interpolate to observation location
// -----------------------------------------------------------------------------
void QgState::interpolate(const LocQG & locs, GomQG & cols) const {
  fields_->interpolate(locs, cols);
}
// -----------------------------------------------------------------------------
/// Interpolate full fields
// -----------------------------------------------------------------------------
void QgState::changeResolution(const QgState & rhs) {
  fields_->changeResolution(*rhs.fields_);
}
// -----------------------------------------------------------------------------
/// Interactions with Increments
// -----------------------------------------------------------------------------
void QgState::diff(const QgState & other, QgFields & fld) const {
  fld.diff(*fields_, *other.fields_);
}
// -----------------------------------------------------------------------------
QgState & QgState::operator+=(const QgIncrement & dx) {
  ASSERT(this->validTime() == dx.validTime());
  dx.updateFields(*fields_);
  return *this;
}
// -----------------------------------------------------------------------------
/// I/O and diagnostics
// -----------------------------------------------------------------------------
void QgState::read(const util::Config & files) {
  fields_->read(files);
}
// -----------------------------------------------------------------------------
void QgState::write(const util::Config & files) const {
  fields_->write(files);
}
// -----------------------------------------------------------------------------
void QgState::prtNorms(const std::string & cgrep) const {
  LOG(Trace) << cgrep << " Valid time: " << validTime();
  fields_->prtNorms(cgrep);
}
// -----------------------------------------------------------------------------
/// Run forecast TL and AD
// -----------------------------------------------------------------------------
void QgState::init(const QgModel & model) {
  model.init(*fields_);
}
// -----------------------------------------------------------------------------
void QgState::step(const QgModel & model, const oops::NullModelAux & merr) {
  model.step(*fields_, merr);
  fields_->timeUpdate(model.timestep());
}
// -----------------------------------------------------------------------------
F90traj * QgState::saveTrajectory(const QgModel & model) {
  F90traj * ftraj = model.stepTraj(*fields_);
  return ftraj;
}
// -----------------------------------------------------------------------------
/// For accumulator
// -----------------------------------------------------------------------------
void QgState::zero() {
  fields_->zero();
}
// -----------------------------------------------------------------------------
void QgState::accumul(const double & zz, const QgState & xx) {
  fields_->axpy(zz, *xx.fields_);
}
// -----------------------------------------------------------------------------
// Reverse axpy
void QgState::xpay(const double & zz, QgFields & xx) const {
  xx.axpy(zz, *fields_);
}
// -----------------------------------------------------------------------------

}  // namespace oops
