!
! Copyright 2011 ECMWF
! 
! This software was developed at ECMWF for evaluation
! and may be used for academic and research purposes only.
! The software is provided as is without any warranty.
! 
! This software can be used, copied and modified but not
! redistributed or sold. This notice must be reproduced
! on each copy made.
!

!> Handle observations for the QG model

module qg_obs_data

use datetime_mod
use qg_obs_vectors
use c_logger_mod, only : c_logger, logger_trace, logger_info, logger_error
use kinds

implicit none
private

public :: obs_data, obs_setup, obs_delete, obs_get, obs_put, &
        & obs_create, obs_time_get, obs_count, max_string

! ------------------------------------------------------------------------------
integer, parameter :: max_string=50
! ------------------------------------------------------------------------------

!> A type to represent observation data
type obs_data
  integer :: ngrp
  character(len=max_string) :: filein, fileout
  type(group_data), pointer :: grphead => null()
end type obs_data

! ------------------------------------------------------------------------------

!> A type to represent a linked list of observation group data
type group_data
  character(len=max_string) :: grpname
  type(group_data), pointer :: next => null()
  integer :: nobs
  integer, allocatable :: seqnos(:)
  type(datetime), allocatable :: times(:)
  type(column_data), pointer :: colhead => null()
end type group_data

! ------------------------------------------------------------------------------

!> A type to represent a linked list of observation columns
type column_data
  character(len=max_string) :: colname
  type(column_data), pointer :: next => null()
  integer :: ncol
  real(kind=kind_real), allocatable :: values(:,:)
end type column_data

! ------------------------------------------------------------------------------

!> Fortran generic
interface obs_count
  module procedure obs_count_time, obs_count_all, obs_count_indx
end interface obs_count

! ------------------------------------------------------------------------------
contains
! ------------------------------------------------------------------------------

subroutine obs_setup(fin, fout, self)
implicit none
type(obs_data), intent(inout) :: self
character(len=*), intent(in) :: fin, fout

self%ngrp=0
self%filein =fin
self%fileout=fout

if (self%filein/="") call obs_read(self)
call c_logger(logger_trace,"qg_obs_data:obs_setup: done")

end subroutine obs_setup

! ------------------------------------------------------------------------------

subroutine obs_delete(self)
implicit none
type(obs_data), intent(inout) :: self
type(group_data), pointer :: jgrp
type(column_data), pointer :: jcol
integer :: jo

if (self%fileout/="") call obs_write(self)

do while (associated(self%grphead))
  jgrp=>self%grphead
  self%grphead=>jgrp%next
  do jo=1,jgrp%nobs
    call datetime_delete(jgrp%times(jo))
  enddo
  deallocate(jgrp%times)
  do while (associated(jgrp%colhead))
    jcol=>jgrp%colhead
    jgrp%colhead=>jcol%next
    deallocate(jcol%values)
    deallocate(jcol)
  enddo
  deallocate(jgrp)
enddo

end subroutine obs_delete

! ------------------------------------------------------------------------------

subroutine obs_get(self, req, col, ovec)
implicit none
type(obs_data), intent(in) :: self
character(len=*), intent(in) :: req, col
type(obs_vect), intent(inout) :: ovec

type(group_data),  pointer :: jgrp
type(column_data), pointer :: jcol
integer :: jo, jc
character(len=250) :: record

write(record,*)"obs_get req=",req
call c_logger(logger_info,trim(record))
write(record,*)"obs_get col=",col
call c_logger(logger_info,trim(record))

! Find obs group
call findgroup(self,req,jgrp)
if (.not.associated(jgrp)) then
  jgrp=>self%grphead
  do while (associated(jgrp))
    write(record,*)"Group ",jgrp%grpname," exists."
    call c_logger(logger_info,trim(record))
    jgrp=>jgrp%next
  enddo
  write(record,*)"Cannot find ",req," ."
  call c_logger(logger_error,trim(record))
  call abor1_ftn("qg_obs_get: obs group not found")
endif

! Find obs column
call findcolumn(jgrp,col,jcol)
if (.not.associated(jcol)) call abor1_ftn("qg_obs_get: obs column not found")

! Get data
!if (ovec%nobs/=jgrp%nobs .or. ovec%ncol/=jcol%ncol) then
  if (allocated(ovec%values)) deallocate(ovec%values)
  ovec%nobs=jgrp%nobs
  ovec%ncol=jcol%ncol
  allocate(ovec%values(ovec%ncol,ovec%nobs))
!endif

do jo=1,jgrp%nobs
  do jc=1,jcol%ncol
    ovec%values(jc,jo)=jcol%values(jc,jo)
  enddo
enddo
write(record,*)"obs_get nobs, ncol=",jgrp%nobs,jcol%ncol
call c_logger(logger_info,trim(record))

end subroutine obs_get

! ------------------------------------------------------------------------------

subroutine obs_put(self, req, col, ovec)
implicit none
type(obs_data), intent(inout) :: self
character(len=*), intent(in) :: req, col
type(obs_vect), intent(in) :: ovec

type(group_data),  pointer :: jgrp
type(column_data), pointer :: jcol
integer :: jo, jc
character(len=250) :: record

! Find obs group
call findgroup(self,req,jgrp)
if (.not.associated(jgrp)) then
  jgrp=>self%grphead
  do while (associated(jgrp))
    write(record,*)"Group ",jgrp%grpname," exists."
    call c_logger(logger_info,trim(record))
    jgrp=>jgrp%next
  enddo
  write(record,*)"Cannot find ",req," ."
  call c_logger(logger_error,trim(record))
  call abor1_ftn("qg_obs_put: obs group not found")
endif

! Find obs column (and add it if not there)
call findcolumn(jgrp,col,jcol)
if (.not.associated(jcol)) then
  if (.not.associated(jgrp%colhead)) call abor1_ftn("qg_obs_put: no locations")
  jcol=>jgrp%colhead
  do while (associated(jcol%next))
    jcol=>jcol%next
  enddo
  allocate(jcol%next)
  jcol=>jcol%next
  
  jcol%colname=col
  jcol%ncol=ovec%ncol
  allocate(jcol%values(jcol%ncol,jgrp%nobs))
endif

! Put data
if (ovec%nobs/=jgrp%nobs) call abor1_ftn("qg_obs_put: error obs number")
if (ovec%ncol/=jcol%ncol) call abor1_ftn("qg_obs_put: error col number")
do jo=1,jgrp%nobs
  do jc=1,jcol%ncol
    jcol%values(jc,jo)=ovec%values(jc,jo)
  enddo
enddo

end subroutine obs_put

! ------------------------------------------------------------------------------

subroutine obs_time_get(self, req, col, t1, t2, ovec)
implicit none
type(obs_data), intent(in)    :: self
character(len=*), intent(in)  :: req, col
type(datetime), intent(in)    :: t1, t2
type(obs_vect), intent(inout) :: ovec

type(group_data),  pointer :: jgrp
type(column_data), pointer :: jcol
integer :: jo, jc, iobs

! Find obs group
call findgroup(self,req,jgrp)
if (.not.associated(jgrp)) call abor1_ftn("obs_time_get: obs group not found")

! Find obs column
call findcolumn(jgrp,col,jcol)
if (.not.associated(jcol)) call abor1_ftn("obs_time_get: obs column not found")

! Time selection
iobs=0
do jo=1,jgrp%nobs
  if (t1<jgrp%times(jo) .and. jgrp%times(jo)<=t2) iobs=iobs+1
enddo

! Get data
if (ovec%nobs/=iobs .or. ovec%ncol/=jcol%ncol) then
  if (allocated(ovec%values)) deallocate(ovec%values)
  ovec%nobs=iobs
  ovec%ncol=jcol%ncol
  allocate(ovec%values(ovec%ncol,ovec%nobs))
endif

iobs=0
do jo=1,jgrp%nobs
  if (t1<jgrp%times(jo) .and. jgrp%times(jo)<=t2) then
    iobs=iobs+1
    do jc=1,jcol%ncol
      ovec%values(jc,iobs)=jcol%values(jc,jo)
    enddo
  endif
enddo

end subroutine obs_time_get

! ------------------------------------------------------------------------------

subroutine obs_count_time(self, req, t1, t2, kobs)
implicit none
type(obs_data), intent(in)   :: self
character(len=*), intent(in) :: req
type(datetime), intent(in)   :: t1, t2
integer, intent(out)         :: kobs

type(group_data),  pointer :: jgrp
integer :: jo

! Find obs group
call findgroup(self,req,jgrp)
if (.not.associated(jgrp)) call abor1_ftn("obs_count: obs group not found")

! Time selection
kobs=0
do jo=1,jgrp%nobs
  if (t1<jgrp%times(jo) .and. jgrp%times(jo)<=t2) kobs=kobs+1
enddo

end subroutine obs_count_time

! ------------------------------------------------------------------------------

subroutine obs_count_indx(self, req, t1, t2, kobs)
implicit none
type(obs_data), intent(in)   :: self
character(len=*), intent(in) :: req
type(datetime), intent(in)   :: t1, t2
integer, intent(inout)       :: kobs(:)

type(group_data),  pointer :: jgrp
integer :: jo, io

! Find obs group
call findgroup(self,req,jgrp)
if (.not.associated(jgrp)) call abor1_ftn("obs_count: obs group not found")

! Time selection
io=0
do jo=1,jgrp%nobs
  if (t1<jgrp%times(jo) .and. jgrp%times(jo)<=t2) then
    io=io+1
    kobs(io)=jo
  endif
enddo

end subroutine obs_count_indx

! ------------------------------------------------------------------------------

subroutine obs_count_all(self, req, kobs)
implicit none
type(obs_data), intent(in) :: self
character(len=*), intent(in) :: req
integer, intent(out) :: kobs
type(group_data),  pointer :: jgrp

call findgroup(self,req,jgrp)
if (.not.associated(jgrp)) call abor1_ftn("obs_count: obs group not found")
kobs=jgrp%nobs

end subroutine obs_count_all

! ------------------------------------------------------------------------------

subroutine obs_create(self, req, times, locs)
implicit none
type(obs_data), intent(inout) :: self
character(len=*), intent(in) :: req
type(datetime), intent(in) :: times(:)
type(obs_vect), intent(in) :: locs
type(group_data), pointer :: igrp
integer :: jo, jc

call findgroup(self,req,igrp)
if (associated(igrp)) call abor1_ftn("obs_create: obs group already exists")

if (associated(self%grphead)) then
  igrp=>self%grphead
  do while (associated(igrp%next))
    igrp=>igrp%next
  enddo
  allocate(igrp%next)
  igrp=>igrp%next
else
  allocate(self%grphead)
  igrp=>self%grphead
endif

igrp%grpname=req
igrp%nobs=size(times)
allocate(igrp%times(igrp%nobs))
igrp%times(:)=times(:)

allocate(igrp%colhead)
igrp%colhead%colname="Location"
igrp%colhead%ncol=3
allocate(igrp%colhead%values(3,igrp%nobs))
if (locs%ncol/=3) call abor1_ftn("obs_create: error locations not 3D")
if (locs%nobs/=igrp%nobs) call abor1_ftn("obs_create: error locations number")
do jo=1,igrp%nobs
  do jc=1,3
    igrp%colhead%values(jc,jo)=locs%values(jc,jo)
  enddo
enddo

self%ngrp=self%ngrp+1

end subroutine obs_create

! ------------------------------------------------------------------------------
!  Private
! ------------------------------------------------------------------------------

subroutine obs_read(self)
implicit none
type(obs_data), intent(inout) :: self
integer :: iin, icol, jo, jc, jg, ncol
type(group_data), pointer :: jgrp
type(column_data), pointer :: jcol
real(kind=kind_real), allocatable :: ztmp(:)
character(len=20) :: stime
character(len=250) :: record

iin=90
write(record,*)'obs_read: opening ',self%filein
call c_logger(logger_info,trim(record))
open(unit=iin, file=trim(self%filein), form='formatted', action='read')

read(iin,*)self%ngrp
do jg=1,self%ngrp
  if (jg==1) then
    allocate(self%grphead)
    jgrp=>self%grphead
  else
    allocate(jgrp%next)
    jgrp=>jgrp%next
  endif
  read(iin,*)jgrp%grpname
  read(iin,*)jgrp%nobs
  write(record,*)'obs_read: reading ',jgrp%nobs,' ',jgrp%grpname,' observations.'
  call c_logger(logger_info,trim(record))
  allocate(jgrp%times(jgrp%nobs))

  read(iin,*)ncol
  icol=0
  do jc=1,ncol
    if (jc==1) then
      allocate(jgrp%colhead)
      jcol=>jgrp%colhead
    else
      allocate(jcol%next)
      jcol=>jcol%next
    endif
    read(iin,*)jcol%colname, jcol%ncol
    icol=icol+jcol%ncol
    allocate(jcol%values(jcol%ncol,jgrp%nobs))
  enddo

  allocate(ztmp(icol))
  do jo=1,jgrp%nobs
    read(iin,*)stime,ztmp(:)
    call datetime_create(stime,jgrp%times(jo))
    icol=0
    jcol=>jgrp%colhead
    do while (associated(jcol))
      do jc=1,jcol%ncol
        icol=icol+1
        jcol%values(jc,jo)=ztmp(icol)
      enddo
      jcol=>jcol%next
    enddo
  enddo
  deallocate(ztmp)
enddo

close(iin)

end subroutine obs_read

! ------------------------------------------------------------------------------

subroutine obs_write(self)
implicit none
type(obs_data), intent(in) :: self
integer :: iout, icol, jc, jo
type(group_data), pointer :: jgrp
type(column_data), pointer :: jcol
real(kind=kind_real), allocatable :: ztmp(:)
character(len=20) :: stime

iout=91
open(unit=iout, file=self%fileout, form='formatted', action='write')

write(iout,*)self%ngrp
jgrp=>self%grphead
do while (associated(jgrp))
  write(iout,*)jgrp%grpname
  write(iout,*)jgrp%nobs

  icol=0
  jcol=>jgrp%colhead
  do while (associated(jcol))
    icol=icol+1
    jcol=>jcol%next
  enddo
  write(iout,*)icol

  icol=0
  jcol=>jgrp%colhead
  do while (associated(jcol))
    write(iout,*)jcol%colname, jcol%ncol
    icol=icol+jcol%ncol
    jcol=>jcol%next
  enddo

  allocate(ztmp(icol))
  do jo=1,jgrp%nobs
    icol=0
    jcol=>jgrp%colhead
    do while (associated(jcol))
      do jc=1,jcol%ncol
        icol=icol+1
        ztmp(icol)=jcol%values(jc,jo)
      enddo
      jcol=>jcol%next
    enddo
    call datetime_to_string(jgrp%times(jo),stime)
    write(iout,*)stime,ztmp(:)
  enddo
  deallocate(ztmp)

  jgrp=>jgrp%next
enddo

close(iout)

end subroutine obs_write

! ------------------------------------------------------------------------------

subroutine findgroup(self,req,find)
type(obs_data), intent(in) :: self
character(len=*), intent(in) :: req
type(group_data), pointer, intent(out) :: find

find=>self%grphead
do while (associated(find))
  if (find%grpname==req) exit
  find=>find%next
enddo

end subroutine findgroup

! ------------------------------------------------------------------------------

subroutine findcolumn(grp,col,find)
type(group_data), intent(in) :: grp
character(len=*), intent(in) :: col
type(column_data), pointer, intent(out) :: find

find=>grp%colhead
do while (associated(find))
  if (find%colname==col) exit
  find=>find%next
enddo

end subroutine findcolumn

! ------------------------------------------------------------------------------

end module qg_obs_data
