!
! Copyright 2011 ECMWF
! 
! This software was developed at ECMWF for evaluation
! and may be used for academic and research purposes only.
! The software is provided as is without any warranty.
! 
! This software can be used, copied and modified but not
! redistributed or sold. This notice must be reproduced
! on each copy made.
!

!> Fortran module handling observation vectors

module qg_obs_vectors

use iso_c_binding
use random_vectors_mod
use kinds

implicit none
private
public :: obs_vect, obsvec_setup

! ------------------------------------------------------------------------------

!> Fortran derived type to represent an observation vector
type obs_vect
  integer :: nobs=0
  integer :: ncol=0
  real(kind=kind_real), allocatable :: values(:,:)
end type obs_vect

! ------------------------------------------------------------------------------
contains
! ------------------------------------------------------------------------------

subroutine obsvec_setup(self, nc, no)
implicit none
type(obs_vect), intent(inout) :: self
integer(c_int), intent(in)    :: nc, no

self%ncol=nc
self%nobs=no
if (allocated(self%values)) deallocate(self%values)
allocate(self%values(self%ncol,self%nobs))

end subroutine obsvec_setup

! ------------------------------------------------------------------------------

subroutine c_qg_obsvec_clone(c_self, c_other) bind(c,name='qg_obsvec_clone_f90')
implicit none
type(c_ptr), intent(in)    :: c_self
type(c_ptr), intent(inout) :: c_other
type(obs_vect), pointer :: self, other

call c_f_pointer(c_self,self)
allocate(other)
other%ncol=self%ncol
other%nobs=self%nobs
allocate(other%values(other%ncol,other%nobs))
other%values(:,:)=self%values(:,:)
c_other = c_loc(other)

end subroutine c_qg_obsvec_clone

! ------------------------------------------------------------------------------

subroutine c_qg_obsvec_delete(c_self) bind(c,name='qg_obsvec_delete_f90')
implicit none
type(c_ptr), intent(inout) :: c_self
type(obs_vect), pointer :: self

call c_f_pointer(c_self,self)
deallocate(self%values)
deallocate(self)

end subroutine c_qg_obsvec_delete

! ------------------------------------------------------------------------------
subroutine c_qg_obsvec_assign(c_self, c_rhs) bind(c,name='qg_obsvec_assign_f90')
implicit none
type(c_ptr), intent(inout) :: c_self
type(c_ptr), intent(in) :: c_rhs
type(obs_vect), pointer :: self, rhs

call c_f_pointer(c_self,self)
call c_f_pointer(c_rhs,rhs)
if (rhs%ncol/=self%ncol .or. rhs%nobs/=self%nobs) then
  deallocate(self%values)
  self%ncol=rhs%ncol
  self%nobs=rhs%nobs
  allocate(self%values(self%ncol,self%nobs))
endif
self%values(:,:)=rhs%values(:,:)

end subroutine c_qg_obsvec_assign
! ------------------------------------------------------------------------------
subroutine c_qg_obsvec_zero(c_self) bind(c,name='qg_obsvec_zero_f90')
implicit none
type(c_ptr), intent(inout) :: c_self
type(obs_vect), pointer :: self

call c_f_pointer(c_self,self)
self%values(:,:)=0.0_kind_real

end subroutine c_qg_obsvec_zero
! ------------------------------------------------------------------------------
subroutine c_qg_obsvec_mul_scal(c_self, zz) bind(c,name='qg_obsvec_mul_scal_f90')
implicit none
type(c_ptr), intent(inout) :: c_self
real(c_double), intent(in) :: zz
type(obs_vect), pointer :: self

call c_f_pointer(c_self,self)
self%values(:,:)=zz*self%values(:,:)

end subroutine c_qg_obsvec_mul_scal
! ------------------------------------------------------------------------------
subroutine c_qg_obsvec_add(c_self, c_other) bind(c,name='qg_obsvec_add_f90')
implicit none
type(c_ptr), intent(inout) :: c_self
type(c_ptr), intent(in) :: c_other
type(obs_vect), pointer :: self, other

call c_f_pointer(c_self,self)
call c_f_pointer(c_other,other)
self%values(:,:)=self%values(:,:)+other%values(:,:)

end subroutine c_qg_obsvec_add
! ------------------------------------------------------------------------------
subroutine c_qg_obsvec_sub(c_self, c_other) bind(c,name='qg_obsvec_sub_f90')
implicit none
type(c_ptr), intent(inout) :: c_self
type(c_ptr), intent(in) :: c_other
type(obs_vect), pointer :: self, other

call c_f_pointer(c_self,self)
call c_f_pointer(c_other,other)
self%values(:,:)=self%values(:,:)-other%values(:,:)

end subroutine c_qg_obsvec_sub
! ------------------------------------------------------------------------------
subroutine c_qg_obsvec_mul(c_self, c_other) bind(c,name='qg_obsvec_mul_f90')
implicit none
type(c_ptr), intent(inout) :: c_self
type(c_ptr), intent(in) :: c_other
type(obs_vect), pointer :: self, other

call c_f_pointer(c_self,self)
call c_f_pointer(c_other,other)
self%values(:,:)=self%values(:,:)*other%values(:,:)

end subroutine c_qg_obsvec_mul
! ------------------------------------------------------------------------------
subroutine c_qg_obsvec_div(c_self, c_other) bind(c,name='qg_obsvec_div_f90')
implicit none
type(c_ptr), intent(inout) :: c_self
type(c_ptr), intent(in) :: c_other
type(obs_vect), pointer :: self, other

call c_f_pointer(c_self,self)
call c_f_pointer(c_other,other)
self%values(:,:)=self%values(:,:)/other%values(:,:)

end subroutine c_qg_obsvec_div
! ------------------------------------------------------------------------------
subroutine c_qg_obsvec_axpy(c_self, zz, c_other) bind(c,name='qg_obsvec_axpy_f90')
implicit none
type(c_ptr), intent(inout) :: c_self
real(c_double), intent(in) :: zz
type(c_ptr), intent(in) :: c_other
type(obs_vect), pointer :: self, other

call c_f_pointer(c_self,self)
call c_f_pointer(c_other,other)
self%values(:,:)=self%values(:,:)+zz*other%values(:,:)

end subroutine c_qg_obsvec_axpy
! ------------------------------------------------------------------------------
subroutine c_qg_obsvec_invert(c_self) bind(c,name='qg_obsvec_invert_f90')
implicit none
type(c_ptr), intent(inout) :: c_self
type(obs_vect), pointer :: self

call c_f_pointer(c_self,self)
self%values(:,:)=1.0_kind_real/self%values(:,:)

end subroutine c_qg_obsvec_invert
! ------------------------------------------------------------------------------
subroutine c_qg_obsvec_random(c_self) bind(c,name='qg_obsvec_random_f90')
implicit none
type(c_ptr), intent(inout) :: c_self
type(obs_vect), pointer :: self

call c_f_pointer(c_self,self)
call random_vector(self%values)

end subroutine c_qg_obsvec_random
! ------------------------------------------------------------------------------
subroutine c_qg_obsvec_dotprod(c_self, c_other, zz) bind(c,name='qg_obsvec_dotprod_f90')
implicit none
type(c_ptr), intent(in) :: c_self
type(c_ptr), intent(in) :: c_other
real(c_double), intent(out) :: zz
type(obs_vect), pointer :: self, other
integer :: jc, jo

zz=0.0_kind_real
call c_f_pointer(c_self,self)
call c_f_pointer(c_other,other)
do jo=1,self%nobs
do jc=1,self%ncol
  zz = zz + self%values(jc,jo)*other%values(jc,jo)
enddo
enddo

end subroutine c_qg_obsvec_dotprod
! ------------------------------------------------------------------------------
subroutine c_qg_obsvec_minmaxavg(c_self, zmin, zmax, zavg) bind(c,name='qg_obsvec_minmaxavg_f90')
implicit none
type(c_ptr), intent(in) :: c_self
real(c_double), intent(out) :: zmin, zmax, zavg
type(obs_vect), pointer :: self
integer :: jc, jo

call c_f_pointer(c_self,self)
if (.not.allocated(self%values)) call abor1_ftn("obsvec_minmax: obs vector not allocated")
if (self%nobs==0.or.self%ncol==0) call abor1_ftn("obsvec_minmax: obs vector size 0")
zmin=self%values(1,1)
zmax=self%values(1,1)
zavg=0.0_kind_real
do jo=1,self%nobs
do jc=1,self%ncol
  zmin=MIN(self%values(jc,jo),zmin)
  zmax=MAX(self%values(jc,jo),zmax)
  zavg=zavg+self%values(jc,jo)
enddo
enddo
zavg=zavg/real(self%ncol*self%nobs,kind_real)

end subroutine c_qg_obsvec_minmaxavg
! ------------------------------------------------------------------------------
subroutine c_qg_obsvec_nobs(c_self, kobs) bind(c,name='qg_obsvec_nobs_f90')
implicit none
type(c_ptr), intent(in) :: c_self
integer(c_int), intent(out) :: kobs
type(obs_vect), pointer :: self

call c_f_pointer(c_self,self)
kobs=self%nobs

end subroutine c_qg_obsvec_nobs
! ------------------------------------------------------------------------------

end module qg_obs_vectors
