!
! Copyright 2011 ECMWF
! 
! This software was developed at ECMWF for evaluation
! and may be used for academic and research purposes only.
! The software is provided as is without any warranty.
! 
! This software can be used, copied and modified but not
! redistributed or sold. This notice must be reproduced
! on each copy made.
!

!> Fortran module for streamfunction observations for the QG model
module qg_stream_mod

use iso_c_binding
use c_config_mod
use datetime_mod
use duration_mod
use qg_obs_data
use qg_obs_vectors
use qg_vars_mod
use qg_locs_mod
use qg_goms_mod
use kinds

implicit none
private
public :: qg_strm

! ------------------------------------------------------------------------------

!> Fortran derived type for stream function observations for the QG model
type :: qg_strm
  type(obs_data), pointer :: obsdb
  character(len=30) :: request
  type(qg_vars) :: varin
end type qg_strm

! ------------------------------------------------------------------------------
contains
! ------------------------------------------------------------------------------

subroutine c_qg_stream_setup(c_self, c_conf, c_obsdb) bind(c,name='qg_stream_setup_f90')
implicit none
type(c_ptr), intent(inout) :: c_self
type(c_ptr), intent(in)    :: c_conf
type(c_ptr), intent(in)    :: c_obsdb

type(qg_strm), pointer :: self
character(len=1) :: svars(1) = (/"x"/)

allocate(self)

call c_f_pointer(c_obsdb,self%obsdb)
self%request = c_config_get_data(c_conf, len(self%request), "ObsType")

call qg_vars_setup(self%varin, svars)

c_self = c_loc(self)

end subroutine c_qg_stream_setup

! ------------------------------------------------------------------------------

subroutine c_qg_stream_delete(c_self) bind(c,name='qg_stream_delete_f90')
implicit none
type(c_ptr), intent(inout) :: c_self

type(qg_strm), pointer :: self
call c_f_pointer(c_self,self)
deallocate(self%varin%fldnames)
deallocate(self)

end subroutine c_qg_stream_delete

! ------------------------------------------------------------------------------
subroutine qg_stream_equiv(c_gom, c_hofx) &
 & bind(c,name='qg_stream_equiv_f90')
implicit none
type(c_ptr), intent(in)    :: c_gom
type(c_ptr), intent(inout) :: c_hofx
type(qg_goms), pointer  :: gom
type(obs_vect), pointer :: hofx
integer :: io, jo

call c_f_pointer(c_gom,gom)
call c_f_pointer(c_hofx,hofx)

do jo=1,gom%nobs
  io=gom%indx(jo)
  hofx%values(1,io)=gom%values(1,jo)
enddo

end subroutine qg_stream_equiv
! ------------------------------------------------------------------------------
subroutine qg_stream_equiv_tl(c_gom, c_hofx) &
 & bind(c,name='qg_stream_equiv_tl_f90')
implicit none
type(c_ptr), intent(in)    :: c_gom
type(c_ptr), intent(inout) :: c_hofx
type(qg_goms), pointer  :: gom
type(obs_vect), pointer :: hofx
integer :: io, jo

call c_f_pointer(c_gom,gom)
call c_f_pointer(c_hofx,hofx)

do jo=1,gom%nobs
  io=gom%indx(jo)
  hofx%values(1,io)=gom%values(1,jo)
enddo

end subroutine qg_stream_equiv_tl
! ------------------------------------------------------------------------------
subroutine qg_stream_equiv_ad(c_gom, c_hofx) &
 & bind(c,name='qg_stream_equiv_ad_f90')
implicit none
type(c_ptr), intent(inout) :: c_gom
type(c_ptr), intent(in)    :: c_hofx
type(qg_goms), pointer  :: gom
type(obs_vect), pointer :: hofx
integer :: io, jo

call c_f_pointer(c_gom,gom)
call c_f_pointer(c_hofx,hofx)

do jo=1,gom%nobs
  io=gom%indx(jo)
  gom%values(1,jo)=hofx%values(1,io)
enddo

end subroutine qg_stream_equiv_ad
! ------------------------------------------------------------------------------
subroutine qg_stream_locations(c_self, c_t1, c_t2, c_pos) bind(c,name='qg_stream_locations_f90')
implicit none
type(c_ptr), intent(in) :: c_self
type(c_ptr), intent(in) :: c_t1, c_t2
type(c_ptr), intent(inout) :: c_pos

type(qg_strm), pointer :: self
type(datetime) :: t1, t2
type(qg_locs), pointer :: locs
type(obs_vect) :: ovec
character(len=8) :: col="Location"

call c_f_pointer(c_self,self)
call c_f_datetime(c_t1, t1)
call c_f_datetime(c_t2, t2)

call obs_time_get(self%obsdb, self%request, col, t1, t2, ovec)

allocate(locs)
call qg_loc_setup(locs, ovec)
c_pos = c_loc(locs)

deallocate(ovec%values)

end subroutine qg_stream_locations
! ------------------------------------------------------------------------------
subroutine qg_stream_getgom(c_self, c_t1, c_t2, c_gom) bind(c,name='qg_stream_getgom_f90')
implicit none
type(c_ptr), intent(in) :: c_self
type(c_ptr), intent(in) :: c_t1, c_t2
type(c_ptr), intent(inout) :: c_gom

type(qg_strm), pointer :: self
type(datetime) :: t1, t2
type(qg_goms), pointer :: gom
integer :: nobs
integer, allocatable :: mobs(:)

call c_f_pointer(c_self,self)
call c_f_pointer(c_gom,gom)
call c_f_datetime(c_t1, t1)
call c_f_datetime(c_t2, t2)

call obs_count(self%obsdb, self%request, t1, t2, nobs)
allocate(mobs(nobs))
call obs_count(self%obsdb, self%request, t1, t2, mobs)

call gom_setup(gom, self%varin, mobs)

deallocate(mobs)
end subroutine qg_stream_getgom
! ------------------------------------------------------------------------------
subroutine qg_stream_generate(c_self, c_conf, c_bgn, c_step, ktimes, c_hofx) bind(c,name='qg_stream_generate_f90')
implicit none
type(c_ptr), intent(in)    :: c_self
type(c_ptr), intent(in)    :: c_conf
type(c_ptr), intent(in)    :: c_bgn
type(c_ptr), intent(in)    :: c_step
integer(c_int), intent(in)  :: ktimes
type(c_ptr), intent(inout) :: c_hofx
type(qg_strm),  pointer :: self
type(datetime) :: bgn
type(duration) :: step
type(obs_vect), pointer :: hofx

integer :: nlocs, nobs
type(datetime), allocatable :: times(:)
type(obs_vect) :: obsloc, obserr

call c_f_pointer(c_self,self)
call c_f_datetime(c_bgn, bgn)
call c_f_duration(c_step, step)

nlocs  = c_config_get_int(c_conf, "obs_density");
nobs=nlocs*ktimes;
write(*,*)'qg_stream_generate: ktimes, nlocs, nobs=', ktimes, nlocs, nobs

allocate(times(nobs))

call generate_locations(c_conf, nlocs, ktimes, bgn, step, times, obsloc)
call obs_create(self%obsdb, trim(self%request), times, obsloc)

deallocate(times)
deallocate(obsloc%values)

call obsvec_setup(obserr,1,nobs)
obserr%values(:,:)=0.4_kind_real
call obs_put(self%obsdb, trim(self%request), "ObsErr", obserr)
deallocate(obserr%values)

allocate(hofx)
call obsvec_setup(hofx,1,nobs)
c_hofx = c_loc(hofx)

end subroutine qg_stream_generate
! ------------------------------------------------------------------------------

end module qg_stream_mod
