!
! Copyright 2011 ECMWF
! 
! This software was developed at ECMWF for evaluation
! and may be used for academic and research purposes only.
! The software is provided as is without any warranty.
! 
! This software can be used, copied and modified but not
! redistributed or sold. This notice must be reproduced
! on each copy made.
!

!> Fortran module to handle wind speed observations for the QG model
module qg_wspeed_mod

use iso_c_binding
use c_config_mod
use datetime_mod
use duration_mod
use qg_obs_data
use qg_obs_vectors
use qg_vars_mod
use qg_locs_mod
use qg_goms_mod
use kinds

implicit none
private
public :: qg_wind

! ------------------------------------------------------------------------------

!> Fortran derived type representing wind speed observations for the QG model
type :: qg_wind
  type(obs_data), pointer :: obsdb
  character(len=30) :: request
  type(qg_vars) :: varin
end type qg_wind

! ------------------------------------------------------------------------------
contains
! ------------------------------------------------------------------------------

subroutine c_qg_wspeed_setup(c_self, c_conf, c_obsdb) bind(c,name='qg_wspeed_setup_f90')
implicit none
type(c_ptr), intent(inout) :: c_self
type(c_ptr), intent(in)    :: c_conf
type(c_ptr), intent(in)    :: c_obsdb

type(qg_wind), pointer :: self
character(len=1) :: svars(2) = (/"u","v"/)

allocate(self)

call c_f_pointer(c_obsdb,self%obsdb)
self%request = c_config_get_data(c_conf, len(self%request), "ObsType")

call qg_vars_setup(self%varin, svars)

c_self = c_loc(self)

end subroutine c_qg_wspeed_setup

! ------------------------------------------------------------------------------

subroutine c_qg_wspeed_delete(c_self) bind(c,name='qg_wspeed_delete_f90')
implicit none
type(c_ptr), intent(inout) :: c_self

type(qg_wind), pointer :: self
call c_f_pointer(c_self,self)
deallocate(self%varin%fldnames)
deallocate(self)

end subroutine c_qg_wspeed_delete

! ------------------------------------------------------------------------------
subroutine qg_wspeed_eqv(c_gom, c_hofx) bind(c,name='qg_wspeed_eqv_f90')
implicit none
type(c_ptr), intent(in)    :: c_gom
type(c_ptr), intent(inout) :: c_hofx
type(qg_goms), pointer  :: gom
type(obs_vect), pointer :: hofx
integer :: io, jo
real(kind=kind_real) :: zz

call c_f_pointer(c_gom,gom)
call c_f_pointer(c_hofx,hofx)

do jo=1,gom%nobs
  io=gom%indx(jo)
  zz=gom%values(1,jo)*gom%values(1,jo)+gom%values(2,jo)*gom%values(2,jo)
  hofx%values(1,io)=sqrt(zz)
enddo

end subroutine qg_wspeed_eqv
! ------------------------------------------------------------------------------
subroutine qg_wspeed_tra(c_gom, c_hofx, c_traj) bind(c,name='qg_wspeed_tra_f90')
use c_logger_mod, only : c_logger, logger_trace
implicit none
type(c_ptr), intent(in)    :: c_gom
type(c_ptr), intent(inout) :: c_hofx
type(c_ptr), intent(inout) :: c_traj
type(qg_goms), pointer  :: gom
type(obs_vect), pointer :: hofx
type(qg_goms), pointer  :: traj
integer :: io, jo
integer :: imin, imax
character(len=250) :: record
real(kind=kind_real) :: zz

imin=999999
imax=-99999

call c_f_pointer(c_gom,gom)
call c_f_pointer(c_hofx,hofx)
call c_f_pointer(c_traj,traj)

do jo=1,gom%nobs
  io=gom%indx(jo)
  zz=gom%values(1,jo)*gom%values(1,jo)+gom%values(2,jo)*gom%values(2,jo)
  traj%values(1,io)=gom%values(1,jo)
  traj%values(2,io)=gom%values(2,jo)
  hofx%values(1,io)=sqrt(zz)
  if (io<imin) imin=io
  if (io>imax) imax=io
enddo
write (record,*) "qg_wspeed_tra: obstraj setup from ",imin," to ",imax
call c_logger(logger_trace,trim(record))

end subroutine qg_wspeed_tra
! ------------------------------------------------------------------------------
subroutine qg_wspeed_equiv_tl(c_gom, c_hofx, c_traj) &
 & bind(c,name='qg_wspeed_equiv_tl_f90')
implicit none
type(c_ptr), intent(in)    :: c_gom
type(c_ptr), intent(inout) :: c_hofx
type(c_ptr), intent(in)    :: c_traj
type(qg_goms), pointer  :: gom
type(obs_vect), pointer :: hofx
type(qg_goms), pointer  :: traj
integer :: io, jo
real(kind=kind_real) :: zz, zu, zv, zt

call c_f_pointer(c_gom,gom)
call c_f_pointer(c_hofx,hofx)
call c_f_pointer(c_traj,traj)

do jo=1,gom%nobs
  io=gom%indx(jo)
  zu=traj%values(1,io)
  zv=traj%values(2,io)
  zt=sqrt(zu*zu+zv*zv)
  if (zt>epsilon(zt)) then
    zz=2.0_kind_real*zu*gom%values(1,jo) &
      +2.0_kind_real*zv*gom%values(2,jo)
    hofx%values(1,io)=zz/(2.0_kind_real*zt)
  else
    hofx%values(1,io)=0.0_kind_real
  endif
enddo

end subroutine qg_wspeed_equiv_tl
! ------------------------------------------------------------------------------
subroutine qg_wspeed_equiv_ad(c_gom, c_hofx, c_traj) &
 & bind(c,name='qg_wspeed_equiv_ad_f90')
implicit none
type(c_ptr), intent(inout) :: c_gom
type(c_ptr), intent(in)    :: c_hofx
type(c_ptr), intent(in)    :: c_traj
type(qg_goms), pointer  :: gom
type(obs_vect), pointer :: hofx
type(qg_goms), pointer  :: traj
integer :: io, jo
real(kind=kind_real) :: zz, zu, zv, zt

call c_f_pointer(c_gom,gom)
call c_f_pointer(c_hofx,hofx)
call c_f_pointer(c_traj,traj)

do jo=1,gom%nobs
  io=gom%indx(jo)
  zu=traj%values(1,io)
  zv=traj%values(2,io)
  zt=sqrt(zu*zu+zv*zv)
  if (zt>epsilon(zt)) then
    zz=hofx%values(1,io)/(2.0_kind_real*zt)
    gom%values(1,jo)=2.0_kind_real*zu*zz
    gom%values(2,jo)=2.0_kind_real*zv*zz
  else
    gom%values(1,jo)=0.0_kind_real
    gom%values(2,jo)=0.0_kind_real
  endif
enddo

end subroutine qg_wspeed_equiv_ad
! ------------------------------------------------------------------------------
subroutine qg_wspeed_locations(c_self, c_t1, c_t2, c_pos) bind(c,name='qg_wspeed_locations_f90')
implicit none
type(c_ptr), intent(in) :: c_self
type(c_ptr), intent(in) :: c_t1, c_t2
type(c_ptr), intent(inout) :: c_pos

type(qg_wind), pointer :: self
type(datetime) :: t1, t2
type(qg_locs), pointer :: locs
type(obs_vect) :: ovec
character(len=8) :: col="Location"

call c_f_pointer(c_self,self)
call c_f_datetime(c_t1, t1)
call c_f_datetime(c_t2, t2)

call obs_time_get(self%obsdb, self%request, col, t1, t2, ovec)

allocate(locs)
call qg_loc_setup(locs, ovec)
c_pos = c_loc(locs)

deallocate(ovec%values)

end subroutine qg_wspeed_locations
! ------------------------------------------------------------------------------
subroutine qg_wspeed_getgom(c_self, c_t1, c_t2, c_gom) bind(c,name='qg_wspeed_getgom_f90')
implicit none
type(c_ptr), intent(in) :: c_self
type(c_ptr), intent(in) :: c_t1, c_t2
type(c_ptr), intent(inout) :: c_gom

type(qg_wind), pointer :: self
type(datetime) :: t1, t2
type(qg_goms), pointer :: gom
integer :: nobs
integer, allocatable :: mobs(:)

call c_f_pointer(c_self,self)
call c_f_pointer(c_gom,gom)
call c_f_datetime(c_t1, t1)
call c_f_datetime(c_t2, t2)

call obs_count(self%obsdb, self%request, t1, t2, nobs)
allocate(mobs(nobs))
call obs_count(self%obsdb, self%request, t1, t2, mobs)

call gom_setup(gom, self%varin, mobs)

deallocate(mobs)
end subroutine qg_wspeed_getgom
! ------------------------------------------------------------------------------
subroutine qg_wspeed_gettra(c_self, c_gom) bind(c,name='qg_wspeed_gettra_f90')
use c_logger_mod, only : c_logger, logger_trace
implicit none
type(c_ptr), intent(in) :: c_self
type(c_ptr), intent(inout) :: c_gom

type(qg_wind), pointer :: self
type(qg_goms), pointer :: gom
integer :: nobs, jj
character(len=250) :: record
integer, allocatable :: mobs(:)

call c_f_pointer(c_self,self)
call c_f_pointer(c_gom,gom)

call obs_count(self%obsdb, self%request, nobs)
allocate(mobs(nobs))
do jj=1,nobs
  mobs(jj)=jj
enddo

call gom_setup(gom, self%varin, mobs)
write (record,*) "qg_wspeed_gettra: obstraj allocated ",nobs
call c_logger(logger_trace,trim(record))

deallocate(mobs)
end subroutine qg_wspeed_gettra
! ------------------------------------------------------------------------------
subroutine qg_wspeed_generate(c_self, c_conf, c_bgn, c_step, ktimes, c_hofx) bind(c,name='qg_wspeed_generate_f90')
implicit none
type(c_ptr), intent(in)    :: c_self
type(c_ptr), intent(in)    :: c_conf
type(c_ptr), intent(in)    :: c_bgn
type(c_ptr), intent(in)    :: c_step
integer(c_int), intent(in)  :: ktimes
type(c_ptr), intent(inout) :: c_hofx
type(qg_wind),  pointer :: self
type(datetime) :: bgn
type(duration) :: step
type(obs_vect), pointer :: hofx

integer :: nlocs, nobs
type(datetime), allocatable :: times(:)
type(obs_vect) :: obsloc, obserr

call c_f_pointer(c_self,self)
call c_f_datetime(c_bgn, bgn)
call c_f_duration(c_step, step)

nlocs  = c_config_get_int(c_conf, "obs_density");
nobs=nlocs*ktimes;
write(*,*)'qg_wspeed_generate: ktimes, nlocs, nobs=', ktimes, nlocs, nobs

allocate(times(nobs))

call generate_locations(c_conf, nlocs, ktimes, bgn, step, times, obsloc)
call obs_create(self%obsdb, trim(self%request), times, obsloc)

deallocate(times)
deallocate(obsloc%values)

call obsvec_setup(obserr,1,nobs)
obserr%values(:,:)=1.2_kind_real
call obs_put(self%obsdb, trim(self%request), "ObsErr", obserr)
deallocate(obserr%values)

allocate(hofx)
call obsvec_setup(hofx,1,nobs)
c_hofx = c_loc(hofx)

end subroutine qg_wspeed_generate
! ------------------------------------------------------------------------------

end module qg_wspeed_mod
