/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef LOGGER_LOGGER_H_
#define LOGGER_LOGGER_H_

#include <sstream>
#include <string>

#include <boost/noncopyable.hpp>

/// Namespace for Logger class and associated classes/functions.
namespace logger {

/// Logging categories. (Note: MAX is only there for array dimensioning, etc.)
/*!
 * Important: If you change these categories, you must also change the
 * initialisations of Logger::selected_ and Logger::CategoryNames_
 * in Logger.cc
 */
enum Category {
  Trace,
  Debug,
  Configs,
  Info,
  Warning,
  Error,
  Fatal,
  MAX
};

/// \brief A simple logger class.
/*!
 * A simple logger class based on the DrDobbs article:
 * http://drdobbs.com/cpp/201804215?pgno=1.
 */

class Logger : private boost::noncopyable {
 public:
  // Constructor
  inline Logger() : msgCategory_(Info) {}

  // Destructor
  ~Logger();

  // Methods
  /// Return true if the message category is selected for output.
  static inline bool selected(Category cat) {return selected_[cat];}

  /// Select/Deselect a message category for output.
  static inline void setLogging(const Category cat, const bool ok) {
      selected_[cat]=ok;
  }

  /// Turn off all logging
  static void disableAll();

  /// Convert a string to a Category. Return false if the lookup fails.
  static bool lookup(const std::string&, Category&);

  /// Return the stringstream message buffer.
  std::ostringstream & get(Category);

 private:
  static bool selected_[MAX];
  static const std::string CategoryNames_[MAX];
  std::ostringstream os_;
  Category msgCategory_;
};

}  // namespace logger

// This is the tricksy bit. Note the else is there to avoid dangling-else
// problems if the macro call occurs inside an if block.

#define LOG(cat) \
  if (!logger::Logger::selected(logger::cat)) ; else \
      logger::Logger().get(logger::cat)

// Use the following macros for debugging only. They should not appear in
// production code!

#define LOG_ENABLE_LOGGING_FOR(cat) \
  logger::Logger::setLogging(logger::cat, true)

#define LOG_DISABLE_LOGGING_FOR(cat) \
  logger::Logger::setLogging(logger::cat, false)


#endif  // LOGGER_LOGGER_H_
