/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include <cmath>
#include <sstream>
#include <string>
#include <vector>

#include "logger/Logger.h"
#include "util/Config.h"
#include "util/asser1.h"
#include "util/dot_product.h"

namespace oops {
// -----------------------------------------------------------------------------
template<typename TRAIT>
ControlIncrement<TRAIT>::ControlIncrement(boost::ptr_vector<Increment> & incr4d,
                                          const util::Duration & dt)
  : modbias_(), obsbias_(), interval_(dt)
{
  incr4d_.transfer(incr4d_.end(), incr4d.begin(), incr4d.end(), incr4d);
  LOG(Trace) << "ControlIncrement:ControlIncrement created.";
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
ControlIncrement<TRAIT>::ControlIncrement(const ControlIncrement & other)
  : modbias_(other.modbias_), obsbias_(other.obsbias_), interval_(other.interval_)
{
  BOOST_FOREACH(const Increment & jx, other.incr4d_) {
    incr4d_.push_back(new Increment(jx));
  }
  LOG(Trace) << "ControlIncrement:ControlIncrement copied.";
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
ControlIncrement<TRAIT>::ControlIncrement(const ControlIncrement & other,
                                          const util::Config & tlConf)
  : modbias_(other.modbias_), obsbias_(other.obsbias_), interval_(other.interval_)
{
  ASSERT(tlConf.empty());  // Change of resolution not implemented yet
  BOOST_FOREACH(const Increment & jx, other.incr4d_) {
    incr4d_.push_back(new Increment(jx));
  }
  LOG(Trace) << "ControlIncrement:ControlIncrement copied.";
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
ControlIncrement<TRAIT>::~ControlIncrement() {}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void ControlIncrement<TRAIT>::diff(const CtrlVar & cv1, const CtrlVar & cv2)
{
  for (unsigned jsub = 0; jsub < incr4d_.size(); ++jsub) {
    incr4d_[jsub].diff(cv1[jsub], cv2[jsub]);
  }
}
// -----------------------------------------------------------------------------
template<typename TRAIT> ControlIncrement<TRAIT> &
  ControlIncrement<TRAIT>::operator=(const ControlIncrement & rhs)
{
  incr4d_ = rhs.incr4d_;
  return *this;
}
// -----------------------------------------------------------------------------
template<typename TRAIT> ControlIncrement<TRAIT> &
  ControlIncrement<TRAIT>::operator+=(const ControlIncrement & rhs)
{
  for (unsigned jsub = 0; jsub < incr4d_.size(); ++jsub) {
    incr4d_[jsub] += rhs[jsub];
  }
  return *this;
}
// -----------------------------------------------------------------------------
template<typename TRAIT> ControlIncrement<TRAIT> &
  ControlIncrement<TRAIT>::operator-=(const ControlIncrement & rhs)
{
  for (unsigned jsub = 0; jsub < incr4d_.size(); ++jsub) {
    incr4d_[jsub] -= rhs[jsub];
  }
  return *this;
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
ControlIncrement<TRAIT> & ControlIncrement<TRAIT>::operator*=(const double zz) {
  BOOST_FOREACH(Increment & jx, incr4d_) jx *= zz;
  return *this;
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void ControlIncrement<TRAIT>::zero() {
  BOOST_FOREACH(Increment & jx, incr4d_) jx.zero();
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void ControlIncrement<TRAIT>::zeroAD() {
  BOOST_FOREACH(Increment & jx, incr4d_) jx.zero(interval_);
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void ControlIncrement<TRAIT>::shift_front() {
  LOG(Debug) << "ControlIncrement:shift_front by " << interval_;
  incr4d_.pop_back();
  Increment * init = new Increment(incr4d_.front());
  init->zero(-interval_);
  incr4d_.push_front(init);
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void ControlIncrement<TRAIT>::shift_back() {
  LOG(Debug) << "ControlIncrement:shift_back by = " << interval_;
  incr4d_.pop_front();
  Increment * final = new Increment(incr4d_.back());
  final->zero(interval_);
  incr4d_.push_back(final);
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void ControlIncrement<TRAIT>::axpy(const double zz, const ControlIncrement & rhs) {
  for (unsigned jsub = 0; jsub < incr4d_.size(); ++jsub) {
    incr4d_[jsub].axpy(zz, rhs[jsub]);
  }
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void ControlIncrement<TRAIT>::updateControl(CtrlVar & xx) const {
  for (unsigned jsub = 0; jsub < incr4d_.size(); ++jsub) {
    xx[jsub] += incr4d_[jsub];
  }
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void ControlIncrement<TRAIT>::random() {
  BOOST_FOREACH(Increment & jx, incr4d_) jx.random();
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
double ControlIncrement<TRAIT>::norm() const {
// Could write optimized version knowing both args are the same... YT
  const double zz = dot_product(*this, *this);
  return sqrt(zz);
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void ControlIncrement<TRAIT>::read(const util::Config & config) {
  for (unsigned jsub = 0; jsub < incr4d_.size(); ++jsub) {
    std::stringstream ss;
    ss << jsub+1;
    std::string query = "increment[@indx='" + ss.str() + "']";
    util::Config fileConfig(config);
    fileConfig.setCurrentNode(query);

    LOG(Configs) << "ControlIncrement:read reading " << "\n"
                         << fileConfig;
    incr4d_[jsub].read(fileConfig);
    incr4d_[jsub].prtNorms("read");
  }
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void ControlIncrement<TRAIT>::write(const util::Config & config) const {
  BOOST_FOREACH(const Increment & jx, incr4d_) jx.write(config);
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void ControlIncrement<TRAIT>::prtNorms(const std::string & cgrep) const {
  BOOST_FOREACH(const Increment & jx, incr4d_) jx.prtNorms(cgrep);
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
double ControlIncrement<TRAIT>::dot_product_with(const ControlIncrement & x2) const {
  double zz = 0.0;
  for (unsigned jsub = 0; jsub < incr4d_.size(); ++jsub) {
    zz += dot_product(incr4d_[jsub], x2.incr4d_[jsub]);
  }
  return zz;
}
// -----------------------------------------------------------------------------
}  // namespace oops
