/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_ASSIMILATION_CONTROLINCREMENT_H_
#define OOPS_ASSIMILATION_CONTROLINCREMENT_H_

#include <string>

#include <boost/ptr_container/ptr_deque.hpp>
#include <boost/ptr_container/ptr_vector.hpp>

#include "oops/assimilation/ControlVariable.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/Duration.h"

namespace oops {


/// Control variable increment
/*!
 * The control variable acts as a container for the inputs of the variational
 * data assimilation cost functions in physical space.
 * That includes the states at the start the assimilation window or of each
 * sub-window but also additional variables such as model bias, VarBC
 * coefficients, or other control variables for algorithms that use them.
 * The control variable increment contains variations of the
 * control variable.
 */

// -----------------------------------------------------------------------------
template<typename TRAIT> class ControlIncrement {
  typedef typename TRAIT::IncrementBase Increment;
  typedef typename TRAIT::ModelAuxIncr  ModelAuxIncr;
  typedef typename TRAIT::ObsAuxIncr    ObsAuxIncr;
  typedef ControlVariable<TRAIT>        CtrlVar;

 public:
/// Constructor, destructor
  ControlIncrement(boost::ptr_vector<Increment> &, const util::Duration &);
  ControlIncrement(const ControlIncrement &);
  ControlIncrement(const ControlIncrement &, const util::Config &);
  ~ControlIncrement();

/// Linear algebra operators
  void diff(const CtrlVar &, const CtrlVar &);
  ControlIncrement & operator=(const ControlIncrement &);
  ControlIncrement & operator+=(const ControlIncrement &);
  ControlIncrement & operator-=(const ControlIncrement &);
  ControlIncrement & operator*=(const double);
  void zero();
  void zeroAD();
  void axpy(const double, const ControlIncrement &);
  void updateControl(CtrlVar &) const;
  double dot_product_with(const ControlIncrement &) const;
  void random();

/// I/O and diagnostics
  void read(const util::Config &);
  void write(const util::Config &) const;
  void prtNorms(const std::string &) const;
  double norm() const;

/// Get model space control variable
  Increment & operator[](const int ii) {return incr4d_[ii];}
  const Increment & operator[](const int ii) const {return incr4d_[ii];}
  void shift_front();
  void shift_back();

/// Get augmented model control variable
  ModelAuxIncr & modVar() {return modbias_;}
  const ModelAuxIncr & modVar() const {return modbias_;}

/// Get augmented observation control variable
  ObsAuxIncr & obsVar() {return obsbias_;}
  const ObsAuxIncr & obsVar() const {return obsbias_;}

 private:

  boost::ptr_deque<Increment> incr4d_;
  ModelAuxIncr modbias_;   // not only for bias, better name?
  ObsAuxIncr   obsbias_;   // not only for bias, better name?

  util::Duration interval_;
};
// -----------------------------------------------------------------------------
}  // namespace oops

#include "oops/assimilation/ControlIncrement.cc"

#endif  // OOPS_ASSIMILATION_CONTROLINCREMENT_H_
