/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include <sstream>

#include <boost/foreach.hpp>

#include "oops/base/Geometry.h"

#include "logger/Logger.h"
#include "util/Config.h"

namespace oops {
// -----------------------------------------------------------------------------

template<typename TRAIT>
ControlVariable<TRAIT>::ControlVariable(const util::Config & ctlConf,
                                        const Geometry & resol,
                                        const util::Config & bkgConf):
  modbias_(ctlConf), obsbias_(ctlConf)
{
  util::Config fileConfig(bkgConf);
  fileConfig.setCurrentNode("states");
  unsigned nbg = fileConfig.getElementSize("state");
  LOG(Debug) << "There are " << nbg << " states";

  for (unsigned jsub = 0; jsub < nbg; ++jsub) {
    util::Config file(fileConfig, "state", jsub);
    LOG(Debug) << "ControlVariable:reading " << "\n" << file;
    State * js = new State(resol, file);
    LOG(Info) << "ControlVariable:ControlVariable: read bg at " << js->validTime();
    js->prtNorms("read");
    state4d_.push_back(js);
  }

  LOG(Trace) << "ControlVariable:ControlVariable done.";
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
ControlVariable<TRAIT>::ControlVariable(const ControlVariable & other)
  : modbias_(other.modbias_), obsbias_(other.obsbias_)
{
  BOOST_FOREACH(const State & js, other.state4d_)
    state4d_.push_back(new State(js));
  LOG(Trace) << "ControlVariable:ControlVariable copied.";
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
ControlVariable<TRAIT>::~ControlVariable() {
  LOG(Trace) << "ControlVariable:~ControlVariable";
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
void ControlVariable<TRAIT>::read(const util::Config & config) {
  util::Config fileConfig(config);
  fileConfig.setCurrentNode("states");
  unsigned nb_states = fileConfig.getElementSize("state");
  LOG(Configs) << "There are " << nb_states << " states";
  unsigned jsub = 0;
  BOOST_FOREACH(State & js, state4d_) {
    util::Config stateConfig(fileConfig, "state", jsub);
    LOG(Configs) << "ControlVariable:read " << "\n" << stateConfig;
    js.read(stateConfig);
    js.prtNorms("read");
    ++jsub;
  }
  if (jsub>nb_states) {
      LOG(Error) << "more states were looped than configured";
  }
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
void ControlVariable<TRAIT>::write(const util::Config & config) const {
  util::Config fileConfig(config);
  fileConfig.setCurrentNode("states");
  unsigned nb_states = fileConfig.getElementSize("state");
  LOG(Configs) << "There are " << nb_states << " states";
  unsigned int jsub = 0;
  BOOST_FOREACH(const State & js, state4d_) {
    util::Config stateConfig(fileConfig, "state", jsub);

    LOG(Configs) << "ControlVariable:write " << "\n" << stateConfig;
    js.write(stateConfig);
    js.prtNorms("written");
    ++jsub;
  }
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
void ControlVariable<TRAIT>::prtNorms(const std::string & cgrep) const {
  BOOST_FOREACH(const State & js, state4d_) js.prtNorms(cgrep);
}

// -----------------------------------------------------------------------------
}  // namespace oops
