/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_ASSIMILATION_CONTROLVARIABLE_H_
#define OOPS_ASSIMILATION_CONTROLVARIABLE_H_

#include <string>

#include <boost/ptr_container/ptr_deque.hpp>

// Forward declarations
namespace util {
  class Config;
}

namespace oops {
  class Geometry;

/// Control variable
/*!
 * The control variable acts as a container for the inputs of the variational
 * data assimilation cost functions in physical space.
 * That includes the states at the start the assimilation window or of each
 * sub-window but also additional variables such as model bias, VarBC
 * coefficients, or other control variables for algorithms that use them.
 * This is mostly a convenience class that is used to keep things together
 * and reduce the number of arguments to be passed around.
 */


// -----------------------------------------------------------------------------
template<typename TRAIT> class ControlVariable {
  typedef typename TRAIT::ModelAuxCtrl  ModelAuxCtrl;
  typedef typename TRAIT::ObsAuxCtrl    ObsAuxCtrl;
  typedef typename TRAIT::StateBase     State;

 public:
/// The arguments define the number of sub-windows and the resolution
  ControlVariable(const util::Config &, const Geometry &, const util::Config &);
  explicit ControlVariable(const ControlVariable &);
  ~ControlVariable();

/// I/O and diagnostics
  void read(const util::Config &);
  void write(const util::Config &) const;
  void prtNorms(const std::string &) const;

/// Get model space control variable
  unsigned nsubwin() const {return state4d_.size();}
  State & operator[](const int ii) {return state4d_[ii];}
  const State & operator[](const int ii) const {return state4d_[ii];}

/// Get augmented model control variable
  ModelAuxCtrl & modVar() {return modbias_;}
  const ModelAuxCtrl & modVar() const {return modbias_;}

/// Get augmented observation control variable
  ObsAuxCtrl & obsVar() {return obsbias_;}
  const ObsAuxCtrl & obsVar() const {return obsbias_;}

 private:
  ControlVariable & operator= (const ControlVariable &);  // No assignment

  boost::ptr_deque<State> state4d_;
  ModelAuxCtrl modbias_;  // not only for bias, better name?
  ObsAuxCtrl obsbias_;    // not only for bias, better name?
};
// -----------------------------------------------------------------------------
}  // namespace oops

#include "oops/assimilation/ControlVariable.cc"

#endif  // OOPS_ASSIMILATION_CONTROLVARIABLE_H_
