/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include "logger/Logger.h"
#include "oops/assimilation/CostJcDFI.h"
#include "oops/base/ModelIncrement.h"
#include "oops/base/ModelState.h"
#include "util/Config.h"
#include "util/abor1_cpp.h"
#include "util/asser1.h"

namespace oops {

// -----------------------------------------------------------------------------

template<typename TRAIT>
CostFct4DEnsVar<TRAIT>::CostFct4DEnsVar(const util::Config & config)
  : CostFunction<TRAIT>::CostFunction(config)
{
  windowLength_ = util::Duration(config.getData("window_length"));
  windowBegin_ = util::DateTime(config.getData("window_begin"));
  windowSub_ = util::Duration(config.getData("window_sub"));

  ncontrol_ = windowLength_.toSeconds() / windowSub_.toSeconds();
  ASSERT(windowLength_.toSeconds() == windowSub_.toSeconds()*ncontrol_);

  LOG(Trace) << "CostFct4DEnsVar constructed";
}

// -----------------------------------------------------------------------------

template <typename TRAIT>
CostJb4D<TRAIT> * CostFct4DEnsVar<TRAIT>::getJb(const Geometry & resol,
                                                const CtrlVar & xb,
                                                const util::Config & jbConf) const {
  return new JbTerm(xb, resol, jbConf, util::Duration(0));
}

// -----------------------------------------------------------------------------

template <typename TRAIT>
CostJo<TRAIT> * CostFct4DEnsVar<TRAIT>::getJo(const Observ & yobs,
                                              const util::Config & joConf) const {
  return new JoTerm(yobs, joConf, windowSub_, false);
}

// -----------------------------------------------------------------------------

template <typename TRAIT>
CostJcBase<TRAIT> * CostFct4DEnsVar<TRAIT>::getJc(const Geometry & resol,
                                                  const util::Config & jcConf) const {
  JcTerm * pjc = 0;
  if (!jcConf.empty()) {
    const util::Config jcdfi(jcConf, "jcdfi", true);
    if (!jcdfi.empty()) {
      const util::DateTime vt(windowBegin_ + windowLength_/2);
      pjc = new CostJcDFI<TRAIT>(resol, jcdfi, vt, windowLength_, windowSub_);
    }
  }
  return pjc;
}

// -----------------------------------------------------------------------------

template <typename TRAIT>
void CostFct4DEnsVar<TRAIT>::runNL(const CtrlVar & xx,
                                   PostProcessor<State> & post) const {
  for (int jsub = 0; jsub <= ncontrol_; ++jsub) {
    util::DateTime now(windowBegin_ + jsub*windowSub_);

    ModelState<TRAIT> zz(xx[jsub], CFbase::getModel());

    ASSERT(zz.validTime() == now);
    zz.forecast(xx.modVar(), now, post);
    ASSERT(zz.validTime() == now);
  }
}

// -----------------------------------------------------------------------------

template <typename TRAIT>
void CostFct4DEnsVar<TRAIT>::runTLM(CtrlInc & dx,
                                    CostProcessor<Increment> & cost,
                                    PostProcessor<Increment> & post) const {
  for (int jsub = 0; jsub <= ncontrol_; ++jsub) {
    util::DateTime now(windowBegin_ + jsub*windowSub_);

    ASSERT(dx[jsub].validTime() == now);
    CFbase::getTLM(jsub).runTL(dx[jsub], dx.modVar(), util::Duration(0), cost, post);
    ASSERT(dx[jsub].validTime() == now);
  }
}

// -----------------------------------------------------------------------------

template <typename TRAIT>
void CostFct4DEnsVar<TRAIT>::runADJ(CtrlInc & dx,
                                    CostProcessor<Increment> & cost,
                                    PostProcessor<Increment> & post) const {
  for (int jsub = ncontrol_; jsub >= 0; --jsub) {
    util::DateTime now(windowBegin_ + jsub*windowSub_);
    LOG(Debug) << "CostFct4DEnsVar:runAD, now = " << now;
    LOG(Debug) << "CostFct4DEnsVar:runAD, vt  = " << dx[jsub].validTime();

    ASSERT(dx[jsub].validTime() == now);
    CFbase::getTLM(jsub).runAD(dx[jsub], dx.modVar(), util::Duration(0), cost, post);
    ASSERT(dx[jsub].validTime() == now);
  }
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
void CostFct4DEnsVar<TRAIT>::addIncr(CtrlVar & xx, const CtrlInc & dx,
                                     PostProcessor<Increment> &) const {
  dx.updateControl(xx);
}

// -----------------------------------------------------------------------------

}  // namespace oops
