/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include "logger/Logger.h"
#include "oops/assimilation/CostJcDFI.h"
#include "oops/base/ModelIncrement.h"
#include "oops/base/ModelState.h"
#include "util/Config.h"
#include "util/abor1_cpp.h"
#include "util/asser1.h"

namespace oops {

// -----------------------------------------------------------------------------

template<typename TRAIT>
CostFct4DVar<TRAIT>::CostFct4DVar(const util::Config & config)
  : CostFunction<TRAIT>::CostFunction(config)
{
  LOG(Trace) << "CostFct4DVar:CostFct4DVar";

  windowLength_ = util::Duration(config.getData("window_length"));
  windowBegin_ = util::DateTime(config.getData("window_begin"));
  windowEnd_ = windowBegin_ + windowLength_;

  LOG(Trace) << "CostFct4DVar constructed";
}

// -----------------------------------------------------------------------------

template <typename TRAIT>
CostJb<TRAIT> * CostFct4DVar<TRAIT>::getJb(const Geometry & resol, const CtrlVar & xb,
                                           const util::Config & jbConf) const {
  ASSERT(xb.nsubwin() == 1);
  return new JbTerm(xb, resol, jbConf, windowEnd_-windowBegin_);
}

// -----------------------------------------------------------------------------

template <typename TRAIT>
CostJo<TRAIT> * CostFct4DVar<TRAIT>::getJo(const Observ & yobs,
                                           const util::Config & joConf) const {
  return new JoTerm(yobs, joConf, util::Duration(0));
}

// -----------------------------------------------------------------------------

template <typename TRAIT>
CostJcBase<TRAIT> * CostFct4DVar<TRAIT>::getJc(const Geometry & resol,
                                               const util::Config & jcConf) const {
  JcTerm * pjc = 0;
  if (!jcConf.empty()) {
    const util::Config jcdfi(jcConf, "jcdfi", true);
    if (!jcdfi.empty()) {
      const util::DateTime vt(windowBegin_ + windowLength_/2);
      pjc = new CostJcDFI<TRAIT>(resol, jcdfi, vt, windowLength_);
    }
  }
  return pjc;
}

// -----------------------------------------------------------------------------

template <typename TRAIT>
void CostFct4DVar<TRAIT>::runNL(const CtrlVar & xx,
                                PostProcessor<State> & post) const {
  ASSERT(xx.nsubwin() == 1);
  ASSERT(xx[0].validTime() == windowBegin_);
  ModelState<TRAIT> zz(xx[0], CFbase::getModel());
  zz.forecast(xx.modVar(), windowEnd_, post);
  ASSERT(zz.validTime() == windowEnd_);
}

// -----------------------------------------------------------------------------

template <typename TRAIT>
void CostFct4DVar<TRAIT>::runTLM(CtrlInc & dx,
                                 CostProcessor<Increment> & cost,
                                 PostProcessor<Increment> & post) const {
  ASSERT(dx[0].validTime() == windowBegin_);
  TLM M(CFbase::getTLM());
  M.runTL(dx[0], dx.modVar(), windowLength_, cost, post);
}

// -----------------------------------------------------------------------------

template <typename TRAIT>
void CostFct4DVar<TRAIT>::runADJ(CtrlInc & dx,
                                 CostProcessor<Increment> & cost,
                                 PostProcessor<Increment> & post) const {
  TLM M(CFbase::getTLM());
  M.runAD(dx[0], dx.modVar(), windowLength_, cost, post);
  ASSERT(dx[0].validTime() == windowBegin_);
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
void CostFct4DVar<TRAIT>::addIncr(CtrlVar & xx, const CtrlInc & dx,
                                  PostProcessor<Increment> &) const {
  ASSERT(xx.nsubwin() == 1);
  dx.updateControl(xx);
}

// -----------------------------------------------------------------------------

}  // namespace oops
