/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include "logger/Logger.h"
#include "oops/assimilation/CostJcDFI.h"
#include "oops/base/ModelIncrement.h"
#include "oops/base/ModelState.h"
#include "util/Config.h"
#include "util/abor1_cpp.h"
#include "util/asser1.h"

namespace oops {

// -----------------------------------------------------------------------------

template<typename TRAIT>
CostFctWeak<TRAIT>::CostFctWeak(const util::Config & config)
  : CostFunction<TRAIT>::CostFunction(config),
    tlforcing_(false)
{
  windowLength_ = util::Duration(config.getData("window_length"));
  windowBegin_ = util::DateTime(config.getData("window_begin"));
  windowSub_ = util::Duration(config.getData("window_sub"));

  nsubwin_ = windowLength_.toSeconds() / windowSub_.toSeconds();
  ASSERT(windowLength_.toSeconds() == windowSub_.toSeconds()*nsubwin_);

  if (config.elementExists("tlforcing")) {
    if (config.getData("tlforcing") == "on") tlforcing_ = true;
  }

  LOG(Trace) << "CostFctWeak constructed";
}

// -----------------------------------------------------------------------------

template <typename TRAIT>
CostJbJq<TRAIT> * CostFctWeak<TRAIT>::getJb(const Geometry & resol, const CtrlVar & xb,
                                            const util::Config & jbConf) const {
  return new JbTerm(xb, resol, jbConf, windowSub_);
}

// -----------------------------------------------------------------------------

template <typename TRAIT>
CostJo<TRAIT> * CostFctWeak<TRAIT>::getJo(const Observ & yobs,
                                          const util::Config & joConf) const {
  return new JoTerm(yobs, joConf, util::Duration(0), true);
}

// -----------------------------------------------------------------------------

template <typename TRAIT>
CostJcBase<TRAIT> * CostFctWeak<TRAIT>::getJc(const Geometry & resol,
                                              const util::Config & jcConf) const {
  JcTerm * pjc = 0;
  if (!jcConf.empty()) {
    const util::Config jcdfi(jcConf, "jcdfi", true);
    if (!jcdfi.empty()) {
      const util::DateTime vt(windowBegin_ + windowLength_/2);
      pjc = new CostJcDFI<TRAIT>(resol, jcdfi, vt, windowLength_);
    }
  }
  return pjc;
}

// -----------------------------------------------------------------------------

template <typename TRAIT>
void CostFctWeak<TRAIT>::runNL(const CtrlVar & xx,
                               PostProcessor<State> & post) const {
  for (int jsub = 0; jsub < nsubwin_; ++jsub) {
    util::DateTime bgn(windowBegin_ + jsub*windowSub_);
    util::DateTime end(bgn + windowSub_);

    ModelState<TRAIT> zz(xx[jsub], CFbase::getModel());

    ASSERT(zz.validTime() == bgn);
    zz.forecast(xx.modVar(), end, post);
    ASSERT(zz.validTime() == end);
  }
}

// -----------------------------------------------------------------------------

template <typename TRAIT>
void CostFctWeak<TRAIT>::runTLM(CtrlInc & dx,
                                CostProcessor<Increment> & cost,
                                PostProcessor<Increment> & post) const {
  for (int jsub = 0; jsub < nsubwin_; ++jsub) {
    util::DateTime bgn(windowBegin_ + jsub*windowSub_);
    util::DateTime end(bgn + windowSub_);

    ASSERT(dx[jsub].validTime() == bgn);
    if (tlforcing_ && jsub > 0) dx[jsub] += dx[jsub-1];
    CFbase::getTLM(jsub).runTL(dx[jsub], dx.modVar(), windowSub_, cost, post);
    ASSERT(dx[jsub].validTime() == end);
  }
}

// -----------------------------------------------------------------------------

template <typename TRAIT>
void CostFctWeak<TRAIT>::runADJ(CtrlInc & dx,
                                CostProcessor<Increment> & cost,
                                PostProcessor<Increment> & post) const {
  for (int jsub = nsubwin_-1; jsub >= 0; --jsub) {
    util::DateTime bgn(windowBegin_ + jsub*windowSub_);
    util::DateTime end(bgn + windowSub_);

    ASSERT(dx[jsub].validTime() == end);
    CFbase::getTLM(jsub).runAD(dx[jsub], dx.modVar(), windowSub_, cost, post);
    if (tlforcing_ && jsub > 0) dx[jsub-1] += dx[jsub];
    ASSERT(dx[jsub].validTime() == bgn);
  }
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
void CostFctWeak<TRAIT>::addIncr(CtrlVar & xx, const CtrlInc & dx,
                                 PostProcessor<Increment> & post) const {
  if (tlforcing_) {
    Increment xi(dx[0]);
    CostProcessor<Increment> cost;
    for (int jsub = 0; jsub < nsubwin_; ++jsub) {
      if (jsub > 0) xi += dx[jsub];
      xx[jsub] += xi;
      if (jsub < nsubwin_-1) {
        CFbase::getTLM(jsub).runTL(xi, dx.modVar(), windowSub_, cost, post);
      }
    }
  } else {
    dx.updateControl(xx);
  }
}

// -----------------------------------------------------------------------------

}  // namespace oops
