/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include <map>
#include <string>

#include <boost/scoped_ptr.hpp>

#include "logger/Logger.h"
#include "oops/base/ObserverTLAD.h"
#include "util/asser1.h"
#include "util/abor1_cpp.h"
#include "util/dot_product.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/Duration.h"

namespace oops {

// -----------------------------------------------------------------------------
//  Factory
// -----------------------------------------------------------------------------

template <typename TRAIT>
CostFactory<TRAIT>::CostFactory(const std::string & name) {
  if (getMakers().find(name) != getMakers().end()) {
    LOG(Error) << name << " already registered in cost function factory.";
    ABORT("Element already registered in CostFactory.");
  }
  getMakers()[name] = this;
}

// -----------------------------------------------------------------------------

template <typename TRAIT>
CostFunction<TRAIT>* CostFactory<TRAIT>::create(const util::Config & config) {
  std::string id = config.getData("cost_type");
  LOG(Trace) << "Variational Assimilation Type=" << id;
  typename std::map<string, CostFactory<TRAIT>*>::iterator j = getMakers().find(id);
  if (j == getMakers().end()) {
    LOG(Error) << id << " does not exist in cost function factory.";
    ABORT("Element does not exist in CostFactory.");
  }
  return (*j).second->make(config);
}

// -----------------------------------------------------------------------------
//  Cost Function
// -----------------------------------------------------------------------------

template<typename TRAIT>
CostFunction<TRAIT>::CostFunction(const util::Config & conf)
  : jb_(0), jterms_(), mtraj_(0), model_(0), tlm_(0),
    jcConf_(conf, "Jc", true)
{
  LOG(Trace) << "CostFunction:created";
}

// -----------------------------------------------------------------------------

// This method could be merged with the constructor. For now
// it saves passing so many arguments through the factory.
template<typename TRAIT>
void CostFunction<TRAIT>::setup(const Geometry & resol, const Model & model,
                                const CtrlVar & xb, const util::Config & jbConf,
                                const Observ & yobs, const util::Config & joConf) {
  model_ = &model;
  jb_.reset(this->getJb(resol, xb, jbConf));
  JoTerm * jo = this->getJo(yobs, joConf);
  jterms_.push_back(jo);
  unsigned njc = 0;
  if (!jcConf_.empty()) {
    njc = jcConf_.getElementSize("jc");
  }
  for (unsigned jj = 0; jj < njc; ++jj) {
    util::Config jconf(jcConf_, "jc", jj);
    JcTerm * jc = this->getJc(resol, jconf);
    jterms_.push_back(jc);
  }
  LOG(Trace) << "CostFunction:setup done";
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
double CostFunction<TRAIT>::eval(const CtrlVar & cvar,
                                 const util::Config & config,
                                 PostProcessor<State> & post) {
// WARNING: inner loop configuration, should not be needed here. YT
// WARNING: lowres and tlm should not be needed here. YT
  const util::Config tlmConf(config, "tlmodel");
  const util::Config resConf(config, "resolution");
  const Geometry lowres(resConf);
// WARNING: lowres and tlm should not be needed here. YT

// Setup observer
  PostProcessor<State> pp(post);
  pp.enroll(jb_->initialize(cvar, lowres, tlmConf));
  BOOST_FOREACH(JTerm & jt, jterms_) {
    pp.enroll(jt.initialize(cvar, config));
  }

// Run NL model
  this->runNL(cvar, pp);

// Cost function value
  double zzz = jb_->finalize();
  BOOST_FOREACH(JTerm & jt, jterms_) {
    zzz += jt.finalize();
  }
  LOG(Info) << "CostFunction: Nonlinear J = " << zzz;
  return zzz;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
double CostFunction<TRAIT>::linearize(const CtrlVar & cvar,
                                      const util::Config & innerConf,
                                      PostProcessor<State> & post) {
// Read inner loop configuration
  const util::Config resConf(innerConf, "resolution");
  const util::Config lowConf(innerConf, "lrmodel");
  const util::Config tlmConf(innerConf, "tlmodel");
  const Geometry lowres(resConf);

// Setup terms of cost function
  PostProcessor<State> pp(post);
  pp.enroll(jb_->initialize(cvar, lowres, tlmConf, true));
  BOOST_FOREACH(JTerm & jt, jterms_) {
    pp.enroll(jt.initialize(cvar, innerConf, true));
  }

// Setup trajectory
  const Model modelLR(lowres, lowConf);
  TrajSaver<TRAIT> * modtraj = new TrajSaver<TRAIT>(innerConf, lowres, modelLR);
  pp.enroll(modtraj);

// Run NL model
  this->runNL(cvar, pp);

// Setup trajectory and TLM
  ASSERT(mtraj_.empty());
  mtraj_.transfer(mtraj_.end(), *modtraj->releaseTraj());
  tlm_.reset(new Model(lowres, tlmConf));

// Cost function value
  double zzz = jb_->finalize();
  BOOST_FOREACH(JTerm & jt, jterms_) {
    zzz += jt.finalize();
  }
  LOG(Info) << "CostFunction: Nonlinear J = " << zzz;
  return zzz;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
ControlIncrement<TRAIT> CostFunction<TRAIT>::getGradientFG() const {
  PostProcessor<Increment> pp;
  CostProcessor<Increment> costad;

  unsigned nb_jterms = jterms_.size();
  for (unsigned i=0; i<nb_jterms; ++i) {
    const JTerm& jt = jterms_[i];
    boost::shared_ptr<const GeneralizedDeparture> tmp(jt.getGradientFG());
    costad.enroll(jt.setupAD(tmp));
  }

  CtrlInc * gg = jb_->newIncrement();
  gg->zeroAD();
  this->runADJ(*gg, costad, pp);

  *gg += jb_->getGradientFG();
  return *gg;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
void CostFunction<TRAIT>::addIncrement(CtrlVar & xx, const CtrlInc & dx,
                                       PostProcessor<Increment> post) const {
  xx.prtNorms("addIncrement: guess norm");
  dx.prtNorms("addIncrement: increment norm");

  this->addIncr(xx, dx, post);

  xx.prtNorms("addIncrement: analysis norm");
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
void CostFunction<TRAIT>::reset() {
  mtraj_.clear();
  BOOST_FOREACH(JTerm & jt, jterms_) {
    jt.reset();
  }
  tlm_.reset();
}

// -----------------------------------------------------------------------------

}  // namespace oops
