/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_ASSIMILATION_COSTFUNCTION_H_
#define OOPS_ASSIMILATION_COSTFUNCTION_H_

#include <map>
#include <string>

#include <boost/noncopyable.hpp>
#include <boost/ptr_container/ptr_vector.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/shared_ptr.hpp>

#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/ControlVariable.h"
#include "oops/assimilation/DualVector.h"
#include "oops/assimilation/CostJbBase.h"
#include "oops/assimilation/CostJcBase.h"
#include "oops/assimilation/CostJo.h"
#include "oops/base/Departure.h"
#include "oops/base/Geometry.h"
#include "oops/base/LinearModel.h"
#include "oops/base/Observation.h"
#include "oops/base/ObsErrorCovariance.h"
#include "oops/base/Observer.h"
#include "oops/base/PostProcessor.h"
#include "oops/base/TrajSaver.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/Duration.h"

namespace oops {

// -----------------------------------------------------------------------------

/// Cost Function
/*!
 * The Cost Function contains all the environment necessary to
 * define and compute the variational data assimilation cost function
 * in physical space.
 * It also contains associated operators such as \f$ H B H^T\f$ and
 * \f$ H^T R^{-1} H\f$ where \f$ B\f$ is meant as a generalized background term,
 * ie the background term for the complete control variable
 * including VarBc, etc...
 */

template<typename TRAIT> class CostFunction : private boost::noncopyable {
  typedef typename TRAIT::StateBase     State;
  typedef typename TRAIT::IncrementBase Increment;
  typedef typename TRAIT::ModelData     Model;
  typedef typename TRAIT::ObsAuxCtrl    ObsAuxCtrl;
  typedef typename TRAIT::ObsAuxIncr    ObsAuxIncr;
  typedef typename TRAIT::ModelTraj     ModelTraj;
  typedef ControlVariable<TRAIT>        CtrlVar;
  typedef ControlIncrement<TRAIT>       CtrlInc;
  typedef CostJbBase<TRAIT>             JbTerm;
  typedef CostJcBase<TRAIT>             JcTerm;
  typedef CostJo<TRAIT>                 JoTerm;
  typedef Observation<TRAIT>            Observ;
  typedef Departure<TRAIT>              Depart;
  typedef LinearModel<TRAIT>            TLM;
  typedef ObsErrorCovariance<TRAIT>     ObsError;
  typedef DualVector<TRAIT>             Items;
  typedef CostTermBase<TRAIT>           JTerm;

 public:
  explicit CostFunction(const util::Config &);
  virtual ~CostFunction() {}

  void setup(const Geometry &, const Model &, const CtrlVar &,
             const util::Config &, const Observ &, const util::Config &);

  double eval(const CtrlVar &, const util::Config &,
              PostProcessor<State> &);
  double linearize(const CtrlVar &, const util::Config &,
                   PostProcessor<State> &);

  virtual void runTLM(CtrlInc &,
                      CostProcessor<Increment>&, PostProcessor<Increment>&) const =0;
  virtual void runADJ(CtrlInc &,
                      CostProcessor<Increment>&, PostProcessor<Increment>&) const =0;

  void addIncrement(CtrlVar &, const CtrlInc &,
                    PostProcessor<Increment> post = PostProcessor<Increment>() ) const;
  void reset();

  /// Get Jb
  const JbTerm & jb() const {return *jb_;}

  /// Get other terms of the cost function
  unsigned nterms() const {return jterms_.size();}
  const JTerm & jterm(const unsigned ii) const {return jterms_[ii];}

  /// Cost function gradient at first guess.
  CtrlInc getGradientFG() const;

 protected:
  const Model & getModel() const {return *model_;}
  const TLM getTLM(const unsigned isub = 0) const {return TLM(*tlm_, mtraj_[isub]);}

 private:
  virtual void runNL(const CtrlVar &, PostProcessor<State>&) const =0;
  virtual void addIncr(CtrlVar &, const CtrlInc &, PostProcessor<Increment>&) const =0;

  virtual JbTerm * getJb(const Geometry &, const CtrlVar &,
                         const util::Config &) const =0;

  virtual JoTerm * getJo(const Observ&, const util::Config &) const =0;

  virtual JcTerm * getJc(const Geometry &, const util::Config &) const =0;

// Data members
  boost::scoped_ptr<JbTerm> jb_;
  boost::ptr_vector<JTerm> jterms_;
  boost::ptr_vector<ModelTraj> mtraj_;
  const Model * model_;  // Should use smart pointer. YT
  boost::scoped_ptr<const Model> tlm_;
  const util::Config jcConf_;
};

// -----------------------------------------------------------------------------

/// Cost Function Factory
template <typename TRAIT>
class CostFactory {
 public:
  static CostFunction<TRAIT> * create(const util::Config &);
  virtual ~CostFactory() { getMakers().clear(); }

 protected:
  explicit CostFactory(const std::string &);
 private:
  virtual CostFunction<TRAIT> * make(const util::Config &) =0;
  static std::map < std::string, CostFactory<TRAIT> * > & getMakers() {
    static std::map < std::string, CostFactory<TRAIT> * > makers_;
    return makers_;
  }
};

template<class TRAIT, class FCT>
class CostMaker : public CostFactory<TRAIT> {
  virtual CostFunction<TRAIT> * make(const util::Config & config)
    {return new FCT(config);}
 public:
  explicit CostMaker(const std::string & name) : CostFactory<TRAIT>(name) {}
};

// -----------------------------------------------------------------------------

}  // namespace oops

#include "oops/assimilation/CostFunction.cc"

#endif  // OOPS_ASSIMILATION_COSTFUNCTION_H_
