/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include <string>

#include "logger/Logger.h"
#include "oops/base/Geometry.h"
#include "util/asser1.h"
#include "util/abor1_cpp.h"
#include "util/dot_product.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/Duration.h"

namespace oops {

// -----------------------------------------------------------------------------
//  Jb Term of Cost Function
// -----------------------------------------------------------------------------

template<typename TRAIT>
CostJb<TRAIT>::CostJb(const CtrlVar & xb, const Geometry & resol,
                      const util::Config & config, const util::Duration & len)
  : B_(0), xb_(xb), dxFG_(0), gradFG_(0), winLength_(len)
{
  ASSERT(xb.nsubwin() == 1);

  util::Config covariances(config, "covariances");
  unsigned nbg = covariances.getElementSize("covariance");
  ASSERT(nbg == 1);

  const util::Config errConfig(covariances, "covariance", unsigned(0));
  B_.reset(CovarianceFactory<TRAIT>::create(errConfig, resol));

  LOG(Trace) << "CostJb:contructor done";
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
ProcessorBase< typename TRAIT::StateBase > * CostJb<TRAIT>::initialize(
                                                 const CtrlVar & fg,
                                                 const Geometry & resol,
                                                 const util::Config & tlm,
                                                 const bool ltraj) {
  ASSERT(fg.nsubwin() == 1);
  if (ltraj) B_->linearize(fg[0], resol, tlm);

// dxFG_ is  x_0 - x_b
  LOG(Trace) << "CostJb:initialize: constructing FG-BG";
  dxFG_.reset(this->newIncrement());
  dxFG_->diff(fg, xb_);

  ProcessorBase<State> * ppb = 0;
  return ppb;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
double CostJb<TRAIT>::finalize() {
  dxFG_->prtNorms("dxFG");

// gradFG_ is B^{-1} ( x_0 - x_b ) as estimated by the minimizer
  if (gradFG_ == 0) {
    gradFG_.reset(this->newIncrement());
    gradFG_->zero();
  }

  const double zjb = 0.5 * dot_product(*dxFG_, *gradFG_);
  LOG(Info) << "CostJb   : Nonlinear Jb = " << zjb;
  return zjb;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
ControlIncrement<TRAIT> * CostJb<TRAIT>::finalTL(const CtrlInc &,
                                                 const CtrlInc & bgns) const {
  CtrlInc * dx = new CtrlInc(bgns);
  return dx;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
ControlIncrement<TRAIT> * CostJb<TRAIT>::setupAD(const CtrlInc &) const {
  CtrlInc * ends = this->newIncrement();
  ends->zeroAD();
  return ends;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
void CostJb<TRAIT>::gradientUpdate(const CtrlInc & di, const bool add) const {
  ASSERT(add == true);
  *gradFG_ += di;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
void CostJb<TRAIT>::Bmult(const CtrlInc & dxin, CtrlInc & dxout) const {
  B_->mult(dxin[0], dxout[0]);
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
void CostJb<TRAIT>::Bminv(const CtrlInc & dxin, CtrlInc & dxout) const {
  LOG(Info) << "*** WARNING *** B inverse might not always exist ***";
  B_->inv_mult(dxin[0], dxout[0]);
  LOG(Info) << "*** WARNING *** B inverse might not always exist ***";
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
oops::ControlIncrement<TRAIT> * CostJb<TRAIT>::newIncrement() const {
  typedef typename TRAIT::IncrementBase Increment;

  boost::ptr_vector<Increment> incr4d;
  incr4d.push_back(B_->getIncrement());

  CtrlInc * dx = new CtrlInc(incr4d, winLength_);
  return dx;
}

// -----------------------------------------------------------------------------

}  // namespace oops
