/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include <string>

#include "logger/Logger.h"
#include "oops/base/Geometry.h"
#include "util/asser1.h"
#include "util/abor1_cpp.h"
#include "util/dot_product.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/Duration.h"

namespace oops {

// -----------------------------------------------------------------------------
//  Generalized Jb Term of Cost Function
// -----------------------------------------------------------------------------

template<typename TRAIT>
CostJb4D<TRAIT>::CostJb4D(const ControlVariable_ & xb, const Geometry & resol,
		const util::Config & config, const util::Duration &) :
		B_(0), xb_(xb), dxFG_(0), gradFG_(0) {
  util::Config covariances(config, "covariances");

  // Create the StaticCovariance just for build Increments
  util::Config stat(covariances, "static");
  incrementBuilder_.reset(new ModelSpaceError(resol, stat));

  // Create one row of blocks of the whole BMatrix
  unsigned nbg = covariances.getElementSize("covariance");
  ASSERT(xb.nsubwin() == nbg);
  LOG(Debug) << "There are " << nbg << " covariances.";

  for (unsigned jsub = 0; jsub < nbg; ++jsub) {
    const util::Config errConfig(covariances, "covariance", jsub);
    B_.push_back(CovarianceFactory<TRAIT>::create(errConfig, resol));
  }

  LOG(Trace) << "CostJb4D:contructor done";
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
ProcessorBase<typename TRAIT::StateBase> * CostJb4D<TRAIT>::initialize(
		const ControlVariable_ & fg, const Geometry & resol, const util::Config & tlm,
		const bool ltraj) {
  if (ltraj) {
	incrementBuilder_->linearize(fg[0], resol, tlm);
	for (unsigned jsub = 0; jsub < B_.size(); ++jsub) {
	  B_[jsub].linearize(fg[jsub], resol, tlm);
	}
  }

// dxFG_ is  x_0 - x_b
  LOG(Trace) << "CostJb4D:initialize: constructing FG-BG";
  dxFG_.reset(this->newIncrement());
  dxFG_->diff(fg, xb_);

  ProcessorBase<State> * ppb = 0;
  return ppb;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
double CostJb4D<TRAIT>::finalize() {
  dxFG_->prtNorms("dxFG");

// gradFG_ is B^{-1} ( x_0 - x_b ) as estimated by the minimizer
  if (gradFG_ == 0) {
    gradFG_.reset(this->newIncrement());
    gradFG_->zero();
  }

  const double zjb = 0.5 * dot_product(*dxFG_, *gradFG_);
  LOG(Info) << "CostJb4D : Nonlinear Jb = " << zjb;
  return zjb;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
ControlIncrement<TRAIT> * CostJb4D<TRAIT>::finalTL(const ControlIncrement_ &,
                                                   const ControlIncrement_ & bgns) const {
  ControlIncrement_ * dx = new ControlIncrement_(bgns);
  return dx;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
ControlIncrement<TRAIT> * CostJb4D<TRAIT>::setupAD(const ControlIncrement_ &) const {
  ControlIncrement_ * ends = this->newIncrement();
  ends->zeroAD();
  return ends;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
void CostJb4D<TRAIT>::gradientUpdate(const ControlIncrement_ & di, const bool add) const {
  ASSERT(add == true);
  *gradFG_ += di;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
void CostJb4D<TRAIT>::Bmult(const ControlIncrement_ & dxin, ControlIncrement_ & dxout) const {
  unsigned n = xb_.nsubwin();
  for (unsigned k1 = 0; k1 < n; ++k1) {
	dxout[k1].zero();
	Increment dout(dxout[k1]);
	for (unsigned k2 = 0; k2 < n; ++k2) {
		dout.zero();
		B_[k2].mult(dxin[k2], dout);
		dxout[k1] += dout;
	}
  }
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
void CostJb4D<TRAIT>::Bminv(const ControlIncrement_ & dxin, ControlIncrement_ & dxout) const {
  LOG(Info) << "*** WARNING *** B inverse might not always exist ***";
  //for (unsigned jsub = 0; jsub < B_.size(); ++jsub) {
  //  B_->inv_mult(dxin[jsub], dxout[jsub]);
  //}
  LOG(Info) << "*** WARNING *** B inverse might not always exist ***";
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
ControlIncrement<TRAIT> * CostJb4D<TRAIT>::newIncrement() const {
  boost::ptr_vector<Increment> incr4d;
  for (unsigned jsub = 0; jsub < xb_.nsubwin(); ++jsub) {
	Increment * dx = new Increment(incrementBuilder_->resolution(),
  	  incrementBuilder_->ctrlvars(), xb_[jsub].validTime());
    incr4d.push_back(dx);
  }
  ControlIncrement_ * dx = new ControlIncrement_(incr4d, Duration(0));
  return dx;
}

// -----------------------------------------------------------------------------

}  // namespace oops
