/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_ASSIMILATION_COSTJB4D_H_
#define OOPS_ASSIMILATION_COSTJB4D_H_

#include <vector>
#include <boost/ptr_container/ptr_vector.hpp>
#include <boost/scoped_ptr.hpp>

#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/ControlVariable.h"
#include "oops/assimilation/CostJbBase.h"
#include "oops/base/Geometry.h"
#include "oops/base/ModelSpaceCovariance.h"
#include "oops/base/StaticCovariance.h"
#include "oops/base/EnsembleCovariance.h"
#include "oops/base/Ensemble.h"

#include "util/Duration.h"

namespace util {
  class Config;
  class DateTime;
}

namespace oops {

// -----------------------------------------------------------------------------

/// 4D Jb Cost Function
/*!
 * CostJb4D encapsulates the generalized four dimensional Jb term of
 * the 4D-Ens-Var cost function.
 */

template<typename TRAIT> class CostJb4D : public CostJbBase<TRAIT> {
  typedef typename TRAIT::StateBase       State;
  typedef typename TRAIT::IncrementBase   Increment;
  typedef typename TRAIT::ModelStateError ModelSpaceError;
  typedef ModelSpaceCovariance<TRAIT>     ModelSpaceCovariance_;
  typedef ControlVariable<TRAIT>          ControlVariable_;
  typedef ControlIncrement<TRAIT>         ControlIncrement_;
  typedef Ensemble<TRAIT>				  Ensemble_;

 public:
/// Construct \f$ J_b\f$.
  CostJb4D(const ControlVariable_&, const Geometry &, const util::Config &,
           const util::Duration &);

/// Destructor
  virtual ~CostJb4D() {}

/// Initialize before the nonlinear model integration.
  ProcessorBase<State> * initialize(const ControlVariable_&, const Geometry &,
                                    const util::Config &, const bool ltraj = false);

/// Finalize after the nonlinear model integration.
  double finalize();

/// Finalize \f$ J_b\f$ after the TL run.
  ControlIncrement_ * finalTL(const ControlIncrement_ &, const ControlIncrement_ &) const;

/// Initialize \f$ J_b\f$ before the AD run.
  ControlIncrement_ * setupAD(const ControlIncrement_ &) const;

/// Multiply by \f$ B\f$ and \f$ B^{-1}\f$.
  void Bmult(const ControlIncrement_ &, ControlIncrement_ &) const;
  void Bminv(const ControlIncrement_ &, ControlIncrement_ &) const;

/// Create new increment (set to 0).
  ControlIncrement_ * newIncrement() const;

/// Return Jb gradient at first guess.
  const ControlIncrement_ & getGradientFG() const {return *gradFG_;}

/// Return first guess \f$ x_0-x_b\f$.
  const ControlIncrement_ & getFirstGuess() const {return *dxFG_;}

/// Update gradient without using \f$ B^{-1}\f$.
  void gradientUpdate(const ControlIncrement_ &, const bool add = true) const;


 private:
  boost::ptr_vector<ModelSpaceCovariance_> B_;
  
  boost::scoped_ptr<util::Config> confCovar_;
  const ControlVariable_ xb_;

/// A StaticCovariance to create Increments
  boost::scoped_ptr<ModelSpaceError> incrementBuilder_;

/// First guess increment \f$x_0-x_b\f$.
  boost::scoped_ptr<ControlIncrement_> dxFG_;

/// Gradient at first guess \f$ B^{-1}dxFG_ = B^{-1}(x-x_b)\f$.
  boost::scoped_ptr<ControlIncrement_> gradFG_;
};

// -----------------------------------------------------------------------------

}  // namespace oops

#include "oops/assimilation/CostJb4D.cc"

#endif  // OOPS_ASSIMILATION_COSTJB4D_H_
