/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include <string>

#include "logger/Logger.h"
#include "oops/assimilation/ProcessorJq.h"
#include "oops/base/Geometry.h"
#include "util/asser1.h"
#include "util/dot_product.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/Duration.h"

namespace oops {

// -----------------------------------------------------------------------------
//  Generalized Jb Term of Cost Function
// -----------------------------------------------------------------------------

template<typename TRAIT>
CostJbJq<TRAIT>::CostJbJq(const CtrlVar & xb, const Geometry & resol,
                          const util::Config & config, const util::Duration & len)
  : B_(0), xb_(xb[0]), dxFG_(0), gradFG_(0), subwin_(len)
{
  util::Config covariances(config, "covariances");
  unsigned nbg = covariances.getElementSize("covariance");
  ASSERT(xb.nsubwin() == nbg);
  LOG(Debug) << "There are " << nbg << " covariances.";

  for (unsigned jsub = 0; jsub < nbg; ++jsub) {
    const util::Config errConfig(covariances, "covariance", jsub);
    B_.push_back(CovarianceFactory<TRAIT>::create(errConfig, resol));
  }

  LOG(Trace) << "CostJbJq:contructor done";
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
ProcessorBase< typename TRAIT::StateBase > * CostJbJq<TRAIT>::initialize(
                                                   const CtrlVar & fg,
                                                   const Geometry & resol,
                                                   const util::Config & tlm,
                                                   const bool ltraj) {
  if (ltraj) {
    for (unsigned jsub = 0; jsub < B_.size(); ++jsub) {
      B_[jsub].linearize(fg[jsub], resol, tlm);
    }
  }

// dxFG_ will contain x_i - M(x_{i-1})
  dxFG_.reset(this->newIncrement());

// Compute x_0 - x_b for Jb
  (*dxFG_)[0].diff(fg[0], xb_);

// Other components will be computed during the model run
  ProcessorJq<State, CtrlVar, CtrlInc> * ppjq =
    new ProcessorJq<State, CtrlVar, CtrlInc>(fg, *dxFG_, subwin_);
  return ppjq;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
double CostJbJq<TRAIT>::finalize() {
// dxFG_ contains x_i - M(x_{i-1})
  dxFG_->prtNorms("dxFG");

// gradFG_ is D^{-1} ( x_i - M(x_{i-1}) ) as estimated by the minimizer
  if (gradFG_ == 0) {
    gradFG_.reset(this->newIncrement());
    gradFG_->zero();
  }

  const double zjb = 0.5 * dot_product(*dxFG_, *gradFG_);
  LOG(Info) << "CostJbJq : Nonlinear Jb = " << zjb;
  return zjb;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
ControlIncrement<TRAIT> * CostJbJq<TRAIT>::finalTL(const CtrlInc & ends,
                                                   const CtrlInc & bgns) const {
  CtrlInc * dx = new CtrlInc(ends);
  dx->shift_front();
  *dx *= -1.0;
  *dx += bgns;
  return dx;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
ControlIncrement<TRAIT> * CostJbJq<TRAIT>::setupAD(const CtrlInc & dx) const {
  CtrlInc * ends = new CtrlInc(dx);
  *ends *= -1.0;
  ends->shift_back();
  return ends;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
void CostJbJq<TRAIT>::gradientUpdate(const CtrlInc & di, const bool add) const {
  if (add) {
    *gradFG_ += di;
  } else {
    *gradFG_ = di;
  }
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
void CostJbJq<TRAIT>::Bmult(const CtrlInc & dxin, CtrlInc & dxout) const {
  for (unsigned jsub = 0; jsub < B_.size(); ++jsub) {
    B_[jsub].mult(dxin[jsub], dxout[jsub]);

//  For diagnostics and debugging...
    if (logger::Logger::selected(logger::Debug)) {
      double zj = 0.5*dot_product(dxin[jsub], dxout[jsub]);
      if (jsub == 0) {
        LOG(Debug) << "CostJbJq:mult Jb is " << zj;
      } else {
        LOG(Debug) << "CostJbJq:mult Jq(" << jsub << ") is " << zj;
      }
    }
//  For diagnostics and debugging...
  }
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
void CostJbJq<TRAIT>::Bminv(const CtrlInc & dxin, CtrlInc & dxout) const {
  LOG(Info) << "*** WARNING *** B inverse might not always exist ***";
  for (unsigned jsub = 0; jsub < B_.size(); ++jsub) {
    B_[jsub].inv_mult(dxin[jsub], dxout[jsub]);

//  For diagnostics and debugging...
    if (logger::Logger::selected(logger::Debug)) {
      LOG(Debug) << "*** WARNING *** B inverse might not always exist ***";
      double zj = 0.5*dot_product(dxin[jsub], dxout[jsub]);
      if (jsub == 0) {
        LOG(Debug) << "CostJbJq:inv_mult Jb is " << zj;
      } else {
        LOG(Debug) << "CostJbJq:inv_mult Jq(" << jsub << ") is " << zj;
      }
      LOG(Debug) << "*** WARNING *** B inverse might not always exist ***";
    }
//  For diagnostics and debugging...
  }
  LOG(Info) << "*** WARNING *** B inverse might not always exist ***";
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
oops::ControlIncrement<TRAIT> * CostJbJq<TRAIT>::newIncrement() const {
  typedef typename TRAIT::IncrementBase Increment;

  boost::ptr_vector<Increment> incr4d;
  for (unsigned jsub = 0; jsub < B_.size(); ++jsub) {
    incr4d.push_back(B_[jsub].getIncrement());
  }

  CtrlInc * dx = new CtrlInc(incr4d, subwin_);
  return dx;
}

// -----------------------------------------------------------------------------

}  // namespace oops
