/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_ASSIMILATION_COSTJBJQ_H_
#define OOPS_ASSIMILATION_COSTJBJQ_H_

#include <boost/ptr_container/ptr_vector.hpp>
#include <boost/scoped_ptr.hpp>

#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/ControlVariable.h"
#include "oops/assimilation/CostJbBase.h"
#include "oops/base/Geometry.h"
#include "oops/base/ModelSpaceCovariance.h"
#include "util/Duration.h"

namespace util {
  class Config;
  class DateTime;
}

namespace oops {

// -----------------------------------------------------------------------------

/// Jb + Jq Cost Function
/*!
 * CostJbJq encapsulates the generalized Jb term of the cost weak
 * constraint 4D-Var function (ie Jb+Jq).
 */

template<typename TRAIT> class CostJbJq : public CostJbBase<TRAIT> {
  typedef typename TRAIT::StateBase     State;
  typedef ModelSpaceCovariance<TRAIT>   StateError;
  typedef ControlVariable<TRAIT>        CtrlVar;
  typedef ControlIncrement<TRAIT>       CtrlInc;

 public:
/// Construct \f$ J_b\f$.
  CostJbJq(const CtrlVar &, const Geometry &, const util::Config &,
           const util::Duration &);

/// Destructor
  virtual ~CostJbJq() {}

/// Initialize before the nonlinear model integration.
  ProcessorBase<State> * initialize(const CtrlVar &, const Geometry &,
                                    const util::Config &, const bool ltraj = false);

/// Finalize after the nonlinear model integration.
  double finalize();

/// Finalize \f$ J_b\f$ after the TL run.
  CtrlInc * finalTL(const CtrlInc &, const CtrlInc &) const;

/// Initialize \f$ J_b\f$ before the AD run.
  CtrlInc * setupAD(const CtrlInc &) const;

/// Multiply by \f$ B\f$ and \f$ B^{-1}\f$.
  void Bmult(const CtrlInc &, CtrlInc &) const;
  void Bminv(const CtrlInc &, CtrlInc &) const;

/// Create new increment (set to 0).
  CtrlInc * newIncrement() const;

/// Return Jb gradient at first guess.
  const CtrlInc & getGradientFG() const {return *gradFG_;}

/// Return first guess \f$ x_0-x_b\f$.
  const CtrlInc & getFirstGuess() const {return *dxFG_;}

/// Update gradient without using \f$ B^{-1}\f$.
  void gradientUpdate(const CtrlInc &, const bool add = true) const;

 private:
  boost::ptr_vector<StateError> B_;
  const State xb_;

/// First guess increment \f$x_0-x_b\f$ or more generally \f$ x_i-M(x_{i-1})\f$.
  boost::scoped_ptr<CtrlInc> dxFG_;

/// Gradient at first guess \f$ B^{-1}dxFG_ = B^{-1}(x_i-M(x_{i-1}))\f$.
  boost::scoped_ptr<CtrlInc> gradFG_;

  const util::Duration subwin_;
};

// -----------------------------------------------------------------------------

}  // namespace oops

#include "oops/assimilation/CostJbJq.cc"

#endif  // OOPS_ASSIMILATION_COSTJBJQ_H_
