/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include <boost/pointer_cast.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/shared_ptr.hpp>

#include "logger/Logger.h"
#include "util/Config.h"
#include "util/asser1.h"

namespace oops {

// -----------------------------------------------------------------------------

template<typename TRAIT>
CostJcDFI<TRAIT>::CostJcDFI(const Geometry & resol, const util::Config & conf,
                            const util::DateTime & vt, const util::Duration & span,
                            const util::Duration & tstep)
  : conf_(conf), vt_(vt), span_(span), alpha_(0), wfct_(), gradFG_(),
    resol_(resol), tstep_(tstep), tlres_(), tlstep_(), filter_(0), ltraj_(false)
{
  alpha_ = conf.getReal("alpha");
  if (conf.elementExists("ftime")) vt_ = util::DateTime(conf.getData("ftime"));
  if (conf.elementExists("span")) span_ = util::Duration(conf.getData("span"));
//  wfct_.reset(WeightFactory::create(config)); YT
  wfct_.reset(new DolphChebyshev(conf_));
  LOG(Debug) << "CostJcDFI created vt = " << vt_ << ", span = " << span_;
  LOG(Trace) << "CostJcDFI created";
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
ProcessorBase< typename TRAIT::StateBase > *
  CostJcDFI<TRAIT>::initialize(const CtrlVar &,
                               const util::Config & innerConf, const bool ltraj) {
  ltraj_ = ltraj;
  if (ltraj_) {
    const util::Config resConf(innerConf, "resolution");
    tlres_.reset(new Geometry(resConf));
    tlstep_ = util::Duration(innerConf.getData("tlmodel/tstep"));
  }

  filter_ = new WeightedDiff<Increment, State>(vt_, span_, resol_, conf_, tstep_, *wfct_);
  filter_->setup();
  return filter_;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
double CostJcDFI<TRAIT>::finalize() {
  double zz = 0.5 * alpha_;
  if (ltraj_) {
    gradFG_.reset(filter_->release());
    zz *= dot_product(*gradFG_, *gradFG_);
    *gradFG_ *= alpha_;
    ltraj_ = false;
  } else {
    boost::scoped_ptr<Increment> dx(filter_->release());
    zz *= dot_product(*dx, *dx);
  }
  LOG(Info) << "CostJcDFI: Nonlinear Jc = " << zz;
  return zz;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
typename TRAIT::IncrementBase * CostJcDFI<TRAIT>::newDualVector() const {
  oops::Variables vars(conf_);
  Increment * dx = new Increment(*tlres_, vars, vt_);
  return dx;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
WeightedDiffTLAD< typename TRAIT::IncrementBase > *
  CostJcDFI<TRAIT>::setupTL(const CtrlInc &) const
{
  WeightedDiffTLAD<Increment> * filterTLAD =
    new WeightedDiffTLAD<Increment>(vt_, span_, *tlres_, conf_, tlstep_, *wfct_);
  filterTLAD->setupTL();
  return filterTLAD;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
WeightedDiffTLAD< typename TRAIT::IncrementBase > *
CostJcDFI<TRAIT>::setupAD(boost::shared_ptr<const GeneralizedDeparture> pv) const
{
  boost::shared_ptr<const Increment>
    dx = boost::dynamic_pointer_cast<const Increment>(pv);
  WeightedDiffTLAD<Increment> * filterTLAD =
    new WeightedDiffTLAD<Increment>(vt_, span_, *tlres_, conf_, tlstep_, *wfct_);
  filterTLAD->setupAD(dx);
  return filterTLAD;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
typename TRAIT::IncrementBase *
CostJcDFI<TRAIT>::CovarMult(const GeneralizedDeparture & dv1) const {
  const Increment & dx1 = dynamic_cast<const Increment &>(dv1);
  Increment * dx2 = new Increment(dx1);
  const double za = 1.0/alpha_;
  *dx2 *= za;
  return dx2;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
typename TRAIT::IncrementBase *
CostJcDFI<TRAIT>::CovarMinv(const GeneralizedDeparture & dv1) const {
  const Increment & dx1 = dynamic_cast<const Increment &>(dv1);
  Increment * dx2 = new Increment(dx1);
  *dx2 *= alpha_;
  return dx2;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
void CostJcDFI<TRAIT>::reset() {
  gradFG_.reset();
}

// -----------------------------------------------------------------------------

}  // namespace oops
