/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_ASSIMILATION_COSTJCDFI_H_
#define OOPS_ASSIMILATION_COSTJCDFI_H_

#include <boost/scoped_ptr.hpp>
#include <boost/shared_ptr.hpp>

#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/ControlVariable.h"
#include "oops/assimilation/CostJcBase.h"
#include "oops/base/ProcessorBase.h"
#include "oops/base/Variables.h"
#include "oops/base/WeightedDiff.h"
#include "oops/base/WeightedDiffTLAD.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/DolphChebyshev.h"
#include "util/WeightingFct.h"

namespace oops {
  class Geometry;

// -----------------------------------------------------------------------------

/// Jc DFI Cost Function
/*!
 * Digital filter based constraint term for the cost function.
 */

template<typename TRAIT> class CostJcDFI : public CostJcBase<TRAIT> {
  typedef typename TRAIT::StateBase      State;
  typedef typename TRAIT::IncrementBase  Increment;
  typedef ControlVariable<TRAIT>         CtrlVar;
  typedef ControlIncrement<TRAIT>        CtrlInc;

 public:
/// Construct \f$ J_c\f$.
  CostJcDFI(const Geometry &, const util::Config &, const util::DateTime &,
            const util::Duration &, const util::Duration & tstep = util::Duration(0));

/// Destructor
  virtual ~CostJcDFI() {}

/// Initialize before nonlinear model integration.
  ProcessorBase<State> * initialize(const CtrlVar &,
                                    const util::Config &, const bool ltraj = false);

/// Finalize computation after nonlinear model integration.
  double finalize();

/// Initialize \f$ J_c\f$ before starting the TL run.
  WeightedDiffTLAD<Increment> * setupTL(const CtrlInc &) const;

/// Initialize \f$ J_c\f$ before starting the AD run.
  WeightedDiffTLAD<Increment> *
    setupAD(boost::shared_ptr<const GeneralizedDeparture>) const;

/// Multiply by \f$ C\f$ and \f$ C^{-1}\f$.
  Increment * CovarMult(const GeneralizedDeparture &) const;
  Increment * CovarMinv(const GeneralizedDeparture &) const;

/// Provide new increment.
  Increment * newDualVector() const;

/// Gradient of \f$ J_c\f$ at first guess.
  Increment * getGradientFG() const {return new Increment(*gradFG_);}

/// Reset trajectory.
  void reset();

 private:
  const util::Config conf_;
  util::DateTime vt_;
  util::Duration span_;
  double alpha_;
  boost::scoped_ptr<WeightingFct> wfct_;
  boost::scoped_ptr<Increment> gradFG_;
  const Geometry resol_;
  const util::Duration tstep_;
  boost::scoped_ptr<Geometry> tlres_;
  util::Duration tlstep_;
  mutable WeightedDiff<Increment, State> * filter_;  // Not safe enough. YT
  bool ltraj_;
};

// -----------------------------------------------------------------------------

}  // namespace oops

#include "oops/assimilation/CostJcDFI.cc"

#endif  // OOPS_ASSIMILATION_COSTJCDFI_H_
