/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include <boost/pointer_cast.hpp>
#include <boost/shared_ptr.hpp>

#include "logger/Logger.h"
#include "util/Config.h"

namespace oops {

// -----------------------------------------------------------------------------

template<typename TRAIT>
CostJo<TRAIT>::CostJo(const Observ & yobs, const util::Config & joConf,
                      const util::Duration & ts, const bool subwindows)
  : yobs_(yobs), R_(yobs, joConf), gradFG_(yobs), subwindows_(subwindows),
    ltraj_(false), traj_(0), conf_(0), yeqv_(0), tslot_(ts)
{
  LOG(Debug) << "CostJo:setup tslot_ = " << tslot_;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
ProcessorBase< typename TRAIT::StateBase > * CostJo<TRAIT>::initialize(
  const CtrlVar & xx, const util::Config & conf, const bool ltraj)
{
  ltraj_ = ltraj;
  conf_ = &conf;
  yeqv_.reset(new Observ(yobs_));
  pobs_ = new Observer<State, Observ, ObsAuxCtrl, ObsTraj>
                (*yeqv_, xx.obsVar(), tslot_, subwindows_);
  if (ltraj_) pobs_->setTraj();
  return pobs_;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
double CostJo<TRAIT>::finalize() {
  LOG(Info) << "Jo Observation Equivalent:" << *yeqv_;
  R_.linearize(*yeqv_);

  Depart ydep(*yeqv_, yobs_);
  LOG(Info) << "Jo Departures:" << ydep;

  const double zz = R_.calcJo(ydep, gradFG_);
  LOG(Info) << "CostJo   : Nonlinear Jo = " << zz;

  if (conf_->elementExists("diagnostics")) {
    const util::Config joConf(*conf_, "diagnostics");
    std::string fname = joConf.getData("departures");
    if (!fname.empty()) ydep.write(fname);
  }

  yeqv_.reset();
  conf_ = 0;
  if (ltraj_) {
    traj_.reset(pobs_->releaseTraj());
    ltraj_ = false;
  }
  return zz;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
ObserverTLAD< typename TRAIT::IncrementBase, TRAIT > * CostJo<TRAIT>::setupTL(
                                                         const CtrlInc & dx) const {
  ObserverTLAD<Increment, TRAIT> * pobs = new
    ObserverTLAD<Increment, TRAIT>(*traj_, tslot_, subwindows_);

  Depart * ydep = this->newDualVector();
  pobs->setupTL(ydep, dx.obsVar());

  return pobs;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
ObserverTLAD< typename TRAIT::IncrementBase, TRAIT > *
  CostJo<TRAIT>::setupAD(boost::shared_ptr<const GeneralizedDeparture> pv) const
{
  ObserverTLAD<Increment, TRAIT> * pobs = new
    ObserverTLAD<Increment, TRAIT>(*traj_, tslot_, subwindows_);

  boost::shared_ptr<const Depart> dy = boost::dynamic_pointer_cast<const Depart>(pv);
  ObsAuxIncr * db = 0;
  pobs->setupAD(dy, db);

  return pobs;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
Departure<TRAIT> * CostJo<TRAIT>::CovarMult(const GeneralizedDeparture & v1) const {
  const Depart & y1 = dynamic_cast<const Depart &>(v1);
  Depart * y2 = new Depart(gradFG_, false);
  R_.mult(y1, *y2);
  return y2;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
Departure<TRAIT> * CostJo<TRAIT>::CovarMinv(const GeneralizedDeparture & v1) const {
  const Depart & y1 = dynamic_cast<const Depart &>(v1);
  Depart * y2 = new Depart(gradFG_, false);
  R_.inv_mult(y1, *y2);
  return y2;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
Departure<TRAIT> * CostJo<TRAIT>::newDualVector() const {
  Depart * ydep = new Depart(gradFG_, false);
  ydep->zero();
  return ydep;
}

// -----------------------------------------------------------------------------

template<typename TRAIT>
void CostJo<TRAIT>::reset() {
  traj_.reset();
}

// -----------------------------------------------------------------------------

}  // namespace oops
