/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_ASSIMILATION_COSTJO_H_
#define OOPS_ASSIMILATION_COSTJO_H_

#include <boost/noncopyable.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/shared_ptr.hpp>

#include "oops/assimilation/ControlVariable.h"
#include "oops/assimilation/ControlIncrement.h"
#include "oops/base/CostBase.h"
#include "oops/base/CostProcessor.h"
#include "oops/base/Departure.h"
#include "oops/base/Observation.h"
#include "oops/base/ObsErrorCovariance.h"
#include "oops/base/Observer.h"
#include "oops/base/ObserverTLAD.h"
#include "oops/base/ObsTrajectory.h"
#include "oops/base/PostProcessor.h"
#include "oops/base/ProcessorBase.h"
#include "util/DateTime.h"
#include "util/Duration.h"

namespace util {
  class Config;
}

namespace oops {

// -----------------------------------------------------------------------------

/// Jo Cost Function
/*!
 * The CostJo class encapsulates the Jo term of the cost function.
 * The Observer to be called during the model integration is managed
 * inside the CostJo class.
 */

template<typename TRAIT> class CostJo : public CostTermBase<TRAIT>,
                                        private boost::noncopyable {
  typedef typename TRAIT::ObsAuxCtrl     ObsAuxCtrl;
  typedef typename TRAIT::ObsAuxIncr     ObsAuxIncr;
  typedef typename TRAIT::ObsTraj        ObsTraject;
  typedef typename TRAIT::StateBase      State;
  typedef typename TRAIT::IncrementBase  Increment;
  typedef ControlVariable<TRAIT>         CtrlVar;
  typedef ControlIncrement<TRAIT>        CtrlInc;
  typedef Observation<TRAIT>             Observ;
  typedef Departure<TRAIT>               Depart;
  typedef ObsErrorCovariance<TRAIT>      ObsError;
  typedef ObsTrajectory<ObsTraject>      ObsTraj;

 public:
  /// Construct \f$ J_o\f$ from \f$ R\f$ and \f$ y_{obs}\f$.
  CostJo(const Observ &, const util::Config &, const util::Duration &,
         const bool subwindows = false);

  /// Destructor
  virtual ~CostJo() {}

  /// Initialize \f$ J_o\f$ before starting the integration of the model.
  ProcessorBase<State> * initialize(const CtrlVar &,
                                    const util::Config &, const bool ltraj = false);
  /// Finalize \f$ J_o\f$ after the integration of the model.
  double finalize();

  /// Initialize \f$ J_o\f$ before starting the TL run.
  ObserverTLAD<Increment, TRAIT> * setupTL(const CtrlInc &) const;

  /// Initialize \f$ J_o\f$ before starting the AD run.
  ObserverTLAD<Increment, TRAIT> *
    setupAD(boost::shared_ptr<const GeneralizedDeparture>) const;

  /// Multiply by \f$ R\f$ and \f$ R^{-1}\f$.
  Depart * CovarMult(const GeneralizedDeparture &) const;
  Depart * CovarMinv(const GeneralizedDeparture &) const;

  /// Provide new departure.
  Depart * newDualVector() const;

  /// Return gradient at first guess ie \f$ R^{-1} {\cal H}(x^t ) - y\f$.
  Depart * getGradientFG() const {return new Depart(gradFG_);}

  /// Reset obs operator trajectory.
  void reset();

 private:
  const Observ & yobs_;
  ObsError R_;

  /// Gradient at first guess : \f$ R^{-1} (H(x_{fg})-y_{obs}) \f$.
  Depart gradFG_;

  /// Trajectory for observation operators.
  const bool subwindows_;
  bool ltraj_;
  boost::scoped_ptr<ObsTraj> traj_;

  /// Observer passed by \f$ J_o\f$ to the model during integration.
  Observer<State, Observ, ObsAuxCtrl, ObsTraj> * pobs_;  // Not safe enough. YT
  const util::Config * conf_;
  boost::scoped_ptr<Observ> yeqv_;

  /// Time slot.
  const util::Duration tslot_;
};

// -----------------------------------------------------------------------------

}  // namespace oops

#include "oops/assimilation/CostJo.cc"

#endif  // OOPS_ASSIMILATION_COSTJO_H_
