/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_ASSIMILATION_DGMRESR_H_
#define OOPS_ASSIMILATION_DGMRESR_H_

#include <cmath>
#include <iostream>
#include <vector>

#include "logger/Logger.h"
#include "util/dot_product.h"

namespace oops {

/*! \file DGMRESR.h
 * \brief Double GMRESR solver for Ax=b.
 *
 * DGMRESR solver for Ax=b.  This is a "double" version of GMRESR
 * (H.A. Van der Vorst  and C. Vuik, 1994, Numerical Linear Algebra with
 * Applications, 1(4), 369-386) following Derber and Rosati (J. Derber and
 * A. Rosati, 1989, J. Phys. Oceanog. 1333-1347).
 * It solves Ax=b for the particular case \f$ A=B^{-1}+C\f$, without
 * requiring the application of \f$ B^{-1}\f$. A must be square, but need
 * not be symmetric. (For a symmetric matrix, and constant preconditioner,
 * DGMRESR is simply Double PCG with full orthogonalisation.)
 * A preconditioner must be supplied that, given a vector q, returns an
 * approximate solution of \f$ (AB)^{-1} q\f$. The preconditioner can be
 * variable. Note that the traditional \f$ B\f$-preconditioning corresponds
 * to precond=\f$I\f$.
 *
 *  On entry:
 * -    x       =  starting point, \f$ X_0 \f$.
 * -    xh      = \f$ B^{-1} x_0\f$.
 * -    B       = \f$ B \f$.
 * -    C       = \f$ C \f$.
 * -    precond = preconditioner matrix \f$ F_k \approx (AB)^{-1} \f$.
 *
 * On exit, x will contain the solution \f$ x \f$ and xh will contain
 * \f$ B^{-1} x\f$.
 *
 * VECTOR must implement:
 *  - dot_product
 *  - operator(=)
 *  - operator(+=),
 *  - operator(-=)
 *  - operator(*=) [double * VECTOR],
 *  - axpy
 *
 *  BMATRIX, CMATRIX and PMATRIX must implement a method:
 *  - void multiply(const VECTOR&, VECTOR&) const
 *
 *  which applies the matrix to the first argument, and returns the
 *  matrix-vector product in the second. (Note: the const is optional, but
 *  recommended.)
 *
 *  Iteration will stop if the maximum iteration limit "maxiter" is reached
 *  or if the residual norm reduces by a factor of "tolerance".
 *
 *  The return value is the achieved reduction in residual norm.
 */

template <typename VECTOR, typename BMATRIX, typename CMATRIX, typename PMATRIX>
double DGMRESR(VECTOR & x, VECTOR & xh, const VECTOR & b,
              const BMATRIX & B, const CMATRIX & C, const PMATRIX & precond,
              const int maxiter, const double tolerance ) {
  std::vector<VECTOR> c;
  std::vector<VECTOR> u;
  std::vector<VECTOR> uh;
  VECTOR cc(x);
  VECTOR z(x);
  VECTOR zh(x);
  VECTOR r(x);

  C.multiply(x, z);  // z=Ax
  r = b;
  r -= xh;
  r -= z;

  double dotRr0  = dot_product(r, r);
  double normReduction = 1.0;

  for (int jiter = 0; jiter < maxiter; ++jiter) {
    LOG(Trace) << " DGMRESR Starting Iteration " << jiter+1;

    precond.multiply(r, zh);  // returns zh as approxmate solve of (BA)zh = r
    B.multiply(zh, z);  // x=B zh

    C.multiply(z, cc);
    cc += zh;      // cc = zh+Cz = Az

    for (int k = 0; k < jiter; ++k) {
      double alpha = -dot_product(c[k], cc);
      cc.axpy(alpha, c[k]);   // cc = cc - <c[k],cc>*c[k];
      z.axpy(alpha, u[k]);    // z  = z  - <c[k],cc>*u[k];
      zh.axpy(alpha, uh[k]);  // zh = zh - <c[k],cc>*uh[k];
    }

    double ccnorm = sqrt(dot_product(cc, cc));
    c.push_back(cc);         // c[jiter]=cc
    c[jiter] *= 1.0/ccnorm;
    u.push_back(z);          // u[jiter]=z
    u[jiter] *= 1.0/ccnorm;
    uh.push_back(zh);          // uh[jiter]=zh
    uh[jiter] *= 1.0/ccnorm;

    double cdotr = dot_product(c[jiter], r);
    xh.axpy(cdotr, uh[jiter]);
    r.axpy(-cdotr, c[jiter]);

    normReduction = sqrt(dot_product(r, r)/dotRr0);
    LOG(Info) << "DGMRESR end of iteration " << jiter+1
                      << ". Norm reduction= " << normReduction;

    if (normReduction < tolerance) {
      LOG(Info) << "DGMRESR: Achieved required reduction in residual norm.";
      break;
    }
  }

  B.multiply(xh, x);

  LOG(Trace) << "DGMRESR: end";

  return normReduction;
}

}  // namespace oops

#endif  // OOPS_ASSIMILATION_DGMRESR_H_
