/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_ASSIMILATION_DIPCG_H_
#define OOPS_ASSIMILATION_DIPCG_H_

#include <cmath>
#include <iostream>

#include "logger/Logger.h"
#include "util/dot_product.h"

namespace oops {

/*! \file DIPCG.h
 * \brief Double Inexact-Preconditioned Conjugate Gradients solver.
 *
 * This solver is based on the Golub-Ye Inexact-Preconditioned Conjugate
 * Gradients solver for Ax=b (G.H. Golub and Q. Ye 1999/00,
 * SIAM J. Sci. Comput. 21(4) 1305-1320), and on the Derber Rosati-Double
 * PCG algorithm (J. Derber and A. Rosati, 1989, J. Phys. Oceanog. 1333-1347).
 * It solves \f$ Ax=b\f$ for the particular case \f$ A=B^{-1}+C\f$,
 * without requiring the application of \f$ B^{-1}\f$.
 *
 * A must be square, symmetric, positive definite.
 *
 * A preconditioner must be supplied that, given a vector q, returns an
 * approximation to \f$ (AB)^{-1} q\f$. The preconditioner can be variable.
 * Note that the traditional \f$ B\f$-preconditioning corresponds to
 * precond=\f$I\f$.
 *
 * On entry:
 * -    x       =  starting point, \f$ X_0 \f$.
 * -    xh      = \f$ B^{-1} x_0\f$.
 * -    b       = right hand side.
 * -    B       = \f$ B \f$.
 * -    C       = \f$ C \f$.
 * -    precond = preconditioner \f$ F_k \approx (AB)^{-1} \f$.
 *
 * On exit, x will contain the solution \f$ x \f$ and xh will contain
 * \f$ B^{-1} x\f$.
 *  The return value is the achieved reduction in residual norm.
 *
 *  Iteration will stop if the maximum iteration limit "maxiter" is reached
 *  or if the residual norm reduces by a factor of "tolerance".
 *
 *  VECTOR must implement:
 *  - dot_product
 *  - operator(=)
 *  - operator(+=),
 *  - operator(-=)
 *  - operator(*=) [double * VECTOR],
 *  - axpy
 *
 *  Each of BMATRIX, CMATRIX amd PMATRIX must implement a method:
 *  - void multiply(const VECTOR&, VECTOR&) const
 *
 *  which applies the matrix to the first argument, and returns the
 *  matrix-vector product in the second. (Note: the const is optonal, but
 *  recommended.)
 */

template <typename VECTOR, typename BMATRIX, typename CMATRIX, typename PMATRIX>
double DIPCG(VECTOR & x, VECTOR & xh, const VECTOR & b,
             const BMATRIX & B, const CMATRIX & C, const PMATRIX & precond,
             const int maxiter, const double tolerance ) {
  VECTOR ap(x);
  VECTOR p(x);
  VECTOR ph(x);
  VECTOR r(x);
  VECTOR s(x);
  VECTOR sh(x);
  VECTOR oldr(x);
  VECTOR w(x);

  C.multiply(x, s);
  r = b;
  r -= xh;
  r -= s;

  double dotRr0  = dot_product(r, r);
  double normReduction = 1.0;
  double rdots;

  for (int jiter = 0; jiter < maxiter; ++jiter) {
    LOG(Info) << " DIPCG Starting Iteration " << jiter+1;

    precond.multiply(r, sh);
    B.multiply(sh, s);

    if (jiter == 0) {
      p =s;
      ph = sh;
    } else {
      w = r;
      w -= oldr;    // w=r-oldr
      double beta = dot_product(s, w)/rdots;

      p *= beta;
      p += s;      // p = s + beta*p

      ph *= beta;
      ph += sh;    // ph = sh + beta*ph
    }

    C.multiply(p, ap);
    ap += ph;

    oldr = r;
    rdots = dot_product(r, s);

    double alpha = rdots/dot_product(p, ap);

    xh.axpy(alpha, ph);  // xh = xh + alpha*ph
    r.axpy(-alpha, ap);  // r = r - alpha*ap

    normReduction = sqrt(dot_product(r, r)/dotRr0);
    LOG(Info) << "DIPCG end of iteration " << jiter+1
              << ". Norm reduction= " << normReduction;

    if (normReduction < tolerance) {
      LOG(Info) << "DIPCG: Achieved required reduction in residual norm.";
      break;
    }
  }

  B.multiply(xh, x);
  LOG(Info) << "DIPCG: Estimated Final Jb = " << 0.5 * dot_product(xh, x);

  LOG(Trace) << "DIPCG: end";

  return normReduction;
}

}  // namespace oops

#endif  // OOPS_ASSIMILATION_DIPCG_H_
