/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_ASSIMILATION_DPCG_H_
#define OOPS_ASSIMILATION_DPCG_H_

#include <cmath>
#include <iostream>

#include "logger/Logger.h"
#include "util/dot_product.h"

namespace oops {

/*! \file DPCG.h
 * \brief Double Preconditioned Conjugate Gradients solver.
 *
 * This solver is based on the standard Preconditioned Conjugate
 * Gradients solver for Ax=b (G. H. Golub and C. F. Van Loan, Matrix
 * Computations), and on the Derber Rosati-Double
 * PCG algorithm (J. Derber and A. Rosati, 1989, J. Phys. Oceanog. 1333-1347).
 * For details see S. Gurol, PhD Manuscript, 2013.
 * It solves \f$ Ax=b\f$ for the particular case \f$ A=B^{-1}+C\f$,
 * without requiring the application of \f$ B^{-1}\f$. This algorithm
 * is similar to DIPCG except it includes standard PCG instead IPCG
 * and stopping criteria is based on the preconditioner norm.
 *
 * A must be square, symmetric, positive definite.
 *
 * A preconditioner must be supplied that, given a vector q, returns an
 * approximation to \f$ (AB)^{-1} q\f$. Possible preconditioning
 * is detailed in S. Gurol, PhD Manuscript, 2013.
 * Note that the traditional \f$ B\f$-preconditioning corresponds to
 * precond=\f$I\f$.
 *
 * On entry:
 * -    x       =  starting point, \f$ X_0 \f$.
 * -    xh      = \f$ B^{-1} x_0\f$.
 * -    b       = right hand side.
 * -    B       = \f$ B \f$.
 * -    C       = \f$ C \f$.
 * -    precond = preconditioner \f$ F_k \approx (AB)^{-1} \f$.
 *
 * On exit, x will contain the solution \f$ x \f$ and xh will contain
 * \f$ B^{-1} x\f$.
 *  The return value is the achieved reduction in residual norm.
 *
 *  Iteration will stop if the maximum iteration limit "maxiter" is reached
 *  or if the residual norm reduces by a factor of "tolerance".
 *
 *  VECTOR must implement:
 *  - dot_product
 *  - operator(=)
 *  - operator(+=),
 *  - operator(-=)
 *  - operator(*=) [double * VECTOR],
 *  - axpy
 *
 *  Each of BMATRIX, CMATRIX amd PMATRIX must implement a method:
 *  - void multiply(const VECTOR&, VECTOR&) const
 *
 *  which applies the matrix to the first argument, and returns the
 *  matrix-vector product in the second. (Note: the const is optonal, but
 *  recommended.)
 */

template <typename VECTOR, typename BMATRIX, typename CMATRIX, typename PMATRIX>
double DPCG (VECTOR & x, VECTOR & xh, const VECTOR & b,
             const BMATRIX & B, const CMATRIX & C, const PMATRIX & precond,
             const int maxiter, const double tolerance ) {
  VECTOR ap(x);
  VECTOR p(x);
  VECTOR ph(x);
  VECTOR r(x);
  VECTOR s(x);
  VECTOR sh(x);
  VECTOR w(x);
  VECTOR v(x);
  VECTOR z(x);

  std::vector<VECTOR> vVEC;  //required for re-orthogonalization
  std::vector<VECTOR> zVEC;  //required for re-orthogonalization

  // Initial residual r = b - Ax
  C.multiply(x, s);
  r = b;
  r -= xh;
  r -= s;

  // s = B precond r
  precond.multiply(r, sh);
  B.multiply(sh, s);

  double dotRr0  = dot_product(r, s);
  double normReduction;
  double rdots = dotRr0;
  double rdots_old;
  
  v = r;
  v  *= 1/sqrt(dotRr0);
  z = s;
  z  *= 1/sqrt(dotRr0);

  vVEC.push_back(v);
  zVEC.push_back(z);  

  for (int jiter = 0; jiter < maxiter; ++jiter) {
    LOG(Info) << " DPCG Starting Iteration " << jiter+1;

    if (jiter == 0) {
      p  = s;
      ph = sh;
    } else {
      w = r;
      
      double beta = dot_product(s, w)/rdots_old;
      LOG(Debug) << "DPCG beta = " << beta;

      p *= beta;
      p += s;      // p = s + beta*p

      ph *= beta;
      ph += sh;    // ph = sh + beta*ph
    }

    //  (Binv + HtRinvH) p
    C.multiply(p, ap);
    ap += ph;

    double alpha = rdots/dot_product(p, ap);

    xh.axpy(alpha, ph);  // xh = xh + alpha*ph
    r.axpy(-alpha, ap);  // r = r - alpha*ap
    

    //Re-orthogonalization
    for (int iiter = 0; iiter < jiter; ++iiter) {
      double proj = dot_product(r, zVEC[iiter]);
      r.axpy(-proj, vVEC[iiter]);
    }
   
    precond.multiply(r, sh);
    B.multiply(sh, s);  // s = B precond r

    rdots_old = rdots;
    rdots = dot_product(r, s);

    v = r;
    v  *= 1/sqrt(rdots);
    z = s;
    z  *= 1/sqrt(rdots);
    vVEC.push_back(v);            
    zVEC.push_back(z);

    normReduction = sqrt(rdots/dotRr0);

    LOG(Info) << "DPCG end of iteration " << jiter+1
              << ". Norm reduction= " << normReduction;

    if (normReduction < tolerance) {
      LOG(Info) << "DPCG: Achieved required reduction in residual norm.";
      break;
    }
  }

  B.multiply(xh, x);
  LOG(Info) << "DPCG: Estimated Final Jb = " << 0.5 * dot_product(xh, x);

  LOG(Trace) << "DPCG: end";

  return normReduction;
}

}  // namespace oops

#endif  // OOPS_ASSIMILATION_DPCG_H_
