/*
* Copyright 2011 ECMWF
*
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_ASSIMILATION_DPLANCZOS_H_
#define OOPS_ASSIMILATION_DPLANCZOS_H_

#include <cmath>
#include <iostream>
#include <vector>

#include <boost/ptr_container/ptr_vector.hpp>

#include "logger/Logger.h"
#include "util/dot_product.h"

#include "oops/assimilation/TriDiagSolve.h"


namespace oops {

/*! \file DPLanczos.h
 * \brief Preconditioned Lanczos Inverse Free solver.
 *
 * This solver is the Lanczos version of the PCGIF algorithm
 * It solves \f$ Ax=b\f$ for the particular case \f$ A=B^{-1}+C\f$,
 * without requiring the application of \f$ B^{-1}\f$.
 *
 * A must be square, symmetric, positive definite.
 *
 * A preconditioner must be supplied that, given a vector q, returns an
 * approximation to \f$ (AB)^{-1} q\f$. Possible preconditioning
 * is detailed in S. Gurol, PhD Manuscript, 2013.
 * Note that the traditional \f$ B\f$-preconditioning corresponds to
 * precond=\f$I\f$.
 *
 * On entry:
 * -    x       =  starting point, \f$ X_0 \f$.
 * -    xh      = \f$ B^{-1} x_0\f$.
 * -    b       = right hand side.
 * -    B       = \f$ B \f$.
 * -    C       = \f$ C \f$.
 * -    precond = preconditioner \f$ F_k \approx (AB)^{-1} \f$.
 *
 * On exit, x will contain the solution \f$ x \f$ and xh will contain
 * \f$ B^{-1} x\f$.
 *  The return value is the achieved reduction in residual norm.
 *
 *  Iteration will stop if the maximum iteration limit "maxiter" is reached
 *  or if the residual norm reduces by a factor of "tolerance".
 *
 *  VECTOR must implement:
 *  - dot_product
 *  - operator(=)
 *  - operator(+=),
 *  - operator(-=)
 *  - operator(*=) [double * VECTOR],
 *  - axpy
 *
 *  Each of BMATRIX, CMATRIX amd PMATRIX must implement a method:
 *  - void multiply(const VECTOR&, VECTOR&) const
 *
 *  which applies the matrix to the first argument, and returns the
 *  matrix-vector product in the second. (Note: the const is optonal, but
 *  recommended.)
 */

template <typename VECTOR, typename BMATRIX, typename CMATRIX, typename PMATRIX>
double DPLanczos(VECTOR & x, VECTOR & xh, const VECTOR & b,
                 const BMATRIX & B, const CMATRIX & C, const PMATRIX & precond,
                 const int maxiter, const double tolerance,
                 std::vector<VECTOR> & zhVEC, std::vector<VECTOR> & zVEC,
                 std::vector<double> & alphas, std::vector<double> & betas) {
  VECTOR r(x);
  VECTOR z(x);
  VECTOR zh(x);
  VECTOR w(x);
  VECTOR v(x);
  VECTOR vold(x);

  std::vector<VECTOR> vVEC;
  std::vector<double> y;
  std::vector<double> d;

  // Initial residual r = b - Ax
  C.multiply(x, z);
  r = b;
  r -= xh;
  r -= z;
  
  // z = B precond r
  precond.multiply(r, zh);
  B.multiply(zh, z);

  double normReduction;
  double beta0 = sqrt(dot_product(r, z));
  double beta = 0.0;

  vold = x;
  vold *= 0;
  v  = r;
  v  *= 1/beta0;
  zh *= 1/beta0;
  z  *= 1/beta0;

  zVEC.clear();
  zhVEC.clear();
  betas.clear();
  alphas.clear();

  zVEC.push_back(z);     // zVEC[0] = z_1 ---> required for re-orthogonalization
  zhVEC.push_back(zh);   // zhVEC[0] = zh_1 --> required for solution
  vVEC.push_back(v);     // vVEC[0] = v_1 ---> required for re-orthogonalization
  
  int jiter;
  for (jiter = 0; jiter < maxiter; ++jiter) {
    LOG(Info) << " DPLanczos Starting Iteration " << jiter+1;

    // w = A z - beta * vold
    C.multiply(z, w);
    w += zh;              // w = A z
    w.axpy(-beta, vold);  // w = w - beta * vold

    double alpha = dot_product(z, w);

    w.axpy(-alpha, v);  // w = w - alpha * v

    // Re-orthogonalization
    for (int iiter = 0; iiter < jiter; ++iiter) {
      double proj = dot_product(w, zVEC[iiter]);
      w.axpy(-proj, vVEC[iiter]);
    }

    precond.multiply(w, zh);
    B.multiply(zh, z);  // z = B precond w

    beta = sqrt(dot_product(z, w));

    vold = v;
    v  = w;
    v  *= 1/beta;
    zh *= 1/beta;
    z  *= 1/beta;

    zhVEC.push_back(zh);  // zhVEC[jiter+1]= zh_jiter
    zVEC.push_back(z);  // zVEC[jiter+1] = z_jiter
    vVEC.push_back(v);  // vVEC[jiter+1] = v_jiter

    alphas.push_back(alpha);

    y.clear();
    if (jiter == 0) {
      y.push_back(beta0/alpha);
    } else {
      // Solve the tridiagonal system T_jiter y_jiter = beta0 * e_1
      d.clear();
      for (int iiter = 0; iiter <= jiter; ++iiter){
        d.push_back(beta0*dot_product(zVEC[0], vVEC[iiter]));
      }
      TriDiagSolve(alphas, betas, d, y);
    } 
     
    // Gradient norm in precond metric --> sqrt(r'z) --> beta * y(jiter)
    double rznorm = beta*fabs(y[jiter]);

    normReduction = rznorm/beta0;

    betas.push_back(beta);


    LOG(Info) << "DPLanczos end of iteration " << jiter+1
              << ". Norm reduction= " << normReduction;

    if (normReduction < tolerance) {
      LOG(Info) << "DPLanczos: Achieved required reduction in residual norm.";
      break;
    }
  }

  // Calculate the solution (xh = Binv x)
  for (int iiter = 0; iiter < jiter; ++iiter) {
      xh.axpy(y[iiter], zhVEC[iiter]);
  }
  B.multiply(xh, x);

  LOG(Info) << "DPLanczos: Estimated Final Jb = " << 0.5 * dot_product(xh, x);
  LOG(Trace) << "DPLanczos: end";

  return normReduction;
}

}  // namespace oops

#endif  // OOPS_ASSIMILATION_DPLANCZOS_H_
