/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_ASSIMILATION_DUALVECTOR_H_
#define OOPS_ASSIMILATION_DUALVECTOR_H_

#include <vector>

#include <boost/scoped_ptr.hpp>
#include <boost/shared_ptr.hpp>

#include "oops/assimilation/ControlIncrement.h"
#include "oops/base/Departure.h"
#include "oops/base/GeneralizedDeparture.h"
#include "util/dot_product.h"

namespace oops {

/// Container of dual space vectors for all terms of the cost function.
/*!
 * Contains dual space vectors for all terms of the cost function,
 * that is Departures for Jo, an Increment for Jc, a ControlIncrement
 * for Jb and Jq.
 */

// -----------------------------------------------------------------------------
template<typename TRAIT> class DualVector {
  typedef typename TRAIT::IncrementBase Increment;
  typedef ControlIncrement<TRAIT>       CtrlInc;
  typedef Departure<TRAIT>              Depart;

 public:
  DualVector(): dxjb_(), dxjo_(), dxjc_(), ijo_(), ijc_(), size_(0) {}
  DualVector(const DualVector &);
  ~DualVector() {}

// Access increment
  void dx(CtrlInc * dx) {dxjb_.reset(dx);}
  const CtrlInc & dx() const {return *dxjb_;}
  CtrlInc & dx() {return *dxjb_;}

// Store and retrieve other elements (takes ownership)
  void append(GeneralizedDeparture *);
  boost::shared_ptr<const GeneralizedDeparture> getv(const unsigned) const;

// Linear algebra
  DualVector & operator=(const DualVector &);
  DualVector & operator+=(const DualVector &);
  DualVector & operator-=(const DualVector &);
  DualVector & operator*=(const double);
  void zero();
  void axpy(const double, const DualVector &);
  double dot_product_with(const DualVector &) const;

// Clear everything
  void clear();

 private:
  bool compatible(const DualVector & other) const;

  boost::scoped_ptr<CtrlInc>   dxjb_;
  std::vector<boost::shared_ptr<Depart> >    dxjo_;
  std::vector<boost::shared_ptr<Increment> > dxjc_;
  std::vector<unsigned> ijo_;
  std::vector<unsigned> ijc_;
  unsigned size_;
};
// -----------------------------------------------------------------------------
}  // namespace oops

#include "oops/assimilation/DualVector.cc"

#endif  // OOPS_ASSIMILATION_DUALVECTOR_H_
