/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_ASSIMILATION_FMATRIX_H_
#define OOPS_ASSIMILATION_FMATRIX_H_

#include <vector>

#include "logger/Logger.h"
#include "util/asser1.h"
#include "util/dot_product.h"

#include "oops/assimilation/TriDiagSpectrum.h"

namespace oops {

  /*  The preconditioner matrix \f$ F \f$ for the system matrix AB = (I + Ht Rinv H B)
   *  It is defined as F_(k+1)  = F_k  +  U (Diag(1/lambda_i) - I) X'
   *  where X = [x_1, x_2, ..., x_l] and U = [u_1, u_2, ..., u_l] are column matrices.
   *  The pairs (lambda_i,x_i) define the Ritz pairs of the matrix PA with respect to
   *  the subspace K(PA,Pr0). (Note that P = BF).
   *  U = Binv X
   */
  /*!
   *  The solvers represent matrices as objects that implement a "multiply"
   *  method. This class defines objects that apply the preconditioner matrix \f$ F \f$.
   */

template<typename VECTOR>
class FMatrix {
 public:
  FMatrix() {}
  ~FMatrix() {}

  void setup(const std::vector<VECTOR> & Zhl, const std::vector<VECTOR> & Zl,
             const std::vector<double> & alphas, const std::vector<double> & betas) {
    LOG(Debug) << "FMatrix size Zl    = " << Zl.size();
    LOG(Debug) << "FMatrix size Zhl   = " << Zhl.size();
    LOG(Debug) << "FMatrix size alpha = " << alphas.size();
    LOG(Debug) << "FMatrix size beta  = " << betas.size();
    const unsigned nvec = Zl.size();
    ASSERT(Zhl.size() == nvec);

    if (nvec != 0) {
      ASSERT(alphas.size() ==  nvec-1);
      ASSERT(betas.size() == nvec-1);
//    Compute spectrum of tri-diagonal matrix
      std::vector<double> evals;
      std::vector< std::vector<double> > evecs;
       
      TriDiagSpectrum(alphas, betas, evals, evecs);
 
//    Determine the converged eigenvalues
      std::vector<unsigned> convIndx;
      for (unsigned jiter = 0; jiter < nvec-1; ++jiter) {
        double erritz = evecs[jiter][nvec-2]*betas[nvec-2];
        double lambda = evals[jiter];
        if (fabs(erritz) <= 0.001*lambda) {
          convIndx.push_back(jiter);
          eigvals_.push_back(lambda);
        }
      }

      VECTOR ww(Zl[0]);

//    Calculate the matrix X
      for (unsigned jiter = 0; jiter < convIndx.size(); ++jiter) {
        ww.zero();
        for (unsigned iiter = 0; iiter < nvec-1; ++iiter) {
          ww.axpy(evecs[convIndx[jiter]][iiter], Zl[iiter]);
        }
        X_.push_back(ww);
      }

//    Calculate the matrix U
      for (unsigned jiter = 0; jiter < convIndx.size(); ++jiter) {
        ww.zero();
        for (unsigned iiter = 0; iiter < nvec-1; ++iiter) {
          ww.axpy(evecs[convIndx[jiter]][iiter], Zhl[iiter]);
        }
        U_.push_back(ww);
      }
    }
  }

  void multiply(const VECTOR & a, VECTOR & b) const {
    b = a;  // F_0 = I
    for (unsigned iiter = 0; iiter < eigvals_.size(); ++iiter) {
      double zeval = std::min(10.0, eigvals_[iiter]);
      double zz = 1.0/zeval - 1.0;
      b.axpy(zz*dot_product(a, X_[iiter]), U_[iiter]);
    }
  }

 private:
  std::vector<VECTOR> X_;
  std::vector<VECTOR> U_;
  std::vector<double> eigvals_;
};

}  // namespace oops

#endif  // OOPS_ASSIMILATION_FMATRIX_H_
