/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_ASSIMILATION_GMRESR_H_
#define OOPS_ASSIMILATION_GMRESR_H_

#include <cmath>
#include <iostream>
#include <vector>

#include "logger/Logger.h"
#include "util/dot_product.h"

namespace oops {

/*! \file GMRESR.h
 * \brief GMRESR solver for Ax=b.
 *
 * GMRESR solver for Ax=b. (H.A. Van der Vorst and C. Vuik, 1994,
 * Numerical Linear Algebra with Applications, 1(4), 369-386.)
 * A must be square, but need not be symmetric. (For a symmetric matrix,
 * and constant preconditioner, GMRESR is simply PCG with full
 * orthogonalisation.) A preconditioner must be supplied that, given a
 * vector q, returns an approximate solution of Ap=q. The preconditioner
 * can be variable.
 *
 * On entry:
 * -    x       =  starting point, \f$ X_0 \f$.
 * -    b       = right hand side.
 * -    A       = \f$ A \f$.
 * -    precond = preconditioner \f$ F_k \approx (A)^{-1} \f$.
 *
 * On exit, x will contain the solution.
 * The return value is the achieved reduction in residual norm.
 *
 * Iteration will stop if the maximum iteration limit "maxiter" is reached
 * or if the residual norm reduces by a factor of "tolerance".
 *
 * VECTOR must implement:
 *  - dot_product
 *  - operator(=)
 *  - operator(+=),
 *  - operator(-=)
 *  - operator(*=) [double * VECTOR],
 *  - axpy
 *
 *  AMATRIX and PMATRIX must implement a method:
 *  - void multiply(const VECTOR&, VECTOR&) const
 *
 *  which applies the matrix to the first argument, and returns the
 *  matrix-vector product in the second. (Note: the const is optional, but
 *  recommended.)
 */

template <typename VECTOR, typename AMATRIX, typename PMATRIX>
double GMRESR(VECTOR & x, const VECTOR & b,
              const AMATRIX & A, const PMATRIX & precond,
              const int maxiter, const double tolerance) {
  std::vector<VECTOR> c;
  std::vector<VECTOR> u;
  VECTOR cc(x);
  VECTOR z(x);
  VECTOR r(x);

  A.multiply(x, z);  // z=Ax
  r = b;
  r -= z;

  double dotRr0  = dot_product(r, r);
  double normReduction = 1.0;
  double cdotr = sqrt(dotRr0);

  for (int jiter = 0; jiter < maxiter; ++jiter) {
    LOG(Trace) << " GMRESR Starting Iteration " << jiter+1;

    // test for stagnation
    double stagthresh = 0.000001;
    if (cdotr*cdotr < stagthresh*normReduction*normReduction*dotRr0) {
      LOG(Info) << "GMRESR stagnated. Doing an LSQR step.";
      A.multiply(r, z);  // should be A^T, but we assume A is symmetric.
    } else {
      precond.multiply(r, z);  // returns z as approxmate solve of Az=r
    }

    A.multiply(z, cc);  // cc=Az

    for (int k = 0; k < jiter; ++k) {
      double alpha = -dot_product(c[k], cc);
      cc.axpy(alpha, c[k]);  // cc = cc - <c[k],cc>*alpha*c[k];
      z.axpy(alpha, u[k]);   // z  = z  - <c[k],cc>*u[k];
    }

    double ccnorm = sqrt(dot_product(cc, cc));
    c.push_back(cc);         // c[jiter]=cc
    c[jiter] *= 1.0/ccnorm;
    u.push_back(z);          // u[jiter]=z
    u[jiter] *= 1.0/ccnorm;

    cdotr = dot_product(c[jiter], r);
    x.axpy(cdotr, u[jiter]);
    r.axpy(-cdotr, c[jiter]);

    normReduction = sqrt(dot_product(r, r)/dotRr0);
    LOG(Info) << "GMRESR end of iteration " << jiter+1
                      << ". Norm reduction= " << normReduction;

    if (normReduction < tolerance) {
      LOG(Info) << "GMRESR: Achieved required reduction in residual norm.";
      break;
    }
  }

  LOG(Trace) << "GMRESR: end";

  return normReduction;
}

}  // namespace oops

#endif  // OOPS_ASSIMILATION_GMRESR_H_
