/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_ASSIMILATION_HBHTMATRIX_H_
#define OOPS_ASSIMILATION_HBHTMATRIX_H_

#include <boost/noncopyable.hpp>
#include <boost/scoped_ptr.hpp>

#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/DualVector.h"

namespace oops {

/// The \f$ H B H^T \f$ matrix.
/*!
 *  The solvers represent matrices as objects that implement a "multiply"
 *  method. This class defines objects that apply a generalized
 *  \f$ H B H ^T\f$ matrix which includes \f$ H \f$ and the equivalent
 *  operators for the other terms of the cost function.
 */

template<typename TRAIT> class HBHtMatrix : private boost::noncopyable {
  typedef typename TRAIT::IncrementBase Increment;
  typedef CostFunction<TRAIT>           CostFct;
  typedef ControlIncrement<TRAIT>       CtrlInc;
  typedef DualVector<TRAIT>             Dual;

 public:
  explicit HBHtMatrix(const CostFct & j): j_(j) {}

  void multiply(const Dual & dy, Dual & dz) const {
    PostProcessor<Increment> post;

//  Run ADJ
    CostProcessor<Increment> costad;
    for (unsigned jj = 0; jj < j_.nterms(); ++jj) {
      costad.enroll(j_.jterm(jj).setupAD(dy.getv(jj)));
    }
    boost::scoped_ptr<CtrlInc> ww(j_.jb().newIncrement());
    ww->zeroAD();
    j_.runADJ(*ww, costad, post);

//  Multiply by B
    boost::scoped_ptr<CtrlInc> zz(j_.jb().newIncrement());
    j_.jb().Bmult(*ww, *zz);

//  Run TLM
    CostProcessor<Increment> costtl;
    for (unsigned jj = 0; jj < j_.nterms(); ++jj) {
      costtl.enroll(j_.jterm(jj).setupTL(*zz));
    }
    j_.runTLM(*zz, costtl, post);

//  Get TLM outputs
    dz.clear();
    for (unsigned jj = 0; jj < j_.nterms(); ++jj) {
      dz.append(costtl.releaseTL(jj));
    }
  }

 private:
  CostFct const & j_;
};

}  // namespace oops

#endif  // OOPS_ASSIMILATION_HBHTMATRIX_H_
