/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_ASSIMILATION_HMATRIX_H_
#define OOPS_ASSIMILATION_HMATRIX_H_

#include <boost/noncopyable.hpp>

#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/DualVector.h"

namespace oops {

/// The \f$ H \f$ matrix.
/*!
 *  The solvers represent matrices as objects that implement a "multiply"
 *  method. This class defines objects that apply a generalized \f$ H \f$
 *  matrix which includes \f$ H \f$ itself and the equivalent operators
 *  for the other terms of the cost function.
 */

template<typename TRAIT> class HMatrix : private boost::noncopyable {
  typedef typename TRAIT::IncrementBase Increment;
  typedef CostFunction<TRAIT>           CostFct;
  typedef ControlIncrement<TRAIT>       CtrlInc;
  typedef DualVector<TRAIT>             Dual;

 public:
  explicit HMatrix(const CostFct & j): j_(j) {}

  void multiply(const CtrlInc & dx, Dual & dy) const {
    PostProcessor<Increment> post;
    CostProcessor<Increment> cost;
    for (unsigned jj = 0; jj < j_.nterms(); ++jj) {
      cost.enroll(j_.jterm(jj).setupTL(dx));
    }

    CtrlInc ww(dx);
    j_.runTLM(ww, cost, post);

    dy.clear();
    for (unsigned jj = 0; jj < j_.nterms(); ++jj) {
      dy.append(cost.releaseTL(jj));
    }
  }

 private:
  CostFct const & j_;
};

}  // namespace oops

#endif  // OOPS_ASSIMILATION_HMATRIX_H_
