/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_ASSIMILATION_HESSIANMATRIX_H_
#define OOPS_ASSIMILATION_HESSIANMATRIX_H_

#include <boost/noncopyable.hpp>

#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/ControlIncrement.h"
#include "oops/base/GeneralizedDeparture.h"

namespace oops {

/// The Hessian matrix: \f$ B^{-1} + H^T R^{-1} H \f$.
/*!
 *  The solvers represent matrices as objects that implement a "multiply"
 *  method. This class defines objects that apply a generalized Hessian
 *  matrix which includes all the terms of the cost function.
 */

template<typename TRAIT> class HessianMatrix : private boost::noncopyable {
  typedef typename TRAIT::IncrementBase Increment;
  typedef CostFunction<TRAIT>           CostFct;
  typedef ControlIncrement<TRAIT>       CtrlInc;

 public:
  explicit HessianMatrix(const CostFct & j): j_(j) {}

  void multiply(const CtrlInc & dx, CtrlInc & dz) const {
    PostProcessor<Increment> post;

//  Setup TL terms of cost function
    CostProcessor<Increment> costtl;
    for (unsigned jj = 0; jj < j_.nterms(); ++jj) {
      costtl.enroll(j_.jterm(jj).setupTL(dx));
    }

//  Run TLM
    dz = dx;
    j_.runTLM(dz, costtl, post);

//  Get TLM outputs, multiply by covariance inverses and setup ADJ forcing terms
    CostProcessor<Increment> costad;
//  Jb
    boost::scoped_ptr<CtrlInc> tmp1(j_.jb().finalTL(dz, dx));
    boost::scoped_ptr<CtrlInc> tmp2(j_.jb().newIncrement());
    j_.jb().Bminv(*tmp1, *tmp2);
    dz = *j_.jb().setupAD(*tmp2);
//  Jo + Jc
    for (unsigned jj = 0; jj < j_.nterms(); ++jj) {
      boost::scoped_ptr<GeneralizedDeparture> ww(costtl.releaseTL(jj));
      boost::shared_ptr<GeneralizedDeparture> zz(j_.jterm(jj).CovarMinv(*ww));
      costad.enroll(j_.jterm(jj).setupAD(zz));
    }

//  Run ADJ
    j_.runADJ(dz, costad, post);
    dz += dx;
  }

 private:
  CostFct const & j_;
};

}  // namespace oops

#endif  // OOPS_ASSIMILATION_HESSIANMATRIX_H_
