/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_ASSIMILATION_HTRINVHMATRIX_H_
#define OOPS_ASSIMILATION_HTRINVHMATRIX_H_

#include <boost/noncopyable.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/shared_ptr.hpp>

#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/DualVector.h"
#include "oops/assimilation/ControlIncrement.h"

namespace oops {

/// The \f$ H^T R^{-1} H \f$ matrix.
/*!
 *  The solvers represent matrices as objects that implement a "multiply"
 *  method. This class defines objects that apply a generalized
 *  \f$ H^T R^{-1} H \f$ matrix that also includes the equivalent
 *  operators for the other terms of the cost function.
 */

template<typename TRAIT> class HtRinvHMatrix : private boost::noncopyable {
  typedef typename TRAIT::IncrementBase Increment;
  typedef CostFunction<TRAIT>           CostFct;
  typedef ControlIncrement<TRAIT>       CtrlInc;
  typedef DualVector<TRAIT>             Items;

 public:
  explicit HtRinvHMatrix(const CostFct & j): j_(j) {}

  void multiply(const CtrlInc & dx, CtrlInc & dz) const {
    PostProcessor<Increment> post;

//  Setup TL terms of cost function
    CostProcessor<Increment> costtl;
    for (unsigned jj = 0; jj < j_.nterms(); ++jj) {
      costtl.enroll(j_.jterm(jj).setupTL(dx));
    }

//  Run TLM
    dz = dx;
    j_.runTLM(dz, costtl, post);

//  Get TLM outputs, multiply by covariance inverses, and setup ADJ forcing terms
    CostProcessor<Increment> costad;
    for (unsigned jj = 0; jj < j_.nterms(); ++jj) {
      boost::scoped_ptr<GeneralizedDeparture> ww(costtl.releaseTL(jj));
      boost::shared_ptr<GeneralizedDeparture> zz(j_.jterm(jj).CovarMinv(*ww));
      costad.enroll(j_.jterm(jj).setupAD(zz));
    }

//  Run ADJ
    dz.zero();
    j_.runADJ(dz, costad, post);
  }

 private:
  CostFct const & j_;
};

}  // namespace oops

#endif  // OOPS_ASSIMILATION_HTRINVHMATRIX_H_
