/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_ASSIMILATION_IPCG_H_
#define OOPS_ASSIMILATION_IPCG_H_

#include <cmath>
#include <iostream>

#include "logger/Logger.h"
#include "util/dot_product.h"

namespace oops {

/*! \file IPCG.h
 * \brief Inexact-Preconditioned Conjugate Gradients solver.
 *
 * Golub-Ye Inexact-Preconditioned Conjugate Gradients solver for Ax=b.
 * (G.H. Golub and Q. Ye 1999/00, SIAM J. Sci. Comput. 21(4) 1305-1320.)
 *
 * A must be square, symmetric, positive definite.
 * A preconditioner must be supplied that, given a vector q, returns an
 * approximate solution of Ap=q. The preconditioner can be variable.
 *
 * On entry:
 * -    x       =  starting point, \f$ X_0 \f$.
 * -    b       = right hand side.
 * -    A       = \f$ A \f$.
 * -    precond = preconditioner \f$ F_k \approx (A)^{-1} \f$.
 *
 * On exit, x will contain the solution.
 * The return value is the achieved reduction in residual norm.
 *
 *  Iteration will stop if the maximum iteration limit "maxiter" is reached
 *  or if the residual norm reduces by a factor of "tolerance".
 *
 * VECTOR must implement:
 *  - dot_product
 *  - operator(=)
 *  - operator(+=),
 *  - operator(-=)
 *  - operator(*=) [double * VECTOR],
 *  - axpy
 *
 *  AMATRIX and PMATRIX must implement a method:
 *  - void multiply(const VECTOR&, VECTOR&) const
 *
 *  which applies the matrix to the first argument, and returns the
 *  matrix-vector product in the second. (Note: the const is optonal, but
 *  recommended.)
 */

template <typename VECTOR, typename AMATRIX, typename PMATRIX>
double IPCG(VECTOR & x, const VECTOR & b,
            const AMATRIX & A, const PMATRIX & precond,
            const int maxiter, const double tolerance ) {
  VECTOR ap(x);
  VECTOR p(x);
  VECTOR r(x);
  VECTOR s(x);
  VECTOR oldr(x);
  VECTOR w(x);

  A.multiply(x, s);
  r = b;
  r -= s;

  double dotRr0  = dot_product(r, r);
  double normReduction = 1.0;
  double rdots;

  for (int jiter = 0; jiter < maxiter; ++jiter) {
    LOG(Trace) << " IPCG Starting Iteration " << jiter+1;

    precond.multiply(r, s);  // returns s as approxmate solve of As=r

    if (jiter == 0) {
      p = s;
    } else {
      w = r;
      w -= oldr;  // w=r-oldr
      double beta = dot_product(s, w)/rdots;
      p *= beta;
      p += s;    // p = s + beta*p
    }

    A.multiply(p, ap);  // ap = Ap

    oldr = r;
    rdots = dot_product(r, s);

    double alpha = rdots/dot_product(p, ap);

    x.axpy(alpha, p);    // x = x + alpha*p;
    r.axpy(-alpha, ap);  // r = r - alpha*ap;

    normReduction = sqrt(dot_product(r, r)/dotRr0);
    LOG(Info) << "IPCG end of iteration " << jiter+1
              << ". Norm reduction= " << normReduction;

    if (normReduction < tolerance) {
      LOG(Info) << "IPCG: Achieved required reduction in residual norm.";
      break;
    }
  }

  LOG(Trace) << "IPGC: end";

  return normReduction;
}

}  // namespace oops

#endif  // OOPS_ASSIMILATION_IPCG_H_
