/*
* Copyright 2011 ECMWF
*
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
*
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_ASSIMILATION_RPCG_H_
#define OOPS_ASSIMILATION_RPCG_H_

#include <cmath>
#include <iostream>

#include "logger/Logger.h"
#include "util/dot_product.h"

namespace oops {

template <typename VECTOR, typename HBHtMATRIX, typename RMATRIX, typename PMATRIX,
          typename PtMATRIX>
double RPCG(VECTOR & vv, double & vvp, VECTOR & rr,
            const HBHtMATRIX & HBHt, const RMATRIX & Rinv,
            const PMATRIX & G, const PtMATRIX & Gt,
            const int maxiter, const double tolerance,
            const VECTOR & aa, const double sigma) {
  VECTOR zz(vv);
  VECTOR ww(vv);
  VECTOR pp(vv);
  VECTOR tt(vv);
  VECTOR ll(vv);
  VECTOR qq(vv);

  zz.zero();
  ww.zero();
  pp.zero();
  tt.zero();
  ll.zero();

  // UPDATE INITIAL RESIDUAL if dx is not a zero vector
  // rr = rr - Rinv H dx0

  // augmented part of the vectors
  double rrp = 1.0;
  double llp;
  double wwp;
  double zzp;
  double ppp = 0.0;
  double ttp = 0.0;
  double qqp;

  // llaug = HBHtaug rraug
  HBHt.multiply(rr, ll);
  ll.axpy(rrp, aa);
  llp = dot_product(aa, rr) + sigma*rrp;

  // wwaug = Gtaug llaug
  Gt.multiply(ll, ww);  // UPDATE WHEN G IS NOT IDENTITY
  wwp = llp;

  // zzaug = Gaug  rraug
  G.multiply(rr, zz);   // UPDATE WHEN G IS NOT IDENTITY
  zzp = rrp;

  double dotwr = dot_product(rr, ww) + rrp*wwp;
  double normReduction;
  double dotw0r0 = dotwr;
  double dotwr_old;

  for (int jiter = 0; jiter < maxiter; ++jiter) {
    LOG(Info) << "RPCG Starting Iteration " << jiter+1;

    if (jiter > 0) {
      double beta = dotwr/dotwr_old;
      LOG(Debug) << "RPCG beta = " << beta;

      pp *= beta;
      ppp *= beta;

      tt *= beta;
      ttp *= beta;
    }

    pp += zz;
    ppp += zzp;  // ppaug = zzaug + beta*ppaug

    tt += ww;
    ttp += wwp;  // ttaug = wwaug + beta*ttaug

    // (RinvHBHt + I) pp
    Rinv.multiply(tt, qq);
    qq += pp;
    qqp = ppp;  // qqaug = ppaug + Rinv_aug ttaug

    double alpha = dotwr/(dot_product(qq, tt) + qqp * ttp);
    LOG(Debug) << "RPCG alpha = " << alpha;

    vv.axpy(alpha, pp);
    vvp = vvp + alpha*ppp;  // vvaug = vvaug + alpha*ppaug

    rr.axpy(-alpha, qq);
    rrp = rrp - alpha*qqp;  // rraug = rraug - alpha*qqaug

    // llaug = HBHt_aug rraug
    HBHt.multiply(rr, ll);
    ll.axpy(rrp, aa);
    llp = dot_product(aa, rr) + sigma * rrp;

    // wwaug = Gt_aug llaug
    Gt.multiply(ll, ww);  // UPDATE WHEN G IS NOT IDENTITY
    wwp = llp;

    // zzaug = Gaug  rraug
    G.multiply(rr, zz);   // UPDATE WHEN G IS NOT IDENTITY
    zzp = rrp;

    dotwr_old = dotwr;
    dotwr = dot_product(ww, rr) + rrp * wwp;

    normReduction = sqrt(dotwr/dotw0r0);

    LOG(Info) << "RPCG end of iteration " << jiter+1
              << ". Norm reduction= " << normReduction;

    if (normReduction < tolerance) {
      LOG(Info) << "RPCG: Achieved required reduction in residual norm.";
      break;
    }
  }
  return normReduction;
}

}  // namespace oops

#endif  // OOPS_ASSIMILATION_RPCG_H_
