/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include "logger/Logger.h"
#include "oops/assimilation/HBHtMatrix.h"
#include "oops/assimilation/HtMatrix.h"
#include "oops/assimilation/HMatrix.h"
#include "oops/assimilation/RinvMatrix.h"
#include "oops/assimilation/RPCG.h"
#include "util/dot_product.h"
#include "util/Config.h"
#include "util/IdentityMatrix.h"

namespace oops {

// -----------------------------------------------------------------------------

template<typename TRAIT>
void RPCGMinimizer<TRAIT>::minimize(const CostFct & J, CtrlInc & dx,
                                    const util::Config & config) {
  int ninner = config.getInt("ninner");
  double gnreduc = config.getReal("gradient_norm_reduction");

  LOG(Info) << "RPCGMinimizer: max iter = " << ninner <<
               ", requested norm reduction = " << gnreduc;

// Define the matrices
  IdentityMatrix<Dual> I;
  const HBHtMatrix<TRAIT> HBHt(J);
  const HtMatrix<TRAIT> Ht(J);
  const RinvMatrix<TRAIT> Rinv(J);
  const HMatrix<TRAIT> H(J);

// Define minimisation starting point in dual space
  Dual vv;
  for (unsigned jj = 0; jj < J.nterms(); ++jj) {
    vv.append(J.jterm(jj).newDualVector());
  }
  double vvp;

// Get R^{-1} d
  Dual rr;
  for (unsigned jj = 0; jj < J.nterms(); ++jj) {
    rr.append(J.jterm(jj).getGradientFG());
  }
  rr *= -1.0;

// Compute a = H (xb - xk)
  Dual dy;
  H.multiply(J.jb().getFirstGuess(), dy);
  dy *= -1.0;

// Compute sigma = (xb - xk)^T Binv (xb - xk)
  double sigma = dot_product(J.jb().getFirstGuess(), J.jb().getGradientFG());

// Solve the linear system
  double reduc = RPCG(vv, vvp, rr, HBHt, Rinv, I, I, ninner, gnreduc, dy, sigma);

  LOG(Info) << "RPCGMinimizer: reduction in residual norm = " << reduc;

// Recover solution in primal space
  CtrlInc dh(dx);
  Ht.multiply(vv, dh);
  dh.axpy(-vvp, J.jb().getGradientFG());
  J.jb().Bmult(dh, dx);    // BHtaug vvaug

  LOG(Info) << "RPCGMinimizer: Estimated Final Jb = " << 0.5 * dot_product(dh, dx);

// Save gradient Jb
  J.jb().gradientUpdate(dh);
  dx.prtNorms("RPCGMinimizer out");
}

// -----------------------------------------------------------------------------

}  // namespace oops
