/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include <boost/scoped_ptr.hpp>

#include "logger/Logger.h"
#include "oops/assimilation/GMRESR.h"
#include "oops/assimilation/SaddlePointVector.h"
#include "oops/assimilation/SaddlePointMatrix.h"
#include "oops/assimilation/SaddlePointPrecondMatrix.h"
#include "util/Config.h"
#include "util/asser1.h"

namespace oops {

// -----------------------------------------------------------------------------

template<typename TRAIT>
void SaddlePointMinimizer<TRAIT>::minimize(const CostFct & J, CtrlInc & dx,
                                           const util::Config & config) {
  int ninner = config.getInt("ninner");
  int gnreduc = config.getReal("gradient_norm_reduction");

  LOG(Trace) << "SaddlePointMinimizer: max iter = " << ninner <<
                ", requested norm reduction = " << gnreduc;

// Define saddle-point control vectors
  Multipliers * pdxx = new Multipliers();
  pdxx->dx(J.jb().newIncrement());
  for (unsigned jj = 0; jj < J.nterms(); ++jj) {
    pdxx->append(J.jterm(jj).newDualVector());
  }
  CtrlInc * tmp1 = J.jb().newIncrement();
  SaddlePointVector<TRAIT> xx(tmp1, pdxx);

// Compute RHS
  Multipliers * pdfg = new Multipliers();
  CtrlInc * tmp3 = new CtrlInc(J.jb().getFirstGuess());
  pdfg->dx(tmp3);
  for (unsigned jj = 0; jj < J.nterms(); ++jj) {
    boost::scoped_ptr<GeneralizedDeparture> ww(J.jterm(jj).getGradientFG());
    pdfg->append(J.jterm(jj).CovarMult(*ww));
  }
  CtrlInc * tmp2 = J.jb().newIncrement();
  SaddlePointVector<TRAIT> rhs(tmp2, pdfg);
  rhs *= -1.0;

// Define the matrices
  SaddlePointMatrix<TRAIT> A(J);
  SaddlePointPrecondMatrix<TRAIT> Pinv(J);

// Solve the linear system
  double reduc = GMRESR(xx, rhs, A, Pinv, ninner, gnreduc);

  LOG(Info) << "SaddlePointMinimizer reduction in residual norm = " << reduc;

// Recover final gradient and solution
  dx = xx.lambda().dx();
  dx *= -1.0;
  J.jb().gradientUpdate(dx, false);
  dx = xx.dx();
  dx.prtNorms("SaddlePointMinimizer out");
}

// -----------------------------------------------------------------------------

}  // namespace oops
