/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_BASE_COSTBASE_H_
#define OOPS_BASE_COSTBASE_H_

#include <boost/noncopyable.hpp>

#include "oops/base/GeneralizedDeparture.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/PostTimer.h"

namespace oops {

// -----------------------------------------------------------------------------

/// Handles post-processing of model fields related to cost function.
/*!
 *  CostBase is the base class for all cost function processors, it
 *  is mostly used so that CostProcessor can hold a vector of such
 *  processors.
 *  The difference with ProcessorBase is that there are TL and AD
 *  methods and that the adjoint one can modify the increment.
 *  By default processing is performed on every call.
 */

template <typename INCR> class CostBase : private boost::noncopyable {
 public:
/// Constructors and basic operators
  CostBase(): timer_() {}
  explicit CostBase(const util::Duration & freq): timer_(freq) {}
  explicit CostBase(const util::Config & conf): timer_(conf) {}
  CostBase(const util::DateTime & start, const util::DateTime & finish,
           const util::Duration & freq = util::Duration(0))
    : timer_(start, finish, freq) {}

  virtual ~CostBase() {}

/// Setup
  void initializeTL(const util::DateTime & bgn, const util::DateTime & end,
                    const util::Duration & step) {
    timer_.initialize(bgn, end, step);
    this->doFirstTL(bgn, end, step);
  }
  void initializeAD(const util::DateTime & bgn, const util::DateTime & end,
                    const util::Duration & step) {
    timer_.initialize(bgn, end, step);
    this->doFirstAD(bgn, end, step);
  }

/// Process increment
  void processTL(const INCR & dx) {
    if (timer_.itIsTime(dx.validTime())) this->doProcessingTL(dx);
  }
  void processAD(INCR & dx) {
    if (timer_.itIsTime(dx.validTime())) this->doProcessingAD(dx);
  }

/// Final call
  void finalizeTL() {
    this->doLastTL();
  }
  void finalizeAD() {
    this->doLastAD();
  }

/// Return observed
  virtual GeneralizedDeparture * finalTL() =0;

 private:
  PostTimer timer_;
/// Actual processing of increment (pure virtual)
  virtual void doFirstTL(const util::DateTime &, const util::DateTime &,
                         const util::Duration &) =0;
  virtual void doProcessingTL(const INCR &) =0;
  virtual void doLastTL() =0;
  virtual void doFirstAD(const util::DateTime &, const util::DateTime &,
                         const util::Duration &) =0;
  virtual void doProcessingAD(INCR &) =0;
  virtual void doLastAD() =0;
};

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_COSTBASE_H_
