/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_BASE_COSTPROCESSOR_H_
#define OOPS_BASE_COSTPROCESSOR_H_

#include <vector>

#include <boost/foreach.hpp>
#include <boost/shared_ptr.hpp>

#include "oops/base/CostBase.h"
#include "oops/base/GeneralizedDeparture.h"

namespace oops {

/// Control model post processing
/*!
 *  This class controls model post processing in the most general sense,
 *  ie all diagnostics computations that do not affect the model integration.
 *  It just calls all the individual processors one by one.
 */

template<typename INCR>
class CostProcessor {
  typedef CostBase<INCR> Process;

 public:
  CostProcessor() {}
  CostProcessor(const CostProcessor & pp): processors_(pp.processors_) {}
  ~CostProcessor() {}

  void enroll(Process * pp) {
    boost::shared_ptr<Process> sp(pp);
    processors_.push_back(sp);
  }

  void enroll(boost::shared_ptr<Process> pp) {
    if (pp != 0) processors_.push_back(pp);
  }

  GeneralizedDeparture * releaseTL(unsigned ii) {
    GeneralizedDeparture * lambda = processors_[ii]->finalTL();
    return lambda;
  }

  void initializeTL(const util::DateTime & bgn, const util::DateTime & end,
                    const util::Duration & step) {
    BOOST_FOREACH(boost::shared_ptr<Process> jp, processors_) {
      jp->initializeTL(bgn, end, step);
    }
  }

  void initializeAD(const util::DateTime & bgn, const util::DateTime & end,
                    const util::Duration & step) {
    BOOST_FOREACH(boost::shared_ptr<Process> jp, processors_) {
      jp->initializeAD(bgn, end, step);
    }
  }

  void processTL(const INCR & dx) {
    BOOST_FOREACH(boost::shared_ptr<Process> jp, processors_) jp->processTL(dx);
  }
  void processAD(INCR & dx) {
    BOOST_FOREACH(boost::shared_ptr<Process> jp, processors_) jp->processAD(dx);
  }

  void finalizeTL() {
    BOOST_FOREACH(boost::shared_ptr<Process> jp, processors_) jp->finalizeTL();
  }
  void finalizeAD() {
    BOOST_FOREACH(boost::shared_ptr<Process> jp, processors_) jp->finalizeAD();
  }

 private:
  std::vector< boost::shared_ptr<Process> > processors_;
  CostProcessor operator= (const CostProcessor &);
};

}  // namespace oops

#endif  // OOPS_BASE_COSTPROCESSOR_H_
