/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include <boost/foreach.hpp>

#include "logger/Logger.h"
#include "util/asser1.h"

namespace oops {
// -----------------------------------------------------------------------------
template<typename TRAIT>
Departure<TRAIT>::Departure(const TheObservation & obs)
  : dep_(0), winbgn_(obs.windowStart()), winend_(obs.windowEnd()),
    bgn_(winbgn_), end_(winend_), hslot_(0)
{
  for (int jj = 0; jj < obs.size(); ++jj) {
    dep_.push_back(new TheDep(obs[jj]));
  }
  LOG(Trace) << "Departure obs-created";
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
Departure<TRAIT>::Departure(const TheObservation & ob1, const TheObservation & ob2)
  : dep_(0), winbgn_(ob1.windowStart()), winend_(ob1.windowEnd()),
    bgn_(winbgn_), end_(winend_), hslot_(0)
{
  ASSERT(TheObservation::compatible(ob1, ob2));
  for (int jj = 0; jj < ob1.size(); ++jj) {
    dep_.push_back(new TheDep(ob1[jj], ob2[jj]));
  }
  LOG(Trace) << "Departure diff-created";
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
Departure<TRAIT>::Departure(const Departure & other, const bool copyDep)
  : dep_(0), winbgn_(other.winbgn_), winend_(other.winend_),
    bgn_(other.bgn_), end_(other.end_), hslot_(0)
{
  BOOST_FOREACH(const TheDep & jdep, other.dep_) {
    dep_.push_back(new TheDep(jdep, copyDep));
  }
  LOG(Trace) << "Departure copy-created";
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
Departure<TRAIT>::~Departure() {
  LOG(Trace) << "Departure destructed";
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
void Departure<TRAIT>::subwindow(const util::DateTime & t1,
                                 const util::DateTime & t2) const {
  bgn_ = t1;
  end_ = t2;
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void Departure<TRAIT>::obsOperatorTLSetup(const util::Duration & dt) {
  hslot_ = dt;
  hslot_ /= 2;
  BOOST_FOREACH(TheDep & jdep, dep_) {
    jdep.setupTL(bgn_, end_);
  }
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void Departure<TRAIT>::obsOperatorTL(const Increment & xx,
                                     const ObsAuxIncr & octl,
                                     const ObsTraj & traj) {
  util::DateTime t1(xx.validTime()-hslot_);
  if (t1 < bgn_) t1 = bgn_;
  util::DateTime t2(xx.validTime()+hslot_);
  if (t2 > end_) t2 = end_;

  for (int jj = 0; jj < size(); ++jj) {
    dep_[jj].obsOperatorTL(xx, octl, t1, t2, traj[jj]);
  }
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void Departure<TRAIT>::obsOperatorTLFinal(const ObsAuxIncr & octl,
                                          const ObsTraj & traj) {
  for (int jj = 0; jj < size(); ++jj) {
    dep_[jj].finalizeTL(octl, traj[jj]);
  }
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void Departure<TRAIT>::obsOperatorADFinal(const util::Duration & dt,
                                          ObsAuxIncr & octl,
                                          const ObsTraj & traj) const {
  hslot_ = dt;
  hslot_ /= 2;
  for (int jj = 0; jj < size(); ++jj) {
    dep_[jj].finalizeAD(bgn_, end_, octl, traj[jj]);
  }
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void Departure<TRAIT>::obsOperatorAD(Increment & xx,
                                     ObsAuxIncr & octl,
                                     const ObsTraj & traj) const {
  util::DateTime t1(xx.validTime()-hslot_);
  if (t1 < bgn_) t1 = bgn_;
  util::DateTime t2(xx.validTime()+hslot_);
  if (t2 > end_) t2 = end_;

  for (int jj = 0; jj < size(); ++jj) {
    dep_[jj].obsOperatorAD(xx, octl, t1, t2, traj[jj]);
  }
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void Departure<TRAIT>::obsOperatorADSetup() const {
  BOOST_FOREACH(const TheDep & jdep, dep_) jdep.setupAD();
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
Departure<TRAIT> & Departure<TRAIT>::operator=(const Departure & rhs) {
  ASSERT(compatible(*this, rhs));
  for (int jj = 0; jj < size(); ++jj) dep_[jj] = rhs.dep_[jj];
  return *this;
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
Departure<TRAIT> & Departure<TRAIT>::operator+=(const Departure & rhs) {
  ASSERT(compatible(*this, rhs));
  for (int jj = 0; jj < size(); ++jj) dep_[jj] += rhs.dep_[jj];
  return *this;
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
Departure<TRAIT> & Departure<TRAIT>::operator-=(const Departure & rhs) {
  ASSERT(compatible(*this, rhs));
  for (int jj = 0; jj < size(); ++jj) dep_[jj] -= rhs.dep_[jj];
  return *this;
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
Departure<TRAIT> & Departure<TRAIT>::operator*=(const double & zz) {
  BOOST_FOREACH(TheDep & jdep, dep_) jdep *= zz;
  return *this;
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
Departure<TRAIT> & Departure<TRAIT>::operator*=(const Departure & rhs) {
  ASSERT(compatible(*this, rhs));
  for (int jj = 0; jj < size(); ++jj) dep_[jj] *= rhs.dep_[jj];
  return *this;
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
Departure<TRAIT> & Departure<TRAIT>::operator/=(const Departure & rhs) {
  ASSERT(compatible(*this, rhs));
  for (int jj = 0; jj < size(); ++jj) dep_[jj] /= rhs.dep_[jj];
  return *this;
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void Departure<TRAIT>::zero() {
  BOOST_FOREACH(TheDep & jdep, dep_) jdep.zero();
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void Departure<TRAIT>::invert() {
  BOOST_FOREACH(TheDep & jdep, dep_) jdep.invert();
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void Departure<TRAIT>::axpy(const double & zz, const Departure & rhs) {
  ASSERT(compatible(*this, rhs));
  for (int jj = 0; jj < size(); ++jj) dep_[jj].axpy(zz, rhs.dep_[jj]);
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void Departure<TRAIT>::random() {
  BOOST_FOREACH(TheDep & jdep, dep_) jdep.random();
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
double Departure<TRAIT>::dot_product_with(const Departure & other) const {
  ASSERT(compatible(*this, other));
  double zz = 0.0;
  for (int jj = 0; jj < size(); ++jj) {
    zz += dot_product(dep_[jj], other.dep_[jj]);
  }
  return zz;
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
void Departure<TRAIT>::read(const std::string & name) {
  BOOST_FOREACH(TheDep & jdep, dep_) jdep.read(name);
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
void Departure<TRAIT>::write(const std::string & name) const {
  BOOST_FOREACH(const TheDep & jdep, dep_) jdep.write(name);
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
std::ostream & operator<<(std::ostream & outs, const Departure<TRAIT> & dy) {
  outs << "Departure Infos for " << dy.dep_.size() << " types:";
  BOOST_FOREACH(const DepType<TRAIT> & jdep, dy.dep_) outs << "\n" << jdep;
  return outs;
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
bool Departure<TRAIT>::compatible(const Departure & d1, const Departure & d2) {
  int ii = d1.size();
  bool same = (d2.size() == ii);
  if (same) {
    for (int jj = 0; jj < ii; ++jj)
      same = same && TheDep::compatible(d1.dep_[jj], d2.dep_[jj]);
  }
  return same;
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
void Departure<TRAIT>::update(TheObservation & obs) const {
  for (int jj = 0; jj < size(); ++jj) {
    dep_[jj].update(obs[jj]);
  }
}
// -----------------------------------------------------------------------------
}  // namespace oops
