/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_BASE_DEPARTURE_H_
#define OOPS_BASE_DEPARTURE_H_

#include <iostream>
#include <string>

#include <boost/ptr_container/ptr_vector.hpp>

#include "oops/base/Observation.h"
#include "oops/base/ObsTrajectory.h"
#include "oops/base/GeneralizedDeparture.h"
#include "oops/observations/DepType.h"
#include "util/dot_product.h"
#include "util/DateTime.h"
#include "util/Duration.h"

namespace oops {

template<typename TRAIT> class Departure;

template<typename TRAIT>
std::ostream & operator<< (std::ostream &, const Departure<TRAIT> &);

/// Difference between two observation vectors.
/*!
 * A departure is the difference between two observations.
 * The archetypal example is \f$ \mathbf{y} - {\cal H}(\mathbf{x}) \f$.
 */

// -----------------------------------------------------------------------------
template <typename TRAIT> class Departure : public GeneralizedDeparture {
  typedef typename TRAIT::ObsAuxIncr    ObsAuxIncr;
  typedef typename TRAIT::ObsTraj       TypeTraj;
  typedef typename TRAIT::IncrementBase Increment;
  typedef Observation<TRAIT>            TheObservation;
  typedef DepType<TRAIT>                TheDep;
  typedef ObsType<TRAIT>                TheObs;
  typedef ObsTrajectory<TypeTraj>       ObsTraj;

 public:
// Constructors and destructor
  explicit Departure(const TheObservation &);
  Departure(const TheObservation &, const TheObservation &);
  Departure(const Departure &, const bool copyDep = true);
  ~Departure();

/// Restrict use to a given subwindow
  void subwindow(const util::DateTime &, const util::DateTime &) const;

// Linear algebra operators
  Departure & operator=(const Departure &);
  Departure & operator+=(const Departure &);
  Departure & operator-=(const Departure &);
  Departure & operator*=(const double &);
  Departure & operator*=(const Departure &);
  Departure & operator/=(const Departure &);
  void zero();
  void invert();
/// \f$ \mathbf{y} = a \mathbf{x} + \mathbf{y} \f$
  void axpy(const double &, const Departure &);
  void random();
  double dot_product_with(const Departure &) const;

/// Setup linear observation operators
  void obsOperatorTLSetup(const util::Duration &);
/// Linear observation operators
  void obsOperatorTL(const Increment &, const ObsAuxIncr &,
                     const ObsTraj &);
/// Finalize linear observation operators
  void obsOperatorTLFinal(const ObsAuxIncr &, const ObsTraj &);

/// Setup adjoint observation operators
  void obsOperatorADFinal(const util::Duration &,
                          ObsAuxIncr &, const ObsTraj &) const;
/// Adjoint observation operators
  void obsOperatorAD(Increment &, ObsAuxIncr &, const ObsTraj &) const;
/// Finalize adjoint observation operators
  void obsOperatorADSetup() const;

/// Read departures
  void read(const std::string &);
/// Write departures
  void write(const std::string &) const;
/// Print human readable departure informations
  friend std::ostream & operator<< <> (std::ostream &, const Departure &);
/// Check that two objects relate to the same observations
  static bool compatible(const Departure &, const Departure &);

/// Add to Observation
  void update(TheObservation &) const;

/// Access departures by type
  TheDep & operator[](const int ii) {return dep_[ii];}
/// Access departures by type
  const TheDep & operator[](const int ii) const {return dep_[ii];}
/// Number of observations types
  int size() const {return dep_.size();}

 private:
  boost::ptr_vector<TheDep> dep_;  //!< Departures for each observation type
  util::DateTime winbgn_;          //!< Begining of assimilation window
  util::DateTime winend_;          //!< End of assimilation window
  mutable util::DateTime bgn_;     //!< Begining of currently active observations
  mutable util::DateTime end_;     //!< End of currently active observations
  mutable util::Duration hslot_;   //!< Half time slot
};
// -----------------------------------------------------------------------------

}  // namespace oops

#include "oops/base/Departure.cc"

#endif  // OOPS_BASE_DEPARTURE_H_
