/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 */

#include <string>

#include "logger/Logger.h"
#include "oops/base/Geometry.h"
#include "util/asser1.h"
#include "util/abor1_cpp.h"
#include "util/dot_product.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/Duration.h"

namespace oops {

template<typename TRAIT>
Ensemble<TRAIT>::Ensemble(const util::DateTime& validTime,
		const util::Config& conf) :
		validTime_(validTime), referenceTime_(
				util::DateTime(conf.getData("reference_time"))), term_(
				util::Duration(validTime_ - referenceTime_)), name_(
				conf.getData("name")), type_(conf.getData("type")), rank_(
				conf.getInt("members")) {
	LOG(Trace) << "Ensemble:contructor done";
}

// -----------------------------------------------------------------------------
template<typename TRAIT>
typename TRAIT::StateBase* Ensemble<TRAIT>::mean(const State& xb,
		const util::Config& confTLM, const Geometry & resol) {
	double r = 1.0 / ((double) rank_);
	util::Config varConf(confTLM);
	varConf.addElement(".", "variables");
	varConf.setElement("./variables", "nl");
	Variables vars(varConf);

	Accumulator<State, State> *avg = new Accumulator<State, State>(resol, vars,
			validTime_);
	for (int jm = 0; jm < rank_; ++jm) {
		State xx(xb);
		std::stringstream sss;
		sss << jm + 1;
		std::string filename = name_ + "." + type_ + "." + sss.str() + "."
				+ referenceTime_.toString() + "." + term_.toString();
		util::Config conf(confTLM);
		conf.addElement(".", "filename");
		conf.setElement("filename", filename);
		conf.addElement(".", "date");
		conf.setElement("date", validTime_.toString());
		xx.read(conf);
		avg->accumul(r, xx);
	}
	return avg;
}

// -----------------------------------------------------------------------------
template<typename TRAIT>
void Ensemble<TRAIT>::linearize(const State& xb, const util::Config& confTLM,
		const Geometry & resol) {
	boost::scoped_ptr<State> avg(mean(xb, confTLM, resol));
	boost::ptr_vector<Increment>* ensemble = new boost::ptr_vector<Increment>();
	double r = 1.0 / sqrt(((double) rank_ - 1.0));

	util::Config varConf(confTLM);
	varConf.addElement(".", "variables");
	varConf.setElement("./variables", "nl");
	Variables vars(varConf);

	for (int jm = 0; jm < rank_; ++jm) {
		State xx(xb);
		std::stringstream sss;
		sss << jm + 1;
		std::string filename = name_ + "." + type_ + "." + sss.str() + "."
				+ referenceTime_.toString() + "." + term_.toString();
		util::Config conf(confTLM);
		conf.addElement(".", "filename");
		conf.setElement("filename", filename);
		conf.addElement(".", "date");
		conf.setElement("date", validTime_.toString());
		xx.read(conf);

		Increment * dx = new Increment(resol, vars, validTime_);
		dx->diff(xx, *avg);
		(*dx) *= r;
		ensemble->push_back(dx);
	}
	ensemblePerturbs_.reset(ensemble);
}

// -----------------------------------------------------------------------------

}// namespace oops
