/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 */

#include <boost/foreach.hpp>

#include "logger/Logger.h"
#include "util/abor1_cpp.h"
#include "util/asser1.h"
#include "util/Duration.h"
#include "oops/base/LocalizationMatrix.h"

using namespace util;

namespace oops {

// -----------------------------------------------------------------------------
/// Constructor, destructor
// -----------------------------------------------------------------------------

template<typename TRAIT>
EnsembleCovariance<TRAIT>::EnsembleCovariance(
		const Geometry & resol, const util::Config & config) :
		config_(config) {
	LOG(Trace) << "EnsembleCovariance created.";
}

// -----------------------------------------------------------------------------
template<typename TRAIT>
EnsembleCovariance<TRAIT>::~EnsembleCovariance() {
	LOG(Trace) << "EnsembleCovariance destructed";
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void EnsembleCovariance<TRAIT>::linearize(const State & xb,
	const Geometry & resol, const util::Config & tlm) {
	EnsemblePtr_ ens_k(new Ensemble_(xb.validTime(), config_));
	ens_k->linearize(xb, tlm, resol);
	EnsemblesCollection_::getInstance().put(xb.validTime(), ens_k);
	resol_.reset(new Geometry(resol));
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void EnsembleCovariance<TRAIT>::mult(const Increment & dxi,
		Increment & dxo) const {
	EnsemblePtr_ e1 = EnsemblesCollection_::getInstance()[dxi.validTime()];
	EnsemblePtr_ e2 = EnsemblesCollection_::getInstance()[dxo.validTime()];

	const util::Config conf(config_, "localization");
	LocalizationMatrix<TRAIT> loc(*resol_, conf);
	for (unsigned m = 0; m < e1->size(); ++m) {
		Increment dx(dxi);
		dx.schur_product_with((*e1)[m]);
		loc.multiply(dx);
		dx.schur_product_with((*e2)[m]);

		// We can't use '=' , dxo and dx aren't at the same time
		dxo.axpy(1.0, dx, false);
	}
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void EnsembleCovariance<TRAIT>::inv_mult(const Increment & dxi,
		Increment &) const {
	ABORT("EnsembleCovariance::inv_mult: not implemented");
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void EnsembleCovariance<TRAIT>::randomize(Increment &) const {
	ABORT("EnsembleCovariance::randomize: Would it make sense?");
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
typename TRAIT::IncrementBase * EnsembleCovariance<TRAIT>::getIncrement() const {
  ABORT("EnsembleCovariance::getIncrement: Would it make sense? (No DateTime in this B)");
  return NULL;
}
// -----------------------------------------------------------------------------
}// namespace oops
