/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include "logger/Logger.h"
#include "util/abor1_cpp.h"
#include "util/asser1.h"

namespace oops {

// -----------------------------------------------------------------------------
/// Constructor, destructor
// -----------------------------------------------------------------------------
template<typename TRAIT>
HybridCovariance<TRAIT>::HybridCovariance(const Geometry & resol,
                                          const util::Config & config)
{
  const util::Config staConf(config, "static");
  static_.reset(new StaticCovariance_(resol, staConf));
  const util::Config ensConf(config, "ensemble");
  ensemble_.reset(new EnsembleCovariance_(resol,  ensConf));
  ensWeight_ = config.getReal("ensemble_weight");
  ASSERT(ensWeight_ > 0.0 && ensWeight_ <= 1.0);
  LOG(Trace) << "HybridCovariance created.";
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
HybridCovariance<TRAIT>::~HybridCovariance() {
  LOG(Trace) << "HybridCovariance destructed";
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void HybridCovariance<TRAIT>::linearize(const State & xb,
                                        const Geometry & resol,
                                        const util::Config & tlConf) {
  ASSERT(xb.validTime() == static_->validTime());
  static_->linearize(xb, resol, tlConf);
  ensemble_->linearize(xb, resol, tlConf);
  LOG(Trace) << "HybridCovariance linearized: time = " << static_->validTime();
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void HybridCovariance<TRAIT>::mult(const Increment & dxi, Increment & dxo) const {
  ASSERT(dxi.validTime() == static_->validTime());
  static_->mult(dxi, dxo);
  dxo *= (1.0-ensWeight_);
  Increment tmp(dxo);
  ensemble_->mult(dxi, tmp);
  dxo.axpy(ensWeight_, tmp);
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void HybridCovariance<TRAIT>::inv_mult(const Increment & dxi, Increment &) const {
  ASSERT(dxi.validTime() == static_->validTime());
  ABORT("HybridCovariance::inv_mult: not implemented");
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void HybridCovariance<TRAIT>::randomize(Increment &) const {
  ABORT("HybridCovariance::randomize: Would it make sense?");
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
typename TRAIT::IncrementBase * HybridCovariance<TRAIT>::getIncrement() const {
  return static_->getIncrement();
}
// -----------------------------------------------------------------------------
}  // namespace oops
