/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_BASE_MODELINCREMENT_H_
#define OOPS_BASE_MODELINCREMENT_H_

#include "oops/base/CostProcessor.h"
#include "oops/base/PostProcessor.h"
#include "util/dot_product.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/Duration.h"

namespace oops {

/// ModelIncrement Class: Difference between two states
/*!
 *  Some fields that are present in a State may not be present in
 *  an Increment. The ModelIncrement contains everything that is needed by
 *  the tangent-linear and adjoint models, and has methods to propagate
 *  itself using these models.
 * 
 * The ModelIncrement class inherits from the model dependent IncrementBase
 * class and adds the TL/AD model functionality to it.
 * The minimum required interface for the IncrementBase class is:
 * ~~~~~~~~~~~~~~~~~~~~{.cc}
 * class IncrementBase {
 *  public:
 * 
 * // Constructor, destructor
 *   IncrementBase(const oops::Geometry &, const oops::Variables &, const util::DateTime &);
 *   IncrementBase(const IncrementBase &);
 *   virtual ~IncrementBase();
 * 
 * // Basic operators
 *   void diff(const StateBase &, const StateBase &);
 *   IncrementBase & operator =(const IncrementBase &);
 *   IncrementBase & operator+=(const IncrementBase &);
 *   IncrementBase & operator-=(const IncrementBase &);
 *   IncrementBase & operator*=(const double &);
 *   void zero(const util::Duration &);
 *   void random();
 *   void axpy(const double &, const IncrementBase &);
 *   double dot_product_with(const IncrementBase &) const;
 * 
 * // Interpolate to observation location
 *   void interpolateTL(const Locations &, LocalModelValues &) const;
 *   void interpolateAD(const Locations &, const LocalModelValues &);
 * 
 * // Interpolate full fields
 *   void changeResolution(const IncrementBase &);
 * 
 * // I/O and diagnostics
 *   void read(const util::Config &);
 *   void write(const util::Config &) const;
 *   void prtNorms(const std::string &) const;
 *   const util::DateTime validTime() const;
 * 
 *  protected:
 *   void initTL(const Model &, const ModelTraj &);
 *   void initAD(const Model &, const ModelTraj &);
 *   void stepTL(const Model &, const ModelTraj &, const ModelAuxIncr &);
 *   void stepAD(const Model &, const ModelTraj &, ModelAuxIncr &);
 * };
 * ~~~~~~~~~~~~~~~~~~~~
 *
 * The resulting public interface of the ModelIncrement class, including inherited methods,
 * will be the following:
 * ~~~~~~~~~~~~~~~~~~~~{.cc}
 * // Constructor, destructor   
 *   ModelIncrement(const IncrementBase &, const Model &);
 *   ModelIncrement(const Model &, const util::DateTime &);
 *   virtual ~ModelIncrement();
 * 
 * // Basic operators
 *   void diff(const StateBase &, const StateBase &);
 *   ModelIncrement & operator =(const ModelIncrement &);
 *   ModelIncrement & operator+=(const ModelIncrement &);
 *   ModelIncrement & operator-=(const ModelIncrement &);
 *   ModelIncrement & operator*=(const double &);
 *   void zero(const util::Duration &);
 *   void random();
 *   void axpy(const double &, const ModelIncrement &);
 *   double dot_product_with(const ModelIncrement &) const;
 * 
 * // Run forecast
 *   void forecastTL(const ModelAuxIncr &, const util::DateTime &, const ModelTraj &,
 *                   PostProcessor<IncrementBase> &, CostProcessor<IncrementBase> &);
 *   void forecastAD(ModelAuxIncr &, const util::DateTime &, const ModelTraj &,
 *                   PostProcessor<IncrementBase> &, CostProcessor<IncrementBase> &);
 * 
 * // Interpolate to observation location
 *   void interpolateTL(const Locations &, LocalModelValues &) const;
 *   void interpolateAD(const Locations &, const LocalModelValues &);
 * 
 * // Interpolate full fields
 *   void changeResolution(const ModelIncrement &);
 * 
 * // I/O and diagnostics
 *   void read(const util::Config &);
 *   void write(const util::Config &) const;
 *   void prtNorms(const std::string &) const;
 *   const util::DateTime validTime() const;
 * ~~~~~~~~~~~~~~~~~~~~
 *
 * The other (protected) methods inherited from TRAIT::IncrementBase will not
 * be visible by OOPS.
 */

// -----------------------------------------------------------------------------

template <typename TRAIT> class ModelIncrement: public TRAIT::IncrementBase {
  typedef typename TRAIT::ModelData        Model;
  typedef typename TRAIT::ModelAuxIncr     ModelAuxIncr;
  typedef typename TRAIT::ModelTraj        ModelTraj;
  typedef typename TRAIT::ModelStateError  ModelSpaceError;
  typedef typename TRAIT::IncrementBase    Increment;

 public:
  ModelIncrement(const Increment &, const Model &);
  ModelIncrement(const Model &, const util::DateTime &);
  ~ModelIncrement();

/// Run the tangent linear forecast
  void forecastTL(const ModelAuxIncr &, const util::DateTime &, const ModelTraj &,
                  PostProcessor<Increment> &, CostProcessor<Increment> &);

/// Run the adjoint forecast
  void forecastAD(ModelAuxIncr &, const util::DateTime &, const ModelTraj &,
                  PostProcessor<Increment> &, CostProcessor<Increment> &);

 private:
  const Model & tlm_;
};

// -----------------------------------------------------------------------------
}  // namespace oops

#include "oops/base/ModelIncrement.cc"

#endif  // OOPS_BASE_MODELINCREMENT_H_
