/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_BASE_MODELSPACECOVARIANCE_H_
#define OOPS_BASE_MODELSPACECOVARIANCE_H_

#include <map>
#include <string>

#include <boost/noncopyable.hpp>

#include "oops/base/Geometry.h"

namespace util {
  class Config;
}

namespace oops {
  class Model;

// -----------------------------------------------------------------------------

/* It would be better if ModelSpaceCovariance inherited from the 
   ModelSpaceErrorBase in the TRAIT definition and other formulations
   like the hybrid derive from that by over-writing what needs to
   be different (but not getIncrement for example).  YT */

/// Abstract base class for model space error covariances.
template <typename TRAIT>
class ModelSpaceCovariance {
  typedef typename TRAIT::StateBase     State;
  typedef typename TRAIT::IncrementBase Increment;

 public:
  virtual ~ModelSpaceCovariance() {}

  virtual void linearize(const State &, const Geometry &, const util::Config &) =0;

  virtual void mult(const Increment &, Increment &) const =0;
  virtual void inv_mult(const Increment &, Increment &) const =0;

  virtual Increment * getIncrement() const =0;
  virtual void randomize(Increment &) const =0;
};

// -----------------------------------------------------------------------------

/// Covariance Factory
template <typename TRAIT>
class CovarianceFactory {
 public:
  static ModelSpaceCovariance<TRAIT> * create(const util::Config &,
                                              const Geometry &);
  virtual ~CovarianceFactory() { makers_.clear(); }
 protected:
  explicit CovarianceFactory(const std::string &);
 private:
  virtual ModelSpaceCovariance<TRAIT> * make(const util::Config &,
                                             const Geometry &) =0;
  static std::map < std::string, CovarianceFactory<TRAIT> * > makers_;
};

template<class TRAIT, class COVAR>
class CovarMaker : public CovarianceFactory<TRAIT> {
  virtual ModelSpaceCovariance<TRAIT> * make(const util::Config & conf,
                                             const Geometry & resol) {
    return new COVAR(resol, conf);
  }
 public:
  explicit CovarMaker(const std::string & name) : CovarianceFactory<TRAIT>(name) {}
};

// -----------------------------------------------------------------------------

}  // namespace oops

#include "oops/base/ModelSpaceCovariance.cc"

#endif  // OOPS_BASE_MODELSPACECOVARIANCE_H_
