/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_BASE_MODELSTATE_H_
#define OOPS_BASE_MODELSTATE_H_

#include "oops/base/PostProcessor.h"

namespace util {
  class Config;
  class DateTime;
  class Duration;
}

namespace oops {

/// Encapsulates the model state.
/*!
 * The ModelState class inherits from the model dependent StateBase class
 * and makes the forecast method available.
 *
 * The minimum required interface for the StateBase class is:
 * ~~~~~~~~~~~~~~~~~~~~{.cc}
 * class StateBase {
 *  public:
 * 
 * // Constructor, destructor
 *   StateBase(const oops::Geometry &, const oops::Variables &, const util::DateTime &);
 *   StateBase(const oops::Geometry &, const util::Config &);
 *   StateBase(const StateBase &);
 *   virtual ~StateBase();
 *   StateBase & operator=(const StateBase &);
 * 
 * // Interpolate to observation location
 *   void interpolate(const Locations &, LocalModelValues &) const;
 * 
 * // Interpolate full fields
 *   void changeResolution(const StateBase &);
 * 
 * // Interactions with Increment
 *   StateBase & operator+=(const IncrementBase &);
 * 
 * // I/O and diagnostics
 *   void read(const util::Config &);
 *   void write(const util::Config &) const;
 *   void prtNorms(const std::string &) const;
 *   const util::DateTime validTime() const;
 * 
 *  protected:
 *    void init(const Model &);
 *    void step(const Model &, const oops::NullModelAux &);
 * };
 * ~~~~~~~~~~~~~~~~~~~~
 *
 * The resulting public interface of the ModelState class, including
 * inherited methods, will be the following:
 * ~~~~~~~~~~~~~~~~~~~~{.cc}
 * // Constructor, destructor   
 *   ModelState(const Geometry &, const Model &, const util::Config &);
 *   ModelState(const StateBase &, const Model &);
 *   ~ModelState();
 * 
 *   ModelState & operator=(const ModelState &);
 *   ModelState & operator+=(const Increment &);
 * 
 * // Run forecast
 *   void forecast(const ModelAuxCtrl &, const util::DateTime &, PostProcessor<StateBase> &);
 * 
 * // Interpolate to observation location
 *   void interpolate(const Locations &, LocalModelValues &) const;
 * 
 * // Interpolate full fields
 *   void changeResolution(const StateBase &);
 * 
 * // I/O and diagnostics
 *   void read(const util::Config &);
 *   void write(const util::Config &) const;
 *   void prtNorms(const std::string &) const;
 *   const util::DateTime validTime() const;
 * ~~~~~~~~~~~~~~~~~~~~
 *
 * The other (protected) methods inherited from TRAIT::StateBase will not
 * be visible by OOPS.
 */

// -----------------------------------------------------------------------------
template <typename TRAIT> class ModelState: public TRAIT::StateBase {
  typedef typename TRAIT::ModelData     Model;
  typedef typename TRAIT::ModelAuxCtrl  ModelAuxCtrl;
  typedef typename TRAIT::StateBase     StateBase;

 public:
  ModelState(const Model &, const util::Config &);
  ModelState(const StateBase &, const Model &);
  ~ModelState();

/// Run a forecast
  void forecast(const ModelAuxCtrl &, const util::DateTime &,
                PostProcessor<StateBase> &);

 private:
  const Model & model_;
};
// -----------------------------------------------------------------------------

}  // namespace oops

#include "oops/base/ModelState.cc"

#endif  // OOPS_BASE_MODELSTATE_H_
