/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include <sstream>
#include <string>

#include "logger/Logger.h"
#include "oops/observations/ObsType.h"
#include "util/Config.h"

namespace oops {

// -----------------------------------------------------------------------------
template <typename TRAIT>
ObsErrorCovariance<TRAIT>::ObsErrorCovariance(const Observ & obs,
                                              const util::Config & conf) {
  const unsigned nobs = conf.getElementSize("Observation");
  for (unsigned jj = 0; jj < nobs; ++jj) {
    const util::Config typeConfig(conf, "Observation", jj);
    LOG(Configs) << "Observation " << jj << " :\n" <<  typeConfig;
    obserr_.push_back(ObsErrorFactory<TRAIT>::create(obs[jj], typeConfig));
  }
  LOG(Trace) << "ObsErrorCovariance created";
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
ObsErrorCovariance<TRAIT>::~ObsErrorCovariance() {
  LOG(Trace) << "ObsErrorCovariance destructed";
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
void ObsErrorCovariance<TRAIT>::linearize(const Observ & ofg) {
  for (unsigned jj = 0; jj < obserr_.size(); ++jj)
    obserr_[jj].linearize(ofg[jj]);
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
double ObsErrorCovariance<TRAIT>::calcJo(const Depart & ydep, Depart & yout) const {
  double zjo = 0.0;
  int nobs = 0;
  int iobs = 0;
  LOG(Info) << "Jo Table Start";
  LOG(Info) << "      Jo      nobs     Jo/n    sigma_o";
  for (unsigned jj = 0; jj < obserr_.size(); ++jj) {
    zjo += obserr_[jj].calcJo(ydep[jj], yout[jj], iobs);
    nobs += iobs;
  }
  if (nobs > 0) {
    LOG(Info) << "Total";
    LOG(Info) << "    " << zjo << "    " << nobs << "    " << zjo/nobs;
  } else {
    LOG(Info) << "WARNING: *** No observations !!! ***";
  }
  LOG(Info) << "Jo Table End";
  return zjo;
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
void ObsErrorCovariance<TRAIT>::mult(const Depart & dyin,
                                     Depart & dyout) const {
  for (unsigned jj = 0; jj < obserr_.size(); ++jj)
    obserr_[jj].mult(dyin[jj], dyout[jj]);
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
void ObsErrorCovariance<TRAIT>::mult_ad(const Depart & dyin,
                                        Depart & dyout) const {
  for (unsigned jj = 0; jj < obserr_.size(); ++jj)
    obserr_[jj].mult_ad(dyin[jj], dyout[jj]);
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
void ObsErrorCovariance<TRAIT>::inv_mult(const Depart & dyin,
                                         Depart & dyout) const {
  for (unsigned jj = 0; jj < obserr_.size(); ++jj)
    obserr_[jj].inv_mult(dyin[jj], dyout[jj]);
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
void ObsErrorCovariance<TRAIT>::inv_mult_ad(const Depart & dyin,
                                            Depart & dyout) const {
  for (unsigned jj = 0; jj < obserr_.size(); ++jj)
    obserr_[jj].inv_mult_ad(dyin[jj], dyout[jj]);
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
void ObsErrorCovariance<TRAIT>::randomize(Depart & dy) const {
  for (unsigned jj = 0; jj < obserr_.size(); ++jj)
    obserr_[jj].randomize(dy[jj]);
}
// -----------------------------------------------------------------------------

}  // namespace oops
