/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include <boost/foreach.hpp>

#include "logger/Logger.h"
#include "util/asser1.h"

namespace oops {
// -----------------------------------------------------------------------------
template <typename TRAIT>
Observation<TRAIT>::Observation(const util::Config & config,
                                const util::DateTime & t1, const util::DateTime & t2)
  : obs_(0), winbgn_(t1), winend_(t2), bgn_(winbgn_), end_(winend_), hslot_(0)
{
  const unsigned nb_obs = config.getElementSize("Observation");
  LOG(Configs) << "Observation:Observation nb_obs= " << nb_obs;
  for (unsigned obs = 0; obs < nb_obs; ++obs) {
    const util::Config typeConfig(config, "Observation", obs);
    LOG(Configs) << "Observation " << obs << " :\n"
                 <<  typeConfig;
    obs_.push_back(new TheObs(typeConfig));
    LOG(Trace) << "Observation:Observation ObsType created";
  }
  LOG(Trace) << "Observation created";
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
Observation<TRAIT>::Observation(const Observation & other, const bool copyObs)
  : obs_(0), winbgn_(other.winbgn_), winend_(other.winend_),
    bgn_(winbgn_), end_(winend_), hslot_(0)
{
  BOOST_FOREACH(const TheObs & jobs, other.obs_) {
    obs_.push_back(new TheObs(jobs, copyObs));
  }
  LOG(Trace) << "Observation copy-created";
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
Observation<TRAIT>::~Observation() {
  LOG(Trace) << "Observation:~Observation destructor";
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
void Observation<TRAIT>::subwindow(const util::DateTime & t1,
                                   const util::DateTime & t2) {
  bgn_ = t1;
  end_ = t2;
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
Observation<TRAIT> & Observation<TRAIT>::operator=(const Observation & rhs) {
  ASSERT(compatible(*this, rhs));
  bgn_ = rhs.bgn_;
  end_ = rhs.end_;
  obs_ = rhs.obs_;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
void Observation<TRAIT>::obsOperatorSetup(const util::Duration & dt) {
  hslot_ = dt;
  hslot_ /= 2;
  BOOST_FOREACH(TheObs & jobs, obs_) {
    jobs.setup(bgn_, end_);
  }
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
void Observation<TRAIT>::obsOperator(const State & xx,
                                     const ObsAuxCtrl & octl) {
  util::DateTime t1(xx.validTime()-hslot_);
  util::DateTime t2(xx.validTime()+hslot_);
  if (t1 < bgn_) t1 = bgn_;
  if (t2 > end_) t2 = end_;

  BOOST_FOREACH(TheObs & jobs, obs_) {
    jobs.obsOperator(xx, octl, t1, t2);
  }
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
void Observation<TRAIT>::obsOperatorFinal(const ObsAuxCtrl & octl) {
  BOOST_FOREACH(TheObs & jobs, obs_) {
    jobs.finalize(octl);
  }
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
void Observation<TRAIT>::obsOperatorSetup(const util::Duration & dt,
                                          ObsTraj & traj) {
  this->obsOperatorSetup(dt);
// The if below is not very safe. YT
  if (traj.size() == 0) {
    BOOST_FOREACH(TheObs & jobs, obs_) traj.append(jobs.getTraj());
  }
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
void Observation<TRAIT>::obsOperator(const State & xx,
                                     const ObsAuxCtrl & octl,
                                     ObsTraj & traj) {
  util::DateTime t1(xx.validTime()-hslot_);
  if (t1 < bgn_) t1 = bgn_;
  util::DateTime t2(xx.validTime()+hslot_);
  if (t2 > end_) t2 = end_;

  for (int jj = 0; jj < size(); ++jj) {
    obs_[jj].obsOperator(xx, octl, t1, t2, traj[jj]);
  }
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
void Observation<TRAIT>::obsOperatorFinal(const ObsAuxCtrl & octl,
                                          ObsTraj & traj) {
  for (int jj = 0; jj < size(); ++jj) {
    obs_[jj].finalize(octl, traj[jj]);
  }
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
void Observation<TRAIT>::generate() {
  BOOST_FOREACH(TheObs & jobs, obs_) jobs.generate(bgn_, end_);
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
void Observation<TRAIT>::read(const util::Config & config) {
  const unsigned nb_obs = config.getElementSize("Observation");
  ASSERT(nb_obs == obs_.size());
  for (unsigned obs = 0; obs < nb_obs; ++obs) {
    const util::Config typeConfig(config, "Observation", obs);
    obs_[obs].read(typeConfig.getData("ObsData/obsvalue"));
  }
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
void Observation<TRAIT>::write(const std::string & name) const {
  BOOST_FOREACH(const TheObs & jobs, obs_) {
    jobs.write(name);
  }
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
std::ostream & operator<<(std::ostream & outs, const Observation<TRAIT> & yy) {
  outs << "Observation Infos for " << yy.obs_.size() << " types:";
  BOOST_FOREACH(const ObsType<TRAIT> & jobs, yy.obs_) outs << "\n" << jobs;
  return outs;
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
bool Observation<TRAIT>::compatible(const Observation & o1,
                                    const Observation & o2) {
  int ii = o1.size();
  bool same = (o2.size() == ii);
  if (same) {
    for (int jj = 0; jj < ii; ++jj)
      same = same && TheObs::compatible(o1.obs_[jj], o2.obs_[jj]);
  }
  return same;
}
// -----------------------------------------------------------------------------
}  // namespace oops
