/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_BASE_OBSERVATION_H_
#define OOPS_BASE_OBSERVATION_H_

#include <iostream>
#include <string>

#include <boost/ptr_container/ptr_vector.hpp>

#include "oops/base/ObsTrajectory.h"
#include "oops/observations/ObsType.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/Duration.h"

namespace oops {

template<typename TRAIT> class Observation;

template<typename TRAIT>
std::ostream & operator<< (std::ostream &, const Observation<TRAIT> &);

/// Observation Class.
/*!
 *  Contains observed values or their model equivalents
 */

// -----------------------------------------------------------------------------
template <typename TRAIT> class Observation {
  typedef typename TRAIT::StateBase   State;
  typedef typename TRAIT::ObsAuxCtrl  ObsAuxCtrl;
  typedef typename TRAIT::ObsTraj     TypeTraj;
  typedef ObsTrajectory<TypeTraj>     ObsTraj;
  typedef ObsType<TRAIT>              TheObs;

 public:
  Observation(const util::Config &, const util::DateTime &, const util::DateTime &);
  Observation(const Observation &, const bool copyObs = true);
  ~Observation();
  Observation & operator=(const Observation &);

/// Restrict use to a given subwindow
  void subwindow(const util::DateTime &, const util::DateTime &);

/// Setup observation operators
  void obsOperatorSetup(const util::Duration &);
/// Create the model equivalents of the observations by applying H
  void obsOperator(const State &, const ObsAuxCtrl &);
/// Finalize observation operators
  void obsOperatorFinal(const ObsAuxCtrl &);

/// Setup observation operators and associated trajectory
  void obsOperatorSetup(const util::Duration &, ObsTraj &);
/// Create the model equivalents of the observations by applying H
  void obsOperator(const State &, const ObsAuxCtrl &, ObsTraj &);
/// Finalize observation operators and associated trajectory
  void obsOperatorFinal(const ObsAuxCtrl &, ObsTraj &);

/// Generate observation distribution
  void generate();

/// Read observations
  void read(const util::Config &);
/// Write observations
  void write(const std::string &) const;
/// Print human readable observation informations
  friend std::ostream & operator<< <> (std::ostream &, const Observation &);
/// Check that two objects relate to the same observations
  static bool compatible(const Observation &, const Observation &);

/// Access observation types
  TheObs & operator[](const int ii) {return obs_[ii];}
/// Access observation types
  const TheObs & operator[](const int ii) const {return obs_[ii];}
/// Number of observations types
  int size() const {return obs_.size();}
/// Start of assimilation window
  const util::DateTime & windowStart() const {return winbgn_;}
/// End of assimilation window
  const util::DateTime & windowEnd() const {return winend_;}

 private:
  boost::ptr_vector<TheObs> obs_;  //!< Store observations by type
  util::DateTime winbgn_;          //!< Begining of assimilation window
  util::DateTime winend_;          //!< End of assimilation window
  util::DateTime bgn_;             //!< Begining of currently active observations
  util::DateTime end_;             //!< End of currently active observations
  util::Duration hslot_;           //!< Half time slot
};
// -----------------------------------------------------------------------------

}  // namespace oops

#include "oops/base/Observation.cc"

#endif  // OOPS_BASE_OBSERVATION_H_
