/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_BASE_PROCESSORBASE_H_
#define OOPS_BASE_PROCESSORBASE_H_

#include <boost/noncopyable.hpp>

#include "util/Config.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/PostTimer.h"

namespace oops {

// -----------------------------------------------------------------------------

/// Handles post-processing of model fields.
/*!
 *  ProcessorBase is the base class for all state post processors, it
 *  is mostly used so that PostProcessor can hold a vector of such
 *  processors.
 *  By default processing is performed on every call.
 */

template <typename FLDS> class ProcessorBase : private boost::noncopyable {
 public:
/// Constructors and basic operators
  ProcessorBase() : timer_() {}
  explicit ProcessorBase(const util::Duration & freq) : timer_(freq) {}
  explicit ProcessorBase(const util::Config & conf) : timer_(conf) {}
  ProcessorBase(const util::DateTime & start, const util::Config & conf)
    : timer_(start, conf) {}
  ProcessorBase(const util::DateTime & start, const util::DateTime & finish,
                const util::Duration & freq = util::Duration(0))
    : timer_(start, finish, freq) {}

  virtual ~ProcessorBase() {}

/// Setup
  void initialize(const util::DateTime & bgn, const util::DateTime & end,
                  const util::Duration & tstep) {
    timer_.initialize(bgn, end, tstep);
    this->doInitialize(bgn, end, tstep);
  }

/// Process state or increment
  void process(const FLDS & xx) {
    if (timer_.itIsTime(xx.validTime())) this->doProcessing(xx);
  }

/// Final
  void finalize() {
    this->doFinalize();
  }

 private:
  PostTimer timer_;

/// Actual processing
  virtual void doProcessing(const FLDS &) =0;
  virtual void doInitialize(const util::DateTime &, const util::DateTime &,
                            const util::Duration &) {}
  virtual void doFinalize() {}
};

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_PROCESSORBASE_H_
