/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include <cmath>

namespace oops {

// -----------------------------------------------------------------------------

template <typename INCR, typename FLDS>
WeightedDiff<INCR, FLDS>::WeightedDiff(const util::DateTime & vt,
                                       const util::Duration & span,
                                       const Geometry & resol,
                                       const util::Config & config,
                                       const util::Duration & tstep,
                                       WeightingFct & wfct)
  : ProcessorBase<FLDS>(vt-span/2, vt+span/2),
    wfct_(wfct), weights_(), avg_(0), sum_(0.0), linit_(false),
    vtime_(vt), bgn_(vt-span/2), end_(vt+span/2), tstep_(tstep),
    bgnleg_(), endleg_(), vars_(config), resol_(resol), lsetup_(false)
{}

// -----------------------------------------------------------------------------

template <typename INCR, typename FLDS>
void WeightedDiff<INCR, FLDS>::setup() {
  avg_ = new Accumulator<INCR, FLDS>(resol_, vars_, vtime_);
  sum_ = 0.0;
  lsetup_ = true;
}

// -----------------------------------------------------------------------------

template <typename INCR, typename FLDS>
INCR * WeightedDiff<INCR, FLDS>::release() {
  LOG(Debug) << "WeightedDiff: release sum = " << sum_
             << ", bgnleg_ = " << bgnleg_ << ", endleg_ = " << endleg_
             << ", bgn_ = " << bgn_ << ", end_ = " << end_;
  ASSERT(linit_);
  ASSERT(std::abs(sum_) < 1.0e-8);
  lsetup_ = false;
  return avg_;
}

// -----------------------------------------------------------------------------

template <typename INCR, typename FLDS>
void WeightedDiff<INCR, FLDS>::doInitialize(const util::DateTime & bgn,
                                            const util::DateTime & end,
                                            const util::Duration & tstep) {
  if (!linit_ && bgn <= end_ && end >= bgn_) {
    if (tstep_ == util::Duration(0)) tstep_ = tstep;
    ASSERT(tstep_ > util::Duration(0));
    weights_ = wfct_.setWeights(bgn_, end_, tstep_);
    linit_ = true;
    ASSERT(weights_.find(vtime_) != weights_.end());
    weights_[vtime_] -= 1.0;
  }
  bgnleg_ = bgn;
  endleg_ = end;
  LOG(Debug) << "WeightedDiff: initialized"
             << " bgnleg_ = " << bgnleg_ << ", endleg_ = " << endleg_
             << ", bgn_ = " << bgn_ << ", end_ = " << end_;
}

// -----------------------------------------------------------------------------

template <typename INCR, typename FLDS>
void WeightedDiff<INCR, FLDS>::doProcessing(const FLDS & xx) {
  const util::DateTime now(xx.validTime());
  if (((bgnleg_ < end_ && endleg_ > bgn_) || bgnleg_ == endleg_) &&
      (now != endleg_ || now == end_ || now == bgnleg_)) {
    ASSERT(weights_.find(now) != weights_.end());
    const double zz = weights_[now];
    avg_->accumul(zz, xx);
    sum_ += zz;
    LOG(Debug) << "WeightedDiff: time = " << now
               << ", weight = " << zz << ", sum = " << sum_;
  }
}

// -----------------------------------------------------------------------------

}  // namespace oops
