/*
* Copyright 2011 ECMWF
*
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
*
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include <cmath>

namespace oops {

// -----------------------------------------------------------------------------

template <typename INCR>
WeightedDiffTLAD<INCR>::WeightedDiffTLAD(const util::DateTime & vt,
                                         const util::Duration & span,
                                         const Geometry & resol,
                                         const util::Config & config,
                                         const util::Duration & tstep,
                                         WeightingFct & wfct)
  : CostBase<INCR>(vt-span/2, vt+span/2),
    wfct_(wfct), weights_(), avg_(0), forcing_(), sum_(0.0), linit_(false),
    vtime_(vt), bgn_(vt-span/2), end_(vt+span/2), tstep_(tstep),
    bgnleg_(), endleg_(), vars_(config), resol_(resol), lsetup_(false)
{}

// -----------------------------------------------------------------------------

template <typename INCR>
void WeightedDiffTLAD<INCR>::setupTL() {
  avg_ = new Accumulator<INCR, INCR>(resol_, vars_, vtime_);
  sum_ = 0.0;
  lsetup_ = true;
}

// -----------------------------------------------------------------------------

template <typename INCR>
INCR * WeightedDiffTLAD<INCR>::finalTL() {
  ASSERT(linit_);
  ASSERT(std::abs(sum_) < 1.0e-8);
  lsetup_ = false;
  return avg_;
}

// -----------------------------------------------------------------------------

template <typename INCR>
void WeightedDiffTLAD<INCR>::setupAD(boost::shared_ptr<const INCR> forcing) {
  forcing_ = forcing;
}

// -----------------------------------------------------------------------------

template <typename INCR>
void WeightedDiffTLAD<INCR>::finalAD() {
  forcing_.reset();
}

// -----------------------------------------------------------------------------

template <typename INCR>
void WeightedDiffTLAD<INCR>::doInitialize(const util::DateTime & bgn,
                                          const util::DateTime & end,
                                          const util::Duration & tstep) {
  ASSERT(bgn <= end);
  if (!linit_ && bgn <= end_ && end >= bgn_) {
    if (tstep_ == util::Duration(0)) tstep_ = tstep;
    ASSERT(tstep_ > util::Duration(0));
    weights_ = wfct_.setWeights(bgn_, end_, tstep_);
    linit_ = true;
    ASSERT(weights_.find(vtime_) != weights_.end());
    weights_[vtime_] -= 1.0;
  }
  bgnleg_ = bgn;
  endleg_ = end;
}

// -----------------------------------------------------------------------------

template <typename INCR>
void WeightedDiffTLAD<INCR>::doProcessingTL(const INCR & xx) {
  const util::DateTime now(xx.validTime());
  if (((bgnleg_ < end_ && endleg_ > bgn_) || bgnleg_ == endleg_) &&
      (now != endleg_ || now == end_ || now == bgnleg_)) {
    ASSERT(weights_.find(now) != weights_.end());
    const double zz = weights_[now];
    avg_->axpy(zz, xx, false);
    sum_ += zz;
  }
}

// -----------------------------------------------------------------------------

template <typename INCR>
void WeightedDiffTLAD<INCR>::doProcessingAD(INCR & xx) {
  const util::DateTime now(xx.validTime());
  if (((bgnleg_ < end_ && endleg_ > bgn_) || bgnleg_ == endleg_) &&
      (now != endleg_ || now == end_ || now == bgnleg_)) {
    ASSERT(weights_.find(now) != weights_.end());
    const double zz = weights_[now];
    xx.axpy(zz, *forcing_, false);
    sum_ += zz;
  }
}

// -----------------------------------------------------------------------------

}  // namespace oops
