/*
* Copyright 2011 ECMWF
*
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
*
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_BASE_WEIGHTEDDIFFTLAD_H_
#define OOPS_BASE_WEIGHTEDDIFFTLAD_H_

#include <map>

#include <boost/scoped_ptr.hpp>
#include <boost/shared_ptr.hpp>

#include "oops/base/CostBase.h"
#include "oops/base/Geometry.h"
#include "oops/base/Variables.h"
#include "util/Accumulator.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/DolphChebyshev.h"
#include "util/asser1.h"

namespace oops {

// -----------------------------------------------------------------------------

/// Compute time average of states or increments during linear model run.
/*!
 *  Derived classes will compute different types of averages (plain
 *  mean, various types of digital filters) by overwriting the weights
 *  computation method.
 *
 *  A lot of code here is common with WeightedDiff and even WeightedMean,
 *  the design could be improved to reduce code duplication. YT
 */

template <typename INCR>
class WeightedDiffTLAD : public CostBase<INCR> {
 public:
  WeightedDiffTLAD(const util::DateTime &, const util::Duration &,
                   const Geometry &, const util::Config &,
                   const util::Duration &, WeightingFct &);

  virtual ~WeightedDiffTLAD() {}

  void setupTL();
  INCR * finalTL();
  void setupAD(boost::shared_ptr<const INCR>);
  void finalAD();

 private:

  void doFirstTL(const util::DateTime & bgn, const util::DateTime & end,
                 const util::Duration & tstep) {
    this->doInitialize(bgn, end, tstep);
  }

  void doFirstAD(const util::DateTime & bgn, const util::DateTime & end,
                 const util::Duration & tstep) {
    this->doInitialize(bgn, end, tstep);
  }

  void doInitialize(const util::DateTime & bgn, const util::DateTime & end,
                    const util::Duration & tstep);

  void doProcessingTL(const INCR &);
  void doProcessingAD(INCR &);

  void doLastTL() {}
  void doLastAD() {}

  WeightingFct & wfct_;
  std::map< util::DateTime, double > weights_;
  Accumulator<INCR, INCR> * avg_;
  boost::shared_ptr<const INCR> forcing_;
  double sum_;
  bool linit_;
  const util::DateTime vtime_;
  const util::DateTime bgn_;
  const util::DateTime end_;
  util::Duration tstep_;
  util::DateTime bgnleg_;
  util::DateTime endleg_;
  oops::Variables vars_;
  oops::Geometry resol_;
  bool lsetup_;
};

// -----------------------------------------------------------------------------

}  // namespace oops

#include "oops/base/WeightedDiffTLAD.cc"

#endif  // OOPS_BASE_WEIGHTEDDIFFTLAD_H_
