/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include <cmath>

namespace oops {

// -----------------------------------------------------------------------------

template <typename FLDS>
WeightedMean<FLDS>::WeightedMean(const util::DateTime & vt, const util::Duration & span,
                                 const Geometry & resol, const util::Config & config)
    : ProcessorBase<FLDS>(vt-span/2, vt+span/2),
      wfct_(), weights_(), avg_(0), sum_(0.0), linit_(false),
      bgn_(vt-span/2), end_(vt+span/2), endleg_()
{
  const oops::Variables vars(config);
  avg_ = new Accumulator<FLDS, FLDS>(resol, vars, vt);
//  wfct_.reset(WeightFactory::create(config)); YT
  wfct_.reset(new DolphChebyshev(config));
}

// -----------------------------------------------------------------------------

template <typename FLDS>
FLDS * WeightedMean<FLDS>::release() {
  ASSERT(linit_);
  ASSERT(std::abs(sum_ - 1.0) < 1.0e-8);
  return avg_;
}

// -----------------------------------------------------------------------------

template <typename FLDS>
void WeightedMean<FLDS>::doInitialize(const util::DateTime & bgn,
                                      const util::DateTime & end,
                                      const util::Duration & tstep) {
  if (!linit_ && bgn <= end_ && end >= bgn_) {
    weights_ = wfct_->setWeights(bgn_, end_, tstep);
    linit_ = true;
  }
  endleg_ = end;
}

// -----------------------------------------------------------------------------

template <typename FLDS>
void WeightedMean<FLDS>::doProcessing(const FLDS & xx) {
  const util::DateTime now(xx.validTime());
  if (now != endleg_ || now == end_) {
    ASSERT(weights_.find(now) != weights_.end());
    const double zz = weights_[now];
    avg_->accumul(zz, xx);
    sum_ += zz;
    LOG(Debug) << "WeightedMean: time = " << now
               << ", weight = " << zz << ", sum = " << sum_;
  }
}

// -----------------------------------------------------------------------------

}  // namespace oops
