/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include <string>

#include "logger/Logger.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/asser1.h"
#include "util/dot_product.h"

using util::DateTime;

namespace oops {
// -----------------------------------------------------------------------------
template<typename TRAIT>
DepType<TRAIT>::DepType(const TheObs & obs)
  : equiv_(obs.getEqv()), oper_(obs.getHop()), depval_(0)
{
  depval_.reset(new ObsVect(obs.getVal(), false));
  depval_->zero();
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
DepType<TRAIT>::DepType(const TheObs & ob1, const TheObs & ob2)
  : equiv_(ob1.getEqv()), oper_(ob1.getHop()), depval_(0)
{
  ASSERT(TheObs::compatible(ob1, ob2));
  depval_.reset(new ObsVect(ob1.getVal(), true));
  *depval_ -= ob2.getVal();
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
DepType<TRAIT>::DepType(const DepType & other, const bool copyObs)
  : equiv_(other.equiv_), oper_(other.oper_), depval_(0)
{
  depval_.reset(new ObsVect(*other.depval_, copyObs));
  LOG(Trace) << "DepType copy-created";
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
DepType<TRAIT>::~DepType() {}
// -----------------------------------------------------------------------------
template <typename TRAIT>
bool DepType<TRAIT>::compatible(const DepType & dt1, const DepType & dt2) {
  bool same = (dt1.equiv_ == dt2.equiv_) && (dt1.oper_ == dt2.oper_);
  return same;
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
DepType<TRAIT> & DepType<TRAIT>::operator=(const DepType & rhs) {
  ASSERT(compatible(*this, rhs));
  *depval_ = *rhs.depval_;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
DepType<TRAIT> & DepType<TRAIT>::operator+=(const DepType & rhs) {
  *depval_ += *rhs.depval_;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
DepType<TRAIT> & DepType<TRAIT>::operator-=(const DepType & rhs) {
  *depval_ -= *rhs.depval_;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
DepType<TRAIT> & DepType<TRAIT>::operator*=(const double & zz) {
  *depval_ *= zz;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
DepType<TRAIT> & DepType<TRAIT>::operator*=(const DepType & rhs) {
  *depval_ *= *rhs.depval_;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
DepType<TRAIT> & DepType<TRAIT>::operator/=(const DepType & rhs) {
  *depval_ /= *rhs.depval_;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
void DepType<TRAIT>::zero() {
  depval_->zero();
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
void DepType<TRAIT>::invert() {
  depval_->invert();
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
void DepType<TRAIT>::axpy(const double & zz, const DepType & rhs) {
  depval_->axpy(zz, *rhs.depval_);
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
void DepType<TRAIT>::random() {
  depval_->random();
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
double DepType<TRAIT>::dot_product_with(const DepType & other) const {
  return dot_product(*depval_, *other.depval_);
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void DepType<TRAIT>::setupTL(const DateTime & t1, const DateTime & t2) {
  oper_->setupTL(t1, t2);
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void DepType<TRAIT>::obsOperatorTL(const Increment & dx,
                                   const ObsAuxIncr & obias,
                                   const DateTime & t1, const DateTime & t2,
                                   const ObsTraj & otraj) {
  oper_->obsOperatorTL(dx, *depval_, obias, t1, t2, otraj);
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void DepType<TRAIT>::finalizeTL(const ObsAuxIncr & obias,
                                const ObsTraj & otraj) {
  oper_->finalTL(*depval_, obias, otraj);
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void DepType<TRAIT>::finalizeAD(const DateTime & t1, const DateTime & t2,
                                ObsAuxIncr & obias,
                                const ObsTraj & otraj) const {
  oper_->finalAD(t1, t2, *depval_, obias, otraj);
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void DepType<TRAIT>::obsOperatorAD(Increment & dx,
                                   ObsAuxIncr & obias,
                                   const DateTime & t1, const DateTime & t2,
                                   const ObsTraj & otraj) const {
  oper_->obsOperatorAD(dx, *depval_, obias, t1, t2, otraj);
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void DepType<TRAIT>::setupAD() const {
  oper_->setupAD();
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void DepType<TRAIT>::read(const std::string & name) {
  depval_.reset(equiv_->read(name));
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void DepType<TRAIT>::write(const std::string & name) const {
  equiv_->save(name, *depval_);
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
std::ostream & operator<<(std::ostream & outs, const DepType<TRAIT> & dy) {
  double zavg, zmin, zmax;
  dy.depval_->minMaxAvg(zmin, zmax, zavg);
  outs << dy.obsname()
       << " nobs= " << dy.depval_->nobs()
       << " Min=" << zmin << ", Max=" << zmax << ", Average=" << zavg;
  return outs;
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
double DepType<TRAIT>::mean() const {
  double zavg, zmin, zmax;
  depval_->minMaxAvg(zmin, zmax, zavg);
  return zavg;
}
// -----------------------------------------------------------------------------

}  // namespace oops
