/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_OBSERVATIONS_DEPTYPE_H_
#define OOPS_OBSERVATIONS_DEPTYPE_H_

#include <iostream>
#include <string>

#include <boost/scoped_ptr.hpp>
#include <boost/shared_ptr.hpp>

#include "oops/observations/ObsOperStrategy.h"
#include "oops/observations/ObsType.h"

// Forward declarations
namespace util {
  class Config;
  class DateTime;
}

namespace oops {

template<typename TRAIT> class DepType;

template<typename TRAIT>
std::ostream & operator<< (std::ostream &, const DepType<TRAIT> &);

/// Generic handling of DepType.
/*!
 *  DepType defines a generic observation operator.
 */

// -----------------------------------------------------------------------------
template<typename TRAIT> class DepType {
  typedef typename TRAIT::IncrementBase Increment;
  typedef typename TRAIT::ObsVector     ObsVect;
  typedef typename TRAIT::ObsAuxIncr    ObsAuxIncr;
  typedef typename TRAIT::ObsTraj       ObsTraj;
  typedef typename TRAIT::ObsOperator   ObsOperator;  // Can be abstract
  typedef oops::ObsType<TRAIT>          TheObs;

 public:
  static bool compatible(const DepType &, const DepType &);

/// Constructors
  explicit DepType(const TheObs &);
  DepType(const TheObs &, const TheObs &);
  DepType(const DepType &, const bool copyObs = true);
  ~DepType();

/// Linear operators
  DepType & operator=(const DepType &);
  DepType & operator+=(const DepType &);
  DepType & operator-=(const DepType &);
  DepType & operator*=(const double &);
  DepType & operator*=(const DepType &);
  DepType & operator/=(const DepType &);
  void zero();
  void invert();
  void axpy(const double &, const DepType &);
  void random();
  double dot_product_with(const DepType &) const;

/// TL and AD operators
  void setupTL(const util::DateTime &, const util::DateTime &);
  void obsOperatorTL(const Increment &, const ObsAuxIncr &,
                     const util::DateTime &, const util::DateTime &,
                     const ObsTraj &);
  void finalizeTL(const ObsAuxIncr &, const ObsTraj &);

  void finalizeAD(const util::DateTime &, const util::DateTime &,
                  ObsAuxIncr &, const ObsTraj &) const;
  void obsOperatorAD(Increment &, ObsAuxIncr &,
                     const util::DateTime &, const util::DateTime &,
                     const ObsTraj &) const;
  void setupAD() const;

// Utilities
  void read(const std::string &);
  void write(const std::string &) const;
  friend std::ostream & operator<< <> (std::ostream &, const DepType &);

  const std::string obsname() const {return equiv_->obsname();}
  int nobs() const {return depval_->nobs();}
  double mean() const;

/// Update Observation (ie implements the += operator)
  void update(TheObs & y) const {y.add_pert(*depval_);}

 private:
  boost::shared_ptr< ObsOperator > equiv_;
  boost::shared_ptr< ObsOperStrategy<TRAIT> > oper_;
  boost::scoped_ptr< ObsVect > depval_;
};
// -----------------------------------------------------------------------------
}  // namespace oops
#include "oops/observations/DepType.cc"

#endif  // OOPS_OBSERVATIONS_DEPTYPE_H_
