/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_OBSERVATIONS_OBSERRORBASE_H_
#define OOPS_OBSERVATIONS_OBSERRORBASE_H_

#include <map>
#include <string>

#include <boost/noncopyable.hpp>

#include "oops/observations/DepType.h"
#include "oops/observations/ObsType.h"

namespace oops {

// -----------------------------------------------------------------------------
/// Base class for observation error covariance matrices.
/*!
 * Base class for observation error covariance matrices for a given
 * observation type.
 */

template<typename TRAIT>
class ObsErrorBase : private boost::noncopyable {
  typedef oops::ObsType<TRAIT> Observ;
  typedef oops::DepType<TRAIT> Depart;

 public:
  virtual ~ObsErrorBase() {}

/// Linearize and reset for inner loop if needed
  virtual void linearize(const Observ &) =0;

/// Multiply a Departure by \f$R\f$
  virtual void mult(const Depart &, Depart &) const =0;
  virtual void mult_ad(const Depart &, Depart &) const =0;

/// Multiply a Departure by \f$R^{-1}\f$
  virtual void inv_mult(const Depart &, Depart &) const =0;
  virtual void inv_mult_ad(const Depart &, Depart &) const =0;

/// Compute and print Jo table
  virtual double calcJo(const Depart &, Depart &, int &) const =0;

/// Generate random perturbation
  virtual void randomize(Depart &) const =0;
};

// -----------------------------------------------------------------------------

/// ObsErrorFactory Factory
template <typename TRAIT>
class ObsErrorFactory {
  typedef oops::ObsType<TRAIT> Observ;
 public:
  static ObsErrorBase<TRAIT> * create(const Observ &, const util::Config &);
  virtual ~ObsErrorFactory() { getMakers().clear(); }
 protected:
  explicit ObsErrorFactory(const string &);
 private:
  virtual ObsErrorBase<TRAIT> * make(const Observ &, const util::Config &) =0;
  static std::map < std::string, ObsErrorFactory<TRAIT> * > & getMakers() {
    static std::map < std::string, ObsErrorFactory<TRAIT> * > makers_;
    return makers_;
  }
};

template<class TRAIT, class T>
class ObsErrorMaker : public ObsErrorFactory<TRAIT> {
  typedef oops::ObsType<TRAIT> Observ;
  virtual ObsErrorBase<TRAIT> * make(const Observ & obs, const util::Config & conf)
    { return new T(obs, conf); }
 public:
  explicit ObsErrorMaker(const string & name) : ObsErrorFactory<TRAIT>(name) {}
};

// -----------------------------------------------------------------------------

}  // namespace oops

#include "oops/observations/ObsErrorBase.cc"

#endif  // OOPS_OBSERVATIONS_OBSERRORBASE_H_
