/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_OBSERVATIONS_OBSOPERSTRATEGY_H_
#define OOPS_OBSERVATIONS_OBSOPERSTRATEGY_H_

#include <map>
#include <string>
#include <vector>

#include <boost/shared_ptr.hpp>

// Forward declarations
namespace util {
  class DateTime;
  class Config;
}

using std::string;
using std::vector;
using util::DateTime;

namespace oops {

// -----------------------------------------------------------------------------

/// A Generic Obs Operator
/*!
 *  ObsOperStrategy defines a generic class for observation operators.
 *  It computes the observation equivalent from the model state,
 *  or departures from an increment.
 */

template <typename TRAIT> class ObsOperStrategy {
  typedef typename TRAIT::ObsVector     ObsVector;
  typedef typename TRAIT::ObsAuxCtrl    ObsAuxCtrl;
  typedef typename TRAIT::ObsAuxIncr    ObsAuxIncr;
  typedef typename TRAIT::ObsTraj       ObsTraj;
  typedef typename TRAIT::ObsOperator   ObsOper;
  typedef typename TRAIT::StateBase     State;
  typedef typename TRAIT::IncrementBase Increment;

 public:
// Constructor, destructor
  ObsOperStrategy() {}
  virtual ~ObsOperStrategy() {}

// NL
  virtual void setup(const DateTime &, const DateTime &) =0;
  virtual void obsOperator(const State &, ObsVector &, const ObsAuxCtrl &,
                           const DateTime &, const DateTime &) =0;
  virtual void final(ObsVector &, const ObsAuxCtrl &) =0;

// NL Trajectory
  virtual void obsOperator(const State &, ObsVector &, const ObsAuxCtrl &,
                           const DateTime &, const DateTime &, ObsTraj &) =0;
  virtual void final(ObsVector &, const ObsAuxCtrl &, ObsTraj &) =0;

// TL
  virtual void setupTL(const DateTime &, const DateTime &) =0;
  virtual void obsOperatorTL(const Increment &, ObsVector &, const ObsAuxIncr &,
                             const DateTime &, const DateTime &, const ObsTraj &) =0;
  virtual void finalTL(ObsVector &, const ObsAuxIncr &, const ObsTraj &) =0;

// AD
  virtual void finalAD(const DateTime &, const DateTime &,
                       const ObsVector &, ObsAuxIncr &, const ObsTraj &) =0;
  virtual void obsOperatorAD(Increment &, const ObsVector &, ObsAuxIncr &,
                             const DateTime &, const DateTime &, const ObsTraj &) =0;
  virtual void setupAD() =0;
};

// -----------------------------------------------------------------------------

/// ObsOperStrategy Factory
template <typename TRAIT>
class ObsOpFactory {
  typedef typename TRAIT::ObsOperator  ObsOper;
 public:
  static ObsOperStrategy<TRAIT> * create(const std::string &, ObsOper &);
  virtual ~ObsOpFactory() { getMakers().clear(); }
 protected:
  explicit ObsOpFactory(const string &);
 private:
  virtual ObsOperStrategy<TRAIT> * make(ObsOper &) =0;
  static std::map < std::string, ObsOpFactory<TRAIT> * > & getMakers() {
    static std::map < std::string, ObsOpFactory<TRAIT> * > makers_;
    return makers_;
  }
};

template<class TRAIT, class T>
class ObsOpMaker : public ObsOpFactory<TRAIT> {
  typedef typename TRAIT::ObsOperator  ObsOper;
  virtual ObsOperStrategy<TRAIT> * make(ObsOper & h)
    { return new T(h); }
 public:
  explicit ObsOpMaker(const string& name) : ObsOpFactory<TRAIT>(name) {}
};

// -----------------------------------------------------------------------------

}  // namespace oops

#include "oops/observations/ObsOperStrategy.cc"

#endif  // OOPS_OBSERVATIONS_OBSOPERSTRATEGY_H_
