/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include <string>

#include "logger/Logger.h"
#include "util/DateTime.h"
#include "util/asser1.h"

using util::DateTime;

namespace oops {
// -----------------------------------------------------------------------------
template<typename TRAIT>
ObsType<TRAIT>::ObsType(const util::Config & conf): obsval_(0), obsConfig_(conf) {
  equiv_.reset(ObsOperator::create(obsConfig_));
  const std::string id = obsConfig_.getData("ObsOperator");
  LOG(Info) << this->obsname() << " asking for " << id << " computation";
  oper_.reset(ObsOpFactory<TRAIT>::create(id, *equiv_));
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
ObsType<TRAIT>::ObsType(const ObsType & other, const bool copyObs)
  : equiv_(other.equiv_), oper_(other.oper_), obsConfig_(other.obsConfig_)
{
  obsval_.reset(new ObsVect(*other.obsval_, copyObs));
  LOG(Trace) << "ObsType copy-created";
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
ObsType<TRAIT>::~ObsType() {}
// -----------------------------------------------------------------------------
template <typename TRAIT>
bool ObsType<TRAIT>::compatible(const ObsType & ot1, const ObsType & ot2) {
  bool same = (ot1.equiv_ == ot2.equiv_) && (ot1.oper_ == ot2.oper_);
  return same;
}
// -----------------------------------------------------------------------------
template <typename TRAIT>
ObsType<TRAIT> & ObsType<TRAIT>::operator=(const ObsType & rhs) {
  ASSERT(compatible(*this, rhs));
  *obsval_ = *rhs.obsval_;
  return *this;
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void ObsType<TRAIT>::setup(const DateTime & t1, const DateTime & t2) {
  oper_->setup(t1, t2);
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void ObsType<TRAIT>::obsOperator(const State & xx, const ObsAuxCtrl & obias,
                                 const DateTime & t1, const DateTime & t2) {
  oper_->obsOperator(xx, *obsval_, obias, t1, t2);
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void ObsType<TRAIT>::finalize(const ObsAuxCtrl & obias) {
  oper_->final(*obsval_, obias);
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void ObsType<TRAIT>::obsOperator(const State & xx, const ObsAuxCtrl & obias,
                                 const DateTime & t1, const DateTime & t2,
                                 ObsTraj & otraj) {
  oper_->obsOperator(xx, *obsval_, obias, t1, t2, otraj);
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void ObsType<TRAIT>::finalize(const ObsAuxCtrl & obias,
                              ObsTraj & otraj) {
  oper_->final(*obsval_, obias, otraj);
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void ObsType<TRAIT>::read(const std::string & name) {
  obsval_.reset(equiv_->read(name));
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void ObsType<TRAIT>::write(const std::string & name) const {
  equiv_->save(name, *obsval_);
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
std::ostream & operator<<(std::ostream & outs, const ObsType<TRAIT> & yy) {
  double zavg, zmin, zmax;
  yy.obsval_->minMaxAvg(zmin, zmax, zavg);
  outs << yy.obsname()
       << " nobs= " << yy.obsval_->nobs()
       << " Min=" << zmin << ", Max=" << zmax << ", Average=" << zavg;
  return outs;
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void ObsType<TRAIT>::generate(const util::DateTime & bgn, const util::DateTime & end) {
  const util::Config genConf(obsConfig_, "Generate");
  obsval_.reset(equiv_->distribute(genConf, bgn, end));
}
// -----------------------------------------------------------------------------
template<typename TRAIT>
void ObsType<TRAIT>::add_pert(const ObsVect & dy) {
  *obsval_ += dy;
}
// -----------------------------------------------------------------------------

}  // namespace oops
