/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_RUNS_EXTERNALDFI_H_
#define OOPS_RUNS_EXTERNALDFI_H_

#include <string>

#include <boost/scoped_ptr.hpp>
#include <boost/shared_ptr.hpp>

#include "logger/Logger.h"
#include "oops/base/Geometry.h"
#include "oops/base/PostProcessor.h"
#include "oops/base/ModelState.h"
#include "oops/base/StateInfo.h"
#include "oops/base/StateWriter.h"
#include "oops/base/WeightedMean.h"
#include "oops/runs/Task.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/abor1_cpp.h"

namespace oops {

template <typename TRAIT> class ExternalDFI: public Task {
  typedef typename TRAIT::ModelAuxCtrl  ModErr;
  typedef typename TRAIT::ModelData     Model;
  typedef typename TRAIT::StateBase     StateBase;

 public:
// -----------------------------------------------------------------------------
  ExternalDFI(int argc, char ** argv): Task(argc, argv) {}
// -----------------------------------------------------------------------------
  ~ExternalDFI() {}
// -----------------------------------------------------------------------------
  void run(const util::Config & fullConfig) {
//  Setup time
    util::Config fcConfig(fullConfig, "forecast");
    const util::DateTime bgndate(fcConfig.getData("begin"));
    const util::DateTime enddate(fcConfig.getData("end"));
    LOG(Info) << "Running forecast from " << bgndate << " to " << enddate;

//  Setup resolution
    const util::Config resolConfig(fullConfig, "resolution");
    const Geometry resol(resolConfig);

//  Setup Model
    const util::Config modelConfig(fullConfig, "model");
    const Model model(resol, modelConfig);

//  Setup initial state
    const util::Config initialConfig(fullConfig, "initial");
    LOG(Configs) << "Initial configuration is:" << "\n" << initialConfig;
    ModelState<TRAIT> xx(model, initialConfig);
    xx.prtNorms("Read");

//  Setup augmented state
    ModErr moderr(initialConfig);

//  Setup post-processing
    PostProcessor<StateBase> post;

    const util::Config prtConfig(fullConfig, "prints", true);
    post.enroll(new StateInfo<StateBase>("fc", prtConfig));

//  Setup DFI
    PostProcessor<StateBase> pp(post);

    const util::Config dfiConf(fullConfig, "dfi");
    const util::Duration dfispan(dfiConf.getData("filter_span"));
    const util::DateTime dfitime(bgndate+dfispan/2);
    boost::shared_ptr< WeightedMean<StateBase> >
      pdfi(new WeightedMean<StateBase>(dfitime, dfispan, resol, dfiConf));
    pp.enroll(pdfi);

//  Run DFI forecast
    const util::DateTime enddfi(bgndate+dfispan);
    xx.forecast(moderr, enddfi, pp);

//  Retrieve initialized state
    boost::scoped_ptr<StateBase> xdfi(pdfi->release());
    xdfi->prtNorms("Initialized");

//  Setup forecast outputs
    const util::Config outConfig(fullConfig, "output");
    post.enroll(new StateWriter<StateBase>(bgndate, outConfig));

//  Run forecast from initialized state
    ModelState<TRAIT> zz(*xdfi, model);
    zz.forecast(moderr, enddate, post);
  }
// -----------------------------------------------------------------------------
};

}  // namespace oops
#endif  // OOPS_RUNS_EXTERNALDFI_H_
