/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_RUNS_GENENSPERTB_H_
#define OOPS_RUNS_GENENSPERTB_H_

#include <sstream>
#include <string>

#include <boost/scoped_ptr.hpp>

#include "logger/Logger.h"
#include "oops/base/Geometry.h"
#include "oops/base/PostProcessor.h"
#include "oops/base/ModelState.h"
#include "oops/base/StaticCovariance.h"
#include "oops/base/StateWriter.h"
#include "oops/runs/Task.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/abor1_cpp.h"

namespace oops {

template <typename TRAIT> class GenEnsPertB: public Task {
  typedef typename TRAIT::ModelAuxCtrl  ModErr;
  typedef typename TRAIT::ModelData     Model;
  typedef typename TRAIT::StateBase     StateBase;
  typedef typename TRAIT::IncrementBase Increment;
  typedef StaticCovariance<TRAIT>       StateError;

 public:
// -----------------------------------------------------------------------------
  GenEnsPertB(int argc, char ** argv): Task(argc, argv) {}
// -----------------------------------------------------------------------------
  ~GenEnsPertB() {}
// -----------------------------------------------------------------------------
  void run(const util::Config & fullConfig) {
//  Setup time loop
    const util::Config fcConfig(fullConfig, "forecast");
    const util::DateTime bgndate(fcConfig.getData("begin"));
    const util::DateTime enddate(fcConfig.getData("end"));
    LOG(Info) << "Running forecast from " << bgndate << " to " << enddate;

//  Setup resolution
    const util::Config resolConfig(fullConfig, "resolution");
    const Geometry resol(resolConfig);

//  Setup Model
    const util::Config modelConfig(fullConfig, "model");
    const Model model(resol, modelConfig);

//  Setup initial state
    const util::Config initialConfig(fullConfig, "initial");
    LOG(Configs) << "Initial configuration is:" << "\n" << initialConfig;
    ModelState<TRAIT> xx(model, initialConfig);
    xx.prtNorms("Read");

//  Setup augmented state
    ModErr moderr(initialConfig);

//  Setup B matrix
    const util::Config covar(fullConfig, "covariance");
    StateError Bmat(resol, covar);
    const util::Config tlConf(fullConfig, "tlmodel");
    Bmat.linearize(xx, resol, tlConf);

//  Generate perturbed states
    const int members = fullConfig.getInt("members");
    for (int jm = 0; jm < members; ++jm) {
      boost::scoped_ptr<Increment> dx(Bmat.getIncrement());
      Bmat.randomize(*dx);
      ModelState<TRAIT> xp(xx);
      xp += *dx;

//    Setup forecast outputs
      PostProcessor<StateBase> post;

      util::Config conf(fullConfig);
      conf.addElement("config/output", "member");
      conf.setElement("config/output/member", jm+1);
      const util::Config outConfig(conf, "config/output");

      post.enroll(new StateWriter<StateBase>(bgndate, outConfig));

//    Run forecast
      xp.forecast(moderr, enddate, post);
    }
  }
// -----------------------------------------------------------------------------
};

}  // namespace oops
#endif  // OOPS_RUNS_GENENSPERTB_H_
