/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_RUNS_HOFX_H_
#define OOPS_RUNS_HOFX_H_

#include <string>

#include "logger/Logger.h"
#include "oops/base/Geometry.h"
#include "oops/base/Observation.h"
#include "oops/base/Observer.h"
#include "oops/base/PostProcessor.h"
#include "oops/base/ModelState.h"
#include "oops/base/StateInfo.h"
#include "oops/runs/Task.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/Duration.h"

namespace oops {

template <typename TRAIT> class HofX: public Task {
  typedef typename TRAIT::ModelAuxCtrl  ModErr;
  typedef typename TRAIT::ModelData     Model;
  typedef typename TRAIT::ObsAuxCtrl    ObsAux;
  typedef typename TRAIT::ObsTraj       ObsTraject;
  typedef typename TRAIT::StateBase     StateBase;
  typedef Observation<TRAIT>            Observ;
  typedef ObsTrajectory<ObsTraject>     ObsTraj;

 public:
// -----------------------------------------------------------------------------
  HofX(int argc, char ** argv): Task(argc, argv) {}
// -----------------------------------------------------------------------------
  ~HofX() {}
// -----------------------------------------------------------------------------
  void run(const util::Config & fullConfig) {
//  Setup observation window
    const util::Config windowConf(fullConfig, "assimilation_window");
    const util::DateTime bgn(windowConf.getData("begin"));
    const util::DateTime end(windowConf.getData("end"));
    LOG(Configs) << "Observation window is:\n" << windowConf;

//  Setup resolution
    const util::Config resolConfig(fullConfig, "resolution");
    const Geometry resol(resolConfig);

//  Setup Model
    const util::Config modelConfig(fullConfig, "model");
    const Model model(resol, modelConfig);

//  Setup initial state
    const util::Config initialConfig(fullConfig, "initial");
    LOG(Configs) << "Initial configuration is:\n" << initialConfig;
    ModelState<TRAIT> xx(model, initialConfig);
    xx.prtNorms("Read");

//  Setup augmented state
    ModErr moderr(initialConfig);

//  Setup observations
    const util::Config obsConfig(fullConfig, "Observations");
    Observ yy(obsConfig, bgn, end);
    yy.read(obsConfig);
    LOG(Info) << "Input observations:" << yy;
    ObsAux varbc(obsConfig);

//  Setup forecast outputs
    PostProcessor<StateBase> post;

    const util::Config prtConfig(fullConfig, "prints");
    post.enroll(new StateInfo<StateBase>("fc", prtConfig));

    Observer<StateBase, Observ, ObsAux, ObsTraj> * pobs = new
      Observer<StateBase, Observ, ObsAux, ObsTraj>(windowConf, yy, varbc);
    post.enroll(pobs);

//  Compute H(x)
    xx.forecast(moderr, end, post);

    LOG(Info) << "MakeObs: Finished computing observation equivalents.";

//  Save H(x)
    LOG(Info) << "H(x):" << yy;
    yy.write("hofx");
  };
// -----------------------------------------------------------------------------
};

}  // namespace oops
#endif  // OOPS_RUNS_HOFX_H_
