/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_RUNS_MAKEOBS_H_
#define OOPS_RUNS_MAKEOBS_H_

#include <string>

#include "logger/Logger.h"
#include "oops/base/Departure.h"
#include "oops/base/Geometry.h"
#include "oops/base/ObsErrorCovariance.h"
#include "oops/base/Observation.h"
#include "oops/base/Observer.h"
#include "oops/base/PostProcessor.h"
#include "oops/base/ModelState.h"
#include "oops/base/StateInfo.h"
#include "oops/observations/instantiateObsErrorFactory.h"
#include "oops/observations/instantiateObsOpFactory.h"
#include "oops/runs/Task.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/Duration.h"

namespace oops {

template <typename TRAIT> class MakeObs: public Task {
  typedef typename TRAIT::ModelAuxCtrl  ModErr;
  typedef typename TRAIT::ModelData     Model;
  typedef typename TRAIT::ObsAuxCtrl    ObsAux;
  typedef typename TRAIT::ObsTraj       ObsTraject;
  typedef typename TRAIT::StateBase     StateBase;
  typedef Departure<TRAIT>              Depart;
  typedef Observation<TRAIT>            Observ;
  typedef ObsErrorCovariance<TRAIT>     ObsError;
  typedef ObsTrajectory<ObsTraject>     ObsTraj;

 public:
// -----------------------------------------------------------------------------
  MakeObs(int argc, char ** argv): Task(argc, argv) {
    instantiateObsErrorFactory<TRAIT>();
    instantiateObsOpFactory<TRAIT>();
  }
// -----------------------------------------------------------------------------
  ~MakeObs() {}
// -----------------------------------------------------------------------------
  void run(const util::Config & fullConfig) {
//  Setup observation window
    const util::Config windowConf(fullConfig, "assimilation_window");
    const util::DateTime bgn(windowConf.getData("begin"));
    const util::DateTime end(windowConf.getData("end"));
    LOG(Configs) << "Observation window is:\n" << windowConf;

//  Setup resolution
    const util::Config resolConfig(fullConfig, "resolution");
    const Geometry resol(resolConfig);

//  Setup Model
    const util::Config modelConfig(fullConfig, "model");
    const Model model(resol, modelConfig);

//  Setup initial "true" state
    const util::Config initialConfig(fullConfig, "initial");
    LOG(Configs) << "Initial configuration is:\n" << initialConfig;
    ModelState<TRAIT> xx(model, initialConfig);
    xx.prtNorms("Read");

//  Setup augmented state
    ModErr moderr(initialConfig);

//  Setup observations
    const util::Config obsConfig(fullConfig, "Observations");
    LOG(Configs) << "Observation configuration is:\n" << obsConfig;
    Observ yy(obsConfig, bgn, end);
    yy.generate();
    ObsAux ybias(obsConfig);

//  Setup forecast outputs
    PostProcessor<StateBase> post;

    const util::Config prtConfig(fullConfig, "prints");
    post.enroll(new StateInfo<StateBase>("fc", prtConfig));

    Observer<StateBase, Observ, ObsAux, ObsTraj> * pobs = new
      Observer<StateBase, Observ, ObsAux, ObsTraj>(windowConf, yy, ybias);
    post.enroll(pobs);

//  Run forecast and generate observations
    xx.forecast(moderr, end, post);
    LOG(Info) << "MakeObs: Finished observation generation.";

//  Perturb observations
    if (obsConfig.elementExists("obspert")) {
      ObsError matR(yy, obsConfig);
      Depart ypert(yy);
      matR.randomize(ypert);
      double opert = obsConfig.getReal("obspert");
      ypert *= opert;
      ypert.update(yy);
    }

//  Save observations
    LOG(Info) << "Generated observation:" << yy;
    yy.write("ObsVal");
  }
// -----------------------------------------------------------------------------
};

}  // namespace oops
#endif  // OOPS_RUNS_MAKEOBS_H_
