/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef OOPS_RUNS_VARIATIONAL_H_
#define OOPS_RUNS_VARIATIONAL_H_

#include <string>

#include <boost/scoped_ptr.hpp>

#include "logger/Logger.h"
#include "oops/assimilation/ControlVariable.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/IncrementalAssimilation.h"
#include "oops/assimilation/Minimizer.h"
#include "oops/assimilation/instantiateCostFactory.h"
#include "oops/assimilation/instantiateMinFactory.h"
#include "oops/base/Geometry.h"
#include "oops/base/Observation.h"
#include "oops/base/PostProcessor.h"
#include "oops/base/StateInfo.h"
#include "oops/base/StateWriter.h"
#include "oops/base/instantiateCovarFactory.h"
#include "oops/observations/instantiateObsErrorFactory.h"
#include "oops/observations/instantiateObsOpFactory.h"
#include "oops/runs/Task.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/Duration.h"

namespace oops {

template <typename TRAIT> class Variational: public Task {
  typedef typename TRAIT::ModelData  Model;
  typedef typename TRAIT::StateBase  State;
  typedef ControlVariable<TRAIT>     Control;
  typedef Observation<TRAIT>         TheObs;
  typedef CostFunction<TRAIT>        CostFct;
  typedef CostFactory<TRAIT>         CFact;
  typedef Minimizer<TRAIT>           Minim;
  typedef MinFactory<TRAIT>          MinFact;

 public:
// -----------------------------------------------------------------------------
  Variational(int argc, char ** argv): Task(argc, argv) {
    instantiateCostFactory<TRAIT>();
    instantiateCovarFactory<TRAIT>();
    instantiateMinFactory<TRAIT>();
    instantiateObsErrorFactory<TRAIT>();
    instantiateObsOpFactory<TRAIT>();
  }
// -----------------------------------------------------------------------------
  ~Variational() {}
// -----------------------------------------------------------------------------
  void run(const util::Config & fullConfig) {
//  Get assimilation window configuration
    const util::Config cfConf(fullConfig, "cost_function");
    const util::Duration winlen(cfConf.getData("window_length"));
    const util::DateTime winbgn(cfConf.getData("window_begin"));
    const util::DateTime winend(winbgn + winlen);

//  Setup resolution
    const util::Config resolConfig(fullConfig, "resolution");
    const Geometry resol(resolConfig);

//  Setup Model
    const util::Config modelConfig(fullConfig, "model");
    const Model model(resol, modelConfig);

//  Setup observations
    const util::Config obsConfig(fullConfig, "Observations");
    TheObs yy(obsConfig, winbgn, winend);
    yy.read(obsConfig);
    LOG(Info) << "Input observations:" << yy;

//  Setup generalized background
    const util::Config controlConfig(fullConfig, "control");
    const util::Config bgConfig(fullConfig, "background");
    Control xx(controlConfig, resol, bgConfig);
    xx.prtNorms("bkg");

//  Setup cost function
    boost::scoped_ptr<CostFct> J(CFact::create(cfConf));
    J->setup(resol, model, xx, controlConfig, yy, obsConfig);

//  Setup minimizer
    const util::Config minimConf(fullConfig, "minimizer");
    boost::scoped_ptr<Minim> minim(MinFact::create(minimConf));

//  Perform Incremental Variational Assimilation
    IncrementalAssimilation<TRAIT>(xx, *J, *minim, fullConfig);

//  Save analysis and final diagnostics
    PostProcessor<State> post;
    const util::Config outConfig(fullConfig, "output");
    post.enroll(new StateWriter<State>(winbgn, outConfig));

    const util::Config finalConfig(fullConfig, "final");
    const util::Config prtConfig(finalConfig, "prints");
    post.enroll(new StateInfo<State>("final", prtConfig));

    J->eval(xx, finalConfig, post);
  };
// -----------------------------------------------------------------------------
};

}  // namespace oops
#endif  // OOPS_RUNS_VARIATIONAL_H_
