/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef TEST_BASE_INCREMENT_H_
#define TEST_BASE_INCREMENT_H_

#include <iostream>
#include <string>
#include <cmath>

#include <boost/test/unit_test.hpp>
#include <boost/test/floating_point_comparison.hpp>

#include "oops/base/ModelIncrement.h"
#include "oops/base/ModelState.h"
#include "oops/base/StaticCovariance.h"
#include "test/TestEnvironment.h"
#include "util/dot_product.h"
#include "util/Config.h"

namespace test {

// A singleton - so we don't re-make States for every test

template <typename TRAIT> class TestIncrementMaker {
  typedef typename TRAIT::IncrementBase Increment;

 public:
  static TestIncrementMaker<TRAIT>& getInstance() {
    static TestIncrementMaker<TRAIT> theTestIncrementMaker;
    return theTestIncrementMaker;
  }

  const Increment * increment() const {return increment_;}
  const oops::StaticCovariance<TRAIT> * covariance() const {return B_;}
 private:
  TestIncrementMaker<TRAIT>() {
    const util::Config * config_ = TestEnvironment<TRAIT>::getInstance().config();

// Setup a state
    const util::Config resolConfig(*config_, "resolution");
    const oops::Geometry resol(resolConfig);
    const util::Config modelConfig(*config_, "model");
    const typename TRAIT::ModelData model(resol, modelConfig);
    const util::Config bgConfig(*config_, "state");
    oops::ModelState<TRAIT> * xx = new oops::ModelState<TRAIT>(model, bgConfig);

//  Setup B matrix
    const util::Config covar(*config_, "covariance");
    B_ = new oops::StaticCovariance<TRAIT>(resol, covar);
    const util::Config tlConf(*config_, "tlmodel");
    B_->linearize(*xx, resol, tlConf);
    increment_ = B_->getIncrement();
  }

  ~TestIncrementMaker<TRAIT>() {
    delete increment_;
    delete B_;
  }

  TestIncrementMaker<TRAIT>(TestIncrementMaker<TRAIT> const &);
  void operator=(TestIncrementMaker<TRAIT> const &);

  Increment * increment_;
  oops::StaticCovariance<TRAIT> * B_;
};

// -----------------------------------------------------------------------------

template <typename TRAIT> void testIncrementConstructFromB() {
  typedef typename TRAIT::IncrementBase Increment;

  const oops::StaticCovariance<TRAIT> *
    B = TestIncrementMaker<TRAIT>::getInstance().covariance();
  const Increment * xx = B->getIncrement();
  BOOST_CHECK(dot_product(*xx, *xx) == 0.0);
}

// -----------------------------------------------------------------------------

template <typename TRAIT> void testIncrementCopyConstructor() {
  typedef typename TRAIT::IncrementBase Increment;

  const Increment * aa = TestIncrementMaker<TRAIT>::getInstance().increment();
  Increment bb(*aa);
  bb.random();

// Copy the increment - copyFields=true
  Increment cc(bb);
  BOOST_CHECK(dot_product(cc, cc) > 0.0);

// Check that the copy is equal to the original
  cc -= bb;
  BOOST_CHECK(dot_product(cc, cc) == 0.0);
}

// -----------------------------------------------------------------------------

template <typename TRAIT> void testIncrementTriangle() {
  typedef typename TRAIT::IncrementBase Increment;

  const Increment * aa = TestIncrementMaker<TRAIT>::getInstance().increment();
  Increment bb(*aa);
  bb.random();
  Increment cc(*aa);
  cc.random();

// test triangle inequality
  double dotbb = dot_product(bb, bb);
  BOOST_CHECK(dotbb > 0.0);

  double dotcc = dot_product(cc, cc);
  BOOST_CHECK(dotcc > 0.0);

  Increment dd(bb);
  dd += cc;
  double dotdd =  dot_product(dd, dd);
  BOOST_CHECK(std::sqrt(dotdd) < std::sqrt(dotbb) + std::sqrt(dotcc));
}

// -----------------------------------------------------------------------------

template <typename TRAIT> void testIncrementOpMinusEq() {
  typedef typename TRAIT::IncrementBase Increment;

  const Increment * aa = TestIncrementMaker<TRAIT>::getInstance().increment();
  Increment bb(*aa);
  bb.random();

// test = and -=
  Increment cc(bb);
  cc = bb;
  cc -= bb;
  BOOST_CHECK(dot_product(cc, cc) == 0.0);
}

// -----------------------------------------------------------------------------

template <typename TRAIT> void testIncrementOpPlusEq() {
  typedef typename TRAIT::IncrementBase Increment;

  const Increment * aa = TestIncrementMaker<TRAIT>::getInstance().increment();
  Increment bb(*aa);
  bb.random();

// test *= and +=
  Increment cc(bb);
  cc *= 2.0;
  Increment dd(bb);
  dd += bb;

  cc -= dd;
  BOOST_CHECK_SMALL(dot_product(cc, cc), 1e-6);
}

// -----------------------------------------------------------------------------

template <typename TRAIT> void testIncrementDotProduct() {
  typedef typename TRAIT::IncrementBase Increment;

  const Increment * aa = TestIncrementMaker<TRAIT>::getInstance().increment();
  Increment bb(*aa);
  bb.random();

// test symmetry of dot product
  Increment cc(bb);
  cc.random();
  BOOST_CHECK_CLOSE(dot_product(bb, cc), dot_product(cc, bb), 1e-4);
}

// -----------------------------------------------------------------------------

template <typename TRAIT> void testIncrementZero() {
  typedef typename TRAIT::IncrementBase Increment;

  const Increment * aa = TestIncrementMaker<TRAIT>::getInstance().increment();
  Increment bb(*aa);
  bb.random();

// test zero
  bb.zero();
  BOOST_CHECK(dot_product(bb, bb) == 0.0);
}

// -----------------------------------------------------------------------------

template <typename TRAIT> void testIncrementAxpy() {
  typedef typename TRAIT::IncrementBase Increment;

  const Increment * aa = TestIncrementMaker<TRAIT>::getInstance().increment();
  Increment bb(*aa);
  bb.random();
  Increment cc(*aa);
  cc.random();

// test axpy
  Increment dd(cc);
  dd.axpy(2.0, bb);

  Increment ee(cc);
  ee += bb;
  ee += bb;

  ee -= dd;
  BOOST_CHECK_SMALL(dot_product(ee, ee), 1e-6);
}

// -----------------------------------------------------------------------------

template <typename TRAIT> void initTestSuiteIncrement() {
  boost::unit_test::test_suite * ts = BOOST_TEST_SUITE("test_increment");

  ts->add(BOOST_TEST_CASE(&testIncrementConstructFromB<TRAIT>));
  ts->add(BOOST_TEST_CASE(&testIncrementCopyConstructor<TRAIT>));
  ts->add(BOOST_TEST_CASE(&testIncrementTriangle<TRAIT>));
  ts->add(BOOST_TEST_CASE(&testIncrementOpMinusEq<TRAIT>));
  ts->add(BOOST_TEST_CASE(&testIncrementOpPlusEq<TRAIT>));
  ts->add(BOOST_TEST_CASE(&testIncrementDotProduct<TRAIT>));
  ts->add(BOOST_TEST_CASE(&testIncrementAxpy<TRAIT>));

  boost::unit_test::framework::master_test_suite().add(ts);
}

// -----------------------------------------------------------------------------

}  // namespace test

#endif  // TEST_BASE_INCREMENT_H_
