/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include <boost/test/unit_test.hpp>

#include "oops/base/Increment.h"
#include "oops/base/Observation.h"
#include "oops/base/ModelState.h"
#include "oops/observations/DepType.h"
#include "test/TestEnvironment.h"
#include "util/dot_product.h"
#include "util/DateTime.h"
#include "util/Config.h"

namespace test {

class DepTypeTestFixture {
 public:
  DepTypeTestFixture() {
    test::TestEnvironment & testEnv = test::TestEnvironment::getInstance();
    const util::Config globalConf(testEnv.config());
    tlConf_ = new util::Config(testEnv.config(), "testInnerLoop");
    yobs_ = &testEnv.getTestObservation();

    xx_ = &testEnv.getTestState();
    traj_ = &testEnv.getTestObsTraj();
    octl_ = &testEnv.getTestObsBias();

    t1_ = testEnv.getTestBgn();
    t2_ = testEnv.getTestEnd();
    tolAD_ = globalConf.getReal("toleranceTestAD");
  }

  ~DepTypeTestFixture() {
  }

// Input data for tests
  const oops::Observation * yobs_;
  const oops::ModelState * xx_;
  const oops::ObservAuxCtrl * octl_;
  const oops::ObsTraj * traj_;
  double tolAD_;
  util::DateTime t1_, t2_;
  const util::Config * tlConf_;
};

// -----------------------------------------------------------------------------
BOOST_FIXTURE_TEST_SUITE(test_deptype, DepTypeTestFixture)
// -----------------------------------------------------------------------------
BOOST_AUTO_TEST_CASE(test_deptype_construct) {
  for (int jj = 0; jj < yobs_->size(); ++jj) {
    oops::DepType * idep = new oops::DepType((*yobs_)[jj]);
    BOOST_CHECK(idep);
    delete idep;
  }
}
// -----------------------------------------------------------------------------
BOOST_AUTO_TEST_CASE(test_deptype_diffconstruct) {
  for (int jj = 0; jj < yobs_->size(); ++jj) {
    oops::DepType * idep = new oops::DepType((*yobs_)[jj], (*yobs_)[jj]);
    const double zz = dot_product(*idep, *idep);
    BOOST_CHECK(zz == 0.0);
    delete idep;
  }
}
// -----------------------------------------------------------------------------
BOOST_AUTO_TEST_CASE(test_deptype_zero) {
  for (int jj = 0; jj < yobs_->size(); ++jj) {
    oops::DepType * idep = new oops::DepType((*yobs_)[jj]);
    idep->zero();
    const double zz = dot_product(*idep, *idep);
    BOOST_CHECK(zz == 0.0);
    delete idep;
  }
}
// -----------------------------------------------------------------------------
BOOST_AUTO_TEST_CASE(test_deptype_random) {
  for (int jj = 0; jj < yobs_->size(); ++jj) {
    oops::DepType * idep = new oops::DepType((*yobs_)[jj]);
    idep->random();
    const double zz = dot_product(*idep, *idep);
    BOOST_CHECK(zz > 0.0);
    delete idep;
  }
}
// -----------------------------------------------------------------------------
BOOST_AUTO_TEST_CASE(test_deptype_copydiff) {
  for (int jj = 0; jj < yobs_->size(); ++jj) {
    oops::DepType * idep = new oops::DepType((*yobs_)[jj]);
    idep->random();
    const double z1 = dot_product(*idep, *idep);
    oops::DepType * ideq = new oops::DepType(*idep);
    BOOST_CHECK(ideq);
    const double z2 = dot_product(*ideq, *ideq);
    BOOST_CHECK(z1 == z2);
    const double z3 = dot_product(*idep, *ideq);
    BOOST_CHECK(z1 == z3);
    *idep -= *ideq;
    const double zz = dot_product(*idep, *idep);
    BOOST_CHECK(zz == 0.0);
    delete ideq;
    BOOST_CHECK(idep);
    delete idep;
  }
}
// -----------------------------------------------------------------------------
BOOST_AUTO_TEST_CASE(test_deptype_axpy) {
  for (int jj = 0; jj < yobs_->size(); ++jj) {
    oops::DepType * idep = new oops::DepType((*yobs_)[jj]);
    oops::DepType * ideq = new oops::DepType((*yobs_)[jj]);
    idep->zero();

    ideq->random();
    const double za = 0.0;
    idep->axpy(za, *ideq);
    double zz = dot_product(*idep, *idep);
    BOOST_CHECK(zz == 0.0);

    ideq->zero();
    const double zb = 999.999;
    idep->axpy(zb, *ideq);
    zz = dot_product(*idep, *idep);
    BOOST_CHECK(zz == 0.0);

    delete idep;
  }
}
// -----------------------------------------------------------------------------
// BOOST_AUTO_TEST_CASE(test_deptype_tlad) {
//   boost::scoped_ptr<oops::Increment> dxtl(new oops::Increment(*xx_, *geom_));
//   boost::scoped_ptr<oops::Increment> dxad(new oops::Increment(*xx_, *geom_));
//
//   oops::TrajICst jtra = traj_->begin();
//   oops::ObservAuxIncr obtl(*octl_, *octl_, *tlConf_);
//   oops::ObservAuxIncr obad(*octl_, *octl_, *tlConf_);
//   oops::DctlICst jotl = obtl.begin();
//   oops::DctlIter joad = obad.begin();
//
//   for (int jj = 0; jj < yobs_->size(); ++jj) {
//     boost::scoped_ptr<oops::DepType> dotl(new oops::DepType((*yobs_)[jj]));
//     boost::scoped_ptr<oops::DepType> doad(new oops::DepType((*yobs_)[jj]));
//
// //  Run TL
//     dxtl->random();
//     dotl->zero();
//     dotl->setupObsOperatorTL(t1_, t2_);
//     dotl->obsOperatorTL(*dxtl, *jotl, t1_, t2_, *jtra);
//     dotl->finalObsOperatorTL(*jotl, *jtra);
//
// //  Run AD
//     dxad->zero();
//     doad->random();
//     doad->finalObsOperatorAD(t1_, t2_, *joad, *jtra);
//     doad->obsOperatorAD(*dxad, *joad, t1_, t2_, *jtra);
//     doad->setupObsOperatorAD();
//
// //  Test output
//     const double zzi = dot_product(*dxtl, *dxad);
//     const double zzd = dot_product(*dotl, *doad);
//     BOOST_CHECK_CLOSE_FRACTION(zzi, zzd, tolAD_);
//
//     ++jotl;
//     ++joad;
//     ++jtra;
//   }
// }
// -----------------------------------------------------------------------------
BOOST_AUTO_TEST_SUITE_END()
// -----------------------------------------------------------------------------

}  // namespace test
