/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef UTIL_CONFIG_H_
#define UTIL_CONFIG_H_

#include <ostream>
#include <string>

#include <boost/shared_ptr.hpp>

// Forward References
namespace util {
  class ConfigImplBase;
}

namespace util {

  /// Helper class. Allows empty Config constructor to have a unique signature.
  class ConfigEmpty {
  };

  /// Configuration
  /*!
   *  A Config represents tree-structured configuration data
   *  that has an Xpath-like syntax for identifying sub-trees and
   *  single elements. This is a bridge class that allows the implementation
   *  to be changed (e.g. from XML to JSON) without major implications
   *  for thesresto of the code.
   */

  class Config {
    friend std::ostream& operator<<(std::ostream& output, const Config& d);

   public:
  // Constructors
    /// Construct from a pointer to something derived from ConfigImplBase
    Config(const std::string & filename, ConfigImplBase * const);

    /// Copy constructor
    explicit Config(const util::Config& d);

    /// Construct from a sub-tree of another Config
    Config(const util::Config& d, const std::string & s, const bool allowEmpty = false);

    /// Construct from a sub-tree of another Config
    Config(const util::Config& d, const std::string & s, const unsigned idx);

    /// Make an empty Config
    explicit Config(ConfigImplBase * const);

    /// Make an empty Config
    explicit Config(const util::Config& d, const util::ConfigEmpty);

  // Destructor
    virtual ~Config() {}

  // Methods
    /// Set the current node for Xpath relative addressing
    void setCurrentNode(const std::string & str);

    /// Return true if the Config is empty
    bool empty() const;

    /// Return true if there is an element satisfying the query string.
    bool elementExists(const std::string &) const;

    /// Get the data from the first element that satisfies the query string.
    std::string getData(const std::string &) const;

    /// Get the data from the first element that satisfies the query string.
    int getInt(const std::string &) const;

    /// Get the data from the first element that satisfies the query string.
    double getReal(const std::string &) const;

    /// Add a new element
    void addElement(const std::string &, const std::string &);

    /// Set/reset the contents of an element
    void setElement(const std::string &, const std::string &);

    /// Set/reset the contents of an element
    void setElement(const std::string &, const int);

    /// Set/reset the contents of an element
    void setElement(const std::string &, const double);

    unsigned getElementSize(const std::string&) const;

   protected:
    /// Print method invoked by operator(<<)
    virtual void print(std::ostream&) const;

   private:
    Config();  // No default constructor (we need to know the implementation)

    Config & operator=(const Config &);  // assignment not allowed

    boost::shared_ptr<ConfigImplBase> pimpl_;
  };

}  // namespace util
#endif  // UTIL_CONFIG_H_
