/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include "util/DolphChebyshev.h"

#include <cmath>
#include <vector>

#include <boost/math/constants/constants.hpp>
#include <boost/math/special_functions/acosh.hpp>

#include "logger/Logger.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/WeightingFct.h"
#include "util/asser1.h"

namespace oops {

// -----------------------------------------------------------------------------

static WeightMaker<DolphChebyshev> makerDolph_("DolphChebyshev");

// -----------------------------------------------------------------------------

DolphChebyshev::DolphChebyshev(const util::Config & config) {
  tau_ = util::Duration(config.getData("cutoff"));
}

// -----------------------------------------------------------------------------

std::map< util::DateTime, double > DolphChebyshev::setWeights(const util::DateTime & bgn,
                                                              const util::DateTime & end,
                                                              const util::Duration & dt)
{
  const double pi = boost::math::constants::pi<double>();
  const util::Duration window(end-bgn);
  const int nstep = window.toSeconds() / dt.toSeconds();
  ASSERT(window.toSeconds() == dt.toSeconds()*nstep);

  const int M = nstep/2;
  const int N = 2*M + 1;
  LOG(Debug) << "DolphChebyshev M = " << M << ", N = " << N;

  const int tt = tau_.toSeconds() / dt.toSeconds();
  ASSERT(tau_.toSeconds() == dt.toSeconds()*tt);
  ASSERT(tt > 1);
  const double thetas = 2.0 * pi / tt;
  const double x0 = 1.0 / std::cos(thetas/2.0);
  const double rr = 1.0 / std::cosh(nstep * boost::math::acosh(x0));
  LOG(Debug) << "DolphChebyshev thetas = " << thetas <<
                ", x0 = " << x0 << ", rr = " << rr;

  std::vector<double> w(M+1);
  for (int n = 0; n <= M; ++n) {
    double tn = 2.0 * pi * n / N;
    double sum = 0.0;
    for (int m = 1; m <= M; ++m) {
      double xx = x0 * std::cos(pi * m / N);
      double t2m = 0.0;
      double tnm2 = 1.0;
      double tnm1 = xx;
      for (int kk = 2; kk <= 2 * M; ++kk) {
        t2m = 2.0 * xx * tnm1 - tnm2;
        tnm2 = tnm1;
        tnm1 = t2m;
      }
//        if (std::abs(xx)<=1.0) {
//          t2m = std::cos(2*M*std::acos(xx));
//        } else {
//          t2m = std::cosh(2*M*boost::math::acosh(xx));
//        }
      sum += t2m * std::cos(m*tn);
    }
    w[n] = (1.0 + 2.0 * rr * sum) / N;
  }

  std::map< util::DateTime, double > weights;
  double checksum = 0.0;
  util::DateTime now(bgn);
  for (int jj = -M; jj <= M; ++jj) {
    int n = std::abs(jj);
    LOG(Debug) << "DolphChebyshev now = " << now << ", n = " << n << ", w = " << w[n];
    weights[now] = w[n];
    checksum += w[n];
    now += dt;
  }
  ASSERT(now == end+dt);
  LOG(Debug) << "DolphChebyshev checksum = " << checksum;
  ASSERT(std::abs(checksum-1.0) < 1.0e-8);

  return weights;
}

// -----------------------------------------------------------------------------

}  // namespace oops

