/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include "util/PostTimer.h"

#include <algorithm>
#include <string>
#include <vector>

#include <boost/tokenizer.hpp>

#include "logger/Logger.h"
#include "util/Config.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/abor1_cpp.h"

namespace oops {
// -----------------------------------------------------------------------------
PostTimer::PostTimer()
  : conf_(0), freq_(0), bgn_(), end_(), start_(0), finish_(0), pptimes_() {}
// -----------------------------------------------------------------------------
PostTimer::PostTimer(const util::Duration & freq)
  : conf_(0), freq_(freq), bgn_(), end_(), start_(0), finish_(0), pptimes_() {}
// -----------------------------------------------------------------------------
PostTimer::PostTimer(const util::Config & conf)
  : conf_(&conf), freq_(0), bgn_(), end_(), start_(0), finish_(0), pptimes_() {}
// -----------------------------------------------------------------------------
PostTimer::PostTimer(const util::DateTime & start, const util::Config & conf)
  : conf_(&conf), freq_(0), bgn_(), end_(), start_(new util::DateTime(start)),
    finish_(0), pptimes_()
{}
// -----------------------------------------------------------------------------
PostTimer::PostTimer(const util::DateTime & start, const util::DateTime & finish,
                     const util::Duration & freq)
  : conf_(0), freq_(freq), bgn_(), end_(), start_(new util::DateTime(start)),
    finish_(new util::DateTime(finish)), pptimes_() {}
// -----------------------------------------------------------------------------
void PostTimer::initialize(const util::DateTime & bgn, const util::DateTime & end,
                           const util::Duration &) {
  util::DateTime start(bgn);
  if (start_ != 0) {
    start = *start_;
  }
  bgn_ = start;
  util::DateTime finish(end);
  if (finish_ != 0) {
    finish = *finish_;
  }
  end_ = finish;

// User specified configuration
  if (conf_ != 0) {
    LOG(Configs) << "PProc configuration is:" << "\n" << *conf_;
//  User specified frequency and start
    if (conf_->elementExists("frequency"))
      freq_ = util::Duration(conf_->getData("frequency"));
    if (conf_->elementExists("first")) {
      const util::Duration first(conf_->getData("first"));
      bgn_ += first;
    }
//  User specified steps
    if (conf_->elementExists("steps")) {
      const std::string steps = conf_->getData("steps");
      boost::tokenizer<> tok(steps);
      for (boost::tokenizer<>::iterator it = tok.begin(); it != tok.end(); ++it) {
        const util::Duration step(*it);
        const util::DateTime tt = start+step;
        if (bgn <= tt && tt <= end) pptimes_.push_back(tt);
      }
    }
  }
}
// -----------------------------------------------------------------------------
bool PostTimer::itIsTime(const util::DateTime & now) {
  bool doit = false;
  if (now >= bgn_ && now <= end_) {
    doit = (freq_.toSeconds() == 0 && pptimes_.empty());
    if (!doit && freq_.toSeconds()>0) {
      const util::Duration dt = now - bgn_;
      doit = (dt >= util::Duration(0) && dt % freq_ == 0);
    }
    if (!doit && !pptimes_.empty()) {
      std::vector<util::DateTime>::iterator it;
      it = find(pptimes_.begin(), pptimes_.end(), now);
      doit = (it != pptimes_.end());
    }
  }
  LOG(Trace) << "In PostTimer:itIsTime, time = " << now << ", doit = " << doit;
  return doit;
}
// -----------------------------------------------------------------------------
}  // namespace oops
