/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include "util/XmlDom.h"

#include <libxml/parser.h>
#include <libxml/xmlreader.h>

#include <algorithm>
#include <sstream>
#include <string>
#include <typeinfo>

#include "logger/Logger.h"
#include "util/asser1.h"
#include "util/abor1_cpp.h"

namespace util {

// -----------------------------------------------------------------------------

int XmlDom::instances_ = 0;                // Static counter of XmlDom instances
std::map<xmlDocPtr, int> XmlDom::docrefs_;  // Static reference counter of docs

// -----------------------------------------------------------------------------

std::ostream& operator<<(std::ostream& output, const XmlDom& d) {
  d.print(output);
  return output;
}

// -----------------------------------------------------------------------------

void XmlDom::print(std::ostream& output) const {   // NOLINT(runtime/references)
  if (doc_ != NULL) {
    xmlBufferPtr buf = xmlBufferCreate();
    xmlNodeDump(buf, doc_, context_->node, 0, 1);

    output << buf->content << std::endl;
    xmlBufferFree(buf);
  }
}

// -----------------------------------------------------------------------------

XmlDom::XmlDom() : context_(NULL), currentNode_("/"), doc_(NULL) {
  if (instances_ == 0) xmlInitParser();
  ++instances_;
  docrefs_[doc_]=docrefs_[doc_]+1;
}

// -----------------------------------------------------------------------------

XmlDom::~XmlDom() {
  if (context_ != NULL) xmlXPathFreeContext(context_);
  decrementDocrefs();
  --instances_;
  if (instances_ == 0) xmlCleanupParser();
}

// -----------------------------------------------------------------------------

XmlDom * XmlDom::makeEmpty() const {
  return new XmlDom();
}

// -----------------------------------------------------------------------------

XmlDom * XmlDom::copy() const {
  XmlDom * cpy = new XmlDom();
  cpy->doc_ = doc_;
  if (doc_ != NULL) {
    docrefs_[doc_]=docrefs_[doc_]+1;
    cpy->context_ = xmlXPathNewContext(doc_);
    cpy->context_->node = context_->node;
    cpy->currentNode_ = currentNode_;
  }
  return cpy;
}

// -----------------------------------------------------------------------------

XmlDom * XmlDom::copySubTree(const std::string & path, const bool allowEmpty,
                             const unsigned idx) const {
  XmlDom * cpy = new XmlDom();

  const xmlChar * x;
  const unsigned char * y;
  ASSERT(typeid(x) == typeid(y));  //  check that reinterpret_cast is safe!

  xmlXPathObjectPtr xpathObj =
      xmlXPathEvalExpression(reinterpret_cast<const xmlChar*>(path.c_str()),
                             context_);

  if (xpathObj == NULL) {
    LOG(Error) << "Error: unable to evaluate xpath expression " << path;
    ABORT("unable to evaluate xpath expression");
  }
//  if (xpathObj->nodesetval->nodeNr > idx) {
//    LOG(Error) << "Xpath \"" << path << "\" does not identify a unique node";
//    ABORT("XPath does not identify a unique node");
//  }

  if (xmlXPathNodeSetIsEmpty(xpathObj->nodesetval)) {
    if (allowEmpty) {
      LOG(Trace) << "XmlDom::copySubTree returning empty XmlDom";
    } else {
      LOG(Error) << "XPath \"" << path << "\" does not match any nodes";
      ABORT("XPath does not match any nodes");
    }
  } else {
    ++instances_;
    cpy->doc_ = doc_;
    docrefs_[doc_]=docrefs_[doc_]+1;
    cpy->context_ = xmlXPathNewContext(doc_);
    cpy->context_->node = xpathObj->nodesetval->nodeTab[idx];
    cpy->currentNode_ = path;
  }
  xmlXPathFreeObject(xpathObj);
  return cpy;
}

// -----------------------------------------------------------------------------

void XmlDom::readFile(const std::string & filename) {
  decrementDocrefs();  // discard former contents

  doc_ = xmlReadFile(filename.c_str(), NULL, XML_PARSE_NOBLANKS);
  if (doc_ == NULL) {
    LOG(Error) << "Failed to parse XML file: " << filename;
    ABORT("Failed to parse XML file");
  }

  docrefs_[doc_]=docrefs_[doc_]+1;
  context_ = xmlXPathNewContext(doc_);

  std::string slash("/");
  this->setCurrentNode(slash);  // there must be an easier way to do this...
}

// -----------------------------------------------------------------------------

void XmlDom::setCurrentNode(const std::string & str) {
  const xmlChar * x;
  const unsigned char * y;
  ASSERT(typeid(x) == typeid(y));  //  check that reinterpret_cast is safe!
  xmlXPathObjectPtr xpathObj =
      xmlXPathEvalExpression(reinterpret_cast<const xmlChar*>(str.c_str()),
                             context_);
  if (xpathObj == NULL) {
    LOG(Error) << "Error: unable to evaluate xpath expression " << str;
    ABORT("unable to evaluate xpath expression");
  }

  if (xmlXPathNodeSetIsEmpty(xpathObj->nodesetval)) {
    LOG(Error) << "XPath \"" << str << "\" does not match any nodes";
    ABORT("XPath does not match any nodes");
  }

  if (xpathObj->nodesetval->nodeNr != 1) {
    LOG(Error) << "Xpath \"" << str
                       << "\" does not identify a unique node";
    ABORT("XPath does not identify a unique node");
  }

  context_->node = xpathObj->nodesetval->nodeTab[0];
  currentNode_ = str;

  xmlXPathFreeObject(xpathObj);
}

// -----------------------------------------------------------------------------

bool XmlDom::elementExists(const std::string & str) const {
  bool b = !this->empty();

  if (b) {
    const xmlChar * x;
    const unsigned char * y;
    ASSERT(typeid(x) == typeid(y));  //  check that reinterpret_cast is safe!
    xmlXPathObjectPtr xpathObj =
        xmlXPathEvalExpression(reinterpret_cast<const xmlChar*>(str.c_str()),
                               context_);
    if (xpathObj == NULL) {
      LOG(Error) << "Error: unable to evaluate xpath \"" << str
                         << "\"";
      ABORT("unable to evaluate xpath expression");
    }

    if (xpathObj->nodesetval->nodeNr > 1) {
      LOG(Error) << "Xpath \"" << str
                         << "\" does not identify a unique node";
      ABORT("XPath does not identify a unique node");
    }

    b = !xmlXPathNodeSetIsEmpty(xpathObj->nodesetval);
    xmlXPathFreeObject(xpathObj);
  }

  return b;
}

// -----------------------------------------------------------------------------

std::string XmlDom::getData(const std::string & str) const {
  std::string result("");
  if (!elementExists(str)) {
    LOG(Error) << "Xpath \"" << str
                       << "\" does not match any nodes";
    ABORT("XPath does not match any nodes");
  } else {
    const xmlChar * x;
    const unsigned char * y;
    ASSERT(typeid(x) == typeid(y));  //  check that reinterpret_cast is safe!

    xmlXPathObjectPtr xpathObj =
        xmlXPathEvalExpression(reinterpret_cast<const xmlChar*>(str.c_str()),
                               context_);

    if (xpathObj->nodesetval->nodeNr < 1) {
      LOG(Error) << "Element \"" << str
                         << "\" does not contain any data";
      ABORT("Element does not contain any data");
    }

    if (xpathObj->nodesetval->nodeTab[0]->children == NULL) {
      LOG(Error) << "Element \"" << str
                         << "\" does not contain any data";
      ABORT("Element does not contain any data");
    }

    xmlChar * resultstr =
        xmlNodeListGetString(doc_,
                       xpathObj->nodesetval->nodeTab[0]->children, 1);

    if (resultstr == NULL) {
      LOG(Error) << "Element \"" << str
                         << "\" does not contain any data";
      ABORT("Element does not contain any data");
    }

    result = std::string(reinterpret_cast<char*>(resultstr));
    xmlFree(resultstr);

  // strip leading and trailing blanks
    size_t start = result.find_first_not_of(" ");

    if (start == std::string::npos) {
      result="";
    } else {
      size_t end;
      end = result.find_last_not_of(" ");
      result = result.substr(start, end-start+1);
    }

    xmlXPathFreeObject(xpathObj);
  }

  return result;
}

// -----------------------------------------------------------------------------

int XmlDom::getInt(const std::string & str) const {
  int ii = 0;
  std::string sd = getData(str);
  if (!sd.empty()) {
    std::stringstream ss(sd);
    ss >> ii;
  }
  return ii;
}

// -----------------------------------------------------------------------------

double XmlDom::getReal(const std::string & str) const {
  double zz = 0.0;
  std::string sd = getData(str);
  if (!sd.empty()) {
    std::stringstream ss(sd);
    ss >> zz;
  }
  return zz;
}

// -----------------------------------------------------------------------------

void XmlDom::decrementDocrefs() {
  docrefs_[doc_] = docrefs_[doc_]-1;
  if (doc_ != NULL && docrefs_[doc_] == 0) {
    docrefs_.erase(doc_);
    xmlFreeDoc(doc_);
  }
}

// -----------------------------------------------------------------------------

void XmlDom::makeDocWithEmptyRootElement(const std::string & element) {
  // Create a proper libxml2 document with root element "element"
  xmlDocPtr newdoc = xmlNewDoc(reinterpret_cast<const xmlChar*>("1.0"));
  decrementDocrefs();
  doc_ = newdoc;
  docrefs_[doc_]=docrefs_[doc_]+1;

  xmlNodePtr root_node = xmlNewNode(NULL,
                            reinterpret_cast<const xmlChar*>(element.c_str()));
  xmlDocSetRootElement(doc_, root_node);

  xmlXPathFreeContext(context_);
  context_ = xmlXPathNewContext(doc_);
  std::string slash("/");
  setCurrentNode(slash);
}

// -----------------------------------------------------------------------------

void XmlDom::addElement(const std::string & path,
                        const std::string & element) {
  if (doc_ == NULL) {
    if (path != std::string("/")) {
       LOG(Error) << "You are adding the root element, so path must be \"/\"";
       ABORT("You are adding the root element, so path must be \"/\"");
    }

    makeDocWithEmptyRootElement(element);

  } else {
    // Add "element" to the existing tree
    xmlDocPtr newdoc = xmlCopyDoc(doc_, 1);
    xmlXPathFreeContext(context_);
    context_ = xmlXPathNewContext(newdoc);
    std::string slash("/");
    setCurrentNode(slash);
    decrementDocrefs();
    doc_ = newdoc;
    docrefs_[doc_]=docrefs_[doc_]+1;
    std::string cn(currentNode_);
    setCurrentNode(cn);

    if (!elementExists(path)) {
      LOG(Error) << "Xpath \"" << path
                         << "\" does not match any nodes";
      ABORT("XPath does not match any nodes");

    } else {
      const xmlChar * x;
      const unsigned char * y;
      ASSERT(typeid(x) == typeid(y));  //  check that reinterpret_cast is safe!

      xmlXPathObjectPtr xpathObj =
          xmlXPathEvalExpression(reinterpret_cast<const xmlChar*>(path.c_str()),
                               context_);

      if (xpathObj->nodesetval->nodeNr < 1) {
        LOG(Error) << "Element \"" << path
                           << "\" has no nodes";
        ABORT("Element has no nodes");
      }

      xmlNodePtr newNode = xmlNewNode(NULL,
                             reinterpret_cast<const xmlChar*>(element.c_str()));

      if (!xmlAddChild(xpathObj->nodesetval->nodeTab[0], newNode)) {
         LOG(Error) << "Unable to add element \"" << element
                           << "\" to \"" << path << "\"";
        ABORT("Unable to add element to XmlDom");
      }

      xmlXPathFreeObject(xpathObj);
    }
  }
}

// -----------------------------------------------------------------------------

void XmlDom::setElement(const std::string & path,
                        const std::string & content) {
  if (!elementExists(path)) {
    LOG(Error) << "Xpath \"" << path
                       << "\" does not match any nodes";
    ABORT("XPath does not match any nodes");

  } else {
    const xmlChar * x;
    const unsigned char * y;
    ASSERT(typeid(x) == typeid(y));  //  check that reinterpret_cast is safe!

    xmlXPathObjectPtr xpathObj =
        xmlXPathEvalExpression(reinterpret_cast<const xmlChar*>(path.c_str()),
                             context_);

    if (xpathObj->nodesetval->nodeNr < 1) {
      LOG(Error) << "Element \"" << path
                         << "\" has no nodes";
      ABORT("Element has no nodes");
    }

    xmlNodeSetContent(xpathObj->nodesetval->nodeTab[0],
                      reinterpret_cast<const xmlChar*>(content.c_str()));

    xmlXPathFreeObject(xpathObj);
  }
}

// -----------------------------------------------------------------------------

void XmlDom::setElement(const std::string & path, const int i) {
  std::stringstream ss;
  ss << i;
  setElement(path, ss.str());
}

// -----------------------------------------------------------------------------

void XmlDom::setElement(const std::string & path, const double d) {
  std::stringstream ss;
  ss << d;
  setElement(path, ss.str());
}

// -----------------------------------------------------------------------------

unsigned XmlDom::getElementSize(const std::string& path) const
{
  xmlXPathObjectPtr xpathObj =
      xmlXPathEvalExpression(reinterpret_cast<const xmlChar*>(path.c_str()),
                             context_);
  return xmlXPathNodeSetGetLength(xpathObj->nodesetval);
//  return xpathObj->nodesetval->nodeMax;
}


}  // namespace util
