/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#ifndef UTIL_XMLDOM_H_
#define UTIL_XMLDOM_H_

#include <map>
#include <string>

#include <libxml/tree.h>
#include <libxml/xpath.h>

#include "util/ConfigImplBase.h"

namespace util {

/// Simple wrapper for the Gnome libxml2 XML parser.
/*!
 * An XmlDom is a simplified representation of an XML document, together
 * with an XPath context representing the current node (for relative
 * location paths).
 *
 * Methods are provided to set the current node, and to access the
 * contents of elements as strings, ints and doubles. The access methods
 * return the content of a single element. Xpaths must be unique.
 *
 * New nodes can be added and the contents of nodes can be changed using
 * the addElement and setElement methods. Note that it is not currently
 * possible to add XML attributes to a node.
 *
 * An XmlDom is a small object.
 *
 * Each XmlDom contains a pointer to a libxml2 xmlDoc object. Many XmlDom
 * objects can point to the same xmlDoc. We maintain a reference count
 * for each xmlDoc in docs_ so we can free it when no longer used.
 *
 * We also maintain a counter of the total number of XmlDom objects.
 * The first XmlDom created does an xmlInitParser(). The last XmlDom destructed
 * does a xmlCleanupParser().
 *
 */

class XmlDom : public util::ConfigImplBase {
 public:
  XmlDom();
  XmlDom(const XmlDom &, const std::string &);

  virtual ~XmlDom();

/// Make an empty XmlDom
  virtual XmlDom * makeEmpty() const;

/// Copy method
  virtual XmlDom * copy() const;

/// Copy from sub-tree
  virtual XmlDom * copySubTree(const std::string & s, const bool,
                               const unsigned idx = 0) const;

/// Read configuration from a file (discarding former contents)
  virtual void readFile(const std::string & filename);

/// Set the current node for Xpath relative addressing
  virtual void setCurrentNode(const std::string & str);

/// Return true if the document is empty - i.e. was created by XmlDom();
  virtual bool empty() const {return doc_ == NULL;}

/// Return true if there is an element satisfying the query string.
  virtual bool elementExists(const std::string &) const;

/// Get the data contained in the first element that satisfies the query string.
  virtual std::string getData(const std::string &) const;

/// Get the data contained in the first element that satisfies the query string.
  virtual int getInt(const std::string &) const;

/// Get the data contained in the first element that satisfies the query string.
  virtual double getReal(const std::string &) const;

/// Add a new element
  virtual void addElement(const std::string &, const std::string &);

/// Set/reset the contents of an element
  virtual void setElement(const std::string &, const std::string &);
  virtual void setElement(const std::string &, const int);
  virtual void setElement(const std::string &, const double);

/// Get the number of occurences of a node with same name (list)
  virtual unsigned getElementSize(const std::string&) const;

/// Print
  virtual void print(std::ostream&) const;

 private:
  explicit XmlDom(const XmlDom& d);  // no copy
  XmlDom & operator=(const XmlDom &);  // no assignment

  void decrementDocrefs();

  void makeDocWithEmptyRootElement(const std::string &);

  /// number of instances of XmlDom.
  static int instances_;

  /// reference counters for documents.
  static std::map<xmlDocPtr, int> docrefs_;

  /// context for Xpath addressing
  xmlXPathContextPtr context_;

  /// string version of current node
  std::string currentNode_;

  /// Pointer to the document.
  xmlDocPtr doc_;
};

}  // namespace oops
#endif  // UTIL_XMLDOM_H_
