!
! Copyright 2011 ECMWF
! 
! This software was developed at ECMWF for evaluation
! and may be used for academic and research purposes only.
! The software is provided as is without any warranty.
! 
! This software can be used, copied and modified but not
! redistributed or sold. This notice must be reproduced
! on each copy made.
!

!>  Fortran module handling the interface to utils::DateTime

module datetime_mod

use iso_c_binding
use string_f_c_mod
use duration_mod

implicit none
private
public datetime, datetime_create, datetime_delete, assignment(=), &
     & f_c_datetime, c_f_datetime, datetime_to_string, &
     & operator(<), operator(<=), operator(>=), operator(>), &
     & datetime_update, datetime_diff_string

!>  Derived type encapsulating a C++ DateTime pointer.

type datetime
  private
  type(C_PTR) ptr
end type datetime

interface assignment(=)
  module procedure datetime_assign
end interface assignment(=)

interface operator(<)
  module procedure datetime_lt
end interface operator(<)

interface operator(<=)
  module procedure datetime_le
end interface operator(<=)

interface operator(>=)
  module procedure datetime_ge
end interface operator(>=)

interface operator(>)
  module procedure datetime_gt
end interface operator(>)

#include "datetime.intfb.h"

!-------------------------------------------------------------------------------
contains
!-------------------------------------------------------------------------------

!>  Create a DateTime from an ISO8601 string.

subroutine datetime_create(fstring, self)
implicit none
type(datetime), intent(out)  :: self
character(len=*), intent(in) :: fstring
character(kind=C_CHAR,len=1) :: cstring(21)

call f_c_string(fstring, cstring)
self%ptr = datetime_construct_f(cstring)

end subroutine datetime_create

!-------------------------------------------------------------------------------

!> Delete a DateTime

subroutine datetime_delete(self)
implicit none
type(datetime), intent(inout) :: self

call datetime_destruct_f(self%ptr)

end subroutine datetime_delete

!-------------------------------------------------------------------------------

!> Delete a DateTime

subroutine datetime_assign(self, other)
implicit none
type(datetime), intent(out) :: self
type(datetime), intent(in)  :: other

! Possible memory leak...
self%ptr = datetime_clone_f(other%ptr)

end subroutine datetime_assign

!-------------------------------------------------------------------------------

!> Convert a Fortran datetime to a C++ DateTime

subroutine f_c_datetime(fdt, cdt)
implicit none
type(datetime), intent(in) :: fdt
type(C_PTR), intent(out)   :: cdt
cdt = fdt%ptr
end subroutine f_c_datetime

!-------------------------------------------------------------------------------

!> Convert a C++ DateTime to a Fortran datetime

subroutine c_f_datetime(cdt, fdt)
implicit none
type(C_PTR), intent(in)     :: cdt
type(datetime), intent(out) :: fdt
fdt%ptr = cdt
end subroutine c_f_datetime

!-------------------------------------------------------------------------------

!> Get DateTime as string

subroutine datetime_to_string(fdt, fstring)
implicit none
type(datetime), intent(in)      :: fdt
character(len=*), intent(inout) :: fstring
character(kind=C_CHAR,len=1) :: cstring(21)

if (len(fstring)<20) call abor1_ftn('datetime_to_string: fstring too short')
call datetime_string_f(fdt%ptr,cstring)
call c_f_string(cstring,fstring)

end subroutine datetime_to_string

!-------------------------------------------------------------------------------

subroutine datetime_diff_string(dt1, dt2, fstring)
implicit none
type(datetime), intent(in)      :: dt1, dt2
character(len=*), intent(inout) :: fstring
character(kind=C_CHAR,len=1) :: cstring(21)

if (len(fstring)<20) call abor1_ftn('datetime_to_string: fstring too short')
call datetime_diff_str_f(dt1%ptr, dt2%ptr, cstring)
call c_f_string(cstring,fstring)

end subroutine datetime_diff_string

!-------------------------------------------------------------------------------

logical function datetime_lt(dt1, dt2)
implicit none
type(datetime), intent(in) :: dt1, dt2
real(c_double) :: zdiff

call datetime_diff_f(dt1%ptr, dt2%ptr, zdiff)
datetime_lt = (zdiff.lt.0.0_c_double)

end function datetime_lt

!-------------------------------------------------------------------------------

logical function datetime_le(dt1, dt2)
implicit none
type(datetime), intent(in) :: dt1, dt2
real(c_double) :: zdiff

call datetime_diff_f(dt1%ptr, dt2%ptr, zdiff)
datetime_le = (zdiff.le.0.0_c_double)

end function datetime_le

!-------------------------------------------------------------------------------

logical function datetime_ge(dt1, dt2)
implicit none
type(datetime), intent(in) :: dt1, dt2
real(c_double) :: zdiff

call datetime_diff_f(dt1%ptr, dt2%ptr, zdiff)
datetime_ge = (zdiff.ge.0.0_c_double)

end function datetime_ge

!-------------------------------------------------------------------------------

logical function datetime_gt(dt1, dt2)
implicit none
type(datetime), intent(in) :: dt1, dt2
real(c_double) :: zdiff

call datetime_diff_f(dt1%ptr, dt2%ptr, zdiff)
datetime_gt = (zdiff.gt.0.0_c_double)

end function datetime_gt

!-------------------------------------------------------------------------------

!> Advance a DateTime

subroutine datetime_update(self, dt)
implicit none
type(datetime), intent(inout) :: self
type(duration), intent(in)    :: dt

call datetime_update_f(self%ptr, dt%ptr)

end subroutine datetime_update

!-------------------------------------------------------------------------------

end module datetime_mod
