/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include <iostream>
#include <string>

#include "test/TestEnvironment.h"
#include "util/Config.h"
#include "util/XmlDom.h"

#include <boost/test/unit_test.hpp>
#include <boost/test/test_case_template.hpp>
#include <boost/mpl/list.hpp>
#include <boost/scoped_ptr.hpp>

namespace {

// List of implementations of Config to be tested. Add JSON, YAML, etc. here!
  typedef boost::mpl::list<util::XmlDom> test_types; // NOLINT

BOOST_AUTO_TEST_SUITE(test_Config)

// -----------------------------------------------------------------------------
  BOOST_AUTO_TEST_CASE_TEMPLATE(test_construct_empty, T, test_types) {
    util::Config x(new T());
    BOOST_CHECK(x.empty());
    util::Config y(x, util::ConfigEmpty());
    BOOST_CHECK(y.empty());
  }
// -----------------------------------------------------------------------------
  BOOST_AUTO_TEST_CASE_TEMPLATE(test_addElement, T, test_types) {
    util::Config x(new T());
    x.addElement("/", "grandparent");
    x.addElement("grandparent", "son");
    x.addElement("grandparent", "daughter");
    x.addElement("grandparent/daughter", "grandson");
    x.addElement("grandparent/son", "granddaughter");
    BOOST_CHECK(x.elementExists("grandparent"));
    BOOST_CHECK(x.elementExists("grandparent/son"));
    BOOST_CHECK(x.elementExists("grandparent/daughter"));
    BOOST_CHECK(x.elementExists("grandparent/daughter/grandson"));
    BOOST_CHECK(x.elementExists("grandparent/son/granddaughter"));
  }
// -----------------------------------------------------------------------------
  BOOST_AUTO_TEST_CASE_TEMPLATE(test_setElement, T, test_types) {
    util::Config x(new T());
    x.addElement("/", "root");

    x.addElement("root", "string_contents");
    x.setElement("root/string_contents", "artichoke");

    x.addElement("root", "int_contents");
    x.setElement("root/int_contents", 43);

    x.addElement("root", "double_contents");
    x.setElement("root/double_contents", 3.14159);

    BOOST_CHECK(x.getData("root/string_contents") == std::string("artichoke"));
    BOOST_CHECK(x.getInt("root/int_contents") == 43);
    BOOST_CHECK(x.getReal("root/double_contents") == 3.14159);
  }
// -----------------------------------------------------------------------------
  BOOST_AUTO_TEST_CASE_TEMPLATE(test_copy_constructor, T, test_types) {
    util::Config x(new T());
    x.addElement("/", "a");
    x.setElement("a", 99);
    util::Config y(x);
    BOOST_CHECK(y.getInt("a") == 99);
  }
// -----------------------------------------------------------------------------
  BOOST_AUTO_TEST_CASE_TEMPLATE(test_copy_subtree, T, test_types) {
    util::Config x(new T());
    x.addElement("/", "a");
    x.addElement("a", "b");
    x.addElement("a/b", "c");
    x.setElement("a/b/c", 99);
    util::Config y(x, "a/b");
    BOOST_CHECK(y.getInt("c") == 99);
  }
// -----------------------------------------------------------------------------
  BOOST_AUTO_TEST_CASE_TEMPLATE(test_set_current_node, T, test_types) {
    util::Config x(new T());
    x.addElement("/", "a");
    x.addElement("a", "b");
    x.addElement("a/b", "c");
    x.setElement("a/b/c", 76);
    x.setCurrentNode("a/b");
    BOOST_CHECK(x.getInt("c") == 76);
  }
// -----------------------------------------------------------------------------
  BOOST_AUTO_TEST_CASE_TEMPLATE(test_output_to_stream, T, test_types) {
    util::Config x(new T());
    x.addElement("/", "teststring");
    x.setElement("teststring", "Hello World!");
    std::stringstream strm;
    strm << x;
    // check that the output stream contains the string "Hello World!"
    BOOST_CHECK(strm.str().find("Hello World!") != std::string::npos);
  }
// -----------------------------------------------------------------------------
  BOOST_AUTO_TEST_CASE_TEMPLATE(test_copy_and_add_element, T, test_types) {
    util::Config x(new T());
    x.addElement("/", "a");
    x.addElement("a", "b");
    x.addElement("a/b", "c");
    x.setElement("a/b/c", 76);
    util::Config y(x);
    y.addElement("a/b", "new_element");
    BOOST_CHECK(!x.elementExists("a/b/new_element"));
    BOOST_CHECK(y.elementExists("a/b/new_element"));
  }
BOOST_AUTO_TEST_SUITE_END()

}  // anonymous namespace
