/*
* Copyright 2011 ECMWF
* 
* This software was developed at ECMWF for evaluation
* and may be used for academic and research purposes only.
* The software is provided as is without any warranty.
* 
* This software can be used, copied and modified but not
* redistributed or sold. This notice must be reproduced
* on each copy made.
*/

#include <sstream>
#include "util/DateTime.h"
#include "util/Duration.h"

#include <string>

#include <boost/test/unit_test.hpp>

namespace {

BOOST_AUTO_TEST_SUITE(test_DateTime)

// -----------------------------------------------------------------------------

  BOOST_AUTO_TEST_CASE(test_construct_from_ints) {
    int year1   = 2011;
    int month1  = 9;
    int day1    = 16;
    int hour1   = 13;
    int minute1 = 55;
    int second1 = 20;
    util::DateTime d1(year1, month1, day1, hour1, minute1, second1);

    int year2    = 0;
    int month2   = 0;
    int day2     = 0;
    int hour2    = 0;
    int minute2  = 0;
    int second2  = 0;
    d1.get(year2, month2, day2, hour2, minute2, second2);

    BOOST_CHECK(year2   == year1);
    BOOST_CHECK(month2  == month1);
    BOOST_CHECK(day2    == day1);
    BOOST_CHECK(hour2   == hour1);
    BOOST_CHECK(minute2 == minute1);
    BOOST_CHECK(second2 == second1);
  }

// -----------------------------------------------------------------------------

  BOOST_AUTO_TEST_CASE(test_set_from_ints) {
    int year1   = 2011;
    int month1  = 9;
    int day1    = 16;
    int hour1   = 13;
    int minute1 = 55;
    int second1 = 20;
    util::DateTime d1(2012, 12, 5, 15, 42, 55);
    d1.set(year1, month1, day1, hour1, minute1, second1);

    int year2   = 0;
    int month2  = 0;
    int day2    = 0;
    int hour2   = 0;
    int minute2 = 0;
    int second2 = 0;
    d1.get(year2, month2, day2, hour2, minute2, second2);

    BOOST_CHECK(year2   == year1);
    BOOST_CHECK(month2  == month1);
    BOOST_CHECK(day2    == day1);
    BOOST_CHECK(hour2   == hour1);
    BOOST_CHECK(minute2 == minute1);
    BOOST_CHECK(second2 == second1);
  }

// -----------------------------------------------------------------------------

  BOOST_AUTO_TEST_CASE(test_construct_from_string) {
    std::string s1("2011-09-16T13:55:20Z");
    util::DateTime d1(s1);
    std::string s2(d1.toString());
    BOOST_CHECK(s2 == s1);
  }

// -----------------------------------------------------------------------------

  BOOST_AUTO_TEST_CASE(test_set_from_string) {
    util::DateTime d1("2012-12-05T15:42:55Z");
    std::string s1("2011-09-16T13:55:20Z");
    d1.set(s1);
    std::string s2(d1.toString());
    BOOST_CHECK(s2 == s1);
  }

// -----------------------------------------------------------------------------

  BOOST_AUTO_TEST_CASE(test_copy_constructor) {
    std::string s1("2011-09-16T13:55:20Z");
    util::DateTime d1(s1);
    util::DateTime d2(d1);
    std::string s2(d2.toString());
    BOOST_CHECK(s2 == s1);
  }

// -----------------------------------------------------------------------------

  BOOST_AUTO_TEST_CASE(test_operators) {
     util::DateTime d1(2011, 9, 16, 13, 55, 20);
     util::DateTime d2(2011, 9, 16, 13, 55, 21);
     util::DateTime d3(2011, 9, 16, 13, 55, 21);
     BOOST_CHECK(d2 == d3);
     BOOST_CHECK(d1 != d2);
     BOOST_CHECK(d1 <  d2);
     BOOST_CHECK(d3 <= d2);
     BOOST_CHECK(d2 >  d1);
     BOOST_CHECK(d3 >= d2);

     util::Duration dur(d2-d1);  // one second

     BOOST_CHECK((d1+dur) == d2);
     BOOST_CHECK((d2-dur) == d1);

     util::DateTime d4(d1);
     d4 += dur;
     BOOST_CHECK(d4 == d2);
     d4 -= dur;
     BOOST_CHECK(d4 == d1);
  }

// -----------------------------------------------------------------------------

  BOOST_AUTO_TEST_CASE(test_stream_write) {
    std::string s1("2011-09-16T13:55:20Z");
    util::DateTime d1(s1);
    std::stringstream strm;
    strm << d1;
    std::string s2;
    strm >> s2;
    BOOST_CHECK(s2 == s1);
  }

// -----------------------------------------------------------------------------

  BOOST_AUTO_TEST_CASE(test_stream_read) {
    std::string s1("2011-09-16T13:55:20Z");
    std::stringstream strm;
    strm << s1;

    util::DateTime d1("2012-12-05T15:42:55Z");
    strm >> d1;
    BOOST_CHECK(s1 == d1.toString());
  }

// -----------------------------------------------------------------------------

  BOOST_AUTO_TEST_CASE(test_special_dates) {
    //  Y2K tests
    util::Duration oneDay("P1D");  // one day
    util::DateTime d1(2000, 2, 28, 0, 0, 0);
    BOOST_CHECK((d1+oneDay).toString() == "2000-02-29T00:00:00Z");

    util::DateTime d2(2000, 2, 29, 0, 0, 0);
    BOOST_CHECK((d2-oneDay).toString() == "2000-02-28T00:00:00Z");

    d1.set(2000, 3, 1, 0, 0, 0);
    BOOST_CHECK((d1-oneDay).toString() == "2000-02-29T00:00:00Z");

    // Far in the past/future
    util::Duration longDur("P345126D");
    util::DateTime d3(2011, 9, 16, 9, 0, 0);
    BOOST_CHECK((d3 - longDur).toString() == "1066-10-14T09:00:00Z");
    BOOST_CHECK((d3 + longDur).toString() == "2956-08-18T09:00:00Z");
  }

// -----------------------------------------------------------------------------

BOOST_AUTO_TEST_SUITE_END()

}  // anonymous namespace
